using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

using TC = System.ComponentModel.TypeConverterAttribute;
using EXP = System.ComponentModel.ExpandableObjectConverter;


namespace CodeWalker.GameFiles
{
    public static class PsoTypes
    {
        //for parsing schema info in PSO files to generate structs for PSO parsing.
        //equivalent of MetaTypes but for PSO.

        public static Dictionary<MetaName, PsoEnumInfo> EnumDict = new Dictionary<MetaName, PsoEnumInfo>();
        public static Dictionary<MetaName, PsoStructureInfo> StructDict = new Dictionary<MetaName, PsoStructureInfo>();



        public static void Clear()
        {
            EnumDict.Clear();
            StructDict.Clear();
        }

        public static void EnsurePsoTypes(PsoFile pso)
        {

            if ((pso.SchemaSection == null) || (pso.SchemaSection.Entries == null) || (pso.SchemaSection.EntriesIdx == null))
            {
                return;
            }


            for (int i = 0; i < pso.SchemaSection.Entries.Length; i++)
            {
                var entry = pso.SchemaSection.Entries[i];
                var enuminfo = entry as PsoEnumInfo;
                var structinfo = entry as PsoStructureInfo;

                if (enuminfo != null)
                {
                    if (!EnumDict.ContainsKey(enuminfo.IndexInfo.NameHash))
                    {
                        EnumDict.Add(enuminfo.IndexInfo.NameHash, enuminfo);
                    }
                    else
                    {
                        PsoEnumInfo oldei = EnumDict[enuminfo.IndexInfo.NameHash];
                        if (!ComparePsoEnumInfos(oldei, enuminfo))
                        {
                            if (enuminfo.EntriesCount > oldei.EntriesCount)//assume this is newer...
                            {
                                EnumDict[enuminfo.IndexInfo.NameHash] = enuminfo;
                            }
                        }
                    }
                }
                else if (structinfo != null)
                {
                    if (!StructDict.ContainsKey(structinfo.IndexInfo.NameHash))
                    {
                        StructDict.Add(structinfo.IndexInfo.NameHash, structinfo);
                    }
                    else
                    {
                        PsoStructureInfo oldsi = StructDict[structinfo.IndexInfo.NameHash];
                        if (!ComparePsoStructureInfos(oldsi, structinfo))
                        {
                            if (structinfo.EntriesCount > oldsi.EntriesCount) //assume more entries is newer.. maybe not correct
                            {
                                StructDict[structinfo.IndexInfo.NameHash] = structinfo;
                            }
                        }
                    }
                }

            }

        }

        public static bool ComparePsoEnumInfos(PsoEnumInfo a, PsoEnumInfo b)
        {
            //returns true if they are the same.

            if (a.Entries.Length != b.Entries.Length)
            {
                return false;
            }

            for (int i = 0; i < a.Entries.Length; i++)
            {
                if ((a.Entries[i].EntryNameHash != b.Entries[i].EntryNameHash) ||
                    (a.Entries[i].EntryKey != b.Entries[i].EntryKey))
                {
                    return false;
                }
            }

            return true;
        }
        public static bool ComparePsoStructureInfos(PsoStructureInfo a, PsoStructureInfo b)
        {
            //returns true if they are the same.

            if (a.Entries.Length != b.Entries.Length)
            {
                return false;
            }

            for (int i = 0; i < a.Entries.Length; i++)
            {
                if ((a.Entries[i].EntryNameHash != b.Entries[i].EntryNameHash) ||
                    (a.Entries[i].DataOffset != b.Entries[i].DataOffset) ||
                    (a.Entries[i].Type != b.Entries[i].Type))
                {
                    return false;
                }
            }

            return true;
        }


        public static string GetTypesString()
        {
            StringBuilder sbe = new StringBuilder();
            StringBuilder sbs = new StringBuilder();

            sbe.AppendLine("//Enum infos");
            sbs.AppendLine("//Struct infos");


            foreach (var kvp in EnumDict)
            {
                var ei = kvp.Value;
                string name = GetSafeName(ei.IndexInfo.NameHash, ei.Type);
                sbe.AppendLine("public enum " + name + " //Type:" + ei.Type.ToString());
                sbe.AppendLine("{");
                foreach (var entry in ei.Entries)
                {
                    string eename = GetSafeName(entry.EntryNameHash, (uint)entry.EntryKey);
                    sbe.AppendFormat("   {0} = {1},", eename, entry.EntryKey);
                    sbe.AppendLine();
                }
                sbe.AppendLine("}");
                sbe.AppendLine();
            }

            foreach (var kvp in StructDict)
            {
                var si = kvp.Value;
                string name = GetSafeName(si.IndexInfo.NameHash, si.Type);
                sbs.AppendLine("public struct " + name + " //" + si.StructureLength.ToString() + " bytes, Type:" + si.Type.ToString());
                sbs.AppendLine("{");
                for (int i = 0; i < si.Entries.Length; i++)
                {
                    var entry = si.Entries[i];

                    if ((entry.DataOffset == 0) && (entry.EntryNameHash == (MetaName)MetaTypeName.ARRAYINFO)) //referred to by array
                    {
                    }
                    else
                    {
                        string sename = GetSafeName(entry.EntryNameHash, entry.ReferenceKey);
                        string fmt = "   public {0} {1}; //{2}   {3}";

                        if (entry.Type == PsoDataType.Array)
                        {
                            if (entry.ReferenceKey >= si.Entries.Length)
                            {
                                sbs.AppendFormat(fmt, entry.Type.ToString(), sename, entry.DataOffset, entry.ToString() + "  { unexpected key! " + entry.ReferenceKey.ToString() + "}");
                                sbs.AppendLine();
                            }
                            else
                            {
                                var structentry = si.Entries[(int)entry.ReferenceKey];
                                var typename = "Array_" + PsoDataTypes.GetCSharpTypeName(structentry.Type);
                                sbs.AppendFormat(fmt, typename, sename, entry.DataOffset, entry.ToString() + "  {" + structentry.ToString() + "}");
                                sbs.AppendLine();
                            }
                        }
                        else if (entry.Type == PsoDataType.Structure)
                        {
                            var typename = GetSafeName((MetaName)entry.ReferenceKey, entry.ReferenceKey);
                            sbs.AppendFormat(fmt, typename, sename, entry.DataOffset, entry.ToString());
                            sbs.AppendLine();
                        }
                        else
                        {
                            var typename = PsoDataTypes.GetCSharpTypeName(entry.Type);
                            sbs.AppendFormat(fmt, typename, sename, entry.DataOffset, entry);
                            sbs.AppendLine();
                        }
                    }
                }
                sbs.AppendLine("}");
                sbs.AppendLine();
            }


            sbe.AppendLine();
            sbe.AppendLine();
            sbe.AppendLine();
            sbe.AppendLine();
            sbe.AppendLine();
            sbe.Append(sbs.ToString());

            string result = sbe.ToString();

            return result;
        }


        public static string GetTypesInitString()
        {
            StringBuilder sb = new StringBuilder();

            foreach (var si in StructDict.Values)
            {
                AddStructureInfoString(si, sb);
            }

            sb.AppendLine();

            foreach (var ei in EnumDict.Values)
            {
                AddEnumInfoString(ei, sb);
            }

            string str = sb.ToString();
            return str;
        }
        private static void AddStructureInfoString(PsoStructureInfo si, StringBuilder sb)
        {
            var ns = GetMetaNameString(si.IndexInfo.NameHash);
            sb.AppendFormat("case " + ns + ":");
            sb.AppendLine();
            sb.AppendFormat("return new PsoStructureInfo({0}, {1}, {2}, {3},", ns, si.Type, si.Unk, si.StructureLength);
            sb.AppendLine();
            for (int i = 0; i < si.Entries.Length; i++)
            {
                var e = si.Entries[i];
                string refkey = "0";
                if (e.ReferenceKey != 0)
                {
                    refkey = GetMetaNameString((MetaName)e.ReferenceKey);
                }
                sb.AppendFormat(" new PsoStructureEntryInfo({0}, PsoDataType.{1}, {2}, {3}, {4})", GetMetaNameString(e.EntryNameHash), e.Type, e.DataOffset, e.Unk_5h, refkey);
                if (i < si.Entries.Length - 1) sb.Append(",");
                sb.AppendLine();
            }
            sb.AppendFormat(");");
            sb.AppendLine();
        }
        private static void AddEnumInfoString(PsoEnumInfo ei, StringBuilder sb)
        {
            var ns = GetMetaNameString(ei.IndexInfo.NameHash);
            sb.AppendFormat("case " + ns + ":");
            sb.AppendLine();
            sb.AppendFormat("return new PsoEnumInfo({0}, {1},", ns, ei.Type);
            sb.AppendLine();
            for (int i = 0; i < ei.Entries.Length; i++)
            {
                var e = ei.Entries[i];
                sb.AppendFormat(" new PsoEnumEntryInfo({0}, {1})", GetMetaNameString(e.EntryNameHash), e.EntryKey);
                if (i < ei.Entries.Length - 1) sb.Append(",");
                sb.AppendLine();
            }
            sb.AppendFormat(");");
            sb.AppendLine();
        }
        private static string GetMetaNameString(MetaName n)
        {
            if (Enum.IsDefined(typeof(MetaName), n))
            {
                return "MetaName." + n.ToString();
            }
            else
            {
                return "(MetaName)" + n.ToString();
            }
        }



        public static PsoStructureInfo GetStructureInfo(MetaName name)
        {
            //to generate structinfos
            switch (name)
            {
                case MetaName.CJunctionTemplateArray:
                    return new PsoStructureInfo(MetaName.CJunctionTemplateArray, 0, 0, 170688,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CJunctionTemplate),
                     new PsoStructureEntryInfo(MetaName.Entries, PsoDataType.Array, 0, 1, (MetaName)9830400),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2111264667),
                     new PsoStructureEntryInfo(MetaName.AutoJunctionAdjustments, PsoDataType.Array, 39344, 129, (MetaName)524290)
                    );
                case MetaName.CJunctionTemplate:
                    return new PsoStructureInfo(MetaName.CJunctionTemplate, 0, 0, 1136,
                     new PsoStructureEntryInfo(MetaName.iFlags, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iNumJunctionNodes, PsoDataType.SInt, 4, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iNumEntrances, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iNumPhases, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iNumTrafficLightLocations, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSearchDistance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPhaseOffset, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vJunctionMin, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vJunctionMax, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float3, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vJunctionNodePositions, PsoDataType.Array, 64, 2, (MetaName)524297),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1057820426),
                     new PsoStructureEntryInfo(MetaName.Entrances, PsoDataType.Array, 192, 2, (MetaName)1048587),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)38018516),
                     new PsoStructureEntryInfo(MetaName.PhaseTimings, PsoDataType.Array, 960, 2, (MetaName)1048589),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2279781599),
                     new PsoStructureEntryInfo(MetaName.TrafficLightLocations, PsoDataType.Array, 1088, 2, (MetaName)524303)
                    );
                case (MetaName)1057820426:
                    return new PsoStructureInfo((MetaName)1057820426, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.vNodePosition, PsoDataType.Float3, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iPhase, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStoppingDistance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fOrientation, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAngleFromCenter, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bCanTurnRightOnRedLight, PsoDataType.Bool, 33, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bLeftLaneIsAheadOnly, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bRightLaneIsRightOnly, PsoDataType.Bool, 34, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iLeftFilterLanePhase, PsoDataType.SInt, 36, 0, 0)
                    );
                case (MetaName)38018516:
                    return new PsoStructureInfo((MetaName)38018516, 0, 0, 8,
                     new PsoStructureEntryInfo(MetaName.fStartTime, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDuration, PsoDataType.Float, 4, 0, 0)
                    );
                case (MetaName)2279781599:
                    return new PsoStructureInfo((MetaName)2279781599, 0, 0, 6,
                     new PsoStructureEntryInfo(MetaName.iPosX, PsoDataType.SShort, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iPosY, PsoDataType.SShort, 2, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iPosZ, PsoDataType.SShort, 4, 0, 0)
                    );
                case (MetaName)2111264667:
                    return new PsoStructureInfo((MetaName)2111264667, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.vLocation, PsoDataType.Float3a, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCycleOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCycleDuration, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CVehicleModelInfoVarGlobal: //carcols.ymt root
                    return new PsoStructureInfo(MetaName.CVehicleModelInfoVarGlobal, 0, 0, 392,
                     new PsoStructureEntryInfo(MetaName.VehiclePlates, PsoDataType.Structure, 8, 0, (MetaName)465922034),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModelColor),
                     new PsoStructureEntryInfo(MetaName.Colors, PsoDataType.Array, 48, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleMetallicSetting),
                     new PsoStructureEntryInfo(MetaName.MetallicSettings, PsoDataType.Array, 64, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleWindowColor),
                     new PsoStructureEntryInfo(MetaName.WindowColors, PsoDataType.Array, 80, 0, (MetaName)5),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.vehicleLightSettings),
                     new PsoStructureEntryInfo(MetaName.Lights, PsoDataType.Array, 96, 0, (MetaName)MetaTypeName.POINTER),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.sirenSettings),
                     new PsoStructureEntryInfo(MetaName.Sirens, PsoDataType.Array, 112, 0, (MetaName)9),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleKit),
                     new PsoStructureEntryInfo(MetaName.Kits, PsoDataType.Array, 128, 0, (MetaName)11),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleWheel),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Array, 0, 0, (MetaName)13),
                     new PsoStructureEntryInfo(MetaName.Wheels, PsoDataType.Array, 144, 4, (MetaName)851982),
                     new PsoStructureEntryInfo(MetaName.GlobalVariationData, PsoDataType.Structure, 352, 0, (MetaName)3062246906),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleXenonLightColor),
                     new PsoStructureEntryInfo(MetaName.XenonLightColors, PsoDataType.Array, 376, 0, (MetaName)MetaTypeName.BYTE)
                    );
                case (MetaName)465922034: //CVehicleModelInfoVarGlobal VehiclePlates
                    return new PsoStructureInfo((MetaName)465922034, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3027500557),
                     new PsoStructureEntryInfo(MetaName.Textures, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTexureIndex, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NumericOffset, PsoDataType.UByte, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AlphabeticOffset, PsoDataType.UByte, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpaceOffset, PsoDataType.UByte, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RandomCharOffset, PsoDataType.UByte, 31, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NumRandomChar, PsoDataType.UByte, 32, 0, 0)
                    );
                case (MetaName)3027500557: //CVehicleModelInfoVarGlobal VehiclePlates Texture
                    return new PsoStructureInfo((MetaName)3027500557, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.TextureSetName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DiffuseMapName, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.NormalMapName, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FontExtents, PsoDataType.Float4, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLettersOnPlate, PsoDataType.Float2, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FontColor, PsoDataType.UInt, 56, 1, 0),
                     new PsoStructureEntryInfo(MetaName.FontOutlineColor, PsoDataType.UInt, 60, 1, 0),
                     new PsoStructureEntryInfo(MetaName.IsFontOutlineEnabled, PsoDataType.Bool, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FontOutlineMinMaxDepth, PsoDataType.Float2, 68, 0, 0)
                    );
                case (MetaName)3062246906: //CVehicleModelInfoVarGlobal GlobalVariationData
                    return new PsoStructureInfo((MetaName)3062246906, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.xenonLightColor, PsoDataType.UInt, 8, 1, 0),
                     new PsoStructureEntryInfo(MetaName.xenonCoronaColor, PsoDataType.UInt, 12, 1, 0),
                     new PsoStructureEntryInfo(MetaName.xenonLightIntensityModifier, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.xenonCoronaIntensityModifier, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CVehicleXenonLightColor: //CVehicleModelInfoVarGlobal XenonLightColors
                    return new PsoStructureInfo(MetaName.CVehicleXenonLightColor, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.lightColor, PsoDataType.UInt, 8, 1, 0),
                     new PsoStructureEntryInfo(MetaName.coronaColor, PsoDataType.UInt, 12, 1, 0),
                     new PsoStructureEntryInfo(MetaName.lightIntensityModifier, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.coronaIntensityModifier, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CVehicleModelColor: //CVehicleModelInfoVarGlobal Colors
                    return new PsoStructureInfo(MetaName.CVehicleModelColor, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.color, PsoDataType.UInt, 8, 1, 0),
                     new PsoStructureEntryInfo(MetaName.metallicID, PsoDataType.Enum, 12, 2, (MetaName)360458334),
                     new PsoStructureEntryInfo(MetaName.audioColor, PsoDataType.Enum, 13, 2, (MetaName)544262540),
                     new PsoStructureEntryInfo(MetaName.audioPrefix, PsoDataType.Enum, 14, 2, (MetaName)2065815796),
                     new PsoStructureEntryInfo(MetaName.audioColorHash, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.audioPrefixHash, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.colorName, PsoDataType.String, 24, 1, 0)
                    );
                case MetaName.CVehicleMetallicSetting: //CVehicleModelInfoVarGlobal MetallicSettings
                    return new PsoStructureInfo(MetaName.CVehicleMetallicSetting, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.specInt, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.specFalloff, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.specFresnel, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CVehicleWindowColor: //CVehicleModelInfoVarGlobal WindowColors
                    return new PsoStructureInfo(MetaName.CVehicleWindowColor, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.color, PsoDataType.UInt, 8, 1, 0),
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 12, 7, 0)
                    );
                case MetaName.vehicleLightSettings: //CVehicleModelInfoVarGlobal Lights
                    return new PsoStructureInfo(MetaName.vehicleLightSettings, 0, 0, 552,
                     new PsoStructureEntryInfo(MetaName.id, PsoDataType.UByte, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.indicator, PsoDataType.Structure, 16, 0, MetaName.vehicleLight),
                     new PsoStructureEntryInfo(MetaName.rearIndicatorCorona, PsoDataType.Structure, 64, 0, MetaName.vehicleCorona),
                     new PsoStructureEntryInfo(MetaName.frontIndicatorCorona, PsoDataType.Structure, 120, 0, MetaName.vehicleCorona),
                     new PsoStructureEntryInfo(MetaName.tailLight, PsoDataType.Structure, 176, 0, MetaName.vehicleLight),
                     new PsoStructureEntryInfo(MetaName.tailLightCorona, PsoDataType.Structure, 224, 0, MetaName.vehicleCorona),
                     new PsoStructureEntryInfo(MetaName.tailLightMiddleCorona, PsoDataType.Structure, 280, 0, MetaName.vehicleCorona),
                     new PsoStructureEntryInfo(MetaName.headLight, PsoDataType.Structure, 336, 0, MetaName.vehicleLight),
                     new PsoStructureEntryInfo(MetaName.headLightCorona, PsoDataType.Structure, 384, 0, MetaName.vehicleCorona),
                     new PsoStructureEntryInfo(MetaName.reversingLight, PsoDataType.Structure, 440, 0, MetaName.vehicleLight),
                     new PsoStructureEntryInfo(MetaName.reversingLightCorona, PsoDataType.Structure, 488, 0, MetaName.vehicleCorona),
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 544, 1, 0)
                    );
                case MetaName.vehicleLight: //CVehicleModelInfoVarGlobal Lights Light info
                    return new PsoStructureInfo(MetaName.vehicleLight, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.intensity, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.falloffMax, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.falloffExponent, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.innerConeAngle, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.outerConeAngle, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.emmissiveBoost, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.color, PsoDataType.UInt, 32, 1, 0),
                     new PsoStructureEntryInfo(MetaName.textureName, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.mirrorTexture, PsoDataType.Bool, 40, 0, 0)
                    );
                case MetaName.vehicleCorona: //CVehicleModelInfoVarGlobal Lights Corona info
                    return new PsoStructureInfo(MetaName.vehicleCorona, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.size, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.size_far, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.intensity, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.intensity_far, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.color, PsoDataType.UInt, 24, 1, 0),
                     new PsoStructureEntryInfo(MetaName.numCoronas, PsoDataType.UByte, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.distBetweenCoronas, PsoDataType.UByte, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.distBetweenCoronas_far, PsoDataType.UByte, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.xRotation, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.yRotation, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.zRotation, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.zBias, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pullCoronaIn, PsoDataType.Bool, 48, 0, 0)
                    );
                case MetaName.sirenSettings:
                    return new PsoStructureInfo(MetaName.sirenSettings, 0, 0, 2376,
                     new PsoStructureEntryInfo(MetaName.id, PsoDataType.UByte, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 16, 1, 0),
                     new PsoStructureEntryInfo(MetaName.timeMultiplier, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lightFalloffMax, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lightFalloffExponent, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lightInnerConeAngle, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lightOuterConeAngle, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lightOffset, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.textureName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.sequencerBpm, PsoDataType.UInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leftHeadLight, PsoDataType.Structure, 56, 0, (MetaName)188820339),
                     new PsoStructureEntryInfo(MetaName.rightHeadLight, PsoDataType.Structure, 72, 0, (MetaName)188820339),
                     new PsoStructureEntryInfo(MetaName.leftTailLight, PsoDataType.Structure, 88, 0, (MetaName)188820339),
                     new PsoStructureEntryInfo(MetaName.rightTailLight, PsoDataType.Structure, 104, 0, (MetaName)188820339),
                     new PsoStructureEntryInfo(MetaName.leftHeadLightMultiples, PsoDataType.UByte, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rightHeadLightMultiples, PsoDataType.UByte, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leftTailLightMultiples, PsoDataType.UByte, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rightTailLightMultiples, PsoDataType.UByte, 123, 0, 0),
                     new PsoStructureEntryInfo(MetaName.useRealLights, PsoDataType.Bool, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.sirenLight),
                     new PsoStructureEntryInfo(MetaName.sirens, PsoDataType.Array, 128, 1, (MetaName)1310739)
                    );
                case (MetaName)188820339:
                    return new PsoStructureInfo((MetaName)188820339, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.sequencer, PsoDataType.UInt, 8, 0, 0)
                    );
                case MetaName.sirenLight:
                    return new PsoStructureInfo(MetaName.sirenLight, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.rotation, PsoDataType.Structure, 8, 0, (MetaName)1356743507),
                     new PsoStructureEntryInfo(MetaName.flashiness, PsoDataType.Structure, 40, 0, (MetaName)1356743507),
                     new PsoStructureEntryInfo(MetaName.corona, PsoDataType.Structure, 72, 0, MetaName.sirenCorona),
                     new PsoStructureEntryInfo(MetaName.color, PsoDataType.UInt, 96, 1, 0),
                     new PsoStructureEntryInfo(MetaName.intensity, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lightGroup, PsoDataType.UByte, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rotate, PsoDataType.Bool, 105, 0, 0),
                     new PsoStructureEntryInfo(MetaName.scale, PsoDataType.Bool, 106, 0, 0),
                     new PsoStructureEntryInfo(MetaName.scaleFactor, PsoDataType.UByte, 107, 0, 0),
                     new PsoStructureEntryInfo(MetaName.flash, PsoDataType.Bool, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.light, PsoDataType.Bool, 109, 0, 0),
                     new PsoStructureEntryInfo(MetaName.spotLight, PsoDataType.Bool, 110, 0, 0),
                     new PsoStructureEntryInfo(MetaName.castShadows, PsoDataType.Bool, 111, 0, 0)
                    );
                case (MetaName)1356743507:
                    return new PsoStructureInfo((MetaName)1356743507, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.delta, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.start, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.speed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.sequencer, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.multiples, PsoDataType.UByte, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.direction, PsoDataType.Bool, 25, 0, 0),
                     new PsoStructureEntryInfo(MetaName.syncToBpm, PsoDataType.Bool, 26, 0, 0)
                    );
                case MetaName.sirenCorona:
                    return new PsoStructureInfo(MetaName.sirenCorona, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.intensity, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.size, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pull, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.faceCamera, PsoDataType.Bool, 20, 0, 0)
                    );
                case MetaName.CVehicleKit:
                    return new PsoStructureInfo(MetaName.CVehicleKit, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.kitName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.id, PsoDataType.UShort, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.kitType, PsoDataType.Enum, 16, 0, MetaName.eModKitType),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModVisible),
                     new PsoStructureEntryInfo(MetaName.visibleMods, PsoDataType.Array, 24, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModLink),
                     new PsoStructureEntryInfo(MetaName.linkMods, PsoDataType.Array, 40, 0, (MetaName)5),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModStat),
                     new PsoStructureEntryInfo(MetaName.statMods, PsoDataType.Array, 56, 0, (MetaName)7),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)427606548),
                     new PsoStructureEntryInfo(MetaName.slotNames, PsoDataType.Array, 72, 0, (MetaName)9),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 2, 0),
                     new PsoStructureEntryInfo(MetaName.liveryNames, PsoDataType.Array, 88, 0, (MetaName)11),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 2, 0),
                     new PsoStructureEntryInfo(MetaName.livery2Names, PsoDataType.Array, 104, 0, (MetaName)13)
                    );
                case MetaName.CVehicleModStat:
                    return new PsoStructureInfo(MetaName.CVehicleModStat, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.identifier, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.modifier, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.audioApply, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.weight, PsoDataType.UByte, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.type, PsoDataType.Enum, 24, 0, MetaName.eVehicleModType)
                    );
                case MetaName.CVehicleModVisible:
                    return new PsoStructureInfo(MetaName.CVehicleModVisible, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.modelName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.modShopLabel, PsoDataType.String, 16, 2, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.linkedModels, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3635907608),
                     new PsoStructureEntryInfo(MetaName.turnOffBones, PsoDataType.Array, 40, 0, (MetaName)4),
                     new PsoStructureEntryInfo(MetaName.type, PsoDataType.Enum, 56, 0, MetaName.eVehicleModType),
                     new PsoStructureEntryInfo(MetaName.bone, PsoDataType.Enum, 60, 0, (MetaName)3635907608),
                     new PsoStructureEntryInfo(MetaName.collisionBone, PsoDataType.Enum, 64, 0, (MetaName)3635907608),
                     new PsoStructureEntryInfo(MetaName.cameraPos, PsoDataType.Enum, 68, 0, MetaName.eVehicleModCameraPos),
                     new PsoStructureEntryInfo(MetaName.audioApply, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.weight, PsoDataType.UByte, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.turnOffExtra, PsoDataType.Bool, 77, 0, 0),
                     new PsoStructureEntryInfo(MetaName.disableBonnetCamera, PsoDataType.Bool, 78, 0, 0),
                     new PsoStructureEntryInfo(MetaName.allowBonnetSlide, PsoDataType.Bool, 79, 0, 0),
                     new PsoStructureEntryInfo(MetaName.weaponSlot, PsoDataType.SByte, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.weaponSlotSecondary, PsoDataType.SByte, 81, 0, 0),
                     new PsoStructureEntryInfo(MetaName.disableProjectileDriveby, PsoDataType.Bool, 82, 0, 0),
                     new PsoStructureEntryInfo(MetaName.disableDriveby, PsoDataType.Bool, 83, 0, 0),
                     new PsoStructureEntryInfo(MetaName.disableDrivebySeat, PsoDataType.SInt, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.disableDrivebySeatSecondary, PsoDataType.SInt, 88, 0, 0)
                    );
                case (MetaName)427606548:
                    return new PsoStructureInfo((MetaName)427606548, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.slot, PsoDataType.Enum, 8, 0, MetaName.eVehicleModType),
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 16, 2, 0)
                    );
                case MetaName.CVehicleModLink:
                    return new PsoStructureInfo(MetaName.CVehicleModLink, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.modelName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.bone, PsoDataType.Enum, 12, 0, (MetaName)3635907608),
                     new PsoStructureEntryInfo(MetaName.turnOffExtra, PsoDataType.Bool, 16, 0, 0)
                    );
                case MetaName.CVehicleWheel:
                    return new PsoStructureInfo(MetaName.CVehicleWheel, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.wheelName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.wheelVariation, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.modShopLabel, PsoDataType.String, 16, 2, 0),
                     new PsoStructureEntryInfo(MetaName.rimRadius, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rear, PsoDataType.Bool, 28, 0, 0)
                    );
                case MetaName.CVehicleModColors: //carmodcols.ymt root
                    return new PsoStructureInfo(MetaName.CVehicleModColors, 0, 0, 128,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModColor),
                     new PsoStructureEntryInfo(MetaName.metallic, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModColor),
                     new PsoStructureEntryInfo(MetaName.classic, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModColor),
                     new PsoStructureEntryInfo(MetaName.matte, PsoDataType.Array, 40, 0, (MetaName)4),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModColor),
                     new PsoStructureEntryInfo(MetaName.metals, PsoDataType.Array, 56, 0, (MetaName)6),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModColor),
                     new PsoStructureEntryInfo(MetaName.chrome, PsoDataType.Array, 72, 0, (MetaName)8),
                     new PsoStructureEntryInfo(MetaName.pearlescent, PsoDataType.Structure, 88, 0, MetaName.CVehicleModPearlescentColors)
                    );
                case MetaName.CVehicleModPearlescentColors:
                    return new PsoStructureInfo(MetaName.CVehicleModPearlescentColors, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModColor),
                     new PsoStructureEntryInfo(MetaName.baseCols, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CVehicleModColor),
                     new PsoStructureEntryInfo(MetaName.specCols, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case MetaName.CVehicleModColor:
                    return new PsoStructureInfo(MetaName.CVehicleModColor, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 2, 0),
                     new PsoStructureEntryInfo(MetaName.col, PsoDataType.UByte, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.spec, PsoDataType.UByte, 17, 0, 0)
                    );
                case MetaName.CVehicleModelInfoVariation: //carvariations.ymt root
                    return new PsoStructureInfo(MetaName.CVehicleModelInfoVariation, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)418053801),
                     new PsoStructureEntryInfo(MetaName.variationData, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)418053801: //CVehicleModelInfoVariation variationData
                    return new PsoStructureInfo((MetaName)418053801, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.modelName, PsoDataType.String, 8, 1, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2575850962),
                     new PsoStructureEntryInfo(MetaName.colors, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.kits, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.windowsWithExposedEdges, PsoDataType.Array, 48, 0, (MetaName)5),
                     new PsoStructureEntryInfo(MetaName.plateProbabilities, PsoDataType.Structure, 64, 0, MetaName.PlateProbabilities),
                     new PsoStructureEntryInfo(MetaName.lightSettings, PsoDataType.UByte, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.sirenSettings, PsoDataType.UByte, 89, 0, 0)
                    );
                case (MetaName)2575850962: //CVehicleModelInfoVariation variationData colors
                    return new PsoStructureInfo((MetaName)2575850962, 0, 0, 44 /*40*/,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.indices, PsoDataType.Array, 8, 4, (MetaName)393216 /*262144*/),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Bool, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.liveries, PsoDataType.Array, 14 /*12*/, 4, (MetaName)1966082 /*1638402*/)
                    );
                case MetaName.PlateProbabilities:
                    return new PsoStructureInfo(MetaName.PlateProbabilities, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)938618322),
                     new PsoStructureEntryInfo(MetaName.Probabilities, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)938618322:
                    return new PsoStructureInfo((MetaName)938618322, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.UInt, 12, 0, 0)
                    );
                case MetaName.CCreditArray: //credits.ymt root
                    return new PsoStructureInfo(MetaName.CCreditArray, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CCreditItem),
                     new PsoStructureEntryInfo(MetaName.CreditItems, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.CCreditItem:
                    return new PsoStructureInfo(MetaName.CCreditItem, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.LineType, PsoDataType.Enum, 8, 0, (MetaName)1795685103),
                     new PsoStructureEntryInfo(MetaName.cTextId1, PsoDataType.String, 16, 3, 0),
                     new PsoStructureEntryInfo(MetaName.cTextId2, PsoDataType.String, 32, 3, 0)
                    );
                case MetaName.CDataFileMgr__ContentsOfDataFileXml:
                    return new PsoStructureInfo(MetaName.CDataFileMgr__ContentsOfDataFileXml, 0, 0, 120,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.disabledFiles, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CDataFileMgr__DataFileArray),
                     new PsoStructureEntryInfo(MetaName.includedXmlFiles, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.includedDataFiles, PsoDataType.Array, 40, 0, (MetaName)4),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CDataFileMgr__DataFile),
                     new PsoStructureEntryInfo(MetaName.dataFiles, PsoDataType.Array, 56, 0, (MetaName)6),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CDataFileMgr__ContentChangeSet),
                     new PsoStructureEntryInfo(MetaName.contentChangeSets, PsoDataType.Array, 72, 0, (MetaName)8),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.patchFiles, PsoDataType.Array, 88, 0, (MetaName)10),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.allowedFolders, PsoDataType.Array, 104, 0, (MetaName)MetaTypeName.PsoPOINTER)
                    );
                case MetaName.CDataFileMgr__DataFile:
                    return new PsoStructureInfo(MetaName.CDataFileMgr__DataFile, 0, 0, 176,
                     new PsoStructureEntryInfo(MetaName.filename, PsoDataType.String, 8, 0, (MetaName)8388608),
                     new PsoStructureEntryInfo(MetaName.fileType, PsoDataType.Enum, 136, 0, MetaName.DataFileType),
                     new PsoStructureEntryInfo(MetaName.registerAs, PsoDataType.String, 144, 3, 0),
                     new PsoStructureEntryInfo(MetaName.locked, PsoDataType.Bool, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.loadCompletely, PsoDataType.Bool, 161, 0, 0),
                     new PsoStructureEntryInfo(MetaName.overlay, PsoDataType.Bool, 162, 0, 0),
                     new PsoStructureEntryInfo(MetaName.patchFile, PsoDataType.Bool, 163, 0, 0),
                     new PsoStructureEntryInfo(MetaName.disabled, PsoDataType.Bool, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.persistent, PsoDataType.Bool, 165, 0, 0),
                     new PsoStructureEntryInfo(MetaName.enforceLsnSorting, PsoDataType.Bool, 166, 0, 0),
                     new PsoStructureEntryInfo(MetaName.contents, PsoDataType.Enum, 168, 0, MetaName.DataFileContents),
                     new PsoStructureEntryInfo(MetaName.installPartition, PsoDataType.Enum, 172, 0, MetaName.InstallPartition)
                    );
                case MetaName.CLevelData: //levels.ymt root
                    return new PsoStructureInfo(MetaName.CLevelData, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.sLevelData),
                     new PsoStructureEntryInfo(MetaName.aLevelsData, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.sLevelData:
                    return new PsoStructureInfo(MetaName.sLevelData, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.cFriendlyName, PsoDataType.String, 8, 3, 0),
                     new PsoStructureEntryInfo(MetaName.cTitle, PsoDataType.String, 24, 3, 0),
                     new PsoStructureEntryInfo(MetaName.cFilename, PsoDataType.String, 40, 3, 0),
                     new PsoStructureEntryInfo(MetaName.cBugstarName, PsoDataType.String, 56, 3, 0)
                    );
                case (MetaName)1056751197: //mapdatacullboxes.ymt root
                    return new PsoStructureInfo((MetaName)1056751197, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)322201475),
                     new PsoStructureEntryInfo(MetaName.entries, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)322201475:
                    return new PsoStructureInfo((MetaName)322201475, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.aabb, PsoDataType.Structure, 16, 0, MetaName.rage__spdAABB),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4022397219, PsoDataType.Array, 48, 0, (MetaName)2),
                     new PsoStructureEntryInfo(MetaName.bEnabled, PsoDataType.Bool, 64, 0, 0)
                    );
                case MetaName.rage__spdAABB:
                    return new PsoStructureInfo(MetaName.rage__spdAABB, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.min, PsoDataType.Float4a, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.max, PsoDataType.Float4a, 16, 0, 0)
                    );
                case MetaName.CPedModelInfo__InitDataList:
                    return new PsoStructureInfo(MetaName.CPedModelInfo__InitDataList, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.residentTxd, PsoDataType.String, 8, 3, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.residentAnims, PsoDataType.Array, 24, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPedModelInfo__InitData),
                     new PsoStructureEntryInfo(MetaName.InitDatas, PsoDataType.Array, 40, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CTxdRelationship),
                     new PsoStructureEntryInfo(MetaName.txdRelationships, PsoDataType.Array, 56, 0, (MetaName)5),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CMultiTxdRelationship),
                     new PsoStructureEntryInfo(MetaName.multiTxdRelationships, PsoDataType.Array, 72, 0, (MetaName)7)
                    );
                case MetaName.CPedModelInfo__InitData:
                    return new PsoStructureInfo(MetaName.CPedModelInfo__InitData, 0, 0, 296,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 2, 0),
                     new PsoStructureEntryInfo(MetaName.PropsName, PsoDataType.String, 16, 2, 0),
                     new PsoStructureEntryInfo(MetaName.ClipDictionaryName, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BlendShapeFileName, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ExpressionSetName, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ExpressionDictionaryName, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ExpressionName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Pedtype, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MovementClipSet, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MovementClipSets, PsoDataType.Array, 56, 0, (MetaName)9),
                     new PsoStructureEntryInfo(MetaName.StrafeClipSet, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MovementToStrafeClipSet, PsoDataType.String, 76, 7, 0),
                     new PsoStructureEntryInfo(MetaName.InjuredStrafeClipSet, PsoDataType.String, 80, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FullBodyDamageClipSet, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AdditiveDamageClipSet, PsoDataType.String, 88, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultGestureClipSet, PsoDataType.String, 92, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FacialClipsetGroupName, PsoDataType.String, 96, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultVisemeClipSet, PsoDataType.String, 100, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SidestepClipSet, PsoDataType.String, 104, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PoseMatcherName, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PoseMatcherProneName, PsoDataType.String, 112, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GetupSetHash, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CreatureMetadataName, PsoDataType.String, 120, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DecisionMakerName, PsoDataType.String, 124, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MotionTaskDataSetName, PsoDataType.String, 128, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTaskDataSetName, PsoDataType.String, 132, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PedCapsuleName, PsoDataType.String, 136, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PedLayoutName, PsoDataType.String, 140, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PedComponentSetName, PsoDataType.String, 144, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PedComponentClothName, PsoDataType.String, 148, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PedIKSettingsName, PsoDataType.String, 152, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TaskDataName, PsoDataType.String, 156, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IsStreamedGfx, PsoDataType.Bool, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmbulanceShouldRespondTo, PsoDataType.Bool, 161, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CanRideBikeWithNoHelmet, PsoDataType.Bool, 162, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CanSpawnInCar, PsoDataType.Bool, 163, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsHeadBlendPed, PsoDataType.Bool, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bOnlyBulkyItemVariations, PsoDataType.Bool, 165, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelationshipGroup, PsoDataType.String, 168, 7, 0),
                     new PsoStructureEntryInfo(MetaName.NavCapabilitiesName, PsoDataType.String, 172, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PerceptionInfo, PsoDataType.String, 176, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBrawlingStyle, PsoDataType.String, 180, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultUnarmedWeapon, PsoDataType.String, 184, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Personality, PsoDataType.String, 188, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CombatInfo, PsoDataType.String, 192, 7, 0),
                     new PsoStructureEntryInfo(MetaName.VfxInfoName, PsoDataType.String, 196, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AmbientClipsForFlee, PsoDataType.String, 200, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Radio1, PsoDataType.Enum, 204, 0, MetaName.ePedRadioGenre),
                     new PsoStructureEntryInfo(MetaName.Radio2, PsoDataType.Enum, 208, 0, MetaName.ePedRadioGenre),
                     new PsoStructureEntryInfo(MetaName.FUpOffset, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RUpOffset, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FFrontOffset, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RFrontOffset, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinActivationImpulse, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Stubble, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HDDist, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetingThreatModifier, PsoDataType.Float, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KilledPerceptionRangeModifer, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.eSexinessFlags),
                     new PsoStructureEntryInfo(MetaName.Sexiness, PsoDataType.Flags, 248, 0, (MetaName)2097211),
                     new PsoStructureEntryInfo(MetaName.Age, PsoDataType.UByte, 252, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPassengersInCar, PsoDataType.UByte, 253, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.eExternallyDrivenDOFs),
                     new PsoStructureEntryInfo(MetaName.ExternallyDrivenDOFs, PsoDataType.Flags, 254, 2, (MetaName)327743),
                     new PsoStructureEntryInfo(MetaName.PedVoiceGroup, PsoDataType.String, 256, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AnimalAudioObject, PsoDataType.String, 260, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AbilityType, PsoDataType.Enum, 264, 0, MetaName.SpecialAbilityType),
                     new PsoStructureEntryInfo(MetaName.ThermalBehaviour, PsoDataType.Enum, 268, 0, MetaName.ThermalBehaviour),
                     new PsoStructureEntryInfo(MetaName.SuperlodType, PsoDataType.Enum, 272, 0, MetaName.eSuperlodType),
                     new PsoStructureEntryInfo(MetaName.ScenarioPopStreamingSlot, PsoDataType.Enum, 276, 0, MetaName.eScenarioPopStreamingSlot),
                     new PsoStructureEntryInfo(MetaName.DefaultSpawningPreference, PsoDataType.Enum, 280, 0, MetaName.DefaultSpawnPreference),
                     new PsoStructureEntryInfo(MetaName.DefaultRemoveRangeMultiplier, PsoDataType.Float, 284, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowCloseSpawning, PsoDataType.Bool, 288, 0, 0)
                    );
                case MetaName.CMultiTxdRelationship:
                    return new PsoStructureInfo(MetaName.CMultiTxdRelationship, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.parent, PsoDataType.String, 8, 3, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.children, PsoDataType.Array, 24, 0, (MetaName)1)
                    );
                case MetaName.CPlayerSpecialAbilityManager:
                    return new PsoStructureInfo(MetaName.CPlayerSpecialAbilityManager, 0, 0, 64,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)4125685004),
                     new PsoStructureEntryInfo(MetaName.specialAbilities, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.smallCharge, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.mediumCharge, PsoDataType.SInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.largeCharge, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.continuousCharge, PsoDataType.SInt, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fadeCurveType, PsoDataType.Enum, 40, 0, MetaName.eFadeCurveType),
                     new PsoStructureEntryInfo(MetaName.halfSigmoidConstant, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.sigmoidConstant, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fadeInTime, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fadeOutTime, PsoDataType.Float, 56, 0, 0)
                    );
                case (MetaName)4125685004:
                    return new PsoStructureInfo((MetaName)4125685004, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.duration, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.initialUnlockedCap, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.timeWarpScale, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.damageMultiplier, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.defenseMultiplier, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.depletionMultiplier, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.chargeMultiplier, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fxName, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.outFxName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.activeAnimSet, PsoDataType.String, 44, 7, 0)
                    );
                case MetaName.CDataFileMgr__ContentChangeSet:
                    return new PsoStructureInfo(MetaName.CDataFileMgr__ContentChangeSet, 0, 0, 184,
                     new PsoStructureEntryInfo(MetaName.changeSetName, PsoDataType.String, 8, 3, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CDataFileMgr__ChangeSetData),
                     new PsoStructureEntryInfo(MetaName.mapChangeSetData, PsoDataType.Array, 24, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.filesToInvalidate, PsoDataType.Array, 40, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.filesToDisable, PsoDataType.Array, 56, 0, (MetaName)5),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.filesToEnable, PsoDataType.Array, 72, 0, (MetaName)MetaTypeName.POINTER),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.txdToLoad, PsoDataType.Array, 88, 0, (MetaName)9),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.txdToUnload, PsoDataType.Array, 104, 0, (MetaName)11),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CDataFileMgr__ResourceReference),
                     new PsoStructureEntryInfo(MetaName.residentResources, PsoDataType.Array, 120, 0, (MetaName)13),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CDataFileMgr__ResourceReference),
                     new PsoStructureEntryInfo(MetaName.unregisterResources, PsoDataType.Array, 136, 0, (MetaName)15),
                     new PsoStructureEntryInfo(MetaName.requiresLoadingScreen, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.executionConditions, PsoDataType.Structure, 160, 0, MetaName.ExecutionConditions)
                    );
                case MetaName.ExecutionConditions:
                    return new PsoStructureInfo(MetaName.ExecutionConditions, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.ExecutionCondition),
                     new PsoStructureEntryInfo(MetaName.activeChangesetConditions, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.fwClipSetManager:
                    return new PsoStructureInfo(MetaName.fwClipSetManager, 0, 0, 128,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.clipSets, PsoDataType.Map, 8, 1, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.fwClipDictionaryMetadata),
                     new PsoStructureEntryInfo(MetaName.clipDictionaryMetadatas, PsoDataType.Map, 32, 1, (MetaName)196612),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.fwMemoryGroupMetadata),
                     new PsoStructureEntryInfo(MetaName.memoryGroupMetadatas, PsoDataType.Map, 56, 1, (MetaName)393223),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3219912345),
                     new PsoStructureEntryInfo((MetaName)3532340754, PsoDataType.Map, 80, 1, (MetaName)589834),//memory group modifiers
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.clipVariationSets, PsoDataType.Map, 104, 1, (MetaName)786445)//move network flags dictionary
                    );
                case (MetaName)MetaTypeName.ARRAYINFO://default map node entry, with a structure pointer
                    return new PsoStructureInfo((MetaName)MetaTypeName.ARRAYINFO, 0, 2, 16,
                     new PsoStructureEntryInfo(MetaName.Key, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Item, PsoDataType.Structure, 8, 3, 0)
                    );
                case MetaName.fwClipSet:
                    return new PsoStructureInfo(MetaName.fwClipSet, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.fallbackId, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.clipDictionaryName, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.clipItems, PsoDataType.Map, 16, 1, (MetaName)131075),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.moveNetworkFlags, PsoDataType.Array, 40, 0, (MetaName)5)
                    );
                case MetaName.fwDirectedClipItemWithProps:
                    return new PsoStructureInfo(MetaName.fwDirectedClipItemWithProps, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.eAnimPlayerFlag),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.Flags, 8, 0, (MetaName)2097152),
                     new PsoStructureEntryInfo(MetaName.priority, PsoDataType.Enum, 12, 0, MetaName.eAnimPriority),
                     new PsoStructureEntryInfo(MetaName.boneMask, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.direction, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.fwClipItemWithProps:
                    return new PsoStructureInfo(MetaName.fwClipItemWithProps, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.eAnimPlayerFlag),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.Flags, 8, 0, (MetaName)2097152),
                     new PsoStructureEntryInfo(MetaName.priority, PsoDataType.Enum, 12, 0, MetaName.eAnimPriority),
                     new PsoStructureEntryInfo(MetaName.boneMask, PsoDataType.String, 16, 7, 0)
                    );
                case MetaName.fwClipItem:
                    return new PsoStructureInfo(MetaName.fwClipItem, 0, 0, 8   //empty struct... possibly abstract/base class
                    );
                case MetaName.fwClipSetWithGetup: //creature/animal clip set?
                    return new PsoStructureInfo(MetaName.fwClipSetWithGetup, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.fallbackId, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.clipDictionaryName, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.clipItems, PsoDataType.Map, 16, 1, (MetaName)131075),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.moveNetworkFlags, PsoDataType.Array, 40, 0, (MetaName)5),
                     new PsoStructureEntryInfo(MetaName.GetupSet, PsoDataType.String, 56, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DeadBlendOutSet, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AllowAnimatedDeadFall, PsoDataType.Bool, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AnimatedDeadFallSet, PsoDataType.String, 68, 7, 0)
                    );
                case MetaName.fwClipDictionaryMetadata:
                    return new PsoStructureInfo(MetaName.fwClipDictionaryMetadata, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.eStreamingPolicy),
                     new PsoStructureEntryInfo(MetaName.streamingPolicy, PsoDataType.Flags, 8, 0, (MetaName)2097152),
                     new PsoStructureEntryInfo(MetaName.memoryGroup, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.streamingPriority, PsoDataType.Enum, 16, 0, MetaName.eStreamingPriority)
                    );
                case MetaName.fwMemoryGroupMetadata:
                    return new PsoStructureInfo(MetaName.fwMemoryGroupMetadata, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)676191063, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1661025772, PsoDataType.UInt, 12, 0, 0)
                    );
                case (MetaName)3219912345://clip sets memory group modifier?
                    return new PsoStructureInfo((MetaName)3219912345, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2356519750),
                     new PsoStructureEntryInfo(MetaName.Adjustments, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)2356519750: //clip sets memory group modifier adjustment
                    return new PsoStructureInfo((MetaName)2356519750, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MemoryGroup, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Amount, PsoDataType.SInt, 12, 0, 0)
                    );
                case MetaName.fwClipVariationSet: //clip sets move network flags list/group?
                    return new PsoStructureInfo(MetaName.fwClipVariationSet, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.clips, PsoDataType.Array, 8, 0, 0)
                    );
                //case (MetaName)257://this shouldn't really be in here, because it's dynamically generated, see PsoBuilder.AddMapNodeStructureInfo
                //    return new PsoStructureInfo((MetaName)257, 0, 2, 32,
                //     new PsoStructureEntryInfo(MetaName.Key, PsoDataType.String, 0, 7, 0),
                //     new PsoStructureEntryInfo(MetaName.Item, PsoDataType.Structure, 8, 0, MetaName.fwClipDictionaryMetadata)
                //    );
                //case (MetaName)258://this shouldn't really be in here, because it's dynamically generated, see PsoBuilder.AddMapNodeStructureInfo
                //    return new PsoStructureInfo((MetaName)258, 0, 2, 24,
                //     new PsoStructureEntryInfo(MetaName.Key, PsoDataType.String, 0, 7, 0),
                //     new PsoStructureEntryInfo(MetaName.Item, PsoDataType.Structure, 8, 0, MetaName.fwMemoryGroupMetadata)
                //    );
                //case (MetaName)259://this shouldn't really be in here, because it's dynamically generated, see PsoBuilder.AddMapNodeStructureInfo
                //    return new PsoStructureInfo((MetaName)259, 0, 2, 32,
                //     new PsoStructureEntryInfo(MetaName.Key, PsoDataType.String, 0, 7, 0),
                //     new PsoStructureEntryInfo(MetaName.Item, PsoDataType.Structure, 8, 0, (MetaName)3219912345)
                //    );
                case MetaName.AnimPostFXManager:
                    return new PsoStructureInfo(MetaName.AnimPostFXManager, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2319197961),
                     new PsoStructureEntryInfo(MetaName.RegisteredStacks, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)2319197961:
                    return new PsoStructureInfo((MetaName)2319197961, 0, 0, 384,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FXStack, PsoDataType.Structure, 16, 0, (MetaName)2675557154),
                     new PsoStructureEntryInfo(MetaName.Priority, PsoDataType.UByte, 376, 0, 0)
                    );
                case (MetaName)2675557154:
                    return new PsoStructureInfo((MetaName)2675557154, 0, 0, 360,
                     new PsoStructureEntryInfo(MetaName.LayerBlend, PsoDataType.Structure, 8, 0, (MetaName)2367955297),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.AnimatedPostFX),
                     new PsoStructureEntryInfo(MetaName.Layers, PsoDataType.Array, 48, 1, (MetaName)393217),
                     new PsoStructureEntryInfo(MetaName.GroupId, PsoDataType.SInt, 344, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1991133938, PsoDataType.UInt, 348, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4185090704, PsoDataType.UInt, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EventType, PsoDataType.Enum, 356, 0, (MetaName)2168155209)
                    );
                case (MetaName)2367955297:
                    return new PsoStructureInfo((MetaName)2367955297, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.LayerA, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LayerB, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Disabled, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FrequencyNoise, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmplitudeNoise, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Frequency, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Bias, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.AnimatedPostFX:
                    return new PsoStructureInfo(MetaName.AnimatedPostFX, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.ModifierName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AnimMode, PsoDataType.Enum, 12, 0, (MetaName)2476021071),
                     new PsoStructureEntryInfo(MetaName.StartDelayDuration, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InDuration, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HoldDuration, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OutDuration, PsoDataType.UInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Disabled, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LoopMode, PsoDataType.Enum, 36, 0, (MetaName)3696221852),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 40, 0, 0)
                    );
                case (MetaName)1300198536:
                    return new PsoStructureInfo((MetaName)1300198536, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.LensArtefact),
                     new PsoStructureEntryInfo(MetaName.layers, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.LensArtefact:
                    return new PsoStructureInfo(MetaName.LensArtefact, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.blurType, PsoDataType.Enum, 12, 0, (MetaName)3261591327),
                     new PsoStructureEntryInfo(MetaName.scale, PsoDataType.Float2, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.offset, PsoDataType.Float2, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.colorShift, PsoDataType.UInt, 32, 1, 0),
                     new PsoStructureEntryInfo(MetaName.opacity, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2600565536, PsoDataType.Enum, 40, 0, (MetaName)516897642),
                     new PsoStructureEntryInfo((MetaName)1250212400, PsoDataType.Float2, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.streakDirection, PsoDataType.Enum, 52, 0, (MetaName)2665868448),
                     new PsoStructureEntryInfo(MetaName.enabled, PsoDataType.Bool, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.sortIndex, PsoDataType.SInt, 60, 0, 0)
                    );
                case MetaName.CPtFxAssetInfoMgr:
                    return new PsoStructureInfo(MetaName.CPtFxAssetInfoMgr, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.ptfxAssetDependencyInfos, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.CPtFxAssetDependencyInfo:
                    return new PsoStructureInfo(MetaName.CPtFxAssetDependencyInfo, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.parentName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.childName, PsoDataType.String, 12, 7, 0)
                    );
                case MetaName.CVfxFogVolumeInfoMgr:
                    return new PsoStructureInfo(MetaName.CVfxFogVolumeInfoMgr, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.vfxFogVolumeInfos, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.CVfxFogVolumeInfo:
                    return new PsoStructureInfo(MetaName.CVfxFogVolumeInfo, 0, 0, 112,
                     new PsoStructureEntryInfo((MetaName)4135301713, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1768347469, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.position, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rotation, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.scale, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.colR, PsoDataType.UByte, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.colG, PsoDataType.UByte, 65, 0, 0),
                     new PsoStructureEntryInfo(MetaName.colB, PsoDataType.UByte, 66, 0, 0),
                     new PsoStructureEntryInfo(MetaName.colA, PsoDataType.UByte, 67, 0, 0),
                     new PsoStructureEntryInfo(MetaName.hdrMult, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.range, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.density, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.falloff, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.interiorHash, PsoDataType.Long, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.isUnderwater, PsoDataType.Bool, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lightingType, PsoDataType.Enum, 100, 0, (MetaName)2628996780)
                    );
                case MetaName.CVfxInteriorInfoMgr:
                    return new PsoStructureInfo(MetaName.CVfxInteriorInfoMgr, 0, 0, 56,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.vfxInteriorSetups, PsoDataType.Map, 8, 1, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.vfxInteriorInfos, PsoDataType.Map, 32, 1, (MetaName)196612)
                    );
                case MetaName.CVfxInteriorSetup:
                    return new PsoStructureInfo(MetaName.CVfxInteriorSetup, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.interiorInfoName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.vfxRoomSetups, PsoDataType.Map, 16, 1, (MetaName)65538)
                    );
                case MetaName.CVfxRoomSetup:
                    return new PsoStructureInfo(MetaName.CVfxRoomSetup, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.interiorInfoName, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CVfxInteriorInfo:
                    return new PsoStructureInfo(MetaName.CVfxInteriorInfo, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.smokePtFxEnabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.smokePtFxName, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.smokePtFxLevelOverride, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.dustPtFxEnabled, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.dustPtFxName, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.dustPtFxEvo, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CVfxPedInfoMgr:
                    return new PsoStructureInfo(MetaName.CVfxPedInfoMgr, 0, 0, 224,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)1639857157, PsoDataType.Map, 8, 1, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)190681942, PsoDataType.Map, 32, 1, (MetaName)196612),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)866013965, PsoDataType.Map, 56, 1, (MetaName)393223),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)619497315, PsoDataType.Map, 80, 1, (MetaName)589834),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)1719780883, PsoDataType.Map, 104, 1, (MetaName)786445),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)2436902422, PsoDataType.Map, 128, 1, (MetaName)983056),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)670699231, PsoDataType.Map, 152, 1, (MetaName)1179667),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)161646760, PsoDataType.Map, 176, 1, (MetaName)1376278),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.vfxPedInfos, PsoDataType.Map, 200, 1, (MetaName)1572889)
                    );
                case (MetaName)3745427162:
                    return new PsoStructureInfo((MetaName)3745427162, 0, 0, 104,
                     new PsoStructureEntryInfo((MetaName)2328909272, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3928010395, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1667415425, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3934831401, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2811575072, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3485910056, PsoDataType.SInt, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)924601972, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2748666322, PsoDataType.SInt, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2717336085, PsoDataType.SInt, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)383606187, PsoDataType.SInt, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1201928922, PsoDataType.SInt, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2683338794, PsoDataType.SInt, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)244388834, PsoDataType.SInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3603110778, PsoDataType.SInt, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1531021558, PsoDataType.SInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1193506380, PsoDataType.SInt, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1561972343, PsoDataType.SInt, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2527519128, PsoDataType.SInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.decalWidth, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.decalLength, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1369041264, PsoDataType.UByte, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1201263928, PsoDataType.UByte, 89, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1596752989, PsoDataType.UByte, 90, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2768457985, PsoDataType.UByte, 91, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1822646338, PsoDataType.UByte, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2318441304, PsoDataType.UByte, 93, 0, 0),
                     new PsoStructureEntryInfo(MetaName.decalLife, PsoDataType.Float, 96, 0, 0)
                    );
                case (MetaName)1299390333:
                    return new PsoStructureInfo((MetaName)1299390333, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)634726437, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1754146930, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4279228573, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxScale, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)2425920550:
                    return new PsoStructureInfo((MetaName)2425920550, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)3741812782, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2206302043, PsoDataType.String, 12, 7, 0)
                    );
                case (MetaName)1811001251:
                    return new PsoStructureInfo((MetaName)1811001251, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.decalId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3806811108, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3913505542, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3041920342, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2666183704, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1369041264, PsoDataType.UByte, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1201263928, PsoDataType.UByte, 29, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1596752989, PsoDataType.UByte, 30, 0, 0)
                    );
                case (MetaName)1559486096:
                    return new PsoStructureInfo((MetaName)1559486096, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ptFxName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxDepthEvoMin, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxDepthEvoMax, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxSpeedEvoMin, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxSpeedEvoMax, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)4188307664:
                    return new PsoStructureInfo((MetaName)4188307664, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)3741812782, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2206302043, PsoDataType.String, 12, 7, 0)
                    );
                case (MetaName)1073939452:
                    return new PsoStructureInfo((MetaName)1073939452, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.sizeEvo, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.depthMult, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.speedMult, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.widthRatio, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)2217981082:
                    return new PsoStructureInfo((MetaName)2217981082, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.sampleSize, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boneSize, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2990477260, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxEnabled, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxEnabled, PsoDataType.Bool, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxEnabled, PsoDataType.Bool, 19, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxEnabled, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2545172352, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxRange, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxRange, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxRange, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxRange, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1146082049, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CVfxPedInfo:
                    return new PsoStructureInfo(MetaName.CVfxPedInfo, 0, 0, 512,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)231867887),
                     new PsoStructureEntryInfo((MetaName)293516458, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1904273735),
                     new PsoStructureEntryInfo((MetaName)4005470263, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)1938733863, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2108095349, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)898188953, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)815580680, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)728045365, PsoDataType.Bool, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1340027467, PsoDataType.Bool, 57, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1185708987, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3067401171, PsoDataType.Bool, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)267862759, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)481653295, PsoDataType.Bool, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3695423274, PsoDataType.Bool, 73, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1954970508, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1273209835, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1481826432, PsoDataType.Bool, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1210003497, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3343556845, PsoDataType.Bool, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)300193613, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2234775971, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2648212357, PsoDataType.Bool, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2189031336, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1537566823, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2106749959, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)621919678, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1944160314, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)537443923, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1293634249, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1694886060, PsoDataType.Bool, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3575961872, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)423449081, PsoDataType.String, 144, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4063560113, PsoDataType.Bool, 148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1810355587),
                     new PsoStructureEntryInfo((MetaName)1379225165, PsoDataType.Array, 152, 0, (MetaName)34),
                     new PsoStructureEntryInfo((MetaName)2002559884, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo((MetaName)717291862, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo((MetaName)606684208, PsoDataType.Bool, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2292742472, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1437576219, PsoDataType.String, 184, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3754033729, PsoDataType.Enum, 188, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo((MetaName)1020967481, PsoDataType.Enum, 192, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo((MetaName)2592389483, PsoDataType.Bool, 196, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4054422403, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo((MetaName)740137113, PsoDataType.String, 204, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1841095686, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo((MetaName)886073686, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1055815172, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2913384203, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3977079346, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3136327436, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2168759960, PsoDataType.Bool, 232, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2723487263, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1349459576, PsoDataType.String, 240, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2366514623, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo((MetaName)194784077, PsoDataType.Float, 248, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1052325933, PsoDataType.Float, 252, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2774545125, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo((MetaName)431439807, PsoDataType.Bool, 260, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2591692617, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2631634259, PsoDataType.String, 268, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3959299203, PsoDataType.Float, 272, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2821295083, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3995586056, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3353201139, PsoDataType.Float, 284, 0, 0),
                     new PsoStructureEntryInfo((MetaName)412619038, PsoDataType.Bool, 288, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1889989706, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3972116805, PsoDataType.Bool, 296, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4143286255, PsoDataType.String, 300, 7, 0),
                     new PsoStructureEntryInfo((MetaName)30217815, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1079949132, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4224134658, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo((MetaName)901763385, PsoDataType.Bool, 316, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3950900138, PsoDataType.String, 320, 7, 0),
                     new PsoStructureEntryInfo((MetaName)10501246, PsoDataType.Float, 324, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4131076032, PsoDataType.Float, 328, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1327278534, PsoDataType.Float, 332, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxName, PsoDataType.String, 336, 7, 0),
                     new PsoStructureEntryInfo((MetaName)374673880, PsoDataType.Float, 340, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1335117116, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4258146773, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxName, PsoDataType.String, 352, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4020845257, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1518647754, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxName, PsoDataType.String, 364, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4021444798, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2149475580, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxName, PsoDataType.String, 376, 7, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxSpeedEvoMin, PsoDataType.Float, 380, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxSpeedEvoMax, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4096431642, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo((MetaName)415880732, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3919360765, PsoDataType.String, 396, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3723675330, PsoDataType.Float, 400, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3295859425, PsoDataType.Bool, 404, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1659367240, PsoDataType.String, 408, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2238873027, PsoDataType.String, 412, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3354804163, PsoDataType.String, 416, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1800302370, PsoDataType.String, 420, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2345868328, PsoDataType.String, 424, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2719523535, PsoDataType.Float, 428, 0, 0),
                     new PsoStructureEntryInfo((MetaName)460277230, PsoDataType.Float, 432, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4222549442, PsoDataType.Float, 436, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2205812843, PsoDataType.Bool, 440, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2543787987, PsoDataType.Float, 444, 0, 0),
                     new PsoStructureEntryInfo((MetaName)789559375, PsoDataType.String, 448, 7, 0),
                     new PsoStructureEntryInfo((MetaName)610648776, PsoDataType.Enum, 452, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo((MetaName)2321865573, PsoDataType.Bool, 456, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3482316648, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3414327374, PsoDataType.String, 464, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3991249477, PsoDataType.Enum, 468, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo((MetaName)3633780286, PsoDataType.Bool, 472, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3663974917),
                     new PsoStructureEntryInfo((MetaName)1614411627, PsoDataType.Array, 480, 0, (MetaName)113),
                     new PsoStructureEntryInfo((MetaName)2158250314, PsoDataType.Bool, 496, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2801240542, PsoDataType.Float, 500, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1702406422, PsoDataType.String, 504, 7, 0)
                    );
                case (MetaName)231867887:
                    return new PsoStructureInfo((MetaName)231867887, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.vfxGroup, PsoDataType.Enum, 8, 0, MetaName.eVfxGroup),
                     new PsoStructureEntryInfo((MetaName)3557178741, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3535009775, PsoDataType.String, 16, 7, 0)
                    );
                case (MetaName)1904273735:
                    return new PsoStructureInfo((MetaName)1904273735, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)2366957880, PsoDataType.Enum, 8, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo((MetaName)1407284946, PsoDataType.Enum, 12, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo(MetaName.limbId, PsoDataType.Enum, 16, 0, (MetaName)726948528),
                     new PsoStructureEntryInfo((MetaName)1061041502, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2989606868, PsoDataType.String, 24, 7, 0)
                    );
                case (MetaName)1810355587:
                    return new PsoStructureInfo((MetaName)1810355587, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ptFxName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2366957880, PsoDataType.Enum, 12, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo((MetaName)1407284946, PsoDataType.Enum, 16, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo(MetaName.limbId, PsoDataType.SInt, 20, 0, 0)
                    );
                case (MetaName)3663974917:
                    return new PsoStructureInfo((MetaName)3663974917, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ptFxName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2366957880, PsoDataType.Enum, 12, 0, MetaName.eAnimBoneTag),
                     new PsoStructureEntryInfo((MetaName)1407284946, PsoDataType.Enum, 16, 0, MetaName.eAnimBoneTag)
                    );
                case MetaName.CVfxRegionInfoMgr:
                    return new PsoStructureInfo(MetaName.CVfxRegionInfoMgr, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.vfxRegionInfos, PsoDataType.Map, 8, 1, (MetaName)1)
                    );
                case MetaName.CVfxRegionInfo:
                    return new PsoStructureInfo(MetaName.CVfxRegionInfo, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.windDebrisPtFxEnabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.windDebrisPtFxName, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)215011533),
                     new PsoStructureEntryInfo((MetaName)157134441, PsoDataType.Array, 16, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)215011533),
                     new PsoStructureEntryInfo((MetaName)2975635968, PsoDataType.Array, 32, 0, (MetaName)4),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)215011533),
                     new PsoStructureEntryInfo((MetaName)3751751493, PsoDataType.Array, 48, 0, (MetaName)6)
                    );
                case (MetaName)215011533:
                    return new PsoStructureInfo((MetaName)215011533, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.gpuPtFxEnabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)28713766, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3099312282, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3614737265, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2437695423, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1775217766, PsoDataType.SInt, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1202806570, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4204794351, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)634319578, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)552536380, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.CVfxVehicleInfoMgr:
                    return new PsoStructureInfo(MetaName.CVfxVehicleInfoMgr, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.vfxVehicleInfos, PsoDataType.Map, 8, 1, (MetaName)1)
                    );
                case MetaName.CVfxVehicleInfo:
                    return new PsoStructureInfo(MetaName.CVfxVehicleInfo, 0, 0, 880,
                     new PsoStructureEntryInfo(MetaName.mtlBangPtFxVehicleEvo, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.mtlBangPtFxVehicleScale, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.mtlScrapePtFxVehicleEvo, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.mtlScrapePtFxVehicleScale, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxEnabled, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxName, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxCutOffSpeed, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxRange, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxScale, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxSpeedEvoMin, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxSpeedEvoMax, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxTempEvoMin, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxTempEvoMax, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.exhaustPtFxThrottleEvoOnGearChange, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.engineStartupPtFxEnabled, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.engineStartupPtFxName, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo(MetaName.engineStartupPtFxRange, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.misfirePtFxEnabled, PsoDataType.Bool, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.misfirePtFxName, PsoDataType.String, 76, 7, 0),
                     new PsoStructureEntryInfo(MetaName.misfirePtFxRange, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.backfirePtFxEnabled, PsoDataType.Bool, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.backfirePtFxName, PsoDataType.String, 88, 7, 0),
                     new PsoStructureEntryInfo(MetaName.backfirePtFxRange, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxEnabled, PsoDataType.Bool, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxHasPanel, PsoDataType.Bool, 97, 0, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxHasRotorEvo, PsoDataType.Bool, 98, 0, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxNoPanelName, PsoDataType.String, 100, 7, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxPanelOpenName, PsoDataType.String, 104, 7, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxPanelShutName, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxRange, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxSpeedEvoMin, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.engineDamagePtFxSpeedEvoMax, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.overturnedSmokePtFxEnabled, PsoDataType.Bool, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.overturnedSmokePtFxName, PsoDataType.String, 128, 7, 0),
                     new PsoStructureEntryInfo(MetaName.overturnedSmokePtFxRange, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.overturnedSmokePtFxAngleThresh, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.overturnedSmokePtFxSpeedThresh, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.overturnedSmokePtFxEngineHealthThresh, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leakPtFxEnabled, PsoDataType.Bool, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leakPtFxOilName, PsoDataType.String, 152, 7, 0),
                     new PsoStructureEntryInfo(MetaName.leakPtFxPetrolName, PsoDataType.String, 156, 7, 0),
                     new PsoStructureEntryInfo(MetaName.leakPtFxRange, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leakPtFxSpeedEvoMin, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leakPtFxSpeedEvoMax, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelGenericPtFxSet, PsoDataType.SInt, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelGenericDecalSet, PsoDataType.SInt, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelGenericRangeMult, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelSkidmarkRearOnly, PsoDataType.Bool, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelSkidmarkSlipMult, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelSkidmarkPressureMult, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelFrictionPtFxFricMult, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelDisplacementPtFxDispMult, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelBurnoutPtFxFricMult, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelBurnoutPtFxTempMult, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelLowLodPtFxScale, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelPuncturePtFxName, PsoDataType.String, 216, 7, 0),
                     new PsoStructureEntryInfo(MetaName.wheelPuncturePtFxRange, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelBurstPtFxName, PsoDataType.String, 224, 7, 0),
                     new PsoStructureEntryInfo(MetaName.wheelBurstPtFxRange, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelFirePtFxName, PsoDataType.String, 232, 7, 0),
                     new PsoStructureEntryInfo(MetaName.wheelFirePtFxRange, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelFirePtFxSpeedEvoMin, PsoDataType.Float, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wheelFirePtFxSpeedEvoMax, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFirePtFxEnabled, PsoDataType.Bool, 248, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFirePtFxName, PsoDataType.String, 252, 7, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFirePtFxDurationMin, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFirePtFxDurationMax, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFirePtFxRadius, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire2PtFxEnabled, PsoDataType.Bool, 268, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire2PtFxName, PsoDataType.String, 272, 7, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire2PtFxDurationMin, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire2PtFxDurationMax, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire2PtFxRadius, PsoDataType.Float, 284, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire2UseOverheatBone, PsoDataType.Bool, 288, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire2OffsetPos, PsoDataType.Float3, 304, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire3PtFxEnabled, PsoDataType.Bool, 320, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire3PtFxName, PsoDataType.String, 324, 7, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire3PtFxDurationMin, PsoDataType.Float, 328, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire3PtFxDurationMax, PsoDataType.Float, 332, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire3PtFxRadius, PsoDataType.Float, 336, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire3UseOverheatBone, PsoDataType.Bool, 340, 0, 0),
                     new PsoStructureEntryInfo(MetaName.wreckedFire3OffsetPos, PsoDataType.Float3, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.petrolTankFirePtFxName, PsoDataType.String, 368, 7, 0),
                     new PsoStructureEntryInfo(MetaName.petrolTankFirePtFxRange, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.petrolTankFirePtFxSpeedEvoMin, PsoDataType.Float, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.petrolTankFirePtFxSpeedEvoMax, PsoDataType.Float, 380, 0, 0),
                     new PsoStructureEntryInfo(MetaName.petrolTankFirePtFxRadius, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatEntryPtFxEnabled, PsoDataType.Bool, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatEntryPtFxRange, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatEntryPtFxName, PsoDataType.String, 396, 7, 0),
                     new PsoStructureEntryInfo(MetaName.boatEntryPtFxSpeedEvoMin, PsoDataType.Float, 400, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatEntryPtFxSpeedEvoMax, PsoDataType.Float, 404, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatEntryPtFxScale, PsoDataType.Float, 408, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatExitPtFxEnabled, PsoDataType.Bool, 412, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatExitPtFxRange, PsoDataType.Float, 416, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatExitPtFxName, PsoDataType.String, 420, 7, 0),
                     new PsoStructureEntryInfo(MetaName.boatExitPtFxSpeedEvoMin, PsoDataType.Float, 424, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatExitPtFxSpeedEvoMax, PsoDataType.Float, 428, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatExitPtFxScale, PsoDataType.Float, 432, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxEnabled, PsoDataType.Bool, 436, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxRange, PsoDataType.Float, 440, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxForwardName, PsoDataType.String, 444, 7, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxReverseName, PsoDataType.String, 448, 7, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxForwardMountedName, PsoDataType.String, 452, 7, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxForwardMountedOffset, PsoDataType.Float3, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxSpeedEvoMin, PsoDataType.Float, 480, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxSpeedEvoMax, PsoDataType.Float, 484, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxKeelEvoMin, PsoDataType.Float, 488, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxKeelEvoMax, PsoDataType.Float, 492, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxScale, PsoDataType.Float, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatBowPtFxReverseOffset, PsoDataType.Float, 500, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatWashPtFxEnabled, PsoDataType.Bool, 504, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatWashPtFxRange, PsoDataType.Float, 508, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatWashPtFxName, PsoDataType.String, 512, 7, 0),
                     new PsoStructureEntryInfo(MetaName.boatWashPtFxSpeedEvoMin, PsoDataType.Float, 516, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatWashPtFxSpeedEvoMax, PsoDataType.Float, 520, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatWashPtFxScale, PsoDataType.Float, 524, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxEnabled, PsoDataType.Bool, 528, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxRange, PsoDataType.Float, 532, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxName, PsoDataType.String, 536, 7, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxForwardSpeedEvoMin, PsoDataType.Float, 540, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxForwardSpeedEvoMax, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxBackwardSpeedEvoMin, PsoDataType.Float, 548, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxBackwardSpeedEvoMax, PsoDataType.Float, 552, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxDepthEvoMin, PsoDataType.Float, 556, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxDepthEvoMax, PsoDataType.Float, 560, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatPropellerPtFxScale, PsoDataType.Float, 564, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatLowLodWakePtFxEnabled, PsoDataType.Bool, 568, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatLowLodWakePtFxRangeMin, PsoDataType.Float, 572, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatLowLodWakePtFxRangeMax, PsoDataType.Float, 576, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatLowLodWakePtFxName, PsoDataType.String, 580, 7, 0),
                     new PsoStructureEntryInfo(MetaName.boatLowLodWakePtFxSpeedEvoMin, PsoDataType.Float, 584, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatLowLodWakePtFxSpeedEvoMax, PsoDataType.Float, 588, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boatLowLodWakePtFxScale, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeAfterburnerPtFxEnabled, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeAfterburnerPtFxName, PsoDataType.String, 600, 7, 0),
                     new PsoStructureEntryInfo(MetaName.planeAfterburnerPtFxRange, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeAfterburnerPtFxScale, PsoDataType.Float, 608, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeWingTipPtFxEnabled, PsoDataType.Bool, 612, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeWingTipPtFxName, PsoDataType.String, 616, 7, 0),
                     new PsoStructureEntryInfo(MetaName.planeWingTipPtFxRange, PsoDataType.Float, 620, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeWingTipPtFxSpeedEvoMin, PsoDataType.Float, 624, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeWingTipPtFxSpeedEvoMax, PsoDataType.Float, 628, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeDamageFirePtFxEnabled, PsoDataType.Bool, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeDamageFirePtFxName, PsoDataType.String, 636, 7, 0),
                     new PsoStructureEntryInfo(MetaName.planeDamageFirePtFxRange, PsoDataType.Float, 640, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeDamageFirePtFxSpeedEvoMin, PsoDataType.Float, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeDamageFirePtFxSpeedEvoMax, PsoDataType.Float, 648, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxEnabled, PsoDataType.Bool, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxNameDefault, PsoDataType.String, 656, 7, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxNameSand, PsoDataType.String, 660, 7, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxNameDirt, PsoDataType.String, 664, 7, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxNameWater, PsoDataType.String, 668, 7, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxNameFoliage, PsoDataType.String, 672, 7, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxRange, PsoDataType.Float, 676, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxDist, PsoDataType.Float, 680, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxSpeedEvoMin, PsoDataType.Float, 684, 0, 0),
                     new PsoStructureEntryInfo(MetaName.planeGroundDisturbPtFxSpeedEvoMax, PsoDataType.Float, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftSectionDamageSmokePtFxEnabled, PsoDataType.Bool, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftSectionDamageSmokePtFxName, PsoDataType.String, 696, 7, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftSectionDamageSmokePtFxRange, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftSectionDamageSmokePtFxSpeedEvoMin, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftSectionDamageSmokePtFxSpeedEvoMax, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxEnabled, PsoDataType.Bool, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxNameDefault, PsoDataType.String, 716, 7, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxNameSand, PsoDataType.String, 720, 7, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxNameDirt, PsoDataType.String, 724, 7, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxNameWater, PsoDataType.String, 728, 7, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxNameFoliage, PsoDataType.String, 732, 7, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxRange, PsoDataType.Float, 736, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxDist, PsoDataType.Float, 740, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxSpeedEvoMin, PsoDataType.Float, 744, 0, 0),
                     new PsoStructureEntryInfo(MetaName.aircraftDownwashPtFxSpeedEvoMax, PsoDataType.Float, 748, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxEnabled, PsoDataType.Bool, 752, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxRange, PsoDataType.Float, 756, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxName, PsoDataType.String, 760, 7, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxSizeEvoMax, PsoDataType.Float, 764, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxSpeedDownwardThresh, PsoDataType.Float, 768, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxSpeedLateralEvoMin, PsoDataType.Float, 772, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxSpeedLateralEvoMax, PsoDataType.Float, 776, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxSpeedDownwardEvoMin, PsoDataType.Float, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashInPtFxSpeedDownwardEvoMax, PsoDataType.Float, 784, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxEnabled, PsoDataType.Bool, 788, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxRange, PsoDataType.Float, 792, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxName, PsoDataType.String, 796, 7, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxSizeEvoMax, PsoDataType.Float, 800, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxSpeedLateralEvoMin, PsoDataType.Float, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxSpeedLateralEvoMax, PsoDataType.Float, 808, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxSpeedUpwardEvoMin, PsoDataType.Float, 812, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashOutPtFxSpeedUpwardEvoMax, PsoDataType.Float, 816, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxEnabled, PsoDataType.Bool, 820, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxRange, PsoDataType.Float, 824, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxName, PsoDataType.String, 828, 7, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxSizeEvoMax, PsoDataType.Float, 832, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxSpeedVehicleEvoMin, PsoDataType.Float, 836, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxSpeedVehicleEvoMax, PsoDataType.Float, 840, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxSpeedRiverEvoMin, PsoDataType.Float, 844, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashWadePtFxSpeedRiverEvoMax, PsoDataType.Float, 848, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxEnabled, PsoDataType.Bool, 852, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxRange, PsoDataType.Float, 856, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxName, PsoDataType.String, 860, 7, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxSizeEvoMax, PsoDataType.Float, 864, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxSpeedEvoMin, PsoDataType.Float, 868, 0, 0),
                     new PsoStructureEntryInfo(MetaName.splashTrailPtFxSpeedEvoMax, PsoDataType.Float, 872, 0, 0)
                    );
                case MetaName.CVfxWeaponInfoMgr:
                    return new PsoStructureInfo(MetaName.CVfxWeaponInfoMgr, 0, 0, 56,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)3938144758, PsoDataType.Map, 8, 1, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo((MetaName)735539988, PsoDataType.Map, 32, 1, (MetaName)196612)
                    );
                case (MetaName)2452529929:
                    return new PsoStructureInfo((MetaName)2452529929, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)294307720, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.probability, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.scale, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.colTint, PsoDataType.UInt, 20, 1, 0)
                    );
                case (MetaName)1794703310:
                    return new PsoStructureInfo((MetaName)1794703310, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)1914084434, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.probability, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3089927233, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2658317066, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CProceduralInfo:
                    return new PsoStructureInfo(MetaName.CProceduralInfo, 0, 0, 4120,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2143317860),
                     new PsoStructureEntryInfo(MetaName.procObjInfos, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPlantInfo),
                     new PsoStructureEntryInfo(MetaName.plantInfos, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3343541253),
                     new PsoStructureEntryInfo(MetaName.procTagTable, PsoDataType.Array, 40, 4, (MetaName)16711684)
                    );
                case (MetaName)3343541253:
                    return new PsoStructureInfo((MetaName)3343541253, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1599521754),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 8, 2, (MetaName)524288)
                    );
                case (MetaName)2143317860:
                    return new PsoStructureInfo((MetaName)2143317860, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Tag, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PlantTag, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ModelName, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Spacing, PsoDataType.HFloat, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinXRotation, PsoDataType.HFloat, 22, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxXRotation, PsoDataType.HFloat, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinYRotation, PsoDataType.HFloat, 26, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxYRotation, PsoDataType.HFloat, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinZRotation, PsoDataType.HFloat, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZRotation, PsoDataType.HFloat, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinScale, PsoDataType.HFloat, 34, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxScale, PsoDataType.HFloat, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinScaleZ, PsoDataType.HFloat, 38, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxScaleZ, PsoDataType.HFloat, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinZOffset, PsoDataType.HFloat, 42, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZOffset, PsoDataType.HFloat, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistance, PsoDataType.HFloat, 46, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistance, PsoDataType.HFloat, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTintPalette, PsoDataType.UByte, 50, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTintPalette, PsoDataType.UByte, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)956281369),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 52, 2, (MetaName)524308)
                    );
                case MetaName.CPlantInfo:
                    return new PsoStructureInfo(MetaName.CPlantInfo, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Tag, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Color, PsoDataType.UInt, 12, 1, 0),
                     new PsoStructureEntryInfo(MetaName.GroundColor, PsoDataType.UInt, 16, 1, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleXY, PsoDataType.HFloat, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleZ, PsoDataType.HFloat, 22, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleVariationXY, PsoDataType.HFloat, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleVariationZ, PsoDataType.HFloat, 26, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleRangeXYZ, PsoDataType.HFloat, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleRangeZ, PsoDataType.HFloat, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MicroMovementsScaleH, PsoDataType.HFloat, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MicroMovementsScaleV, PsoDataType.HFloat, 34, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MicroMovementsFreqH, PsoDataType.HFloat, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MicroMovementsFreqV, PsoDataType.HFloat, 38, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WindBendScale, PsoDataType.HFloat, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WindBendVariation, PsoDataType.HFloat, 42, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.HFloat, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Density, PsoDataType.HFloat, 46, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DensityRange, PsoDataType.HFloat, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ModelId, PsoDataType.UByte, 50, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TextureId, PsoDataType.UByte, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1247137385),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 52, 1, (MetaName)1048596),
                     new PsoStructureEntryInfo(MetaName.Intensity, PsoDataType.UByte, 54, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IntensityVar, PsoDataType.UByte, 55, 0, 0)
                    );
                case MetaName.camMetadataStore:
                    return new PsoStructureInfo(MetaName.camMetadataStore, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.MetadataList, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.DirectorList, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case MetaName.camDepthOfFieldSettingsMetadata:
                    return new PsoStructureInfo(MetaName.camDepthOfFieldSettingsMetadata, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3872687437, PsoDataType.Float2, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)809440628, PsoDataType.Float2, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)490969727, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3622576788, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1424443550, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3810611648, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1068831663, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1181720025, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2941105367, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1760535059, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)244310625, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2726944087, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)366578954, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1584584282, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4112080817, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4019248724, PsoDataType.Bool, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3170603601, PsoDataType.Bool, 85, 0, 0),
                     new PsoStructureEntryInfo((MetaName)367424479, PsoDataType.Bool, 86, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3748087913, PsoDataType.Bool, 87, 0, 0),
                     new PsoStructureEntryInfo((MetaName)981218312, PsoDataType.Bool, 88, 0, 0)
                    );
                case MetaName.camCinematicAnimatedCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicAnimatedCameraMetadata, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0)
                    );
                case MetaName.camCinematicCameraManCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicCameraManCameraMetadata, 0, 0, 328,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2962759393, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3005507009, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)114561573, PsoDataType.Bool, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1535052963, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3023548272, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)846082931, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3539146194, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3668836939, PsoDataType.UInt, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1699827350, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1265046558, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1161122588, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3190591149, PsoDataType.UInt, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2884448793, PsoDataType.UInt, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)69722150, PsoDataType.Bool, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScanRadius, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScanDistance, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)199666231, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InVehicleLookAtDampingRef, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OnFootLookAtDampingRef, PsoDataType.String, 112, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4200791860, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3308026067, PsoDataType.String, 120, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1279708504, PsoDataType.Structure, 128, 0, MetaName.camCinematicCameraOperatorShakeSettings),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1035890821, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3518053347, PsoDataType.Bool, 272, 0, 0),
                     new PsoStructureEntryInfo((MetaName)568093394, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo((MetaName)781245272, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2596681112, PsoDataType.Bool, 284, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1468609247, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727781899, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3140771983, PsoDataType.Float, 296, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3196312636, PsoDataType.Bool, 300, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3061937367, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1900366054, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4085704227, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3250401979, PsoDataType.Float, 316, 0, 0),
                     new PsoStructureEntryInfo((MetaName)618136263, PsoDataType.Float, 320, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2469116692, PsoDataType.Float, 324, 0, 0)
                    );
                case MetaName.camCinematicCameraOperatorShakeSettings:
                    return new PsoStructureInfo(MetaName.camCinematicCameraOperatorShakeSettings, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.UncertaintySettings, PsoDataType.Structure, 8, 0, MetaName.camCinematicCameraOperatorShakeUncertaintySettings),
                     new PsoStructureEntryInfo(MetaName.TurbulenceSettings, PsoDataType.Structure, 88, 0, MetaName.camCinematicCameraOperatorShakeTurbulenceSettings)
                    );
                case MetaName.camCinematicCameraOperatorShakeUncertaintySettings:
                    return new PsoStructureInfo(MetaName.camCinematicCameraOperatorShakeUncertaintySettings, 0, 0, 80,
                     new PsoStructureEntryInfo((MetaName)784117005, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4083769430, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4138212913, PsoDataType.Float2, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2793659114, PsoDataType.Float2, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2611684493, PsoDataType.Float2, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3850080675, PsoDataType.Float2, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2902509544, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3004590170, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScalingFactor, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2280675285, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)933422390, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3253772488, PsoDataType.Float, 72, 0, 0)
                    );
                case MetaName.camCinematicCameraOperatorShakeTurbulenceSettings:
                    return new PsoStructureInfo(MetaName.camCinematicCameraOperatorShakeTurbulenceSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1741343660, PsoDataType.Float2, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceRange, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2902509544, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3004590170, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camCinematicHeliChaseCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicHeliChaseCameraMetadata, 0, 0, 392,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDuration, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3800346187, PsoDataType.UInt, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2004927681, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)408757490, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1805018741, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1929531797, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1117152888, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3610993313, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2885533396, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1162205528, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2128811877, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3504182592, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2040715445, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1346377414, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)495502725, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3725520797, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)832331471, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)953228499, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)199666231, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1288561522, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4220570274, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)613448395, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)106130617, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3190591149, PsoDataType.UInt, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2884448793, PsoDataType.UInt, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1265046558, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1699827350, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1161122588, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)781245272, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)568093394, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2289962831, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3658577234, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3483998043, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NearClip, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2217923367, PsoDataType.Bool, 168, 0, 0),
                     new PsoStructureEntryInfo((MetaName)114561573, PsoDataType.Bool, 169, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2187445607, PsoDataType.Bool, 170, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1144995443, PsoDataType.Bool, 171, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4185802977, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo((MetaName)315713489, PsoDataType.Bool, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchLimits, PsoDataType.Float2, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InVehicleLookAtDampingRef, PsoDataType.String, 188, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OnFootLookAtDampingRef, PsoDataType.String, 192, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4200791860, PsoDataType.String, 196, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3308026067, PsoDataType.String, 200, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1279708504, PsoDataType.Structure, 208, 0, MetaName.camCinematicCameraOperatorShakeSettings),
                     new PsoStructureEntryInfo((MetaName)983508410, PsoDataType.Float, 336, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3451936358, PsoDataType.Float, 340, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2352608445, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo((MetaName)618136263, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2469116692, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4085704227, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3250401979, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3574069037, PsoDataType.Float, 364, 0, 0),
                     new PsoStructureEntryInfo((MetaName)761184934, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1444706843, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo((MetaName)218641622, PsoDataType.Float, 376, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3061937367, PsoDataType.Float, 380, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1900366054, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1035890821, PsoDataType.Float, 388, 0, 0)
                    );
                case MetaName.camCinematicVehicleOrbitCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleOrbitCameraMetadata, 0, 0, 136,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchLimits, PsoDataType.Float2, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FovLimits, PsoDataType.Float2, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)315073442, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleDelta, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)559738595, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3983210518, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForOcclusionTest, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2153696126, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)765619120, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3644129030, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.UInt, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3991802629, PsoDataType.UInt, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicVehicleOrbitCameraInitalSettings),
                     new PsoStructureEntryInfo((MetaName)1539141183, PsoDataType.Array, 104, 0, (MetaName)20),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 128, 0, 0)
                    );
                case MetaName.camCinematicVehicleOrbitCameraInitalSettings:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleOrbitCameraInitalSettings, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Heading, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingDelta, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camCinematicVehicleLowOrbitCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleLowOrbitCameraMetadata, 0, 0, 256,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchLimits, PsoDataType.Float2, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FovLimits, PsoDataType.Float2, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)315073442, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleDelta, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2037218535, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4219845621, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1088437692, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3983210518, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForOcclusionTest, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)565251953, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2153696126, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)765619120, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.UInt, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicVehicleOrbitCameraInitalSettings),
                     new PsoStructureEntryInfo((MetaName)1539141183, PsoDataType.Array, 112, 0, (MetaName)MetaTypeName.UINT),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1958709726, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2815616753, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2843406703, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4109249654, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo((MetaName)565527739, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3889226535, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3920334627, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3189979615, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1031927866, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2223609026, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo((MetaName)752307799, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1612512934, PsoDataType.Structure, 192, 0, MetaName.camSpeedRelativeShakeSettingsMetadata),
                     new PsoStructureEntryInfo(MetaName.HighSpeedShakeSettings, PsoDataType.Structure, 224, 0, MetaName.camSpeedRelativeShakeSettingsMetadata)
                    );
                case MetaName.camSpeedRelativeShakeSettingsMetadata:
                    return new PsoStructureInfo(MetaName.camSpeedRelativeShakeSettingsMetadata, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardSpeed, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3092190689, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camCinematicIdleCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicIdleCameraMetadata, 0, 0, 168,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1305640980, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)653716069, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)887335192, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WideShot, PsoDataType.Structure, 48, 0, MetaName.camCinematicIdleShots),
                     new PsoStructureEntryInfo(MetaName.MediumShot, PsoDataType.Structure, 72, 0, MetaName.camCinematicIdleShots),
                     new PsoStructureEntryInfo((MetaName)2026925651, PsoDataType.Structure, 96, 0, MetaName.camCinematicIdleShots),
                     new PsoStructureEntryInfo((MetaName)2443187932, PsoDataType.Float2, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3983210518, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForOcclusionTest, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)762567518, PsoDataType.Float2, 148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)348135302, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.Float, 160, 0, 0)
                    );
                case MetaName.camCinematicIdleShots:
                    return new PsoStructureInfo(MetaName.camCinematicIdleShots, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.PitchLimits, PsoDataType.Float2, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceLimits, PsoDataType.Float2, 16, 0, 0)
                    );
                case MetaName.camCinematicFirstPersonIdleCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicFirstPersonIdleCameraMetadata, 0, 0, 200,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.tacticalXml, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2146055434, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)498489070, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)642179285, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3833690476, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)91642876, PsoDataType.UInt, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3423831703, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1365351828, PsoDataType.UInt, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1799833343, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2451425179, PsoDataType.UInt, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2398830068, PsoDataType.UInt, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4013543182, PsoDataType.UInt, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4060544092, PsoDataType.UInt, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1863497759, PsoDataType.UInt, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2865768466, PsoDataType.UInt, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2935498108, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3267214601, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)25343385, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2453008156, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)333391552, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3472598366, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3523922750, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)27172088, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2656885705, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3821389091, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3137663245, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2149837453, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1352549573, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)559353307, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)340096669, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3997020725, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1330235813, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3625015755, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo((MetaName)815735892, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2653018133, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3754548217, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo((MetaName)216085311, PsoDataType.UInt, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)811132412, PsoDataType.UInt, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1561520371, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2384466151, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1800176594, PsoDataType.Float, 192, 0, 0)
                    );
                case MetaName.camCinematicStuntCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicStuntCameraMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FovLimits, PsoDataType.Float2, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3686787728, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2332427685, PsoDataType.Float2, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3195921543, PsoDataType.Float2, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2033229500, PsoDataType.Float2, 60, 0, 0)
                    );
                case MetaName.camCinematicGroupCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicGroupCameraMetadata, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FovLimits, PsoDataType.Float2, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3686787728, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4150576777, PsoDataType.Float2, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.UInt, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3434528744, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2373655501, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)400251000, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1035890821, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1856660360, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 92, 0, 0)
                    );
                case MetaName.camCinematicTrainTrackingCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicTrainTrackingCameraMetadata, 0, 0, 400,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2962759393, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3005507009, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)114561573, PsoDataType.Bool, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1535052963, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3023548272, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)846082931, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3539146194, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3668836939, PsoDataType.UInt, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1699827350, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1265046558, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1161122588, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3190591149, PsoDataType.UInt, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2884448793, PsoDataType.UInt, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)69722150, PsoDataType.Bool, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScanRadius, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScanDistance, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)199666231, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InVehicleLookAtDampingRef, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OnFootLookAtDampingRef, PsoDataType.String, 112, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4200791860, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3308026067, PsoDataType.String, 120, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1279708504, PsoDataType.Structure, 128, 0, MetaName.camCinematicCameraOperatorShakeSettings),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1035890821, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3518053347, PsoDataType.Bool, 272, 0, 0),
                     new PsoStructureEntryInfo((MetaName)568093394, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo((MetaName)781245272, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2596681112, PsoDataType.Bool, 284, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1468609247, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727781899, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3140771983, PsoDataType.Float, 296, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3196312636, PsoDataType.Bool, 300, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3061937367, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1900366054, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4085704227, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3250401979, PsoDataType.Float, 316, 0, 0),
                     new PsoStructureEntryInfo((MetaName)618136263, PsoDataType.Float, 320, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2469116692, PsoDataType.Float, 324, 0, 0),
                     new PsoStructureEntryInfo((MetaName)886237220, PsoDataType.Float3, 336, 0, 0),
                     new PsoStructureEntryInfo((MetaName)649866655, PsoDataType.Float2, 352, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3049827950, PsoDataType.Float2, 360, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1066491626, PsoDataType.Float2, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3832386323, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo((MetaName)412149987, PsoDataType.Float2, 384, 0, 0)
                    );
                case MetaName.camCinematicMountedCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicMountedCameraMetadata, 0, 0, 1344,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MobilePhoneCameraControlHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3464513788, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachSpringConstantEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollBlendEnvelopeRef, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4071354708, PsoDataType.String, 52, 7, 0),
                     new PsoStructureEntryInfo((MetaName)972815273, PsoDataType.String, 56, 7, 0),
                     new PsoStructureEntryInfo((MetaName)844800904, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo((MetaName)648641770, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo((MetaName)722540604, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3400674238, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo((MetaName)885223840, PsoDataType.String, 76, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SpringMountRef, PsoDataType.String, 80, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TripleHeadNearClip, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachPosition, PsoDataType.Float3, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1792533284, PsoDataType.Float3, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BoneRelativeAttachOffset, PsoDataType.Float3, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3162861957, PsoDataType.Float3, 144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2981089011, PsoDataType.Float3, 160, 0, 0),
                     new PsoStructureEntryInfo((MetaName)464677746, PsoDataType.Float3, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2301891370, PsoDataType.Float3, 192, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1452169682, PsoDataType.Float3, 208, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4196591490, PsoDataType.Float3, 224, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1001532487, PsoDataType.Float3, 240, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1862921276, PsoDataType.Float3, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachSpringConstantLimits, PsoDataType.Float2, 272, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachSpringConstantLimitsForPassengers, PsoDataType.Float2, 280, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3248777081, PsoDataType.Float2, 288, 0, 0),
                     new PsoStructureEntryInfo((MetaName)74704022, PsoDataType.Float2, 296, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4088070730, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2782354748, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1994565733, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo((MetaName)383680708, PsoDataType.Float, 316, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2931383793, PsoDataType.Bool, 320, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1919709187, PsoDataType.Float, 324, 0, 0),
                     new PsoStructureEntryInfo((MetaName)945934562, PsoDataType.Float, 328, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultRelativePitch, PsoDataType.Float, 332, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultReversePitch, PsoDataType.Float, 336, 0, 0),
                     new PsoStructureEntryInfo((MetaName)742850166, PsoDataType.Float, 340, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2686299665, PsoDataType.Float2, 344, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1473955583, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo((MetaName)704024974, PsoDataType.Bool, 356, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2135497875, PsoDataType.Bool, 357, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4187867131, PsoDataType.Bool, 358, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2423170684, PsoDataType.Structure, 360, 0, MetaName.camCinematicMountedCameraMetadataRelativePitchScalingToThrottle),
                     new PsoStructureEntryInfo(MetaName.HighSpeedShakeSettings, PsoDataType.Structure, 376, 0, MetaName.camSpeedRelativeShakeSettingsMetadata),
                     new PsoStructureEntryInfo((MetaName)1672349165, PsoDataType.Structure, 408, 0, MetaName.camSpeedRelativeShakeSettingsMetadata),
                     new PsoStructureEntryInfo(MetaName.RocketSettings, PsoDataType.Structure, 440, 0, MetaName.camVehicleRocketSettings),
                     new PsoStructureEntryInfo((MetaName)1913942883, PsoDataType.Bool, 456, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3459873736, PsoDataType.Bool, 457, 0, 0),
                     new PsoStructureEntryInfo((MetaName)880876795, PsoDataType.Bool, 458, 0, 0),
                     new PsoStructureEntryInfo((MetaName)153965986, PsoDataType.Structure, 464, 0, MetaName.camCinematicMountedCameraMetadataLookAroundSettings),
                     new PsoStructureEntryInfo((MetaName)1625111759, PsoDataType.Structure, 504, 0, MetaName.camCinematicMountedCameraMetadataLookAroundSettings),
                     new PsoStructureEntryInfo((MetaName)2701071860, PsoDataType.Structure, 544, 0, MetaName.camCinematicMountedCameraMetadataLookAroundSettings),
                     new PsoStructureEntryInfo((MetaName)1020746469, PsoDataType.Structure, 584, 0, MetaName.camCinematicMountedCameraMetadataFirstPersonRoll),
                     new PsoStructureEntryInfo((MetaName)3266557322, PsoDataType.Structure, 608, 0, MetaName.camCinematicMountedCameraMetadataFirstPersonPitchOffset),
                     new PsoStructureEntryInfo((MetaName)4210630063, PsoDataType.Bool, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachSpringDampingRatio, PsoDataType.Float, 660, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForMaxRelativeAttachSpringConstant, PsoDataType.Float, 664, 0, 0),
                     new PsoStructureEntryInfo((MetaName)17770171, PsoDataType.Float, 668, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2454715147, PsoDataType.Float, 672, 0, 0),
                     new PsoStructureEntryInfo((MetaName)789371927, PsoDataType.Bool, 676, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToFollowPedHead, PsoDataType.Bool, 677, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639714447, PsoDataType.Bool, 678, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToFollowPedSeat, PsoDataType.Bool, 679, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldRestictToFrontSeat, PsoDataType.Bool, 680, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToVehicleExitEntryPoint, PsoDataType.Bool, 681, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToVehicleBone, PsoDataType.Bool, 682, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4227962431, PsoDataType.Bool, 683, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3424726202, PsoDataType.Bool, 684, 0, 0),
                     new PsoStructureEntryInfo((MetaName)883646732, PsoDataType.Bool, 685, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1344883999, PsoDataType.Bool, 686, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)2782414211),
                     new PsoStructureEntryInfo((MetaName)3524553148, PsoDataType.Flags, 688, 0, (MetaName)2097231),
                     new PsoStructureEntryInfo(MetaName.VehicleAttachPart, PsoDataType.Enum, 692, 0, (MetaName)3404480226),
                     new PsoStructureEntryInfo(MetaName.LookAtBehaviour, PsoDataType.Enum, 696, 0, (MetaName)2136742399),
                     new PsoStructureEntryInfo(MetaName.RelativeLookAtPosition, PsoDataType.Float3, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookBehindRelativeAttachPosition, PsoDataType.Float3, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLookBehindCustomPosition, PsoDataType.Bool, 736, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowPedLookAtBoneTag, PsoDataType.SInt, 740, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrientationSpring, PsoDataType.Structure, 744, 0, MetaName.camCinematicMountedCameraMetadataOrientationSpring),
                     new PsoStructureEntryInfo((MetaName)2204528705, PsoDataType.Structure, 808, 0, MetaName.camCinematicMountedCameraMetadataLeadingLookSettings),
                     new PsoStructureEntryInfo(MetaName.MinPitch, PsoDataType.Float, 856, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 860, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3932844561, PsoDataType.Float, 864, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1690462370, PsoDataType.Float, 868, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeHeadingLimitsForMobilePhoneCamera, PsoDataType.Float2, 872, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsBehindVehicleGlass, PsoDataType.Bool, 880, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2385481086, PsoDataType.Bool, 881, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 882, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldMakeFollowPedHeadInvisible, PsoDataType.Bool, 883, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldMakePedInAttachSeatInvisible, PsoDataType.Bool, 884, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldCopyVehicleCameraMotionBlur, PsoDataType.Bool, 885, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LimitAttachParentRelativePitchAndHeading, PsoDataType.Bool, 886, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTerminateForPitchAndHeading, PsoDataType.Bool, 887, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRelativePitch, PsoDataType.Float2, 888, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1537030464, PsoDataType.Float2, 896, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3275096368, PsoDataType.Float2, 904, 0, 0),
                     new PsoStructureEntryInfo((MetaName)923714524, PsoDataType.Float2, 912, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3680927973, PsoDataType.Float2, 920, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRelativeHeading, PsoDataType.Float2, 928, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2362796003, PsoDataType.Float2, 936, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2931436971, PsoDataType.Float2, 944, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4288033461, PsoDataType.Float2, 952, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1148577305, PsoDataType.Float2, 960, 0, 0),
                     new PsoStructureEntryInfo((MetaName)818123698, PsoDataType.Float2, 968, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3197853445, PsoDataType.Float2, 976, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1847367979, PsoDataType.Float2, 984, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1083919137, PsoDataType.Float2, 992, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4130730234, PsoDataType.Float2, 1000, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialRelativePitchLimits, PsoDataType.Float2, 1008, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialRelativeHeadingLimits, PsoDataType.Float2, 1016, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachParentRoll, PsoDataType.Bool, 1024, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTerminateForWorldPitch, PsoDataType.Bool, 1025, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialWorldPitchLimits, PsoDataType.Float2, 1028, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WorldPitchLimits, PsoDataType.Float2, 1036, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTerminateForOcclusion, PsoDataType.Bool, 1044, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.UInt, 1048, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTerminateIfOccludedByAttachParent, PsoDataType.Bool, 1052, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccludedByAttachParent, PsoDataType.UInt, 1056, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 1060, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForOcclusionTest, PsoDataType.Float, 1064, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTerminateForDistanceToTarget, PsoDataType.Bool, 1068, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToTerminate, PsoDataType.Float, 1072, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldCalculateXYDistance, PsoDataType.Bool, 1076, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTestForClipping, PsoDataType.Bool, 1077, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToClipIntoDynamicCollision, PsoDataType.UInt, 1080, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 1084, 0, 0),
                     new PsoStructureEntryInfo((MetaName)816145056, PsoDataType.Bool, 1088, 0, 0),
                     new PsoStructureEntryInfo((MetaName)439797539, PsoDataType.Bool, 1089, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 1092, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 1096, 0, 0),
                     new PsoStructureEntryInfo((MetaName)976391107, PsoDataType.Bool, 1100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 1104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)200247978, PsoDataType.Bool, 1108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1613380386, PsoDataType.Float, 1112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2154681758, PsoDataType.Bool, 1116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InVehicleLookAtDampingRef, PsoDataType.String, 1120, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OnFootLookAtDampingRef, PsoDataType.String, 1124, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4200791860, PsoDataType.String, 1128, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldByPassNearClip, PsoDataType.Bool, 1132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTestForMapPenetrationFromAttachPosition, PsoDataType.Bool, 1133, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseHeading, PsoDataType.Float, 1136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1214392329, PsoDataType.Bool, 1140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2327168839, PsoDataType.Float, 1144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1246262822, PsoDataType.Float, 1148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4216676429, PsoDataType.Float, 1152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2719028802, PsoDataType.Structure, 1160, 0, MetaName.camCinematicMountedCameraMetadataMovementOnAccelerationSettings),
                     new PsoStructureEntryInfo((MetaName)36987752, PsoDataType.Float, 1208, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3434721036, PsoDataType.Float3, 1216, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3664516177, PsoDataType.Float, 1232, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2807019381, PsoDataType.Float, 1236, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1688077651, PsoDataType.Float, 1240, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1499512130, PsoDataType.Float, 1244, 0, 0),
                     new PsoStructureEntryInfo((MetaName)805860083, PsoDataType.Float, 1248, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1793179555, PsoDataType.Float, 1252, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1784170819, PsoDataType.Float, 1256, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2549816384, PsoDataType.Float, 1260, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2482994089, PsoDataType.Float, 1264, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3440187587, PsoDataType.Float, 1268, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2744603068, PsoDataType.Float, 1272, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1387580481, PsoDataType.Float, 1276, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1233327091, PsoDataType.Float, 1280, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3672709376, PsoDataType.Float3, 1296, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2119789086, PsoDataType.Float3, 1312, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2958804846, PsoDataType.Float, 1328, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2107938182, PsoDataType.Float, 1332, 0, 0)
                    );
                case MetaName.camCinematicMountedCameraMetadataRelativePitchScalingToThrottle:
                    return new PsoStructureInfo(MetaName.camCinematicMountedCameraMetadataRelativePitchScalingToThrottle, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)1802668561, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)200823333, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camCinematicMountedCameraMetadataLookAroundSettings:
                    return new PsoStructureInfo(MetaName.camCinematicMountedCameraMetadataLookAroundSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.LeanScale, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)214599438, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2702689596, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2951014073, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2830026870, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3102721723, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)509094775, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3198719034, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camCinematicMountedCameraMetadataFirstPersonRoll:
                    return new PsoStructureInfo(MetaName.camCinematicMountedCameraMetadataFirstPersonRoll, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)455355504, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)694448363, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2004679597, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2782811596, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camCinematicMountedCameraMetadataFirstPersonPitchOffset:
                    return new PsoStructureInfo(MetaName.camCinematicMountedCameraMetadataFirstPersonPitchOffset, 0, 0, 48,
                     new PsoStructureEntryInfo((MetaName)551752221, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxOffset, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1775298710, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2173221242, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)75090556, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.camCinematicMountedCameraMetadataOrientationSpring:
                    return new PsoStructureInfo(MetaName.camCinematicMountedCameraMetadataOrientationSpring, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1464813290, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)12756205, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)967704366, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3344080959, PsoDataType.Float2, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)243857972, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2399868550, PsoDataType.Float2, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2060478858, PsoDataType.Float2, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)131585240, PsoDataType.Float2, 56, 0, 0)
                    );
                case MetaName.camCinematicMountedCameraMetadataLeadingLookSettings:
                    return new PsoStructureInfo(MetaName.camCinematicMountedCameraMetadataLeadingLookSettings, 0, 0, 48,
                     new PsoStructureEntryInfo((MetaName)966483701, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1778767763, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3524239627, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1788878871, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3960905630, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2872708052, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1733937006, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.camCinematicMountedCameraMetadataMovementOnAccelerationSettings:
                    return new PsoStructureInfo(MetaName.camCinematicMountedCameraMetadataMovementOnAccelerationSettings, 0, 0, 48,
                     new PsoStructureEntryInfo((MetaName)3974113312, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendInSpringConstant, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutSpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305119805, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1186910375, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1778767763, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZoomFactor, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3493441660, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1259376027, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.camCinematicPositionCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicPositionCameraMetadata, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchLimits, PsoDataType.Float2, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3983210518, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.UInt, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3715714007, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3434528744, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1699827350, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomDuration, PsoDataType.UInt, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3609378888, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1843629037, PsoDataType.UInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InVehicleLookAtDampingRef, PsoDataType.String, 92, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OnFootLookAtDampingRef, PsoDataType.String, 96, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4200791860, PsoDataType.String, 100, 7, 0)
                    );
                case MetaName.camCinematicWaterCrashCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicWaterCrashCameraMetadata, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchLimits, PsoDataType.Float2, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3106190053, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScale, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DropDistance, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3822313305, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.tacticalXml, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2965889555, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.UInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DropDuration, PsoDataType.UInt, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2920068067, PsoDataType.UInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendDuration, PsoDataType.UInt, 80, 0, 0)
                    );
                case MetaName.camCinematicTwoShotCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicTwoShotCameraMetadata, 0, 0, 136,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeScale, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3072113430, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)654218252, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRelativeVerticalOffset, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1926313598, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1167967077, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)656358172, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)940306022, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4169793326, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistance, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3573176421, PsoDataType.SInt, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLockOn, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1953981041, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdleCmini, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3754548217, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2383824547, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1720877983, PsoDataType.Bool, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1117249844, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3115155794, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1144205032, PsoDataType.Bool, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1450705774, PsoDataType.Float, 132, 0, 0)
                    );
                case MetaName.camCinematicPedCloseUpCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicPedCloseUpCameraMetadata, 0, 0, 192,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachBoneTag, PsoDataType.SInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOffset, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1099610506, PsoDataType.Bool, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3573176421, PsoDataType.SInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookAtOffset, PsoDataType.Float3, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)367186942, PsoDataType.Bool, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3381486527, PsoDataType.UInt, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3185512023, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4280914022, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)430381212, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406519031, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2116722083, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4065237819, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1391390084, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3698583789, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3754548217, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2383824547, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3406809215, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4170206883, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3418665750, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2934396710, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2715105524, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.UInt, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3609378888, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 180, 0, 0)
                    );
                case MetaName.camControlHelperMetadata:
                    return new PsoStructureInfo(MetaName.camControlHelperMetadata, 0, 0, 288,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ViewModes, PsoDataType.Structure, 16, 0, MetaName.camControlHelperMetadataViewModes),
                     new PsoStructureEntryInfo(MetaName.LookAround, PsoDataType.Structure, 128, 0, MetaName.camControlHelperMetadataLookAround),
                     new PsoStructureEntryInfo(MetaName.Zoom, PsoDataType.Structure, 224, 0, MetaName.camControlHelperMetadataZoom),
                     new PsoStructureEntryInfo(MetaName.LookBehindOutroTimeMS, PsoDataType.UInt, 264, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLookBehindInput, PsoDataType.Bool, 268, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseAccurateModeInput, PsoDataType.Bool, 269, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldToggleAccurateModeInput, PsoDataType.Bool, 270, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplySniperControlPref, PsoDataType.Bool, 271, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAimSensitivityPref, PsoDataType.Bool, 272, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimSensitivityScalingLimits, PsoDataType.Float2, 276, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDurationForMultiplayerViewModeActivation, PsoDataType.UInt, 284, 0, 0)
                    );
                case MetaName.camControlHelperMetadataViewModes:
                    return new PsoStructureInfo(MetaName.camControlHelperMetadataViewModes, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.ShouldUseViewModeInput, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldToggleViewModeBetweenThirdAndFirstPerson, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Context, PsoDataType.Enum, 12, 0, (MetaName)513957198),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)2703064232),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 16, 2, (MetaName)327683),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camControlHelperMetadataViewModeSettings),
                     new PsoStructureEntryInfo(MetaName.Settings, PsoDataType.Array, 24, 4, (MetaName)327685),
                     new PsoStructureEntryInfo(MetaName.ViewModeBlendEnvelopeRef, PsoDataType.String, 104, 7, 0)
                    );
                case MetaName.camControlHelperMetadataViewModeSettings:
                    return new PsoStructureInfo(MetaName.camControlHelperMetadataViewModeSettings, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitScaling, PsoDataType.Float2, 8, 0, 0)
                    );
                case MetaName.camControlHelperMetadataLookAround:
                    return new PsoStructureInfo(MetaName.camControlHelperMetadataLookAround, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.InputMagPowerFactor, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Acceleration, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Deceleration, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeadingSpeed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitchSpeed, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MouseMaxHeadingSpeedMin, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MouseMaxHeadingSpeedMax, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MouseMaxPitchSpeedMin, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MouseMaxPitchSpeedMax, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LSDeadZoneAngle, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LSAcceleration, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LSDeceleration, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseGameTime, PsoDataType.Bool, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InputEnvelopeRef, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PrecisionAimSettings, PsoDataType.Structure, 64, 0, MetaName.camControlHelperMetaDataPrecisionAimSettings)
                    );
                case MetaName.camControlHelperMetaDataPrecisionAimSettings:
                    return new PsoStructureInfo(MetaName.camControlHelperMetaDataPrecisionAimSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MinAccelModifier, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAccelModifier, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDeccelModifier, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDeccelModifier, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InputMagToIncreaseDeadZoneMin, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InputMagToIncreaseDeadZoneMax, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.camControlHelperMetadataZoom:
                    return new PsoStructureInfo(MetaName.camControlHelperMetadataZoom, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldUseZoomInput, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDiscreteZoomControl, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseGameTime, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFov, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFovForNetworkPlay, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxFov, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InputMagPowerFactor, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Acceleration, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Deceleration, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeed, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camCinematicVehicleTrackingCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleTrackingCameraMetadata, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookAtOffset, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PositionOffset, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Fov, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1292601136, PsoDataType.Float2, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3554537001, PsoDataType.Float2, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4013358927, PsoDataType.Float2, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)328948722, PsoDataType.Float2, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2198318784, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1142735275, PsoDataType.UInt, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 124, 0, 0)
                    );
                case MetaName.camEnvelopeMetadata:
                    return new PsoStructureInfo(MetaName.camEnvelopeMetadata, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4004354650, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttackDuration, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DecayDuration, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1735638891, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HoldDuration, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReleaseDuration, PsoDataType.UInt, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2014772834, PsoDataType.Enum, 40, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)1004206996, PsoDataType.Enum, 44, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)1484523694, PsoDataType.Bool, 48, 0, 0)
                    );
                case MetaName.camSpringMountMetadata:
                    return new PsoStructureInfo(MetaName.camSpringMountMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AccelerationLimit, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AccelerationForce, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringForce, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DampeningForce, PsoDataType.Float3, 64, 0, 0)
                    );
                case MetaName.camAnimatedCameraMetadata:
                    return new PsoStructureInfo(MetaName.camAnimatedCameraMetadata, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0)
                    );
                case MetaName.camFreeCameraMetadata:
                    return new PsoStructureInfo(MetaName.camFreeCameraMetadata, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartPosition, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForwardAcceleration, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StrafeAcceleration, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerticalAcceleration, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)430698760, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingAcceleration, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchAcceleration, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RollAcceleration, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)991761956, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FovAcceleration, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2276437979, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1380375800, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305680632, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3631492583, PsoDataType.UInt, 100, 0, 0)
                    );
                case MetaName.camCinematicMountedPartCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicMountedPartCameraMetadata, 0, 0, 192,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachSpringConstantEnvelopeRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SpringMountRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPart, PsoDataType.Enum, 48, 0, (MetaName)3404480226),
                     new PsoStructureEntryInfo((MetaName)2797646907, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeLookAtPosition, PsoDataType.Float3, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachSpringConstantLimits, PsoDataType.Float2, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachSpringConstantLimitsForPassengers, PsoDataType.Float2, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForMaxRelativeAttachSpringConstant, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighSpeedShakeSettings, PsoDataType.Structure, 120, 0, MetaName.camSpeedRelativeShakeSettingsMetadata),
                     new PsoStructureEntryInfo(MetaName.IsBehindVehicleGlass, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2385481086, PsoDataType.Bool, 153, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 154, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldMakeFollowPedHeadInvisible, PsoDataType.Bool, 155, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldCopyVehicleCameraMotionBlur, PsoDataType.Bool, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendOccluded, PsoDataType.UInt, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRadius, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForOcclusionTest, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTestForClipping, PsoDataType.Bool, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusScalingForClippingTest, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForWaterClippingTest, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForRiverWaterClippingTest, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 188, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadata:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadata, 0, 0, 2288,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpHelperRef, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseAttachVelocityToIgnoreEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305154880, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldOrbitRelativeToAttachParentOrientation, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistOrbitOrientationRelativeToAttachParent, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringConstant, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringDampingRatio, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406229624, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727876548, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSubmergedLevelToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftGroundSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftContactSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldTimeToBlockFullAttachParentMatrixForRelativeOrbit, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSpeedToClonePitchFromCinematicMountedCameras, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomFramingInTightSpace, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachSpeedToUpdateTightSpaceLevel, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringConstant, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsInTightSpace, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToParentCentreOfGravity, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDynamicCentreOfGravity, PsoDataType.Bool, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVelocityOfAttachParentAttachEntity, PsoDataType.Bool, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomBoundingBoxSettings, PsoDataType.Structure, 128, 0, MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachPedPelvisOffset, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotPosition, PsoDataType.Structure, 176, 0, MetaName.camThirdPersonCameraMetadataBasePivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotPosition, PsoDataType.Structure, 272, 0, MetaName.camThirdPersonCameraMetadataPivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotOverBoundingBoxSettings, PsoDataType.Structure, 320, 0, MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotHeightScalingForFootRoom, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVerticalPivotOffsetForFootRoom, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeOrbitDistanceScalingForExtensions, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)560700931, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomOrbitDistanceLimitsToForce, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringConstant, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringDampingRatio, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingForCustomFirstPersonFallBack, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 400, 0, MetaName.camThirdPersonCameraMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithAttachParent, PsoDataType.Bool, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithFollowVehicle, PsoDataType.Bool, 449, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionOrigin, PsoDataType.Bool, 450, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionRoot, PsoDataType.Bool, 451, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringConstant, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringDampingRatio, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomCollisionOriginRelativePosition, PsoDataType.Float3, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4194152221, PsoDataType.Float3, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomCollisionOrigin, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1998689777, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235540893, PsoDataType.Bool, 514, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3479639237, PsoDataType.Float3, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496115250, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionFallBackPosition, PsoDataType.Structure, 552, 0, MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionFallBackToPivotBlendValue, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConstrainCollisionRootPositionAgainstClippingTypes, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringConstant, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringDampingRatio, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOnTopOfVehicleCollisionSettings, PsoDataType.Structure, 608, 0, MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.IdealHeadingOffsetForLimiting, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffsetInTightSpace, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookOverSettings, PsoDataType.Structure, 664, 0, MetaName.camThirdPersonCameraMetadataLookOverSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentPitchForLookBehind, PsoDataType.Bool, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitsForBasePosition, PsoDataType.Float2, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreToPostCollisionLookAtOrientationBlendValue, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringConstant, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringDampingRatio, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchSoftLimits, PsoDataType.Float2, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchHardLimits, PsoDataType.Float2, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthZoomSettings, PsoDataType.Structure, 728, 0, MetaName.camThirdPersonCameraMetadataStealthZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1947299079, PsoDataType.Structure, 752, 0, MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 776, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2122226771, PsoDataType.Bool, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentInAirEnvelopeRef, PsoDataType.String, 784, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentUpwardSpeedScalingOnGroundEnvelopeRef, PsoDataType.String, 788, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentUpwardSpeedScalingInAirEnvelopeRef, PsoDataType.String, 792, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AimBehaviourEnvelopeRef, PsoDataType.String, 796, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WaterBobShakeRef, PsoDataType.String, 800, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentMovementForOrientation, PsoDataType.Bool, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PullAroundSettings, PsoDataType.Structure, 808, 0, MetaName.camFollowCameraMetadataPullAroundSettings),
                     new PsoStructureEntryInfo(MetaName.PullAroundSettingsForLookBehind, PsoDataType.Structure, 872, 0, MetaName.camFollowCameraMetadataPullAroundSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderAttachParentLocalXYVelocityForPullAround, PsoDataType.Bool, 936, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderAttachParentForwardSpeedForPullAround, PsoDataType.Bool, 937, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundToAttachParentFront, PsoDataType.Bool, 938, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundToBasicAttachParentMatrix, PsoDataType.Bool, 939, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundUsingSimpleSpringDamping, PsoDataType.Bool, 940, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2671392542, PsoDataType.Bool, 941, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachParentApproachSpeedForPitchLock, PsoDataType.Float, 944, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockHeading, PsoDataType.Bool, 948, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMoveSpeedForFollowOrientation, PsoDataType.Float, 952, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLookAroundMoveSpeedMultiplier, PsoDataType.Float, 956, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpeedLimitsForVerticalMoveSpeedScaling, PsoDataType.Float2, 960, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerticalMoveSpeedScaling, PsoDataType.Float, 968, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerticalMoveSpeedScalingAtMaxSpeed, PsoDataType.Float, 972, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpwardMoveSpeedScalingOnGround, PsoDataType.Float, 976, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpwardMoveSpeedScalingInAir, PsoDataType.Float, 980, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMoveOrientationSpeedDuringLookAround, PsoDataType.Float, 984, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowOrientationConing, PsoDataType.Structure, 992, 0, MetaName.camFollowCameraMetadataFollowOrientationConing),
                     new PsoStructureEntryInfo(MetaName.RollSettings, PsoDataType.Structure, 1024, 0, MetaName.camFollowCameraMetadataRollSettings),
                     new PsoStructureEntryInfo(MetaName.HighAltitudeZoomSettings, PsoDataType.Structure, 1064, 0, MetaName.camFollowCameraMetadataHighAltitudeZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1885083576, PsoDataType.Bool, 1104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1213389792, PsoDataType.Bool, 1105, 0, 0),
                     new PsoStructureEntryInfo((MetaName)473739279, PsoDataType.Float, 1108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollBlendEnvelopeRef, PsoDataType.String, 1112, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3575968142, PsoDataType.String, 1116, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2291862537, PsoDataType.UInt, 1120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)459469516, PsoDataType.Float, 1124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4093180426, PsoDataType.Float, 1128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1028171116, PsoDataType.Float, 1132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2578221071, PsoDataType.Float2, 1136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2418079526, PsoDataType.String, 1144, 7, 0),
                     new PsoStructureEntryInfo((MetaName)300074581, PsoDataType.Bool, 1148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camFollowPedCameraMetadataCustomViewModeSettings),
                     new PsoStructureEntryInfo((MetaName)95464861, PsoDataType.Array, 1152, 4, (MetaName)327809),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camFollowPedCameraMetadataCustomViewModeSettings),
                     new PsoStructureEntryInfo((MetaName)741352357, PsoDataType.Array, 1272, 4, (MetaName)327811),
                     new PsoStructureEntryInfo((MetaName)1557873238, PsoDataType.Bool, 1392, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camFollowPedCameraMetadataCustomViewModeSettings),
                     new PsoStructureEntryInfo((MetaName)1607272001, PsoDataType.Array, 1400, 4, (MetaName)327814),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camFollowPedCameraMetadataCustomViewModeSettings),
                     new PsoStructureEntryInfo((MetaName)467721499, PsoDataType.Array, 1520, 4, (MetaName)327816),
                     new PsoStructureEntryInfo((MetaName)4261252982, PsoDataType.String, 1640, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2397521015, PsoDataType.Float, 1644, 0, 0),
                     new PsoStructureEntryInfo((MetaName)886249434, PsoDataType.Float, 1648, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3909973356, PsoDataType.Float, 1652, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4282038076, PsoDataType.Float, 1656, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1414513338, PsoDataType.Float, 1660, 0, 0),
                     new PsoStructureEntryInfo((MetaName)233447402, PsoDataType.Float, 1664, 0, 0),
                     new PsoStructureEntryInfo((MetaName)813655062, PsoDataType.Float, 1668, 0, 0),
                     new PsoStructureEntryInfo((MetaName)859959795, PsoDataType.Float, 1672, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2322941924, PsoDataType.Structure, 1680, 0, MetaName.camFollowPedCameraMetadataAssistedMovementAlignment),
                     new PsoStructureEntryInfo(MetaName.LadderAlignment, PsoDataType.Structure, 1768, 0, MetaName.camFollowPedCameraMetadataLadderAlignment),
                     new PsoStructureEntryInfo(MetaName.RappellingAlignment, PsoDataType.Structure, 1952, 0, MetaName.camFollowPedCameraMetadataRappellingAlignment),
                     new PsoStructureEntryInfo((MetaName)215880757, PsoDataType.Bool, 1968, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4161177976, PsoDataType.Structure, 1976, 0, MetaName.camFollowCameraMetadataPullAroundSettings),
                     new PsoStructureEntryInfo((MetaName)2738607365, PsoDataType.Float2, 2040, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3410646358, PsoDataType.Float2, 2048, 0, 0),
                     new PsoStructureEntryInfo((MetaName)528803049, PsoDataType.Bool, 2056, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3857904181, PsoDataType.Float, 2060, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2773176002, PsoDataType.Float, 2064, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3592715839, PsoDataType.Bool, 2068, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3895855693, PsoDataType.Float2, 2072, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2344836057, PsoDataType.Structure, 2080, 0, MetaName.camFollowPedCameraMetadataOrbitPitchLimitsForOverheadCollision),
                     new PsoStructureEntryInfo((MetaName)3095276846, PsoDataType.Float, 2112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleEntryExitPitchLevelSmoothRate, PsoDataType.Float, 2116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2068541596, PsoDataType.Float, 2120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1699579049, PsoDataType.Structure, 2128, 0, MetaName.camFollowPedCameraMetadataRunningShakeSettings),
                     new PsoStructureEntryInfo((MetaName)498572221, PsoDataType.Structure, 2160, 0, MetaName.camFollowPedCameraMetadataSwimmingShakeSettings),
                     new PsoStructureEntryInfo((MetaName)2385645301, PsoDataType.Structure, 2192, 0, MetaName.camFollowPedCameraMetadataDivingShakeSettings),
                     new PsoStructureEntryInfo((MetaName)136635268, PsoDataType.Structure, 2224, 0, MetaName.camFollowPedCameraMetadataHighFallShakeSettings),
                     new PsoStructureEntryInfo((MetaName)41370383, PsoDataType.Structure, 2248, 0, MetaName.camFollowPedCameraMetadataPushBeyondNearbyVehiclesInRagdollSettings)
                    );
                case MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.HeightScaling, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxExtraHeightForVehicleTrailers, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxExtraHeightForTowedVehicles, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveVehicleDriverSeat, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataBasePivotPosition:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataBasePivotPosition, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.ShouldUseBaseAttachPosition, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentHeightRatioToAttain, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentHeightRatioToAttainInTightSpace, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyInAttachParentLocalSpace, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RollSettings, PsoDataType.Structure, 24, 0, MetaName.camThirdPersonCameraMetadataBasePivotPositionRollSettings),
                     new PsoStructureEntryInfo(MetaName.RelativeOffset, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockVerticalOffset, PsoDataType.Bool, 80, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataBasePivotPositionRollSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataBasePivotPositionRollSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachParentRoll, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardSpeed, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleScalingFactor, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngle, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataPivotPosition:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataPivotPosition, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.CameraRelativeSideOffset, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRelativeVerticalOffset, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentWidthScalingForCameraRelativeSideOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentHeightScalingForCameraRelativeVerticalOffset, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.timeAfterAimingToApplyAlternateScalingMin, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.timeAfterAimingToApplyAlternateScalingMax, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxDistanceAfterAimingToApplyAlternateScalingMax, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.timeAfterAimingToApplyDistanceBlend, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentHeightScalingForCameraRelativeVerticalOffset_AfterAiming, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2888330455, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.BlendLevel, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraCameraRelativeVerticalOffset, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentHeightScalingForExtraCameraRelativeVerticalOffset, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataBuoyancySettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataBuoyancySettings, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.ShouldApplyBuoyancy, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayBetweenBuoyancyStateChanges, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayOnSubmerging, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayOnSurfacing, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeSpentSwimmingToRespectMotionTask, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentDepthUnderWaterToRemainBuoyant, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentDepthUnderWaterToRemainBuoyantOut, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionFallBackBlendLevelToForceUnderWater, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldSetBuoyantWhenAttachParentNotFullySubmerged, PsoDataType.Bool, 40, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.AttachParentHeightRatioToAttain, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachParentHeightRatioToPushAwayFromCollision, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4004666528, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendLevelAfterPushAwayFromCollision, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)31648245, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2702477980, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)20195593, PsoDataType.Bool, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyInAttachParentLocalSpace, PsoDataType.Bool, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ShouldApply, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDurationToTrackVehicles, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToTestDownForVehiclesToReject, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataLookOverSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataLookOverSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.MinHeight, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeight, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchOffsetAtMinHeight, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchOffsetAtMaxHeight, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataStealthZoomSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataStealthZoomSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ShouldApply, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZoomFactor, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camFollowCameraMetadataPullAroundSettings:
                    return new PsoStructureInfo(MetaName.camFollowCameraMetadataPullAroundSettings, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.ShouldBlendOutWhenAttachParentIsInAir, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldBlendOutWhenAttachParentIsOnGround, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldBlendWithAttachParentMatrixForRelativeOrbitBlend, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingPullAroundMinMoveSpeed, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingPullAroundMaxMoveSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingPullAroundSpeedAtMaxMoveSpeed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingPullAroundErrorScalingBlendLevel, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingPullAroundSpringConstant, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingPullAroundSpringDampingRatio, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchPullAroundMinMoveSpeed, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchPullAroundMaxMoveSpeed, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchPullAroundSpeedAtMaxMoveSpeed, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchPullAroundErrorScalingBlendLevel, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchPullAroundSpringConstant, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchPullAroundSpringDampingRatio, PsoDataType.Float, 56, 0, 0)
                    );
                case MetaName.camFollowCameraMetadataFollowOrientationConing:
                    return new PsoStructureInfo(MetaName.camFollowCameraMetadataFollowOrientationConing, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MaxAngle, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AspectRatio, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchOffset, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SmoothRate, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camFollowCameraMetadataRollSettings:
                    return new PsoStructureInfo(MetaName.camFollowCameraMetadataRollSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldApplyRoll, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RollSpringConstant, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RollSpringDampRatio, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardSpeed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardSpeed, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RollAngleScale, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camFollowCameraMetadataHighAltitudeZoomSettings:
                    return new PsoStructureInfo(MetaName.camFollowCameraMetadataHighAltitudeZoomSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MinAltitudeDelta, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAltitudeDelta, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBaseFovScaling, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadataCustomViewModeSettings:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataCustomViewModeSettings, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)4026985674, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadataAssistedMovementAlignment:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataAssistedMovementAlignment, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.ShouldAlign, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2541847736, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3143901835, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2602308947, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PullAroundSettings, PsoDataType.Structure, 24, 0, MetaName.camFollowCameraMetadataPullAroundSettings)
                    );
                case MetaName.camFollowPedCameraMetadataLadderAlignment:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataLadderAlignment, 0, 0, 184,
                     new PsoStructureEntryInfo(MetaName.ShouldAlign, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2541847736, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)261180327, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerticalMoveSpeedScaling, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)666447554, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4131471736, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3980517625, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PullAroundSettings, PsoDataType.Structure, 56, 0, MetaName.camFollowCameraMetadataPullAroundSettings),
                     new PsoStructureEntryInfo((MetaName)738690141, PsoDataType.Structure, 120, 0, MetaName.camFollowCameraMetadataPullAroundSettings)
                    );
                case MetaName.camFollowPedCameraMetadataRappellingAlignment:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataRappellingAlignment, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.ShouldAlign, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)261180327, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadataOrbitPitchLimitsForOverheadCollision:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataOrbitPitchLimitsForOverheadCollision, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)1746008730, PsoDataType.Float2, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)256117257, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadataRunningShakeSettings:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataRunningShakeSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinAmplitude, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAmplitude, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)491218034, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2700987639, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadataSwimmingShakeSettings:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataSwimmingShakeSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinAmplitude, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAmplitude, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)491218034, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1443947683, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2015392943, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadataDivingShakeSettings:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataDivingShakeSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeight, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeight, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAmplitude, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAmplitude, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadataHighFallShakeSettings:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataHighFallShakeSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeed, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAmplitude, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camFollowPedCameraMetadataPushBeyondNearbyVehiclesInRagdollSettings:
                    return new PsoStructureInfo(MetaName.camFollowPedCameraMetadataPushBeyondNearbyVehiclesInRagdollSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.MaxDurationToTrackVehicles, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1159487467, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2913317904, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)752814960, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1115262693, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DetectionRadius, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToTestDownForVehiclesToReject, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camCollisionMetadata:
                    return new PsoStructureInfo(MetaName.camCollisionMetadata, 0, 0, 240,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreOcclusionWithBrokenFragments, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreOcclusionWithBrokenFragmentsOfIgnoredEntities, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldMoveTowardsLos, PsoDataType.Bool, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldSweepToAvoidPopIn, PsoDataType.Bool, 19, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistPopInBehaviour, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullBackByCapsuleRadius, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreOcclusionWithSelectCollision, PsoDataType.Bool, 22, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreOcclusionWithRagdolls, PsoDataType.Bool, 23, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldReportAsCameraTypeTest, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OcclusionSweep, PsoDataType.Structure, 32, 0, MetaName.camCollisionMetadataOcclusionSweep),
                     new PsoStructureEntryInfo(MetaName.PathFinding, PsoDataType.Structure, 72, 0, MetaName.camCollisionMetadataPathFinding),
                     new PsoStructureEntryInfo(MetaName.RotationTowardsLos, PsoDataType.Structure, 88, 0, MetaName.camCollisionMetadataRotationTowardsLos),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceDamping, PsoDataType.Structure, 104, 0, MetaName.camCollisionMetadataOrbitDistanceDamping),
                     new PsoStructureEntryInfo(MetaName.ClippingAvoidance, PsoDataType.Structure, 136, 0, MetaName.camCollisionMetadataClippingAvoidance),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 152, 0, MetaName.camCollisionMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.PushBeyondEntitiesIfClipping, PsoDataType.Structure, 184, 0, MetaName.camCollisionMetadataPushBeyondEntitiesIfClipping),
                     new PsoStructureEntryInfo(MetaName.PullBackTowardsCollision, PsoDataType.Structure, 216, 0, MetaName.camCollisionMetadataPullBackTowardsCollision)
                    );
                case MetaName.camCollisionMetadataOcclusionSweep:
                    return new PsoStructureInfo(MetaName.camCollisionMetadataOcclusionSweep, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.NumCapsuleTests, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionRootSpeedToForcePopIn, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPreCollisionCameraSpeedToForcePopIn, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSweepAngleWhenMovingTowardsLos, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSweepAngleWhenAvoidingPopIn, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinOrientationSpeedToMaintainDirection, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinCameraMoveSpeedToSweepInDirectionOfTravel, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camCollisionMetadataPathFinding:
                    return new PsoStructureInfo(MetaName.camCollisionMetadataPathFinding, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MaxCapsuleTests, PsoDataType.UInt, 8, 0, 0)
                    );
                case MetaName.camCollisionMetadataRotationTowardsLos:
                    return new PsoStructureInfo(MetaName.camCollisionMetadataRotationTowardsLos, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camCollisionMetadataOrbitDistanceDamping:
                    return new PsoStructureInfo(MetaName.camCollisionMetadataOrbitDistanceDamping, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MaxCollisionRootSpeedToPausePullBack, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPreCollisionCameraSpeedToPausePullBack, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceErrorToIgnore, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camCollisionMetadataClippingAvoidance:
                    return new PsoStructureInfo(MetaName.camCollisionMetadataClippingAvoidance, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MaxIterations, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CapsuleLengthForDetection, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camCollisionMetadataBuoyancySettings:
                    return new PsoStructureInfo(MetaName.camCollisionMetadataBuoyancySettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShouldApplyBuoyancy, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreBuoyancyStateAndAvoidSurface, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHitNormalDotWorldUpForRivers, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WaterHeightSmoothRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWaterWhenBuoyant, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDepthUnderWaterWhenNotBuoyant, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camCollisionMetadataPushBeyondEntitiesIfClipping:
                    return new PsoStructureInfo(MetaName.camCollisionMetadataPushBeyondEntitiesIfClipping, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ExtraDistanceToPushAway, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingToApplyWhenPushing, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PullBackSpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PushInSpringConstant, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAllowOtherCollisionToConstrainCameraIntoEntities, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)925224069, PsoDataType.Bool, 29, 0, 0),
                     new PsoStructureEntryInfo((MetaName)631999118, PsoDataType.Bool, 30, 0, 0)
                    );
                case MetaName.camCollisionMetadataPullBackTowardsCollision:
                    return new PsoStructureInfo(MetaName.camCollisionMetadataPullBackTowardsCollision, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.BlendInSpringConstant, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutSpringConstant, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.camHintHelperMetadata:
                    return new PsoStructureInfo(MetaName.camHintHelperMetadata, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FovScalar, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1674069994, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)67329582, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2753181554, PsoDataType.Structure, 40, 0, MetaName.camHintHelperMetadataPivotPositionAdditive),
                     new PsoStructureEntryInfo((MetaName)3051761482, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo((MetaName)399304638, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo((MetaName)315713489, PsoDataType.Bool, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3421755121, PsoDataType.Bool, 73, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLockOn, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1167921072, PsoDataType.Bool, 80, 0, 0)
                    );
                case MetaName.camHintHelperMetadataPivotPositionAdditive:
                    return new PsoStructureInfo(MetaName.camHintHelperMetadataPivotPositionAdditive, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)3517580346, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2171374488, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3803694830, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)187223663, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camInconsistentBehaviourZoomHelperMetadata:
                    return new PsoStructureInfo(MetaName.camInconsistentBehaviourZoomHelperMetadata, 0, 0, 184,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2835430912, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxFov, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)192388977, PsoDataType.Structure, 24, 0, MetaName.camInconsistentBehaviourZoomHelperDetectSuddenMovementSettings),
                     new PsoStructureEntryInfo((MetaName)3137016932, PsoDataType.Structure, 64, 0, MetaName.camInconsistentBehaviourZoomHelperDetectFastCameraTurnSettings),
                     new PsoStructureEntryInfo((MetaName)3922359176, PsoDataType.Structure, 104, 0, MetaName.camInconsistentBehaviourZoomHelperAirborneSettings),
                     new PsoStructureEntryInfo((MetaName)3604184077, PsoDataType.Structure, 144, 0, MetaName.camInconsistentBehaviourZoomHelperLosSettings)
                    );
                case MetaName.camInconsistentBehaviourZoomHelperDetectSuddenMovementSettings:
                    return new PsoStructureInfo(MetaName.camInconsistentBehaviourZoomHelperDetectSuddenMovementSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldDetect, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReactionTime, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)989577416, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2226168954, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2667621860, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2802018341, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camInconsistentBehaviourZoomHelperDetectFastCameraTurnSettings:
                    return new PsoStructureInfo(MetaName.camInconsistentBehaviourZoomHelperDetectFastCameraTurnSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldDetect, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReactionTime, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)989577416, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2226168954, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2667621860, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1513989445, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)423008264, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camInconsistentBehaviourZoomHelperAirborneSettings:
                    return new PsoStructureInfo(MetaName.camInconsistentBehaviourZoomHelperAirborneSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldDetect, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReactionTime, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)989577416, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2226168954, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2667621860, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)424844475, PsoDataType.UInt, 32, 0, 0)
                    );
                case MetaName.camInconsistentBehaviourZoomHelperLosSettings:
                    return new PsoStructureInfo(MetaName.camInconsistentBehaviourZoomHelperLosSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldDetect, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReactionTime, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)989577416, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2226168954, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2667621860, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1918423771, PsoDataType.UInt, 32, 0, 0)
                    );
                case MetaName.camCatchUpHelperMetadata:
                    return new PsoStructureInfo(MetaName.camCatchUpHelperMetadata, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)482340760, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1046944473, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendDuration, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3764337155, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2836572250, PsoDataType.Enum, 32, 0, MetaName.eCurveType)
                    );
                case MetaName.camLookAtDampingHelperMetadata:
                    return new PsoStructureInfo(MetaName.camLookAtDampingHelperMetadata, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3543584583, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2722511334, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3076282416, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3481945794, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1705112434, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)627718911, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3152826625, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3880553368, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1561096582, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1150311187, PsoDataType.Float, 52, 0, 0)
                    );
                case MetaName.camLookAheadHelperMetadata:
                    return new PsoStructureInfo(MetaName.camLookAheadHelperMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)765307181, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)988110708, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeed, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3931376905, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1623903878, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardSpeed, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3766959106, PsoDataType.Bool, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)88070689, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1846985178, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2004585270, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 60, 0, 0)
                    );
                case MetaName.camFollowVehicleCameraMetadata:
                    return new PsoStructureInfo(MetaName.camFollowVehicleCameraMetadata, 0, 0, 1376,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpHelperRef, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseAttachVelocityToIgnoreEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305154880, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldOrbitRelativeToAttachParentOrientation, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistOrbitOrientationRelativeToAttachParent, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringConstant, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringDampingRatio, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406229624, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727876548, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSubmergedLevelToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftGroundSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftContactSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldTimeToBlockFullAttachParentMatrixForRelativeOrbit, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSpeedToClonePitchFromCinematicMountedCameras, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomFramingInTightSpace, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachSpeedToUpdateTightSpaceLevel, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringConstant, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsInTightSpace, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToParentCentreOfGravity, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDynamicCentreOfGravity, PsoDataType.Bool, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVelocityOfAttachParentAttachEntity, PsoDataType.Bool, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomBoundingBoxSettings, PsoDataType.Structure, 128, 0, MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachPedPelvisOffset, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotPosition, PsoDataType.Structure, 176, 0, MetaName.camThirdPersonCameraMetadataBasePivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotPosition, PsoDataType.Structure, 272, 0, MetaName.camThirdPersonCameraMetadataPivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotOverBoundingBoxSettings, PsoDataType.Structure, 320, 0, MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotHeightScalingForFootRoom, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVerticalPivotOffsetForFootRoom, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeOrbitDistanceScalingForExtensions, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)560700931, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomOrbitDistanceLimitsToForce, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringConstant, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringDampingRatio, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingForCustomFirstPersonFallBack, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 400, 0, MetaName.camThirdPersonCameraMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithAttachParent, PsoDataType.Bool, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithFollowVehicle, PsoDataType.Bool, 449, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionOrigin, PsoDataType.Bool, 450, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionRoot, PsoDataType.Bool, 451, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringConstant, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringDampingRatio, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomCollisionOriginRelativePosition, PsoDataType.Float3, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4194152221, PsoDataType.Float3, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomCollisionOrigin, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1998689777, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235540893, PsoDataType.Bool, 514, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3479639237, PsoDataType.Float3, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496115250, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionFallBackPosition, PsoDataType.Structure, 552, 0, MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionFallBackToPivotBlendValue, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConstrainCollisionRootPositionAgainstClippingTypes, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringConstant, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringDampingRatio, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOnTopOfVehicleCollisionSettings, PsoDataType.Structure, 608, 0, MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.IdealHeadingOffsetForLimiting, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffsetInTightSpace, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookOverSettings, PsoDataType.Structure, 664, 0, MetaName.camThirdPersonCameraMetadataLookOverSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentPitchForLookBehind, PsoDataType.Bool, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitsForBasePosition, PsoDataType.Float2, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreToPostCollisionLookAtOrientationBlendValue, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringConstant, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringDampingRatio, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchSoftLimits, PsoDataType.Float2, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchHardLimits, PsoDataType.Float2, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthZoomSettings, PsoDataType.Structure, 728, 0, MetaName.camThirdPersonCameraMetadataStealthZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1947299079, PsoDataType.Structure, 752, 0, MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 776, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2122226771, PsoDataType.Bool, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentInAirEnvelopeRef, PsoDataType.String, 784, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentUpwardSpeedScalingOnGroundEnvelopeRef, PsoDataType.String, 788, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentUpwardSpeedScalingInAirEnvelopeRef, PsoDataType.String, 792, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AimBehaviourEnvelopeRef, PsoDataType.String, 796, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WaterBobShakeRef, PsoDataType.String, 800, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentMovementForOrientation, PsoDataType.Bool, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PullAroundSettings, PsoDataType.Structure, 808, 0, MetaName.camFollowCameraMetadataPullAroundSettings),
                     new PsoStructureEntryInfo(MetaName.PullAroundSettingsForLookBehind, PsoDataType.Structure, 872, 0, MetaName.camFollowCameraMetadataPullAroundSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderAttachParentLocalXYVelocityForPullAround, PsoDataType.Bool, 936, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderAttachParentForwardSpeedForPullAround, PsoDataType.Bool, 937, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundToAttachParentFront, PsoDataType.Bool, 938, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundToBasicAttachParentMatrix, PsoDataType.Bool, 939, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundUsingSimpleSpringDamping, PsoDataType.Bool, 940, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2671392542, PsoDataType.Bool, 941, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachParentApproachSpeedForPitchLock, PsoDataType.Float, 944, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockHeading, PsoDataType.Bool, 948, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMoveSpeedForFollowOrientation, PsoDataType.Float, 952, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLookAroundMoveSpeedMultiplier, PsoDataType.Float, 956, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpeedLimitsForVerticalMoveSpeedScaling, PsoDataType.Float2, 960, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerticalMoveSpeedScaling, PsoDataType.Float, 968, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerticalMoveSpeedScalingAtMaxSpeed, PsoDataType.Float, 972, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpwardMoveSpeedScalingOnGround, PsoDataType.Float, 976, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpwardMoveSpeedScalingInAir, PsoDataType.Float, 980, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMoveOrientationSpeedDuringLookAround, PsoDataType.Float, 984, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowOrientationConing, PsoDataType.Structure, 992, 0, MetaName.camFollowCameraMetadataFollowOrientationConing),
                     new PsoStructureEntryInfo(MetaName.RollSettings, PsoDataType.Structure, 1024, 0, MetaName.camFollowCameraMetadataRollSettings),
                     new PsoStructureEntryInfo(MetaName.HighAltitudeZoomSettings, PsoDataType.Structure, 1064, 0, MetaName.camFollowCameraMetadataHighAltitudeZoomSettings),
                     new PsoStructureEntryInfo(MetaName.HandBrakeSwingSettings, PsoDataType.Structure, 1104, 0, MetaName.camFollowVehicleCameraMetadataHandBrakeSwingSettings),
                     new PsoStructureEntryInfo(MetaName.DuckUnderOverheadCollisionSettings, PsoDataType.Structure, 1136, 0, MetaName.camFollowVehicleCameraMetadataDuckUnderOverheadCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.HighSpeedZoomSettings, PsoDataType.Structure, 1192, 0, MetaName.camFollowVehicleCameraMetadataHighSpeedZoomSettings),
                     new PsoStructureEntryInfo(MetaName.HighSpeedShakeSettings, PsoDataType.Structure, 1224, 0, MetaName.camFollowVehicleCameraMetadataHighSpeedShakeSettings),
                     new PsoStructureEntryInfo(MetaName.WaterEntryShakeSettings, PsoDataType.Structure, 1248, 0, MetaName.camFollowVehicleCameraMetadataWaterEntryShakeSettings),
                     new PsoStructureEntryInfo(MetaName.VerticalFlightModeSettings, PsoDataType.Structure, 1280, 0, MetaName.camFollowVehicleCameraMetadataVerticalFlightModeSettings),
                     new PsoStructureEntryInfo(MetaName.DoorAlignmentSettings, PsoDataType.Structure, 1296, 0, MetaName.camVehicleCustomSettingsMetadataDoorAlignmentSettings),
                     new PsoStructureEntryInfo(MetaName.VehicleEntryExitPitchLevelSmoothRate, PsoDataType.Float, 1336, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraOrbitPitchOffsetForHighAngleMode, PsoDataType.Float, 1340, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraOrbitPitchOffsetForThirdPersonFarViewMode, PsoDataType.Float, 1344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldForceCutToOrbitDistanceLimitsForThirdPersonFarViewMode, PsoDataType.Bool, 1348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThirdPersonVehicleAimCameraRef, PsoDataType.String, 1352, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4075164277, PsoDataType.Bool, 1356, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2625336677, PsoDataType.String, 1360, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3347720841, PsoDataType.Float2, 1364, 0, 0)
                    );
                case MetaName.camFollowVehicleCameraMetadataHandBrakeSwingSettings:
                    return new PsoStructureInfo(MetaName.camFollowVehicleCameraMetadataHandBrakeSwingSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.HandBrakeInputEnvelopeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinLateralSkidSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLateralSkidSpeed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SwingSpeedAtMaxSkidSpeed, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camFollowVehicleCameraMetadataDuckUnderOverheadCollisionSettings:
                    return new PsoStructureInfo(MetaName.camFollowVehicleCameraMetadataDuckUnderOverheadCollisionSettings, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.EnvelopeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDuck, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchOffsetWhenFullyDucked, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToPersist, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CapsuleSettings, PsoDataType.Structure, 32, 0, MetaName.camFollowVehicleCameraMetadataDuckUnderOverheadCollisionSettingsCapsuleSettings)
                    );
                case MetaName.camFollowVehicleCameraMetadataDuckUnderOverheadCollisionSettingsCapsuleSettings:
                    return new PsoStructureInfo(MetaName.camFollowVehicleCameraMetadataDuckUnderOverheadCollisionSettingsCapsuleSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.NumTests, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LengthScaling, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OffsetLimits, PsoDataType.Float2, 16, 0, 0)
                    );
                case MetaName.camFollowVehicleCameraMetadataHighSpeedZoomSettings:
                    return new PsoStructureInfo(MetaName.camFollowVehicleCameraMetadataHighSpeedZoomSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MinForwardSpeed, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardSpeed, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBaseFovScaling, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CutsceneBlendSpringConstant, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CutsceneBlendSpringDampingRatio, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.camFollowVehicleCameraMetadataHighSpeedShakeSettings:
                    return new PsoStructureInfo(MetaName.camFollowVehicleCameraMetadataHighSpeedShakeSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardSpeed, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camFollowVehicleCameraMetadataWaterEntryShakeSettings:
                    return new PsoStructureInfo(MetaName.camFollowVehicleCameraMetadataWaterEntryShakeSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxShakeInstances, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DownwardSpeedLimits, PsoDataType.Float2, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmplitudeLimits, PsoDataType.Float2, 24, 0, 0)
                    );
                case MetaName.camFollowVehicleCameraMetadataVerticalFlightModeSettings:
                    return new PsoStructureInfo(MetaName.camFollowVehicleCameraMetadataVerticalFlightModeSettings, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 8, 0, 0)
                    );
                case MetaName.camVehicleCustomSettingsMetadataDoorAlignmentSettings:
                    return new PsoStructureInfo(MetaName.camVehicleCustomSettingsMetadataDoorAlignmentSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderData, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAlignOnVehicleExit, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo((MetaName)669372144, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AlignmentConeOffsetTowardsVehicleFrontAngle, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AlignmentConeAngle, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AlignmentConeAngleWithTrailer, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinOrientationDeltaToCut, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinOrientationDeltaToCutForReverseAngle, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinOrientationDeltaToCutWithTrailer, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camFollowParachuteCameraMetadata:
                    return new PsoStructureInfo(MetaName.camFollowParachuteCameraMetadata, 0, 0, 1136,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpHelperRef, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseAttachVelocityToIgnoreEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305154880, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldOrbitRelativeToAttachParentOrientation, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistOrbitOrientationRelativeToAttachParent, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringConstant, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringDampingRatio, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406229624, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727876548, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSubmergedLevelToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftGroundSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftContactSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldTimeToBlockFullAttachParentMatrixForRelativeOrbit, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSpeedToClonePitchFromCinematicMountedCameras, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomFramingInTightSpace, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachSpeedToUpdateTightSpaceLevel, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringConstant, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsInTightSpace, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToParentCentreOfGravity, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDynamicCentreOfGravity, PsoDataType.Bool, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVelocityOfAttachParentAttachEntity, PsoDataType.Bool, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomBoundingBoxSettings, PsoDataType.Structure, 128, 0, MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachPedPelvisOffset, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotPosition, PsoDataType.Structure, 176, 0, MetaName.camThirdPersonCameraMetadataBasePivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotPosition, PsoDataType.Structure, 272, 0, MetaName.camThirdPersonCameraMetadataPivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotOverBoundingBoxSettings, PsoDataType.Structure, 320, 0, MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotHeightScalingForFootRoom, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVerticalPivotOffsetForFootRoom, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeOrbitDistanceScalingForExtensions, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)560700931, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomOrbitDistanceLimitsToForce, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringConstant, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringDampingRatio, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingForCustomFirstPersonFallBack, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 400, 0, MetaName.camThirdPersonCameraMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithAttachParent, PsoDataType.Bool, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithFollowVehicle, PsoDataType.Bool, 449, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionOrigin, PsoDataType.Bool, 450, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionRoot, PsoDataType.Bool, 451, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringConstant, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringDampingRatio, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomCollisionOriginRelativePosition, PsoDataType.Float3, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4194152221, PsoDataType.Float3, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomCollisionOrigin, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1998689777, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235540893, PsoDataType.Bool, 514, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3479639237, PsoDataType.Float3, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496115250, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionFallBackPosition, PsoDataType.Structure, 552, 0, MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionFallBackToPivotBlendValue, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConstrainCollisionRootPositionAgainstClippingTypes, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringConstant, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringDampingRatio, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOnTopOfVehicleCollisionSettings, PsoDataType.Structure, 608, 0, MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.IdealHeadingOffsetForLimiting, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffsetInTightSpace, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookOverSettings, PsoDataType.Structure, 664, 0, MetaName.camThirdPersonCameraMetadataLookOverSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentPitchForLookBehind, PsoDataType.Bool, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitsForBasePosition, PsoDataType.Float2, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreToPostCollisionLookAtOrientationBlendValue, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringConstant, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringDampingRatio, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchSoftLimits, PsoDataType.Float2, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchHardLimits, PsoDataType.Float2, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthZoomSettings, PsoDataType.Structure, 728, 0, MetaName.camThirdPersonCameraMetadataStealthZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1947299079, PsoDataType.Structure, 752, 0, MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 776, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2122226771, PsoDataType.Bool, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentInAirEnvelopeRef, PsoDataType.String, 784, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentUpwardSpeedScalingOnGroundEnvelopeRef, PsoDataType.String, 788, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentUpwardSpeedScalingInAirEnvelopeRef, PsoDataType.String, 792, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AimBehaviourEnvelopeRef, PsoDataType.String, 796, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WaterBobShakeRef, PsoDataType.String, 800, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentMovementForOrientation, PsoDataType.Bool, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PullAroundSettings, PsoDataType.Structure, 808, 0, MetaName.camFollowCameraMetadataPullAroundSettings),
                     new PsoStructureEntryInfo(MetaName.PullAroundSettingsForLookBehind, PsoDataType.Structure, 872, 0, MetaName.camFollowCameraMetadataPullAroundSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderAttachParentLocalXYVelocityForPullAround, PsoDataType.Bool, 936, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderAttachParentForwardSpeedForPullAround, PsoDataType.Bool, 937, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundToAttachParentFront, PsoDataType.Bool, 938, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundToBasicAttachParentMatrix, PsoDataType.Bool, 939, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPullAroundUsingSimpleSpringDamping, PsoDataType.Bool, 940, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2671392542, PsoDataType.Bool, 941, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachParentApproachSpeedForPitchLock, PsoDataType.Float, 944, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockHeading, PsoDataType.Bool, 948, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMoveSpeedForFollowOrientation, PsoDataType.Float, 952, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLookAroundMoveSpeedMultiplier, PsoDataType.Float, 956, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpeedLimitsForVerticalMoveSpeedScaling, PsoDataType.Float2, 960, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerticalMoveSpeedScaling, PsoDataType.Float, 968, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerticalMoveSpeedScalingAtMaxSpeed, PsoDataType.Float, 972, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpwardMoveSpeedScalingOnGround, PsoDataType.Float, 976, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpwardMoveSpeedScalingInAir, PsoDataType.Float, 980, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMoveOrientationSpeedDuringLookAround, PsoDataType.Float, 984, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowOrientationConing, PsoDataType.Structure, 992, 0, MetaName.camFollowCameraMetadataFollowOrientationConing),
                     new PsoStructureEntryInfo(MetaName.RollSettings, PsoDataType.Structure, 1024, 0, MetaName.camFollowCameraMetadataRollSettings),
                     new PsoStructureEntryInfo(MetaName.HighAltitudeZoomSettings, PsoDataType.Structure, 1064, 0, MetaName.camFollowCameraMetadataHighAltitudeZoomSettings),
                     new PsoStructureEntryInfo(MetaName.CustomSettings, PsoDataType.Structure, 1104, 0, MetaName.camFollowParachuteCameraMetadataCustomSettings)
                    );
                case MetaName.camFollowParachuteCameraMetadataCustomSettings:
                    return new PsoStructureInfo(MetaName.camFollowParachuteCameraMetadataCustomSettings, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)3501562487, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1514912105, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2028447415, PsoDataType.String, 16, 7, 0)
                    );
                case MetaName.camFirstPersonPedAimCameraMetadata:
                    return new PsoStructureInfo(MetaName.camFirstPersonPedAimCameraMetadata, 0, 0, 208,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinUpdatesBeforeApplyingMotionBlur, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseMotionBlurStrength, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMinFovDelta, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMaxFovDelta, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMaxStrengthForFov, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachRelativeOffset, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachOffsetRelativeToCamera, PsoDataType.Bool, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPitch, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRelativeHeading, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRelativeHeading, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldMakeAttachedEntityInvisible, PsoDataType.Bool, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1959607978, PsoDataType.Bool, 113, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 114, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1120407412, PsoDataType.Bool, 115, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeFirstPersonShootingAbilityLimits, PsoDataType.Float2, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeAmplitudeScalingForShootingAbilityLimits, PsoDataType.Float2, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2758176724, PsoDataType.Structure, 136, 0, MetaName.camFirstPersonAimCameraMetadataHeadingCorrection),
                     new PsoStructureEntryInfo((MetaName)330110937, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachBoneTag, PsoDataType.SInt, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TripleHeadNearClip, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachPositionSmoothRate, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTorsoIkLimitsOverrideOrbitPitchLimits, PsoDataType.Bool, 188, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1847867369, PsoDataType.Bool, 189, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1173900687, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1322555765, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1723254225, PsoDataType.Bool, 200, 0, 0)
                    );
                case MetaName.camFirstPersonAimCameraMetadataHeadingCorrection:
                    return new PsoStructureInfo(MetaName.camFirstPersonAimCameraMetadataHeadingCorrection, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496770623, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DeltaTolerance, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.camFirstPersonShooterCameraMetadata:
                    return new PsoStructureInfo(MetaName.camFirstPersonShooterCameraMetadata, 0, 0, 1664,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinUpdatesBeforeApplyingMotionBlur, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseMotionBlurStrength, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMinFovDelta, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMaxFovDelta, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMaxStrengthForFov, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachRelativeOffset, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachOffsetRelativeToCamera, PsoDataType.Bool, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPitch, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRelativeHeading, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRelativeHeading, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldMakeAttachedEntityInvisible, PsoDataType.Bool, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1959607978, PsoDataType.Bool, 113, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 114, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1120407412, PsoDataType.Bool, 115, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeFirstPersonShootingAbilityLimits, PsoDataType.Float2, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeAmplitudeScalingForShootingAbilityLimits, PsoDataType.Float2, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2758176724, PsoDataType.Structure, 136, 0, MetaName.camFirstPersonAimCameraMetadataHeadingCorrection),
                     new PsoStructureEntryInfo((MetaName)330110937, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 176, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3500215588, PsoDataType.Float3, 192, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1583458555, PsoDataType.Float3, 208, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3253139357, PsoDataType.Float3, 224, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4262893203, PsoDataType.Float3, 240, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2347724525, PsoDataType.Float3, 256, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3438104893, PsoDataType.Float3, 272, 0, 0),
                     new PsoStructureEntryInfo((MetaName)424961570, PsoDataType.Float3, 288, 0, 0),
                     new PsoStructureEntryInfo((MetaName)104465235, PsoDataType.Float3, 304, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4034848485, PsoDataType.Float3, 320, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRelativeHeading, PsoDataType.Float2, 336, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2372600507, PsoDataType.Float2, 344, 0, 0),
                     new PsoStructureEntryInfo((MetaName)971666241, PsoDataType.Float2, 352, 0, 0),
                     new PsoStructureEntryInfo((MetaName)162051482, PsoDataType.Float2, 360, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3789298543, PsoDataType.Float2, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3507003526, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1765689234, PsoDataType.String, 384, 7, 0),
                     new PsoStructureEntryInfo((MetaName)722540604, PsoDataType.String, 388, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnEnvelopeRef, PsoDataType.String, 392, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3102337801, PsoDataType.String, 396, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1038878908, PsoDataType.String, 400, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3684814544, PsoDataType.String, 404, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2377645189, PsoDataType.String, 408, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3664775821, PsoDataType.String, 412, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3075522471, PsoDataType.String, 416, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AttachBoneTag, PsoDataType.SInt, 420, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeAttachPositionSmoothRate, PsoDataType.Float, 424, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3559802057, PsoDataType.Structure, 432, 0, MetaName.camFirstPersonShooterCameraMetadataRelativeAttachOrientationSettings),
                     new PsoStructureEntryInfo((MetaName)3284913244, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo((MetaName)742850166, PsoDataType.Float, 548, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1801708329, PsoDataType.Float, 552, 0, 0),
                     new PsoStructureEntryInfo((MetaName)522025471, PsoDataType.Float, 556, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimFov, PsoDataType.Float, 560, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2685045238, PsoDataType.Float, 564, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1545486036, PsoDataType.Float, 568, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PhoneFov, PsoDataType.Float, 572, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2263644173, PsoDataType.Float, 576, 0, 0),
                     new PsoStructureEntryInfo((MetaName)189771911, PsoDataType.Float, 580, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3266229162, PsoDataType.Float, 584, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3113298498, PsoDataType.Float, 588, 0, 0),
                     new PsoStructureEntryInfo((MetaName)317025866, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo((MetaName)788370107, PsoDataType.Float, 596, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1892114730, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2312570903, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1650788232, PsoDataType.Float, 608, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3142109941, PsoDataType.Float, 612, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1692829019, PsoDataType.Float, 616, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1583361695, PsoDataType.Float, 620, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3530519735, PsoDataType.Float, 624, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1560089378, PsoDataType.Float, 628, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1683046794, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1176445410, PsoDataType.Float, 636, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3761456258, PsoDataType.Float, 640, 0, 0),
                     new PsoStructureEntryInfo((MetaName)413985988, PsoDataType.Float, 644, 0, 0),
                     new PsoStructureEntryInfo((MetaName)731578407, PsoDataType.Float, 648, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3517278862, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LadderPitch, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo((MetaName)83336335, PsoDataType.Float, 660, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3203179444, PsoDataType.Float, 664, 0, 0),
                     new PsoStructureEntryInfo((MetaName)803486985, PsoDataType.Float, 668, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecoilShakeAmplitudeScaling, PsoDataType.Float, 672, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2122499987, PsoDataType.Float, 676, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1515082975, PsoDataType.Float, 680, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1515564685, PsoDataType.Float, 684, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3592059377, PsoDataType.Float, 688, 0, 0),
                     new PsoStructureEntryInfo((MetaName)448454700, PsoDataType.Bool, 692, 0, 0),
                     new PsoStructureEntryInfo((MetaName)545313111, PsoDataType.Float, 696, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1534221961, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo((MetaName)270168093, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2100042799, PsoDataType.Float, 712, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3626441228, PsoDataType.UInt, 716, 0, 0),
                     new PsoStructureEntryInfo((MetaName)700340890, PsoDataType.UInt, 720, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1968480925, PsoDataType.UInt, 724, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2850438294, PsoDataType.UInt, 728, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1608452380, PsoDataType.UInt, 732, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2387142724, PsoDataType.UInt, 736, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2743970065, PsoDataType.UInt, 740, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2452155390, PsoDataType.UInt, 744, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrientationSpring, PsoDataType.Structure, 752, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpring),
                     new PsoStructureEntryInfo((MetaName)415673303, PsoDataType.Structure, 792, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpring),
                     new PsoStructureEntryInfo((MetaName)3397810078, PsoDataType.Structure, 832, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpring),
                     new PsoStructureEntryInfo((MetaName)3440729757, PsoDataType.Structure, 872, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)2454767338, PsoDataType.Structure, 888, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)1962460716, PsoDataType.Structure, 904, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)1766077384, PsoDataType.Structure, 920, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)2053281479, PsoDataType.Structure, 936, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)3083287001, PsoDataType.Structure, 952, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)4060968766, PsoDataType.Structure, 968, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)2437540040, PsoDataType.Structure, 984, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)3391036765, PsoDataType.Structure, 1000, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)2728534787, PsoDataType.Structure, 1016, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)3139215619, PsoDataType.Structure, 1032, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)4087791749, PsoDataType.Structure, 1048, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)2550891007, PsoDataType.Structure, 1064, 0, MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite),
                     new PsoStructureEntryInfo((MetaName)475503552, PsoDataType.Float, 1080, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1215439957, PsoDataType.Float2, 1084, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTorsoIkLimitsOverrideOrbitPitchLimits, PsoDataType.Bool, 1092, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1723254225, PsoDataType.Bool, 1093, 0, 0),
                     new PsoStructureEntryInfo((MetaName)366723406, PsoDataType.String, 1096, 7, 0),
                     new PsoStructureEntryInfo((MetaName)264754789, PsoDataType.String, 1100, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1539647797, PsoDataType.String, 1104, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1519473347, PsoDataType.Float, 1108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3224324247, PsoDataType.Float, 1112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2478271115, PsoDataType.Float, 1116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2317113905, PsoDataType.Float, 1120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3903912151, PsoDataType.Float, 1124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2266019553, PsoDataType.Float, 1128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3706052819, PsoDataType.Float, 1132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2722258227, PsoDataType.Float, 1136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1525907069, PsoDataType.Float, 1140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2570774561, PsoDataType.Float, 1144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1547572758, PsoDataType.Bool, 1148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3957958745, PsoDataType.UInt, 1152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1630855608, PsoDataType.Float, 1156, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3965978499, PsoDataType.Structure, 1160, 0, MetaName.camFirstPersonShooterCameraMetadataStickyAim),
                     new PsoStructureEntryInfo((MetaName)585212636, PsoDataType.Structure, 1272, 0, MetaName.camFirstPersonShooterCameraMetadataStickyAim),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLockOnAiming, PsoDataType.Bool, 1384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldValidateLockOnTargetPosition, PsoDataType.Bool, 1385, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLockOn, PsoDataType.Float, 1388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForFineAimScaling, PsoDataType.Float, 1392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForFineAimScaling, PsoDataType.Float, 1396, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleDeltaForLockOnSwitchBlendScaling, PsoDataType.Float, 1400, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringConstant, PsoDataType.Float, 1404, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringDampingRatio, PsoDataType.Float, 1408, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForLockOnSwitch, PsoDataType.UInt, 1412, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForLockOnSwitch, PsoDataType.UInt, 1416, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForInitialLockOn, PsoDataType.UInt, 1420, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForInitialLockOn, PsoDataType.UInt, 1424, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 1428, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 1432, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2447870594, PsoDataType.Float, 1436, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2225908075, PsoDataType.Float, 1440, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1894138, PsoDataType.Float, 1444, 0, 0),
                     new PsoStructureEntryInfo((MetaName)724885771, PsoDataType.Float, 1448, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2413827894, PsoDataType.Float2, 1452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverSettings, PsoDataType.Structure, 1464, 0, MetaName.camFirstPersonShooterCameraMetadataCoverSettings),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camFirstPersonShooterCameraMetadataSprintBreakOutSettings),
                     new PsoStructureEntryInfo((MetaName)2011532787, PsoDataType.Array, 1584, 1, (MetaName)131236),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 1656, 7, 0)
                    );
                case MetaName.camFirstPersonShooterCameraMetadataRelativeAttachOrientationSettings:
                    return new PsoStructureInfo(MetaName.camFirstPersonShooterCameraMetadataRelativeAttachOrientationSettings, 0, 0, 112,
                     new PsoStructureEntryInfo((MetaName)3727099699, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3120224920, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2408234737, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)869960116, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3340971535, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2197092363, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2198368821, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3606762226, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1647925848, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)490937808, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)379620733, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1441798592, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3840322276, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3383026591, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1824793815, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1259376027, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2964527414, PsoDataType.Float3, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2439730995, PsoDataType.Float3, 96, 0, 0)
                    );
                case MetaName.camFirstPersonShooterCameraMetadataOrientationSpring:
                    return new PsoStructureInfo(MetaName.camFirstPersonShooterCameraMetadataOrientationSpring, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.HeadingLimits, PsoDataType.Float2, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchLimits, PsoDataType.Float2, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1524988618, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1290012653, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite:
                    return new PsoStructureInfo(MetaName.camFirstPersonShooterCameraMetadataOrientationSpringLite, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camFirstPersonShooterCameraMetadataStickyAim:
                    return new PsoStructureInfo(MetaName.camFirstPersonShooterCameraMetadataStickyAim, 0, 0, 112,
                     new PsoStructureEntryInfo((MetaName)934287534, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1890095417, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2911838887, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1168068740, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)943460322, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2440008981, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3546382126, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3702929248, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1390609527, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1952283316, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1859396322, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3508865952, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4071651879, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3012719683, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1163787532, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3148686816, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2294593316, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3264503435, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)432760898, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3118501359, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2024506330, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1824271937, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)721888806, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)903289241, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2991714924, PsoDataType.Float, 104, 0, 0)
                    );
                case MetaName.camFirstPersonShooterCameraMetadataCoverSettings:
                    return new PsoStructureInfo(MetaName.camFirstPersonShooterCameraMetadataCoverSettings, 0, 0, 120,
                     new PsoStructureEntryInfo((MetaName)2970917480, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)256711860, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2716737212, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1153272400, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)11232964, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1680843526, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1294539981, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3548711893, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3299300152, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1497386222, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)783496886, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2616989255, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingCorrection, PsoDataType.Structure, 56, 0, MetaName.camFirstPersonAimCameraMetadataHeadingCorrection),
                     new PsoStructureEntryInfo((MetaName)907428778, PsoDataType.SInt, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4143200730, PsoDataType.SInt, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)105496678, PsoDataType.SInt, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2098811685, PsoDataType.SInt, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1946159554, PsoDataType.SInt, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3595423851, PsoDataType.SInt, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)153370955, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1271201291, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)558015512, PsoDataType.UInt, 112, 0, 0)
                    );
                case MetaName.camFirstPersonShooterCameraMetadataSprintBreakOutSettings:
                    return new PsoStructureInfo(MetaName.camFirstPersonShooterCameraMetadataSprintBreakOutSettings, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)741085985, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)87271889, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4190510062, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1438187592, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1823031239, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2144963791, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.camFirstPersonHeadTrackingAimCameraMetadata:
                    return new PsoStructureInfo(MetaName.camFirstPersonHeadTrackingAimCameraMetadata, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinUpdatesBeforeApplyingMotionBlur, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseMotionBlurStrength, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMinFovDelta, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMaxFovDelta, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomMotionBlurMaxStrengthForFov, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachRelativeOffset, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachOffsetRelativeToCamera, PsoDataType.Bool, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPitch, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRelativeHeading, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRelativeHeading, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachBoneTag, PsoDataType.SInt, 116, 0, 0)
                    );
                case MetaName.camThirdPersonPedAimCameraMetadata:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAimCameraMetadata, 0, 0, 1040,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpHelperRef, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseAttachVelocityToIgnoreEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305154880, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldOrbitRelativeToAttachParentOrientation, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistOrbitOrientationRelativeToAttachParent, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringConstant, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringDampingRatio, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406229624, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727876548, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSubmergedLevelToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftGroundSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftContactSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldTimeToBlockFullAttachParentMatrixForRelativeOrbit, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSpeedToClonePitchFromCinematicMountedCameras, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomFramingInTightSpace, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachSpeedToUpdateTightSpaceLevel, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringConstant, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsInTightSpace, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToParentCentreOfGravity, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDynamicCentreOfGravity, PsoDataType.Bool, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVelocityOfAttachParentAttachEntity, PsoDataType.Bool, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomBoundingBoxSettings, PsoDataType.Structure, 128, 0, MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachPedPelvisOffset, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotPosition, PsoDataType.Structure, 176, 0, MetaName.camThirdPersonCameraMetadataBasePivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotPosition, PsoDataType.Structure, 272, 0, MetaName.camThirdPersonCameraMetadataPivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotOverBoundingBoxSettings, PsoDataType.Structure, 320, 0, MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotHeightScalingForFootRoom, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVerticalPivotOffsetForFootRoom, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeOrbitDistanceScalingForExtensions, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)560700931, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomOrbitDistanceLimitsToForce, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringConstant, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringDampingRatio, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingForCustomFirstPersonFallBack, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 400, 0, MetaName.camThirdPersonCameraMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithAttachParent, PsoDataType.Bool, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithFollowVehicle, PsoDataType.Bool, 449, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionOrigin, PsoDataType.Bool, 450, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionRoot, PsoDataType.Bool, 451, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringConstant, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringDampingRatio, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomCollisionOriginRelativePosition, PsoDataType.Float3, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4194152221, PsoDataType.Float3, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomCollisionOrigin, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1998689777, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235540893, PsoDataType.Bool, 514, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3479639237, PsoDataType.Float3, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496115250, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionFallBackPosition, PsoDataType.Structure, 552, 0, MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionFallBackToPivotBlendValue, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConstrainCollisionRootPositionAgainstClippingTypes, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringConstant, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringDampingRatio, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOnTopOfVehicleCollisionSettings, PsoDataType.Structure, 608, 0, MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.IdealHeadingOffsetForLimiting, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffsetInTightSpace, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookOverSettings, PsoDataType.Structure, 664, 0, MetaName.camThirdPersonCameraMetadataLookOverSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentPitchForLookBehind, PsoDataType.Bool, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitsForBasePosition, PsoDataType.Float2, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreToPostCollisionLookAtOrientationBlendValue, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringConstant, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringDampingRatio, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchSoftLimits, PsoDataType.Float2, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchHardLimits, PsoDataType.Float2, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthZoomSettings, PsoDataType.Structure, 728, 0, MetaName.camThirdPersonCameraMetadataStealthZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1947299079, PsoDataType.Structure, 752, 0, MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 776, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2122226771, PsoDataType.Bool, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnEnvelopeRef, PsoDataType.String, 784, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 788, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticuleDuringInterpolation, PsoDataType.Bool, 789, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAllowInterpolationSourceCameraToPersistReticule, PsoDataType.Bool, 790, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyWeaponFov, PsoDataType.Bool, 791, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLockOnAiming, PsoDataType.Bool, 792, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockOnToTargetEntityPosition, PsoDataType.Bool, 793, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldValidateLockOnTargetPosition, PsoDataType.Bool, 794, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TripleHeadNearClip, PsoDataType.Float, 796, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecoilShakeAmplitudeScaling, PsoDataType.Float, 800, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLockOn, PsoDataType.Float, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForFineAimScaling, PsoDataType.Float, 808, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForFineAimScaling, PsoDataType.Float, 812, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleDeltaForLockOnSwitchBlendScaling, PsoDataType.Float, 816, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForInitialLockOn, PsoDataType.UInt, 820, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForInitialLockOn, PsoDataType.UInt, 824, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForLockOnSwitch, PsoDataType.UInt, 828, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForLockOnSwitch, PsoDataType.UInt, 832, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringConstant, PsoDataType.Float, 836, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringDampingRatio, PsoDataType.Float, 840, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringConstant, PsoDataType.Float, 844, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringDampingRatio, PsoDataType.Float, 848, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldFocusOnLockOnTarget, PsoDataType.Bool, 852, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFovToEmulateWithFocalLengthMultiplier, PsoDataType.Float, 856, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FocusParentToTargetBlendLevel, PsoDataType.Float, 860, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SecondaryFocusParentToTargetBlendLevel, PsoDataType.Float, 864, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFocusToSecondaryFocusDistance, PsoDataType.Float, 868, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnTargetStunnedEnvelopeRef, PsoDataType.String, 880, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnTargetDampingSettings, PsoDataType.Structure, 888, 0, MetaName.camThirdPersonPedAimCameraMetadataLockOnTargetDampingSettings),
                     new PsoStructureEntryInfo(MetaName.LockOnOrbitDistanceSettings, PsoDataType.Structure, 928, 0, MetaName.camThirdPersonPedAimCameraMetadataLockOnOrbitDistanceSettings),
                     new PsoStructureEntryInfo(MetaName.ParentRelativeAttachOffset, PsoDataType.Float3, 976, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParentRelativeAttachOffsetAtOrbitHeadingLimits, PsoDataType.Float3, 992, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachBoneTag, PsoDataType.SInt, 1008, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldScriptedAimTaskOverrideOrbitPitchLimits, PsoDataType.Bool, 1012, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAimSweepOverrideOrbitPitchLimits, PsoDataType.Bool, 1013, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsForMobilePhoneShallowDofMode, PsoDataType.String, 1016, 7, 0),
                     new PsoStructureEntryInfo((MetaName)309842601, PsoDataType.Float, 1020, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3770307856, PsoDataType.Bool, 1024, 0, 0)
                    );
                case MetaName.camThirdPersonPedAimCameraMetadataLockOnTargetDampingSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAimCameraMetadataLockOnTargetDampingSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldApplyDamping, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StunnedHeadingSpringConstant, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StunnedHeadingSpringDampingRatio, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StunnedPitchSpringConstant, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StunnedPitchSpringDampingRatio, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StunnedDistanceSpringConstant, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StunnedDistanceSpringDampingRatio, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camThirdPersonPedAimCameraMetadataLockOnOrbitDistanceSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAimCameraMetadataLockOnOrbitDistanceSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldApplyScaling, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnDistanceLimits, PsoDataType.Float2, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingLimits, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingSpringConstant, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingSpringDampingRatio, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraMetadata:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraMetadata, 0, 0, 1584,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpHelperRef, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseAttachVelocityToIgnoreEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305154880, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldOrbitRelativeToAttachParentOrientation, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistOrbitOrientationRelativeToAttachParent, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringConstant, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringDampingRatio, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406229624, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727876548, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSubmergedLevelToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftGroundSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftContactSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldTimeToBlockFullAttachParentMatrixForRelativeOrbit, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSpeedToClonePitchFromCinematicMountedCameras, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomFramingInTightSpace, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachSpeedToUpdateTightSpaceLevel, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringConstant, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsInTightSpace, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToParentCentreOfGravity, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDynamicCentreOfGravity, PsoDataType.Bool, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVelocityOfAttachParentAttachEntity, PsoDataType.Bool, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomBoundingBoxSettings, PsoDataType.Structure, 128, 0, MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachPedPelvisOffset, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotPosition, PsoDataType.Structure, 176, 0, MetaName.camThirdPersonCameraMetadataBasePivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotPosition, PsoDataType.Structure, 272, 0, MetaName.camThirdPersonCameraMetadataPivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotOverBoundingBoxSettings, PsoDataType.Structure, 320, 0, MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotHeightScalingForFootRoom, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVerticalPivotOffsetForFootRoom, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeOrbitDistanceScalingForExtensions, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)560700931, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomOrbitDistanceLimitsToForce, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringConstant, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringDampingRatio, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingForCustomFirstPersonFallBack, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 400, 0, MetaName.camThirdPersonCameraMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithAttachParent, PsoDataType.Bool, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithFollowVehicle, PsoDataType.Bool, 449, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionOrigin, PsoDataType.Bool, 450, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionRoot, PsoDataType.Bool, 451, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringConstant, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringDampingRatio, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomCollisionOriginRelativePosition, PsoDataType.Float3, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4194152221, PsoDataType.Float3, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomCollisionOrigin, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1998689777, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235540893, PsoDataType.Bool, 514, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3479639237, PsoDataType.Float3, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496115250, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionFallBackPosition, PsoDataType.Structure, 552, 0, MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionFallBackToPivotBlendValue, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConstrainCollisionRootPositionAgainstClippingTypes, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringConstant, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringDampingRatio, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOnTopOfVehicleCollisionSettings, PsoDataType.Structure, 608, 0, MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.IdealHeadingOffsetForLimiting, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffsetInTightSpace, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookOverSettings, PsoDataType.Structure, 664, 0, MetaName.camThirdPersonCameraMetadataLookOverSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentPitchForLookBehind, PsoDataType.Bool, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitsForBasePosition, PsoDataType.Float2, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreToPostCollisionLookAtOrientationBlendValue, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringConstant, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringDampingRatio, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchSoftLimits, PsoDataType.Float2, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchHardLimits, PsoDataType.Float2, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthZoomSettings, PsoDataType.Structure, 728, 0, MetaName.camThirdPersonCameraMetadataStealthZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1947299079, PsoDataType.Structure, 752, 0, MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 776, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2122226771, PsoDataType.Bool, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnEnvelopeRef, PsoDataType.String, 784, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 788, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticuleDuringInterpolation, PsoDataType.Bool, 789, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAllowInterpolationSourceCameraToPersistReticule, PsoDataType.Bool, 790, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyWeaponFov, PsoDataType.Bool, 791, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLockOnAiming, PsoDataType.Bool, 792, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockOnToTargetEntityPosition, PsoDataType.Bool, 793, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldValidateLockOnTargetPosition, PsoDataType.Bool, 794, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TripleHeadNearClip, PsoDataType.Float, 796, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecoilShakeAmplitudeScaling, PsoDataType.Float, 800, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLockOn, PsoDataType.Float, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForFineAimScaling, PsoDataType.Float, 808, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForFineAimScaling, PsoDataType.Float, 812, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleDeltaForLockOnSwitchBlendScaling, PsoDataType.Float, 816, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForInitialLockOn, PsoDataType.UInt, 820, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForInitialLockOn, PsoDataType.UInt, 824, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForLockOnSwitch, PsoDataType.UInt, 828, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForLockOnSwitch, PsoDataType.UInt, 832, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringConstant, PsoDataType.Float, 836, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringDampingRatio, PsoDataType.Float, 840, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringConstant, PsoDataType.Float, 844, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringDampingRatio, PsoDataType.Float, 848, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldFocusOnLockOnTarget, PsoDataType.Bool, 852, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFovToEmulateWithFocalLengthMultiplier, PsoDataType.Float, 856, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FocusParentToTargetBlendLevel, PsoDataType.Float, 860, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SecondaryFocusParentToTargetBlendLevel, PsoDataType.Float, 864, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFocusToSecondaryFocusDistance, PsoDataType.Float, 868, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnTargetStunnedEnvelopeRef, PsoDataType.String, 880, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnTargetDampingSettings, PsoDataType.Structure, 888, 0, MetaName.camThirdPersonPedAimCameraMetadataLockOnTargetDampingSettings),
                     new PsoStructureEntryInfo(MetaName.LockOnOrbitDistanceSettings, PsoDataType.Structure, 928, 0, MetaName.camThirdPersonPedAimCameraMetadataLockOnOrbitDistanceSettings),
                     new PsoStructureEntryInfo(MetaName.ParentRelativeAttachOffset, PsoDataType.Float3, 976, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParentRelativeAttachOffsetAtOrbitHeadingLimits, PsoDataType.Float3, 992, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachBoneTag, PsoDataType.SInt, 1008, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldScriptedAimTaskOverrideOrbitPitchLimits, PsoDataType.Bool, 1012, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAimSweepOverrideOrbitPitchLimits, PsoDataType.Bool, 1013, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsForMobilePhoneShallowDofMode, PsoDataType.String, 1016, 7, 0),
                     new PsoStructureEntryInfo((MetaName)309842601, PsoDataType.Float, 1020, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3770307856, PsoDataType.Bool, 1024, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2612404486, PsoDataType.Bool, 1040, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3464513788, PsoDataType.String, 1044, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3165832172, PsoDataType.Float, 1048, 0, 0),
                     new PsoStructureEntryInfo((MetaName)313901382, PsoDataType.Float, 1052, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3456520914, PsoDataType.Float, 1056, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3227758752, PsoDataType.Float, 1060, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2595201884, PsoDataType.Structure, 1064, 0, MetaName.camThirdPersonPedAssistedAimCameraShakeActivityScalingSettings),
                     new PsoStructureEntryInfo((MetaName)1699579049, PsoDataType.Structure, 1088, 0, MetaName.camThirdPersonPedAssistedAimCameraRunningShakeSettings),
                     new PsoStructureEntryInfo((MetaName)2514985597, PsoDataType.Structure, 1112, 0, MetaName.camThirdPersonPedAssistedAimCameraRunningShakeSettings),
                     new PsoStructureEntryInfo((MetaName)3459655185, PsoDataType.Structure, 1136, 0, MetaName.camThirdPersonPedAssistedAimCameraPivotScalingSettings),
                     new PsoStructureEntryInfo((MetaName)1039438125, PsoDataType.Structure, 1192, 0, MetaName.camThirdPersonPedAssistedAimCameraShootingFocusSettings),
                     new PsoStructureEntryInfo((MetaName)2331386580, PsoDataType.Structure, 1248, 0, MetaName.camThirdPersonPedAssistedAimCameraCinematicMomentSettings),
                     new PsoStructureEntryInfo((MetaName)3975806377, PsoDataType.Structure, 1432, 0, MetaName.camThirdPersonPedAssistedAimCameraRecoilShakeScalingSettings),
                     new PsoStructureEntryInfo((MetaName)2579888707, PsoDataType.Structure, 1464, 0, MetaName.camThirdPersonPedAssistedAimCameraInCoverSettings)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraShakeActivityScalingSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraShakeActivityScalingSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.AmplitudeScale, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendInSpringConstant, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutSpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraRunningShakeSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraRunningShakeSettings, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinAmplitude, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAmplitude, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3083741385, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434404012, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1424595128, PsoDataType.Bool, 22, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraPivotScalingSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraPivotScalingSettings, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3290868314, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZoomFactor, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1239610477, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4266810853, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2490035552, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1582111169, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ErrorThreshold, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SideOffset, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3185812029, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2740464011, PsoDataType.Float, 52, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraShootingFocusSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraShootingFocusSettings, 0, 0, 56,
                     new PsoStructureEntryInfo((MetaName)1778381348, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)352609433, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendInSpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutSpringConstant, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2503271690, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZoomFactor, PsoDataType.Float, 48, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraCinematicMomentSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraCinematicMomentSettings, 0, 0, 184,
                     new PsoStructureEntryInfo((MetaName)3112353702, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3298892127, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3924454742, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2215779446, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)917560913, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1705773296, PsoDataType.UInt, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3495762240, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2011477591, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3944656129, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3273599536, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ZoomFactor, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1293989475, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3934763180, PsoDataType.Structure, 56, 0, MetaName.camThirdPersonPedAssistedAimCameraLockOnAlignmentSettings),
                     new PsoStructureEntryInfo((MetaName)4209943355, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3222932576, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1411402256, PsoDataType.Structure, 120, 0, MetaName.camThirdPersonPedAssistedAimCameraPlayerFramingSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 180, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraLockOnAlignmentSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraLockOnAlignmentSettings, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2192037365, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1207772736, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1953981041, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdleCmini, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3067708566, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3928624067, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4117544351, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3546588287, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2433975532, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2894326732, PsoDataType.Float, 48, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraPlayerFramingSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraPlayerFramingSettings, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.AttackDelay, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttackDuration, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReleaseDelay, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1918481730, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredPitch, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3067708566, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3928624067, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4117544351, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3648026356, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraRecoilShakeScalingSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraRecoilShakeScalingSettings, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)632884413, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3501089975, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1841607681, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)405521980, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2308368369, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camThirdPersonPedAssistedAimCameraInCoverSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAssistedAimCameraInCoverSettings, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3044106186, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2342994624, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2363962395, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3996729234, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2974603259, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1163820100, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1153320978, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)887750583, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.parameterIntensity, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3038154213, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)795086429, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2305231569, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2043206239, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1283135688, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3878807616, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3640542370, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3758922505, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1764223557, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3214654546, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1111260888, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3050977735, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1185331567, PsoDataType.Float, 108, 0, 0)
                    );
                case MetaName.camThirdPersonPedAimInCoverCameraMetadata:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAimInCoverCameraMetadata, 0, 0, 1248,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpHelperRef, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseAttachVelocityToIgnoreEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305154880, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldOrbitRelativeToAttachParentOrientation, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistOrbitOrientationRelativeToAttachParent, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringConstant, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringDampingRatio, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406229624, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727876548, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSubmergedLevelToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftGroundSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftContactSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldTimeToBlockFullAttachParentMatrixForRelativeOrbit, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSpeedToClonePitchFromCinematicMountedCameras, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomFramingInTightSpace, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachSpeedToUpdateTightSpaceLevel, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringConstant, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsInTightSpace, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToParentCentreOfGravity, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDynamicCentreOfGravity, PsoDataType.Bool, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVelocityOfAttachParentAttachEntity, PsoDataType.Bool, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomBoundingBoxSettings, PsoDataType.Structure, 128, 0, MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachPedPelvisOffset, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotPosition, PsoDataType.Structure, 176, 0, MetaName.camThirdPersonCameraMetadataBasePivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotPosition, PsoDataType.Structure, 272, 0, MetaName.camThirdPersonCameraMetadataPivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotOverBoundingBoxSettings, PsoDataType.Structure, 320, 0, MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotHeightScalingForFootRoom, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVerticalPivotOffsetForFootRoom, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeOrbitDistanceScalingForExtensions, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)560700931, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomOrbitDistanceLimitsToForce, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringConstant, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringDampingRatio, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingForCustomFirstPersonFallBack, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 400, 0, MetaName.camThirdPersonCameraMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithAttachParent, PsoDataType.Bool, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithFollowVehicle, PsoDataType.Bool, 449, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionOrigin, PsoDataType.Bool, 450, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionRoot, PsoDataType.Bool, 451, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringConstant, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringDampingRatio, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomCollisionOriginRelativePosition, PsoDataType.Float3, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4194152221, PsoDataType.Float3, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomCollisionOrigin, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1998689777, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235540893, PsoDataType.Bool, 514, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3479639237, PsoDataType.Float3, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496115250, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionFallBackPosition, PsoDataType.Structure, 552, 0, MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionFallBackToPivotBlendValue, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConstrainCollisionRootPositionAgainstClippingTypes, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringConstant, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringDampingRatio, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOnTopOfVehicleCollisionSettings, PsoDataType.Structure, 608, 0, MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.IdealHeadingOffsetForLimiting, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffsetInTightSpace, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookOverSettings, PsoDataType.Structure, 664, 0, MetaName.camThirdPersonCameraMetadataLookOverSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentPitchForLookBehind, PsoDataType.Bool, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitsForBasePosition, PsoDataType.Float2, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreToPostCollisionLookAtOrientationBlendValue, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringConstant, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringDampingRatio, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchSoftLimits, PsoDataType.Float2, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchHardLimits, PsoDataType.Float2, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthZoomSettings, PsoDataType.Structure, 728, 0, MetaName.camThirdPersonCameraMetadataStealthZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1947299079, PsoDataType.Structure, 752, 0, MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 776, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2122226771, PsoDataType.Bool, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnEnvelopeRef, PsoDataType.String, 784, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 788, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticuleDuringInterpolation, PsoDataType.Bool, 789, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAllowInterpolationSourceCameraToPersistReticule, PsoDataType.Bool, 790, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyWeaponFov, PsoDataType.Bool, 791, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLockOnAiming, PsoDataType.Bool, 792, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockOnToTargetEntityPosition, PsoDataType.Bool, 793, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldValidateLockOnTargetPosition, PsoDataType.Bool, 794, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TripleHeadNearClip, PsoDataType.Float, 796, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecoilShakeAmplitudeScaling, PsoDataType.Float, 800, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLockOn, PsoDataType.Float, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForFineAimScaling, PsoDataType.Float, 808, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForFineAimScaling, PsoDataType.Float, 812, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleDeltaForLockOnSwitchBlendScaling, PsoDataType.Float, 816, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForInitialLockOn, PsoDataType.UInt, 820, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForInitialLockOn, PsoDataType.UInt, 824, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForLockOnSwitch, PsoDataType.UInt, 828, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForLockOnSwitch, PsoDataType.UInt, 832, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringConstant, PsoDataType.Float, 836, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringDampingRatio, PsoDataType.Float, 840, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringConstant, PsoDataType.Float, 844, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringDampingRatio, PsoDataType.Float, 848, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldFocusOnLockOnTarget, PsoDataType.Bool, 852, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFovToEmulateWithFocalLengthMultiplier, PsoDataType.Float, 856, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FocusParentToTargetBlendLevel, PsoDataType.Float, 860, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SecondaryFocusParentToTargetBlendLevel, PsoDataType.Float, 864, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFocusToSecondaryFocusDistance, PsoDataType.Float, 868, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnTargetStunnedEnvelopeRef, PsoDataType.String, 880, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnTargetDampingSettings, PsoDataType.Structure, 888, 0, MetaName.camThirdPersonPedAimCameraMetadataLockOnTargetDampingSettings),
                     new PsoStructureEntryInfo(MetaName.LockOnOrbitDistanceSettings, PsoDataType.Structure, 928, 0, MetaName.camThirdPersonPedAimCameraMetadataLockOnOrbitDistanceSettings),
                     new PsoStructureEntryInfo(MetaName.ParentRelativeAttachOffset, PsoDataType.Float3, 976, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParentRelativeAttachOffsetAtOrbitHeadingLimits, PsoDataType.Float3, 992, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachBoneTag, PsoDataType.SInt, 1008, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldScriptedAimTaskOverrideOrbitPitchLimits, PsoDataType.Bool, 1012, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAimSweepOverrideOrbitPitchLimits, PsoDataType.Bool, 1013, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsForMobilePhoneShallowDofMode, PsoDataType.String, 1016, 7, 0),
                     new PsoStructureEntryInfo((MetaName)309842601, PsoDataType.Float, 1020, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3770307856, PsoDataType.Bool, 1024, 0, 0),
                     new PsoStructureEntryInfo((MetaName)124410885, PsoDataType.Float, 1040, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1852506764, PsoDataType.Float, 1044, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1914946004, PsoDataType.Float, 1048, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2125876256, PsoDataType.Float, 1052, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3456376985, PsoDataType.Float, 1056, 0, 0),
                     new PsoStructureEntryInfo(MetaName.streamedMusc, PsoDataType.Float, 1060, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2791057952, PsoDataType.Float, 1064, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1336398630, PsoDataType.Bool, 1068, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1847527244, PsoDataType.Float, 1072, 0, 0),
                     new PsoStructureEntryInfo((MetaName)364456330, PsoDataType.Float, 1076, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3398320415, PsoDataType.Float, 1080, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2699503170, PsoDataType.Float, 1084, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2215910236, PsoDataType.Float, 1088, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2140597790, PsoDataType.Float, 1092, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2141354850, PsoDataType.Float, 1096, 0, 0),
                     new PsoStructureEntryInfo((MetaName)9125054, PsoDataType.Float, 1100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)341585039, PsoDataType.Float, 1104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3090152422, PsoDataType.Float, 1108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3518102089, PsoDataType.Float, 1112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2714979879, PsoDataType.Float, 1116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4047322117, PsoDataType.Float, 1120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)10419895, PsoDataType.Float, 1124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)817394632, PsoDataType.Float2, 1128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3793303691, PsoDataType.Bool, 1136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)891610436, PsoDataType.Float, 1140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1332592052, PsoDataType.Float, 1144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3712243434, PsoDataType.Float, 1148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2126717159, PsoDataType.Float, 1152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)467663222, PsoDataType.Float, 1156, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3920880465, PsoDataType.Float, 1160, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1407257659, PsoDataType.UInt, 1164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1350298358, PsoDataType.Structure, 1168, 0, MetaName.camThirdPersonPedAimInCoverCameraMetadataLowCoverSettings),
                     new PsoStructureEntryInfo(MetaName.AimingSettings, PsoDataType.Structure, 1208, 0, MetaName.camThirdPersonPedAimInCoverCameraMetadataAimingSettings)
                    );
                case MetaName.camThirdPersonPedAimInCoverCameraMetadataLowCoverSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAimInCoverCameraMetadataLowCoverSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.BlendInSpringConstant, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2811394796, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutSpringConstant, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)388057725, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4026985674, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2972801644, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camThirdPersonPedAimInCoverCameraMetadataAimingSettings:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedAimInCoverCameraMetadataAimingSettings, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ShouldApply, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3794685898, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3031220740, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)810186943, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2302950975, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camThirdPersonPedMeleeAimCameraMetadata:
                    return new PsoStructureInfo(MetaName.camThirdPersonPedMeleeAimCameraMetadata, 0, 0, 1088,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpHelperRef, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseAttachVelocityToIgnoreEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305154880, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldOrbitRelativeToAttachParentOrientation, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistOrbitOrientationRelativeToAttachParent, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringConstant, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringDampingRatio, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406229624, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727876548, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSubmergedLevelToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftGroundSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftContactSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldTimeToBlockFullAttachParentMatrixForRelativeOrbit, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSpeedToClonePitchFromCinematicMountedCameras, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomFramingInTightSpace, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachSpeedToUpdateTightSpaceLevel, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringConstant, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsInTightSpace, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToParentCentreOfGravity, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDynamicCentreOfGravity, PsoDataType.Bool, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVelocityOfAttachParentAttachEntity, PsoDataType.Bool, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomBoundingBoxSettings, PsoDataType.Structure, 128, 0, MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachPedPelvisOffset, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotPosition, PsoDataType.Structure, 176, 0, MetaName.camThirdPersonCameraMetadataBasePivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotPosition, PsoDataType.Structure, 272, 0, MetaName.camThirdPersonCameraMetadataPivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotOverBoundingBoxSettings, PsoDataType.Structure, 320, 0, MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotHeightScalingForFootRoom, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVerticalPivotOffsetForFootRoom, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeOrbitDistanceScalingForExtensions, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)560700931, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomOrbitDistanceLimitsToForce, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringConstant, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringDampingRatio, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingForCustomFirstPersonFallBack, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 400, 0, MetaName.camThirdPersonCameraMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithAttachParent, PsoDataType.Bool, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithFollowVehicle, PsoDataType.Bool, 449, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionOrigin, PsoDataType.Bool, 450, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionRoot, PsoDataType.Bool, 451, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringConstant, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringDampingRatio, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomCollisionOriginRelativePosition, PsoDataType.Float3, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4194152221, PsoDataType.Float3, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomCollisionOrigin, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1998689777, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235540893, PsoDataType.Bool, 514, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3479639237, PsoDataType.Float3, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496115250, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionFallBackPosition, PsoDataType.Structure, 552, 0, MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionFallBackToPivotBlendValue, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConstrainCollisionRootPositionAgainstClippingTypes, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringConstant, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringDampingRatio, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOnTopOfVehicleCollisionSettings, PsoDataType.Structure, 608, 0, MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.IdealHeadingOffsetForLimiting, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffsetInTightSpace, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookOverSettings, PsoDataType.Structure, 664, 0, MetaName.camThirdPersonCameraMetadataLookOverSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentPitchForLookBehind, PsoDataType.Bool, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitsForBasePosition, PsoDataType.Float2, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreToPostCollisionLookAtOrientationBlendValue, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringConstant, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringDampingRatio, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchSoftLimits, PsoDataType.Float2, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchHardLimits, PsoDataType.Float2, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthZoomSettings, PsoDataType.Structure, 728, 0, MetaName.camThirdPersonCameraMetadataStealthZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1947299079, PsoDataType.Structure, 752, 0, MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 776, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2122226771, PsoDataType.Bool, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnEnvelopeRef, PsoDataType.String, 784, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 788, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticuleDuringInterpolation, PsoDataType.Bool, 789, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAllowInterpolationSourceCameraToPersistReticule, PsoDataType.Bool, 790, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyWeaponFov, PsoDataType.Bool, 791, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLockOnAiming, PsoDataType.Bool, 792, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockOnToTargetEntityPosition, PsoDataType.Bool, 793, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldValidateLockOnTargetPosition, PsoDataType.Bool, 794, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TripleHeadNearClip, PsoDataType.Float, 796, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecoilShakeAmplitudeScaling, PsoDataType.Float, 800, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLockOn, PsoDataType.Float, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForFineAimScaling, PsoDataType.Float, 808, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForFineAimScaling, PsoDataType.Float, 812, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleDeltaForLockOnSwitchBlendScaling, PsoDataType.Float, 816, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForInitialLockOn, PsoDataType.UInt, 820, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForInitialLockOn, PsoDataType.UInt, 824, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForLockOnSwitch, PsoDataType.UInt, 828, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForLockOnSwitch, PsoDataType.UInt, 832, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringConstant, PsoDataType.Float, 836, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringDampingRatio, PsoDataType.Float, 840, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringConstant, PsoDataType.Float, 844, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringDampingRatio, PsoDataType.Float, 848, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldFocusOnLockOnTarget, PsoDataType.Bool, 852, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFovToEmulateWithFocalLengthMultiplier, PsoDataType.Float, 856, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FocusParentToTargetBlendLevel, PsoDataType.Float, 860, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SecondaryFocusParentToTargetBlendLevel, PsoDataType.Float, 864, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFocusToSecondaryFocusDistance, PsoDataType.Float, 868, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnTargetStunnedEnvelopeRef, PsoDataType.String, 880, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnTargetDampingSettings, PsoDataType.Structure, 888, 0, MetaName.camThirdPersonPedAimCameraMetadataLockOnTargetDampingSettings),
                     new PsoStructureEntryInfo(MetaName.LockOnOrbitDistanceSettings, PsoDataType.Structure, 928, 0, MetaName.camThirdPersonPedAimCameraMetadataLockOnOrbitDistanceSettings),
                     new PsoStructureEntryInfo(MetaName.ParentRelativeAttachOffset, PsoDataType.Float3, 976, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParentRelativeAttachOffsetAtOrbitHeadingLimits, PsoDataType.Float3, 992, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachBoneTag, PsoDataType.SInt, 1008, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldScriptedAimTaskOverrideOrbitPitchLimits, PsoDataType.Bool, 1012, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAimSweepOverrideOrbitPitchLimits, PsoDataType.Bool, 1013, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsForMobilePhoneShallowDofMode, PsoDataType.String, 1016, 7, 0),
                     new PsoStructureEntryInfo((MetaName)309842601, PsoDataType.Float, 1020, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3770307856, PsoDataType.Bool, 1024, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3195731427, PsoDataType.Float, 1040, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2888452072, PsoDataType.Float2, 1044, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2501102835, PsoDataType.Float, 1052, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2903662656, PsoDataType.Float, 1056, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2037218535, PsoDataType.Float, 1060, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4219845621, PsoDataType.Float, 1064, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3549368383, PsoDataType.Float, 1068, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1000603972, PsoDataType.Float, 1072, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1991971403, PsoDataType.Float, 1076, 0, 0)
                    );
                case MetaName.camThirdPersonVehicleAimCameraMetadata:
                    return new PsoStructureInfo(MetaName.camThirdPersonVehicleAimCameraMetadata, 0, 0, 896,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlHelperRef, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HintHelperRef, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpHelperRef, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseAttachVelocityToIgnoreEnvelopeRef, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFov, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseNearClip, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3305154880, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldOrbitRelativeToAttachParentOrientation, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPersistOrbitOrientationRelativeToAttachParent, PsoDataType.Bool, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringConstant, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentMatrixForRelativeOrbitSpringDampingRatio, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406229624, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)727876548, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSubmergedLevelToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftGroundSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAircraftContactSpeedToApplyFullAttachParentMatrixForRelativeOrbit, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldTimeToBlockFullAttachParentMatrixForRelativeOrbit, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAttachParentSpeedToClonePitchFromCinematicMountedCameras, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomFramingInTightSpace, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAttachSpeedToUpdateTightSpaceLevel, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringConstant, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TightSpaceSpringDampingRatio, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettingsInTightSpace, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAttachToParentCentreOfGravity, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseDynamicCentreOfGravity, PsoDataType.Bool, 121, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVelocityOfAttachParentAttachEntity, PsoDataType.Bool, 122, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomBoundingBoxSettings, PsoDataType.Structure, 128, 0, MetaName.camThirdPersonCameraMetadataCustomBoundingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyAttachPedPelvisOffset, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringConstant, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachPedPelvisOffsetSpringDampingRatio, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotPosition, PsoDataType.Structure, 176, 0, MetaName.camThirdPersonCameraMetadataBasePivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotPosition, PsoDataType.Structure, 272, 0, MetaName.camThirdPersonCameraMetadataPivotPosition),
                     new PsoStructureEntryInfo(MetaName.PivotOverBoundingBoxSettings, PsoDataType.Structure, 320, 0, MetaName.camThirdPersonCameraMetadataPivotOverBoungingBoxSettings),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoom, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoom, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMinFootRoomInTightSpace, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScreenRatioForMaxFootRoomInTightSpace, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasePivotHeightScalingForFootRoom, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreVerticalPivotOffsetForFootRoom, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeOrbitDistanceScalingForExtensions, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)560700931, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomOrbitDistanceLimitsToForce, PsoDataType.Float2, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringConstant, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitSpringDampingRatio, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceScalingForCustomFirstPersonFallBack, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BuoyancySettings, PsoDataType.Structure, 400, 0, MetaName.camThirdPersonCameraMetadataBuoyancySettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithAttachParent, PsoDataType.Bool, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreCollisionWithFollowVehicle, PsoDataType.Bool, 449, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionOrigin, PsoDataType.Bool, 450, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreFollowVehicleForCollisionRoot, PsoDataType.Bool, 451, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringConstant, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionTestRadiusSpringDampingRatio, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomCollisionOriginRelativePosition, PsoDataType.Float3, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4194152221, PsoDataType.Float3, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseCustomCollisionOrigin, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1998689777, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235540893, PsoDataType.Bool, 514, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3479639237, PsoDataType.Float3, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1496115250, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionFallBackPosition, PsoDataType.Structure, 552, 0, MetaName.camThirdPersonCameraMetadataCollisionFallBackPosition),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionFallBackToPivotBlendValue, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldConstrainCollisionRootPositionAgainstClippingTypes, PsoDataType.Bool, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringConstant, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRootPositionSpringDampingRatio, PsoDataType.Float, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOnTopOfVehicleCollisionSettings, PsoDataType.Structure, 608, 0, MetaName.camThirdPersonCameraMetadataVehicleOnTopOfVehicleCollisionSettings),
                     new PsoStructureEntryInfo(MetaName.IdealHeadingOffsetForLimiting, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeOrbitHeadingLimits, PsoDataType.Float2, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitPitchLimits, PsoDataType.Float2, 644, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffset, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseOrbitPitchOffsetInTightSpace, PsoDataType.Float, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookOverSettings, PsoDataType.Structure, 664, 0, MetaName.camThirdPersonCameraMetadataLookOverSettings),
                     new PsoStructureEntryInfo(MetaName.ShouldIgnoreAttachParentPitchForLookBehind, PsoDataType.Bool, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OrbitDistanceLimitsForBasePosition, PsoDataType.Float2, 692, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreToPostCollisionLookAtOrientationBlendValue, PsoDataType.Float, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringConstant, PsoDataType.Float, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollSpringDampingRatio, PsoDataType.Float, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchSoftLimits, PsoDataType.Float2, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRollDampingPitchHardLimits, PsoDataType.Float2, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthZoomSettings, PsoDataType.Structure, 728, 0, MetaName.camThirdPersonCameraMetadataStealthZoomSettings),
                     new PsoStructureEntryInfo((MetaName)1947299079, PsoDataType.Structure, 752, 0, MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring),
                     new PsoStructureEntryInfo(MetaName.MotionBlurSettings, PsoDataType.String, 776, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2122226771, PsoDataType.Bool, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnEnvelopeRef, PsoDataType.String, 784, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticule, PsoDataType.Bool, 788, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldDisplayReticuleDuringInterpolation, PsoDataType.Bool, 789, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldAllowInterpolationSourceCameraToPersistReticule, PsoDataType.Bool, 790, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldApplyWeaponFov, PsoDataType.Bool, 791, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLockOnAiming, PsoDataType.Bool, 792, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLockOnToTargetEntityPosition, PsoDataType.Bool, 793, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldValidateLockOnTargetPosition, PsoDataType.Bool, 794, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TripleHeadNearClip, PsoDataType.Float, 796, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecoilShakeAmplitudeScaling, PsoDataType.Float, 800, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLockOn, PsoDataType.Float, 804, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForFineAimScaling, PsoDataType.Float, 808, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForFineAimScaling, PsoDataType.Float, 812, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleDeltaForLockOnSwitchBlendScaling, PsoDataType.Float, 816, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForInitialLockOn, PsoDataType.UInt, 820, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForInitialLockOn, PsoDataType.UInt, 824, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendDurationForLockOnSwitch, PsoDataType.UInt, 828, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBlendDurationForLockOnSwitch, PsoDataType.UInt, 832, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringConstant, PsoDataType.Float, 836, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimBlendSpringDampingRatio, PsoDataType.Float, 840, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringConstant, PsoDataType.Float, 844, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponZoomFactorSpringDampingRatio, PsoDataType.Float, 848, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldFocusOnLockOnTarget, PsoDataType.Bool, 852, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseFovToEmulateWithFocalLengthMultiplier, PsoDataType.Float, 856, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FocusParentToTargetBlendLevel, PsoDataType.Float, 860, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SecondaryFocusParentToTargetBlendLevel, PsoDataType.Float, 864, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFocusToSecondaryFocusDistance, PsoDataType.Float, 868, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLockOnAimingForDriver, PsoDataType.Bool, 880, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldUseLockOnAimingForPassenger, PsoDataType.Bool, 881, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraSideOffsetForHangingOnLeftSide, PsoDataType.Float, 884, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraSideOffsetForHangingOnRightSide, PsoDataType.Float, 888, 0, 0)
                    );
                case MetaName.camMarketingFreeCameraMetadata:
                    return new PsoStructureInfo(MetaName.camMarketingFreeCameraMetadata, 0, 0, 200,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3808044930, PsoDataType.Structure, 32, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2925373184, PsoDataType.Structure, 64, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)31061142, PsoDataType.Structure, 96, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3295350072, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1591024915, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)829837985, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3177251856, PsoDataType.Structure, 144, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3141690306, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4157311158, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1989308054, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 192, 0, 0)
                    );
                case MetaName.camMarketingFreeCameraMetadataInputResponse:
                    return new PsoStructureInfo(MetaName.camMarketingFreeCameraMetadataInputResponse, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)3855676319, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InputMagPowerFactor, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Acceleration, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Deceleration, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeed, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.camMarketingAToBCameraMetadata:
                    return new PsoStructureInfo(MetaName.camMarketingAToBCameraMetadata, 0, 0, 216,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3808044930, PsoDataType.Structure, 32, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2925373184, PsoDataType.Structure, 64, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)31061142, PsoDataType.Structure, 96, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3295350072, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1591024915, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)829837985, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3177251856, PsoDataType.Structure, 144, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3141690306, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4157311158, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1989308054, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1044888659, PsoDataType.String, 200, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1147067733, PsoDataType.UInt, 204, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1825469896, PsoDataType.UInt, 208, 0, 0)
                    );
                case MetaName.camTimedSplineCameraMetadata:
                    return new PsoStructureInfo(MetaName.camTimedSplineCameraMetadata, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFov, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3867080895, PsoDataType.SInt, 40, 0, 0)
                    );
                case MetaName.camMarketingOrbitCameraMetadata:
                    return new PsoStructureInfo(MetaName.camMarketingOrbitCameraMetadata, 0, 0, 208,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3808044930, PsoDataType.Structure, 32, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2925373184, PsoDataType.Structure, 64, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)31061142, PsoDataType.Structure, 96, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3295350072, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1591024915, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)829837985, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3177251856, PsoDataType.Structure, 144, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3141690306, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4157311158, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1989308054, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2308368369, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringMountRef, PsoDataType.String, 204, 7, 0)
                    );
                case MetaName.camMarketingMountedCameraMetadata:
                    return new PsoStructureInfo(MetaName.camMarketingMountedCameraMetadata, 0, 0, 208,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3808044930, PsoDataType.Structure, 32, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2925373184, PsoDataType.Structure, 64, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)31061142, PsoDataType.Structure, 96, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3295350072, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1591024915, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)829837985, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3177251856, PsoDataType.Structure, 144, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3141690306, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4157311158, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1989308054, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2308368369, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringMountRef, PsoDataType.String, 204, 7, 0)
                    );
                case MetaName.camMarketingStickyCameraMetadata:
                    return new PsoStructureInfo(MetaName.camMarketingStickyCameraMetadata, 0, 0, 200,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3808044930, PsoDataType.Structure, 32, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2925373184, PsoDataType.Structure, 64, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)31061142, PsoDataType.Structure, 96, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3295350072, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1591024915, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)829837985, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3177251856, PsoDataType.Structure, 144, 0, MetaName.camMarketingFreeCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3141690306, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4157311158, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1989308054, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 192, 0, 0)
                    );
                case MetaName.camScriptedCameraMetadata:
                    return new PsoStructureInfo(MetaName.camScriptedCameraMetadata, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFov, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camScriptedFlyCameraMetadata:
                    return new PsoStructureInfo(MetaName.camScriptedFlyCameraMetadata, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFov, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1160500952, PsoDataType.Structure, 40, 0, MetaName.camScriptedFlyCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2802203540, PsoDataType.Structure, 64, 0, MetaName.camScriptedFlyCameraMetadataInputResponse),
                     new PsoStructureEntryInfo(MetaName.DefaultPitch, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1176855075, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2736468652, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeight, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CapsuleRadius, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveWater, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2779060394, PsoDataType.UInt, 112, 0, 0)
                    );
                case MetaName.camScriptedFlyCameraMetadataInputResponse:
                    return new PsoStructureInfo(MetaName.camScriptedFlyCameraMetadataInputResponse, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.InputMagPowerFactor, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAcceleration, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDeceleration, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeed, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camCustomTimedSplineCameraMetadata:
                    return new PsoStructureInfo(MetaName.camCustomTimedSplineCameraMetadata, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFov, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3867080895, PsoDataType.SInt, 40, 0, 0)
                    );
                case MetaName.camRoundedSplineCameraMetadata:
                    return new PsoStructureInfo(MetaName.camRoundedSplineCameraMetadata, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFov, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camSmoothedSplineCameraMetadata:
                    return new PsoStructureInfo(MetaName.camSmoothedSplineCameraMetadata, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFov, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3867080895, PsoDataType.SInt, 40, 0, 0)
                    );
                case MetaName.camSwitchCameraMetadata:
                    return new PsoStructureInfo(MetaName.camSwitchCameraMetadata, 0, 0, 144,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1245276384, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2105074563, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Fov, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NearClip, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MotionBlurStrength, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1056782751, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)341706306, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1176855075, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2303904065, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)978298090, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultPitch, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771807090, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1215756626, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1181521957, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2559186523, PsoDataType.UInt, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1045169360, PsoDataType.UInt, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1298516568, PsoDataType.UInt, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3976211620, PsoDataType.UInt, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3193297650, PsoDataType.UInt, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3466836211, PsoDataType.UInt, 132, 0, 0)
                    );
                case MetaName.camReplayRecordedCameraMetadata:
                    return new PsoStructureInfo(MetaName.camReplayRecordedCameraMetadata, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camReplayPresetCameraMetadata:
                    return new PsoStructureInfo(MetaName.camReplayPresetCameraMetadata, 0, 0, 256,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3243156997, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RelativeLookAtPosition, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3869518259, PsoDataType.Structure, 80, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)31061142, PsoDataType.Structure, 104, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3177251856, PsoDataType.Structure, 128, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo(MetaName.MinFov, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxFov, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFov, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo((MetaName)777402839, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1557834870, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1089474876, PsoDataType.Bool, 172, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2480397790, PsoDataType.Bool, 173, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionSettings, PsoDataType.Structure, 176, 0, MetaName.camReplayBaseCameraMetadataCollisionSettings),
                     new PsoStructureEntryInfo((MetaName)549109147, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3394194196, PsoDataType.UInt, 228, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2704093958, PsoDataType.UInt, 232, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1677933530, PsoDataType.UInt, 236, 0, 0),
                     new PsoStructureEntryInfo((MetaName)343964264, PsoDataType.UInt, 240, 0, 0),
                     new PsoStructureEntryInfo((MetaName)681964647, PsoDataType.UInt, 244, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2782721869, PsoDataType.UInt, 248, 0, 0)
                    );
                case MetaName.camReplayBaseCameraMetadataInputResponse:
                    return new PsoStructureInfo(MetaName.camReplayBaseCameraMetadataInputResponse, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.InputMagPowerFactor, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAcceleration, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDeceleration, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeed, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camReplayBaseCameraMetadataCollisionSettings:
                    return new PsoStructureInfo(MetaName.camReplayBaseCameraMetadataCollisionSettings, 0, 0, 48,
                     new PsoStructureEntryInfo((MetaName)2162401385, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WaterHeightSmoothRate, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSafeRadiusReductionWithinPedMoverCapsule, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)653804618, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3232040475, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)819688366, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)122123361, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)557881422, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldPushBeyondAttachParentIfClipping, PsoDataType.Bool, 40, 0, 0)
                    );
                case MetaName.camReplayFreeCameraMetadata:
                    return new PsoStructureInfo(MetaName.camReplayFreeCameraMetadata, 0, 0, 544,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionRef, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DofSettings, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCollisionTestRadius, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NearClip, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1160500952, PsoDataType.Structure, 48, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2802203540, PsoDataType.Structure, 72, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2292446130, PsoDataType.Structure, 96, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)31061142, PsoDataType.Structure, 120, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)3177251856, PsoDataType.Structure, 144, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)1176189471, PsoDataType.Structure, 168, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)4043692651, PsoDataType.Structure, 192, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)2391928883, PsoDataType.Structure, 216, 0, MetaName.camReplayBaseCameraMetadataInputResponse),
                     new PsoStructureEntryInfo((MetaName)1466164024, PsoDataType.Structure, 240, 0, MetaName.camInterpolatorMetadata),
                     new PsoStructureEntryInfo((MetaName)3699879509, PsoDataType.Structure, 280, 0, MetaName.camInterpolatorMetadata),
                     new PsoStructureEntryInfo((MetaName)2461431150, PsoDataType.Structure, 320, 0, MetaName.camInterpolatorMetadata),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFov, PsoDataType.Float, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxFov, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFov, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CapsuleRadius, PsoDataType.Float, 376, 0, 0),
                     new PsoStructureEntryInfo((MetaName)537337364, PsoDataType.Float, 380, 0, 0),
                     new PsoStructureEntryInfo((MetaName)870756038, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1068739939, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3629360459, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo((MetaName)479683481, PsoDataType.Float, 396, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2077539523, PsoDataType.Float3, 400, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionSettings, PsoDataType.Structure, 416, 0, MetaName.camReplayBaseCameraMetadataCollisionSettings),
                     new PsoStructureEntryInfo((MetaName)4113213041, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2101093733, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2090863388, PsoDataType.Float, 472, 0, 0),
                     new PsoStructureEntryInfo((MetaName)109690208, PsoDataType.Float, 476, 0, 0),
                     new PsoStructureEntryInfo((MetaName)425553139, PsoDataType.Float, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2285453310, PsoDataType.UInt, 484, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3186336829, PsoDataType.Float, 488, 0, 0),
                     new PsoStructureEntryInfo((MetaName)858986389, PsoDataType.Float, 492, 0, 0),
                     new PsoStructureEntryInfo((MetaName)168114371, PsoDataType.Float, 496, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2257617130, PsoDataType.Float, 500, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4119877675, PsoDataType.Float, 504, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2969211614, PsoDataType.UInt, 508, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3394194196, PsoDataType.UInt, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2704093958, PsoDataType.UInt, 516, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1677933530, PsoDataType.UInt, 520, 0, 0),
                     new PsoStructureEntryInfo((MetaName)343964264, PsoDataType.UInt, 524, 0, 0),
                     new PsoStructureEntryInfo((MetaName)681964647, PsoDataType.UInt, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2782721869, PsoDataType.UInt, 532, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3026527735, PsoDataType.Float, 536, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1272604977, PsoDataType.Float, 540, 0, 0)
                    );
                case MetaName.camShakeMetadata:
                    return new PsoStructureInfo(MetaName.camShakeMetadata, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1260823161, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1428204256, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3480173900, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camShakeMetadataFrameComponent),
                     new PsoStructureEntryInfo(MetaName.FrameComponents, PsoDataType.Array, 24, 0, (MetaName)4),
                     new PsoStructureEntryInfo(MetaName.Vibration, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.camShakeMetadataFrameComponent:
                    return new PsoStructureInfo(MetaName.camShakeMetadataFrameComponent, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Component, PsoDataType.Enum, 8, 0, (MetaName)4206293824),
                     new PsoStructureEntryInfo((MetaName)3866719227, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnvelopeRef, PsoDataType.String, 16, 7, 0)
                    );
                case MetaName.camOscillatorMetadata:
                    return new PsoStructureInfo(MetaName.camOscillatorMetadata, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Waveform, PsoDataType.Enum, 16, 0, (MetaName)902027429),
                     new PsoStructureEntryInfo(MetaName.Amplitude, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Frequency, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.phase, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2974245218, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Decay, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.camAnimatedShakeMetadata:
                    return new PsoStructureInfo(MetaName.camAnimatedShakeMetadata, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1260823161, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1428204256, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3480173900, PsoDataType.Bool, 21, 0, 0)
                    );
                case MetaName.camCinematicCameraManShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicCameraManShotMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 64, 0, 0)
                    );
                case MetaName.camCinematicCraningCameraManShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicCraningCameraManShotMetadata, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)934463872, PsoDataType.Bool, 65, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2227467637, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)943740099, PsoDataType.Bool, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1606540295, PsoDataType.UInt, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3516181833, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3249173046, PsoDataType.Float2, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2987455099, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScanRadius, PsoDataType.Float, 96, 0, 0)
                    );
                case MetaName.camCinematicInVehicleCrashShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicInVehicleCrashShotMetadata, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRoll, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3669617802, PsoDataType.UInt, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3100384106, PsoDataType.UInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelocity, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3730987855, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4150416286, PsoDataType.UInt, 88, 0, 0)
                    );
                case MetaName.camCinematicVehiclePartShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehiclePartShotMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)233780411, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicHeliTrackingShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicHeliTrackingShotMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 64, 0, 0)
                    );
                case MetaName.camCinematicVehicleOrbitShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleOrbitShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicVehicleLowOrbitShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleLowOrbitShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicTrainRoofMountedShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicTrainRoofMountedShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicTrainStationShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicTrainStationShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicTrainPassengerShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicTrainPassengerShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicTrainTrackShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicTrainTrackShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicPoliceCarMountedShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicPoliceCarMountedShotMetadata, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LimitAttachParentRelativePitchAndHeading, PsoDataType.Bool, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldTerminateForPitchAndHeading, PsoDataType.Bool, 65, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRelativePitch, PsoDataType.Float2, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachParentRelativeHeading, PsoDataType.Float2, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialRelativePitchLimits, PsoDataType.Float2, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialRelativeHeadingLimits, PsoDataType.Float2, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InVehicleLookAtDampingRef, PsoDataType.String, 100, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OnFootLookAtDampingRef, PsoDataType.String, 104, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 108, 0, 0)
                    );
                case MetaName.camCinematicPoliceHeliMountedShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicPoliceHeliMountedShotMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 64, 0, 0)
                    );
                case MetaName.camCinematicPoliceInCoverShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicPoliceInCoverShotMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 64, 0, 0)
                    );
                case MetaName.camCinematicPoliceRoadBlockShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicPoliceRoadBlockShotMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)876494977, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpreadLimit, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 72, 0, 0)
                    );
                case MetaName.camCinematicPoliceExitVehicleShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicPoliceExitVehicleShotMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4197822931, PsoDataType.Bool, 64, 0, 0)
                    );
                case MetaName.camCinematicOnFootIdleShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicOnFootIdleShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicOnFootFirstPersonIdleShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicOnFootFirstPersonIdleShotMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Cameras, PsoDataType.Array, 64, 0, (MetaName)14)
                    );
                case MetaName.camCinematicStuntJumpShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicStuntJumpShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicParachuteHeliShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicParachuteHeliShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicParachuteCameraManShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicParachuteCameraManShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicOnFootAssistedAimingKillShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicOnFootAssistedAimingKillShotMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1965493319, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)513375599, PsoDataType.Float, 68, 0, 0)
                    );
                case MetaName.camCinematicOnFootMeleeShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicOnFootMeleeShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicVehicleBonnetShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleBonnetShotMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1109718587, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4170146926, PsoDataType.UInt, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2661334408, PsoDataType.String, 72, 7, 0)
                    );
                case MetaName.camCinematicBustedShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicBustedShotMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3180920449, PsoDataType.Float, 64, 0, 0)
                    );
                case MetaName.camCinematicVehicleGroupShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleGroupShotMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)189014933, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3490628210, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)456924361, PsoDataType.UInt, 72, 0, 0)
                    );
                case MetaName.camCinematicMissileKillShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicMissileKillShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicWaterCrashShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicWaterCrashShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicFallFromHeliShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicFallFromHeliShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicVehicleConvertibleRoofShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicVehicleConvertibleRoofShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camCinematicScriptRaceCheckPointShotMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicScriptRaceCheckPointShotMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3727857604, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2997337835, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3529626013),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 32, 0, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)841129953, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771050581, PsoDataType.Bool, 37, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2973786865, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1480107235, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2460595867, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3700663941, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2275630497, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60485337, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camLongSwoopSwitchHelperMetadata:
                    return new PsoStructureInfo(MetaName.camLongSwoopSwitchHelperMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2578467489, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1608990953, PsoDataType.Enum, 24, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)2608832721, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2368120393, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3903914248, PsoDataType.Bool, 33, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1595139928, PsoDataType.Bool, 34, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1489449044, PsoDataType.Bool, 35, 0, 0),
                     new PsoStructureEntryInfo((MetaName)796068415, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)417423151, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3958574934, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2939152164, PsoDataType.Enum, 52, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)2707662935, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)26423852, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)322264486, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2868975511, PsoDataType.Enum, 68, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)3659327390, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)833330146, PsoDataType.Float, 76, 0, 0)
                    );
                case MetaName.camShortZoomToHeadSwitchHelperMetadata:
                    return new PsoStructureInfo(MetaName.camShortZoomToHeadSwitchHelperMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2578467489, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1608990953, PsoDataType.Enum, 24, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)2608832721, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2368120393, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3903914248, PsoDataType.Bool, 33, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1595139928, PsoDataType.Bool, 34, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1489449044, PsoDataType.Bool, 35, 0, 0),
                     new PsoStructureEntryInfo((MetaName)864831122, PsoDataType.Bool, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3573176421, PsoDataType.SInt, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1837935839, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2395979106, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4037747409, PsoDataType.Float, 68, 0, 0)
                    );
                case MetaName.camShortRotationSwitchHelperMetadata:
                    return new PsoStructureInfo(MetaName.camShortRotationSwitchHelperMetadata, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2578467489, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1608990953, PsoDataType.Enum, 24, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)2608832721, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2368120393, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3903914248, PsoDataType.Bool, 33, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1595139928, PsoDataType.Bool, 34, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1489449044, PsoDataType.Bool, 35, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2291716575, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.camShortTranslationSwitchHelperMetadata:
                    return new PsoStructureInfo(MetaName.camShortTranslationSwitchHelperMetadata, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2578467489, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1608990953, PsoDataType.Enum, 24, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)2608832721, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2368120393, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3903914248, PsoDataType.Bool, 33, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1595139928, PsoDataType.Bool, 34, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1489449044, PsoDataType.Bool, 35, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3634726737, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.camShortZoomInOutSwitchHelperMetadata:
                    return new PsoStructureInfo(MetaName.camShortZoomInOutSwitchHelperMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2578467489, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1608990953, PsoDataType.Enum, 24, 0, MetaName.eCurveType),
                     new PsoStructureEntryInfo((MetaName)2608832721, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2368120393, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3903914248, PsoDataType.Bool, 33, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1595139928, PsoDataType.Bool, 34, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1489449044, PsoDataType.Bool, 35, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1190587282, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)547488153, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2504913741, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)26423852, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4037747409, PsoDataType.Float, 56, 0, 0)
                    );
                case MetaName.camNearClipScannerMetadata:
                    return new PsoStructureInfo(MetaName.camNearClipScannerMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)494178060, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1713774083, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3006651453, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1502537267, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3831489261, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1103532228, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)639510975, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1947972687, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1176855075, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)529344496, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1434632498, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3130208423, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1525295652, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)304640727, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2477079271, PsoDataType.Float, 72, 0, 0)
                    );
                case MetaName.camVehicleCustomSettingsMetadata:
                    return new PsoStructureInfo(MetaName.camVehicleCustomSettingsMetadata, 0, 0, 192,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DoorAlignmentSettings, PsoDataType.Structure, 16, 0, MetaName.camVehicleCustomSettingsMetadataDoorAlignmentSettings),
                     new PsoStructureEntryInfo((MetaName)2728784918, PsoDataType.Structure, 56, 0, MetaName.camVehicleCustomSettingsMetadataExitSeatPhaseForCameraExitSettings),
                     new PsoStructureEntryInfo((MetaName)556143364, PsoDataType.Structure, 72, 0, MetaName.camVehicleCustomSettingsMetadataMultiplayerPassengerCameraHashSettings),
                     new PsoStructureEntryInfo((MetaName)981761687, PsoDataType.Structure, 104, 0, MetaName.camVehicleCustomSettingsMetadataInvalidCinematcShotsRefsForVehicleSettings),
                     new PsoStructureEntryInfo((MetaName)955775077, PsoDataType.Structure, 136, 0, MetaName.camVehicleCustomSettingsMetadataAdditionalBoundScalingVehicleSettings),
                     new PsoStructureEntryInfo((MetaName)3842334374, PsoDataType.Structure, 152, 0, MetaName.camVehicleCustomSettingsMetadataSeatSpecficCameras),
                     new PsoStructureEntryInfo((MetaName)3299186532, PsoDataType.Bool, 184, 0, 0)
                    );
                case MetaName.camVehicleCustomSettingsMetadataExitSeatPhaseForCameraExitSettings:
                    return new PsoStructureInfo(MetaName.camVehicleCustomSettingsMetadataExitSeatPhaseForCameraExitSettings, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderData, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2549072274, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camVehicleCustomSettingsMetadataMultiplayerPassengerCameraHashSettings:
                    return new PsoStructureInfo(MetaName.camVehicleCustomSettingsMetadataMultiplayerPassengerCameraHashSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderData, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)109781807, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2400069518, PsoDataType.Array, 16, 0, (MetaName)2)
                    );
                case MetaName.camVehicleCustomSettingsMetadataInvalidCinematcShotsRefsForVehicleSettings:
                    return new PsoStructureInfo(MetaName.camVehicleCustomSettingsMetadataInvalidCinematcShotsRefsForVehicleSettings, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderData, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3579087558, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case MetaName.camVehicleCustomSettingsMetadataAdditionalBoundScalingVehicleSettings:
                    return new PsoStructureInfo(MetaName.camVehicleCustomSettingsMetadataAdditionalBoundScalingVehicleSettings, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderData, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeightScaling, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camVehicleCustomSettingsMetadataSeatSpecficCameras:
                    return new PsoStructureInfo(MetaName.camVehicleCustomSettingsMetadataSeatSpecficCameras, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ShouldConsiderData, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camSeatSpecificCameras),
                     new PsoStructureEntryInfo((MetaName)1517635298, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case MetaName.camSeatSpecificCameras:
                    return new PsoStructureInfo(MetaName.camSeatSpecificCameras, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.SeatIndex, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1261448360, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PovCameraOffset, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2254474492, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)273799675, PsoDataType.Bool, 36, 0, 0)
                    );
                case MetaName.camGameplayDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camGameplayDirectorMetadata, 0, 0, 608,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2089690180, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo((MetaName)519684035, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4121580243, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3253949128, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3345137885, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2365172496, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ThirdPersonVehicleAimCameraRef, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3734793071, PsoDataType.String, 52, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2936418982, PsoDataType.String, 56, 7, 0),
                     new PsoStructureEntryInfo((MetaName)595399339, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4266955763, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1070442381, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo((MetaName)838085564, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1252415609, PsoDataType.String, 76, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2466448846, PsoDataType.String, 80, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3845919335, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3326977311, PsoDataType.String, 88, 7, 0),
                     new PsoStructureEntryInfo((MetaName)374047020, PsoDataType.String, 92, 7, 0),
                     new PsoStructureEntryInfo((MetaName)859347369, PsoDataType.String, 96, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1384877759, PsoDataType.String, 100, 7, 0),
                     new PsoStructureEntryInfo((MetaName)399013837, PsoDataType.String, 104, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3929905827, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2078537289, PsoDataType.String, 112, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2211994659, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1839344859, PsoDataType.String, 120, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1168367025, PsoDataType.String, 124, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3316487821, PsoDataType.String, 128, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2772847581, PsoDataType.String, 132, 7, 0),
                     new PsoStructureEntryInfo((MetaName)766173022, PsoDataType.String, 136, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2535543863, PsoDataType.String, 140, 7, 0),
                     new PsoStructureEntryInfo((MetaName)160927927, PsoDataType.String, 144, 7, 0),
                     new PsoStructureEntryInfo((MetaName)478304411, PsoDataType.String, 148, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2995820414, PsoDataType.String, 152, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2615323901, PsoDataType.String, 156, 7, 0),
                     new PsoStructureEntryInfo((MetaName)773221014, PsoDataType.UInt, 160, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1041806552, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2032331076, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1446696509, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3986662918, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)508043832, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2976021060, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3396642792, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3225930628, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1352017718, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo((MetaName)73333937, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1414986019, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3574987574, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo((MetaName)992242800, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2148272557, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1884609047, PsoDataType.UInt, 220, 0, 0),
                     new PsoStructureEntryInfo((MetaName)839256524, PsoDataType.Float2, 224, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3914143003, PsoDataType.Float2, 232, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1935117365, PsoDataType.Float2, 240, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3465208228, PsoDataType.Float2, 248, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2745575061, PsoDataType.UInt, 256, 0, 0),
                     new PsoStructureEntryInfo((MetaName)19236851, PsoDataType.UInt, 260, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3381171437, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4274113098, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3364491568, PsoDataType.UInt, 272, 0, 0),
                     new PsoStructureEntryInfo((MetaName)565980479, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo((MetaName)467608921, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3180328046, PsoDataType.UInt, 284, 0, 0),
                     new PsoStructureEntryInfo((MetaName)924888517, PsoDataType.UInt, 288, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1870040536, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camGameplayDirectorMetadataVehicleCustomSettings),
                     new PsoStructureEntryInfo(MetaName.VehicleCustomSettingsList, PsoDataType.Array, 296, 0, (MetaName)66),
                     new PsoStructureEntryInfo((MetaName)3559905910, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo((MetaName)259392755, PsoDataType.Float, 316, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1523938219, PsoDataType.Float, 320, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3704241863, PsoDataType.UInt, 324, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2185907197, PsoDataType.UInt, 328, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2766631017, PsoDataType.UInt, 332, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4088657953, PsoDataType.UInt, 336, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2940797056, PsoDataType.UInt, 340, 0, 0),
                     new PsoStructureEntryInfo((MetaName)926072750, PsoDataType.UInt, 344, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4071374960, PsoDataType.UInt, 348, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4097214758, PsoDataType.UInt, 352, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4182146981, PsoDataType.UInt, 356, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3961367630, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1597015313, PsoDataType.Float, 364, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2284520874, PsoDataType.Bool, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4217422073, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo((MetaName)636518286, PsoDataType.String, 376, 7, 0),
                     new PsoStructureEntryInfo((MetaName)984379181, PsoDataType.UInt, 380, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2430934664, PsoDataType.UInt, 384, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1190562413, PsoDataType.Float2, 388, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3193796044, PsoDataType.Float2, 396, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3497158180, PsoDataType.Float2, 404, 0, 0),
                     new PsoStructureEntryInfo((MetaName)202581691, PsoDataType.Float2, 412, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1600149759, PsoDataType.Float2, 420, 0, 0),
                     new PsoStructureEntryInfo((MetaName)310864401, PsoDataType.Float2, 428, 0, 0),
                     new PsoStructureEntryInfo((MetaName)910310359, PsoDataType.Float, 436, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1518272577, PsoDataType.Float, 440, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2881277133, PsoDataType.Float, 444, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1283045856, PsoDataType.Float, 448, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1013584657, PsoDataType.UInt, 452, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1494877942, PsoDataType.UInt, 456, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3664775821, PsoDataType.String, 460, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1867117468, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo((MetaName)425967458, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2079587041, PsoDataType.Float, 472, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3495981266, PsoDataType.Float, 476, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3273630905, PsoDataType.Float, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3860060233, PsoDataType.Float, 484, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2204296545, PsoDataType.Float, 488, 0, 0),
                     new PsoStructureEntryInfo((MetaName)111546466, PsoDataType.Float, 492, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2322128139, PsoDataType.Float, 496, 0, 0),
                     new PsoStructureEntryInfo((MetaName)642537232, PsoDataType.Float, 500, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2040447490, PsoDataType.Float, 504, 0, 0),
                     new PsoStructureEntryInfo((MetaName)928667502, PsoDataType.Float, 508, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3216527316, PsoDataType.UInt, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3465879577, PsoDataType.UInt, 516, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3390712027, PsoDataType.UInt, 520, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2683111365, PsoDataType.UInt, 524, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3170020756, PsoDataType.UInt, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4287333189, PsoDataType.UInt, 532, 0, 0),
                     new PsoStructureEntryInfo((MetaName)115056986, PsoDataType.UInt, 536, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1843179030, PsoDataType.UInt, 540, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3134298162, PsoDataType.String, 544, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1564626128, PsoDataType.String, 548, 7, 0),
                     new PsoStructureEntryInfo((MetaName)834090440, PsoDataType.Structure, 552, 0, MetaName.camGameplayDirectorMetadataExplosionShakeSettings)
                    );
                case MetaName.camGameplayDirectorMetadataExplosionShakeSettings:
                    return new PsoStructureInfo(MetaName.camGameplayDirectorMetadataExplosionShakeSettings, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceLimits, PsoDataType.Float2, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxInstances, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RumbleDuration, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3070197011, PsoDataType.UInt, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2337695573, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)815713089, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1985122943, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3938350897, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)951241725, PsoDataType.Float, 48, 0, 0)
                    );
                case MetaName.camGameplayDirectorMetadataVehicleCustomSettings:
                    return new PsoStructureInfo(MetaName.camGameplayDirectorMetadataVehicleCustomSettings, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.ModelName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SettingsRef, PsoDataType.String, 12, 7, 0)
                    );
                case MetaName.camCinematicDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicDirectorMetadata, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)164388064, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1567199172, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4027640984, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)730699960, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4198475225, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3640482549, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2985235183, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)200602575, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2683111365, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1677201089, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2101508607, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1867117468, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)425967458, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2079587041, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.CinematicContexts, PsoDataType.Array, 80, 0, (MetaName)MetaTypeName.STRING),
                     new PsoStructureEntryInfo((MetaName)28440621, PsoDataType.UInt, 96, 0, 0)
                    );
                case MetaName.camCinematicInVehicleMultiplayerPassengerContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicInVehicleMultiplayerPassengerContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicShots:
                    return new PsoStructureInfo(MetaName.camCinematicShots, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Shot, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Priority, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ProbabilityWeighting, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.camCinematicSpectatorNewsChannelContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicSpectatorNewsChannelContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicScriptedMissionCreatorFailContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicScriptedMissionCreatorFailContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicScriptContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicScriptContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicScriptedRaceCheckPointContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicScriptedRaceCheckPointContextMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1650416254, PsoDataType.UInt, 72, 0, 0)
                    );
                case MetaName.camCinematicBustedContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicBustedContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicFallFromHeliContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicFallFromHeliContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicInVehicleOverriddenFirstPersonContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicInVehicleOverriddenFirstPersonContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicInVehicleWantedContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicInVehicleWantedContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camPreferredShotSelectionType:
                    return new PsoStructureInfo(MetaName.camPreferredShotSelectionType, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Shot, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.InputType, PsoDataType.Enum, 12, 0, (MetaName)3889593720)
                    );
                case MetaName.camCinematicInVehicleContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicInVehicleContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicInTrainAtStationContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicInTrainAtStationContextMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2216049071, PsoDataType.UInt, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3119408512, PsoDataType.UInt, 76, 0, 0)
                    );
                case MetaName.camCinematicInTrainContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicInTrainContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicOnFootSpectatingContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicOnFootSpectatingContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicParachuteContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicParachuteContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicStuntJumpContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicStuntJumpContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicOnFootAssistedAimingContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicOnFootAssistedAimingContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicOnFootMeleeContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicOnFootMeleeContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicInVehicleFirstPersonContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicInVehicleFirstPersonContextMetadata, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0)
                    );
                case MetaName.camCinematicOnFootIdleContextMetadata:
                    return new PsoStructureInfo(MetaName.camCinematicOnFootIdleContextMetadata, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camCinematicShots),
                     new PsoStructureEntryInfo(MetaName.Shots, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camPreferredShotSelectionType),
                     new PsoStructureEntryInfo((MetaName)2569180342, PsoDataType.Array, 32, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)2437129562, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4114376857, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2434943359, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2838462683, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2639584784, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3699223405, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3274921789, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3966397175, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2910409886, PsoDataType.UInt, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1879525600, PsoDataType.UInt, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1202637816, PsoDataType.Float, 76, 0, 0)
                    );
                case MetaName.camSyncedSceneDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camSyncedSceneDirectorMetadata, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406434970, PsoDataType.String, 24, 7, 0)
                    );
                case MetaName.camScriptDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camScriptDirectorMetadata, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3308601968, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1044888659, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo((MetaName)609660376, PsoDataType.UInt, 32, 0, 0)
                    );
                case MetaName.camSwitchDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camSwitchDirectorMetadata, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4284026199, PsoDataType.String, 24, 7, 0)
                    );
                case MetaName.camCutsceneDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camCutsceneDirectorMetadata, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1814055656, PsoDataType.String, 24, 7, 0)
                    );
                case MetaName.camReplayDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camReplayDirectorMetadata, 0, 0, 160,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2721634605, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2132680058, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo((MetaName)491767013, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2648380967, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4072288030, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo((MetaName)664486210, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3903471436, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3789126205, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2319741638, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1596110048, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1238798750, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1362065867, PsoDataType.String, 80, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2640324566, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3068943162, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)81027141, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4246618820, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1674718375, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)227690771, PsoDataType.String, 104, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1982856886, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo((MetaName)881665360, PsoDataType.String, 112, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2599823537, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3206570060, PsoDataType.String, 120, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1690145332, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4171384554, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3385138470, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)950635108, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)925408249, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3918000719, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1379084243, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2245689292, PsoDataType.Float, 152, 0, 0)
                    );
                case MetaName.camDebugDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camDebugDirectorMetadata, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2750009688, PsoDataType.String, 24, 7, 0)
                    );
                case MetaName.camMarketingDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camMarketingDirectorMetadata, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.camMarketingDirectorMetadataMode),
                     new PsoStructureEntryInfo(MetaName.Modes, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case MetaName.camMarketingDirectorMetadataMode:
                    return new PsoStructureInfo(MetaName.camMarketingDirectorMetadataMode, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.CameraRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TextLabel, PsoDataType.String, 12, 0, (MetaName)1048576)
                    );
                case MetaName.CExplosionInfoManager:
                    return new PsoStructureInfo(MetaName.CExplosionInfoManager, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2541591167),
                     new PsoStructureEntryInfo(MetaName.aExplosionTagData, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)2541591167:
                    return new PsoStructureInfo((MetaName)2541591167, 0, 0, 136,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 3, 0),
                     new PsoStructureEntryInfo(MetaName.damageAtCentre, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.damageAtEdge, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.networkPlayerModifier, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.networkPedModifier, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.endRadius, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.initSpeed, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.decayFactor, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.forceFactor, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRagdollForceModifier, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSelfForceModifier, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.directedWidth, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.directedLifeTime, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.camShakeName, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo(MetaName.camShake, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.camShakeRollOffScaling, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.shockingEventVisualRangeOverride, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.shockingEventAudioRangeOverride, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.minorExplosion, PsoDataType.Bool, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vfxTagHashName, PsoDataType.String, 96, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fragDamage, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bAppliesContinuousDamage, PsoDataType.Bool, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bPostProcessCollisionsWithNoForce, PsoDataType.Bool, 105, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bDamageVehicles, PsoDataType.Bool, 106, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bDamageObjects, PsoDataType.Bool, 107, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bOnlyAffectsLivePeds, PsoDataType.Bool, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bIgnoreExplodingEntity, PsoDataType.Bool, 109, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bNoOcclusion, PsoDataType.Bool, 110, 0, 0),
                     new PsoStructureEntryInfo(MetaName.explodeAttachEntityWhenFinished, PsoDataType.Bool, 111, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bCanSetPedOnFire, PsoDataType.Bool, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bCanSetPlayerOnFire, PsoDataType.Bool, 113, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bSuppressCrime, PsoDataType.Bool, 114, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bUseDistanceDamageCalc, PsoDataType.Bool, 115, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bPreventWaterExplosionVFX, PsoDataType.Bool, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bIgnoreRatioCheckForFire, PsoDataType.Bool, 117, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bAllowUnderwaterExplosion, PsoDataType.Bool, 118, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bForceVehicleExplosion, PsoDataType.Bool, 119, 0, 0),
                     new PsoStructureEntryInfo(MetaName.midRadius, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.damageAtMid, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bApplyVehicleEMP, PsoDataType.Bool, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bApplyVehicleSlick, PsoDataType.Bool, 129, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bApplyVehicleSlowdown, PsoDataType.Bool, 130, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bApplyVehicleTyrePop, PsoDataType.Bool, 131, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bForcePetrolTankDamage, PsoDataType.Bool, 132, 0, 0)
                    );
                case MetaName.CPedModelInfo__PersonalityDataList:
                    return new PsoStructureInfo(MetaName.CPedModelInfo__PersonalityDataList, 0, 0, 88,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1111461724),
                     new PsoStructureEntryInfo(MetaName.MovementModeUnholsterData, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)29118966),
                     new PsoStructureEntryInfo(MetaName.MovementModes, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)4185391031),
                     new PsoStructureEntryInfo(MetaName.PedPersonalities, PsoDataType.Array, 40, 0, (MetaName)4),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)905918498),
                     new PsoStructureEntryInfo(MetaName.BraveryTypes, PsoDataType.Array, 56, 0, (MetaName)6),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2072330808),
                     new PsoStructureEntryInfo(MetaName.CriminalityTypes, PsoDataType.Array, 72, 0, (MetaName)8)
                    );
                case (MetaName)1111461724:
                    return new PsoStructureInfo((MetaName)1111461724, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2934735155),
                     new PsoStructureEntryInfo(MetaName.UnholsterClips, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case (MetaName)2934735155:
                    return new PsoStructureInfo((MetaName)2934735155, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Weapons, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Clip, PsoDataType.String, 24, 7, 0)
                    );
                case (MetaName)29118966:
                    return new PsoStructureInfo((MetaName)29118966, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1727236615),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Array, 0, 0, (MetaName)1),
                     new PsoStructureEntryInfo(MetaName.MovementModes, PsoDataType.Array, 16, 4, (MetaName)131074),
                     new PsoStructureEntryInfo(MetaName.LastBattleEventHighEnergyStartTime, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LastBattleEventHighEnergyEndTime, PsoDataType.Float, 52, 0, 0)
                    );
                case (MetaName)1727236615:
                    return new PsoStructureInfo((MetaName)1727236615, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Weapons, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3455138791),
                     new PsoStructureEntryInfo(MetaName.ClipSets, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case (MetaName)3455138791:
                    return new PsoStructureInfo((MetaName)3455138791, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.MovementClipSetId, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponClipSetId, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponClipFilterId, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.UpperBodyShadowExpressionEnabled, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpperBodyFeatheredLeanEnabled, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseWeaponAnimsForGrip, PsoDataType.Bool, 22, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseLeftHandIk, PsoDataType.Bool, 23, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdleTransitionBlendOutTime, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IdleTransitions, PsoDataType.Array, 32, 0, (MetaName)8),
                     new PsoStructureEntryInfo(MetaName.UnholsterClipSetId, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.UnholsterClipData, PsoDataType.String, 52, 7, 0)
                    );
                case (MetaName)4185391031:
                    return new PsoStructureInfo((MetaName)4185391031, 0, 0, 184,
                     new PsoStructureEntryInfo(MetaName.PersonalityName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultWeaponLoadout, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Bravery, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AgitatedPersonality, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Criminality, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AgitationTriggers, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HealthConfigHash, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponAnimations, PsoDataType.Array, 40, 0, (MetaName)MetaTypeName.POINTER),
                     new PsoStructureEntryInfo(MetaName.AmbientAudio, PsoDataType.String, 56, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WitnessPersonality, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Agility, PsoDataType.Structure, 64, 0, (MetaName)265080325),
                     new PsoStructureEntryInfo(MetaName.IsMale, PsoDataType.Bool, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsHuman, PsoDataType.Bool, 81, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldRewardMoneyOnDeath, PsoDataType.Bool, 82, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsGang, PsoDataType.Bool, 83, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsSecurity, PsoDataType.Bool, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsWeird, PsoDataType.Bool, 85, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsDangerousAnimal, PsoDataType.Bool, 86, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CausesRumbleWhenCollidesWithPlayer, PsoDataType.Bool, 87, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowSlowCruisingWithMusic, PsoDataType.Bool, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowRoadCrossHurryOnLightChange, PsoDataType.Bool, 89, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)2920442291),
                     new PsoStructureEntryInfo(MetaName.VehicleTypes, PsoDataType.Flags, 92, 0, (MetaName)524310),
                     new PsoStructureEntryInfo(MetaName.AttackStrengthMin, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttackStrengthMax, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StaminaEfficiency, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArmourEfficiency, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HealthRegenEfficiency, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExplosiveDamageMod, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HandGunDamageMod, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RifleDamageMod, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SmgDamageMod, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PopulationFleeMod, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HotwireRate, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MotivationMin, PsoDataType.UInt, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MotivationMax, PsoDataType.UInt, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrivingAbilityMin, PsoDataType.UByte, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrivingAbilityMax, PsoDataType.UByte, 149, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrivingAggressivenessMin, PsoDataType.UByte, 150, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrivingAggressivenessMax, PsoDataType.UByte, 151, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Affluence, PsoDataType.Enum, 152, 0, MetaName.Affluence),
                     new PsoStructureEntryInfo(MetaName.TechSavvy, PsoDataType.Enum, 156, 0, MetaName.TechSavvy),
                     new PsoStructureEntryInfo(MetaName.MovementModes, PsoDataType.String, 160, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponAnimsFPSIdle, PsoDataType.String, 164, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponAnimsFPSRNG, PsoDataType.String, 168, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponAnimsFPSLT, PsoDataType.String, 172, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponAnimsFPSScope, PsoDataType.String, 176, 7, 0)
                    );
                case (MetaName)265080325:
                    return new PsoStructureInfo((MetaName)265080325, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)2675594832),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 8, 0, (MetaName)2097152),
                     new PsoStructureEntryInfo(MetaName.MovementCostModifier, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)905918498:
                    return new PsoStructureInfo((MetaName)905918498, 0, 0, 256,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3079935216),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 12, 0, (MetaName)2097153),
                     new PsoStructureEntryInfo(MetaName.TakedownProbability, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThreatResponseUnarmed, PsoDataType.Structure, 24, 0, (MetaName)3637970520),
                     new PsoStructureEntryInfo(MetaName.ThreatResponseMelee, PsoDataType.Structure, 96, 0, (MetaName)3637970520),
                     new PsoStructureEntryInfo(MetaName.ThreatResponseArmed, PsoDataType.Structure, 168, 0, (MetaName)3637970520),
                     new PsoStructureEntryInfo(MetaName.FleeDuringCombat, PsoDataType.Structure, 240, 0, (MetaName)651017172)
                    );
                case (MetaName)3637970520:
                    return new PsoStructureInfo((MetaName)3637970520, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Action, PsoDataType.Structure, 8, 0, (MetaName)3450423538),
                     new PsoStructureEntryInfo(MetaName.Fight, PsoDataType.Structure, 32, 0, (MetaName)301719986)
                    );
                case (MetaName)3450423538:
                    return new PsoStructureInfo((MetaName)3450423538, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Weights, PsoDataType.Structure, 8, 0, (MetaName)3954379444)
                    );
                case (MetaName)3954379444:
                    return new PsoStructureInfo((MetaName)3954379444, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Fight, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Flee, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)301719986:
                    return new PsoStructureInfo((MetaName)301719986, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Weights, PsoDataType.Structure, 8, 0, (MetaName)272028034),
                     new PsoStructureEntryInfo(MetaName.ProbabilityDrawWeaponWhenLosing, PsoDataType.Float, 32, 0, 0)
                    );
                case (MetaName)272028034:
                    return new PsoStructureInfo((MetaName)272028034, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.KeepWeapon, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MatchTargetWeapon, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EquipBestWeapon, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)651017172:
                    return new PsoStructureInfo((MetaName)651017172, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesWhenBuddyKilledWithScaryWeapon, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)2072330808:
                    return new PsoStructureInfo((MetaName)2072330808, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1639628591),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 12, 0, (MetaName)2097153)
                    );
                case MetaName.sStatsMetadataTuning:
                    return new PsoStructureInfo(MetaName.sStatsMetadataTuning, 0, 0, 72,
                     new PsoStructureEntryInfo((MetaName)1063717096, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AwardVehicleJumpTime, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AwardParachuteJumpTime, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SPLargeAccidenThresold, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MPLargeAccidenThresold, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FreefallThresold, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AwardVehicleJumpDistanceA, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AwardVehicleJumpDistanceB, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AwardParachuteJumpDistanceA, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AwardParachuteJumpDistanceB, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.nonFlyableAreas, PsoDataType.Structure, 48, 0, MetaName.NonFlyableAreaArray)
                    );
                case MetaName.CFiringPatternInfoManager:
                    return new PsoStructureInfo(MetaName.CFiringPatternInfoManager, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.Infos, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.CFiringPatternInfo:
                    return new PsoStructureInfo(MetaName.CFiringPatternInfo, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.NumberOfBurstsMin, PsoDataType.SShort, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NumberOfBurstsMax, PsoDataType.SShort, 14, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NumberOfShotsPerBurstMin, PsoDataType.SShort, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NumberOfShotsPerBurstMax, PsoDataType.SShort, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenShotsMin, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenShotsMax, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenShotsAbsoluteMin, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenBurstsMin, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenBurstsMax, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenBurstsAbsoluteMin, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBeforeFiringMin, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBeforeFiringMax, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2922044807, PsoDataType.Bool, 52, 0, 0)
                    );
                case MetaName.CTuningFile:
                    return new PsoStructureInfo(MetaName.CTuningFile, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.Tunables, PsoDataType.Array, 0, 0, 0)
                    );
                case MetaName.CCombatTaskManager__Tunables:
                    return new PsoStructureInfo(MetaName.CCombatTaskManager__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenUpdates, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iMaxPedsInCombatTask, PsoDataType.SInt, 20, 0, 0)
                    );
                case MetaName.CTaskAnimatedHitByExplosion__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAnimatedHitByExplosion__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.InitialRagdollDelay, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowPitchAndRoll, PsoDataType.Bool, 20, 0, 0)
                    );
                case MetaName.CTaskStandGuard__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskStandGuard__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinStandWaitTimeMS, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStandWaitTimeMS, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDefendPointWaitTimeMS, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDefendPointWaitTimeMS, PsoDataType.SInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinNavmeshPatrolRadiusFactor, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxNavmeshPatrolRadiusFactor, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RouteRadiusFactor, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CTaskCombatAdditionalTask__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCombatAdditionalTask__Tunables, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.iBulletEventResponseLengthMs, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fChanceOfDynamicRun, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxDynamicStrafeDistance, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeInState, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMoveBlendRatioLerpTime, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceToClearCorner, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxDistanceFromCorner, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxLeavingCornerDistance, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fBlockedLosAimTime, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStartAimingDistance, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStopAimingDistance, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinOtherPedDistanceDiff, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeBetweenRunDirectlyChecks, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxTimeStrafing, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeRunning, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fForceStrafeDistance, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2878670581, PsoDataType.Float, 80, 0, 0)
                    );
                case MetaName.CTaskCombatFlank__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCombatFlank__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fInfluenceSphereInnerWeight, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInfluenceSphereOuterWeight, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInfluenceSphereRequestRadius, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInfluenceSphereCheckRouteRadius, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSmallInfluenceSphereRadius, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDistanceBetweenInfluenceSpheres, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAbsoluteMinDistanceToTarget, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCoverPointScoreMultiplier, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.CTaskHelicopterStrafe__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskHelicopterStrafe__Tunables, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FlightHeightAboveTarget, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveTerrain, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetDirectionMinDot, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetOffset, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetMinSpeedToIgnore, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetMaxSpeedToStrafe, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToAvoidTargetAfterDamaged, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidOffsetXY, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidOffsetZ, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToBeConsideredInFront, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BehindRotateAngleLookAhead, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SearchRotateAngleLookAhead, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CircleRotateAngleLookAhead, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BehindTargetAngle, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetOffsetFilter, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenStrafeDirectionChanges, PsoDataType.Float, 76, 0, 0)
                    );
                case MetaName.CTaskMeleeActionResult__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMeleeActionResult__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ActionModeTime, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceRunDelayTime, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskMoveCombatMounted__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMoveCombatMounted__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CircleTestRadii, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo(MetaName.CircleTestsMoveDistToTestNewPos, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeSinceAnyCircleJoined, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeSinceSameCircleJoined, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TransitionReactionTime, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelStartCircling, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelStopCircling, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeWaitingForCircleMs, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.CTaskDyingDead__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskDyingDead__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleForwardInitialScale, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleForwardScale, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToApplyPushFromVehicleForce, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceToApply, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFallingSpeedForAnimatedDyingFall, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SphereTestRadiusForDeadWaterSettle, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollAbortPoseDistanceThreshold, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollAbortPoseMaxVelocity, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToThrowWeaponMS, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToThrowWeaponPlayerMS, PsoDataType.UInt, 52, 0, 0)
                    );
                case MetaName.CTaskDamageElectric__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskDamageElectric__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FallsOutofVehicleVelocity, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskInvestigate__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskInvestigate__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.iTimeToStandAtSearchPoint, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceToUseVehicle, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceSavingToUseVehicle, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeToStandAtPerimeter, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fNewPositionThreshold, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.CTaskCombat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCombat__Tunables, 0, 0, 704,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BuddyShot, PsoDataType.Structure, 16, 0, (MetaName)4144542252),
                     new PsoStructureEntryInfo(MetaName.LackOfHostility, PsoDataType.Structure, 48, 0, (MetaName)455065356),
                     new PsoStructureEntryInfo(MetaName.EnemyAccuracyScaling, PsoDataType.Structure, 144, 0, (MetaName)1818844971),
                     new PsoStructureEntryInfo(MetaName.ChargeTuning, PsoDataType.Structure, 168, 0, (MetaName)3736726978),
                     new PsoStructureEntryInfo(MetaName.ThrowSmokeGrenadeTuning, PsoDataType.Structure, 240, 0, (MetaName)1530415684),
                     new PsoStructureEntryInfo(MetaName.MaxDistToCoverZ, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistToCoverXY, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAmbientAnimsMinDistToTargetSq, PsoDataType.Float, 296, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAmbientAnimsMaxDistToTargetSq, PsoDataType.Float, 300, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fGoToDefAreaTimeOut, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFireContinuouslyDistMin, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFireContinuouslyDistMax, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fLostTargetTime, PsoDataType.Float, 316, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeAfterAimPoseForStateChange, PsoDataType.Float, 320, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxAttemptMoveToCoverDelay, PsoDataType.Float, 324, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinAttemptMoveToCoverDelay, PsoDataType.Float, 328, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxAttemptMoveToCoverDelayGlobal, PsoDataType.Float, 332, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinAttemptMoveToCoverDelayGlobal, PsoDataType.Float, 336, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceForAltCover, PsoDataType.Float, 340, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeStandingAtCover, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeBetweenFrustratedPeds, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxTimeBetweenFrustratedPeds, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRetreatTime, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTargetTooCloseDistance, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenJackingAttempts, PsoDataType.Float, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenCoverSearchesMin, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenCoverSearchesMax, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenAltCoverSearches, PsoDataType.Float, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fShoutTargetPositionInterval, PsoDataType.Float, 380, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fShoutBlockingLosInterval, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenDragsMin, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenSecondaryTargetUsesMin, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenSecondaryTargetUsesMax, PsoDataType.Float, 396, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeToUseSecondaryTargetMin, PsoDataType.Float, 400, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeToUseSecondaryTargetMax, PsoDataType.Float, 404, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenCombatDirectorUpdates, PsoDataType.Float, 408, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenPassiveAnimsMin, PsoDataType.Float, 412, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenPassiveAnimsMax, PsoDataType.Float, 416, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenQuickGlancesMin, PsoDataType.Float, 420, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenQuickGlancesMax, PsoDataType.Float, 424, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenGestureAnimsMin, PsoDataType.Float, 428, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenGestureAnimsMax, PsoDataType.Float, 432, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenFailedGestureMin, PsoDataType.Float, 436, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenFailedGestureMax, PsoDataType.Float, 440, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenGesturesMinGlobal, PsoDataType.Float, 444, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenGesturesMaxGlobal, PsoDataType.Float, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeSinceLastAimedAtForGesture, PsoDataType.Float, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeBeforeReactToExplosion, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxTimeBeforeReactToExplosion, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetInfluenceSphereRadius, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetMinDistanceToRoute, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetMinDistanceToAwayFacingNavLink, PsoDataType.Float, 472, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxWaitForCoverExitTime, PsoDataType.Float, 476, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxDstanceToMoveAwayFromAlly, PsoDataType.Float, 480, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenAllyProximityChecks, PsoDataType.Float, 484, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceFromPrimaryTarget, PsoDataType.Float, 488, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxAngleBetweenTargets, PsoDataType.Float, 492, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceFromPedToHelpPed, PsoDataType.Float, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDotToTargetToHelpPed, PsoDataType.Float, 500, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeadingDifferenceForQuickGlanceInSameDirection, PsoDataType.Float, 504, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenQuickGlancesInSameDirection, PsoDataType.Float, 508, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedToStartJackingVehicle, PsoDataType.Float, 512, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedToContinueJackingVehicle, PsoDataType.Float, 516, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetJackRadius, PsoDataType.Float, 520, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SafetyProportionInDefensiveAreaMin, PsoDataType.Float, 524, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SafetyProportionInDefensiveAreaMax, PsoDataType.Float, 528, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMoveToDefensiveAreaAngleVariation, PsoDataType.Float, 532, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToEnterVehicleIfTargetEntersVehicle, PsoDataType.Float, 536, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToMyVehicleToChase, PsoDataType.Float, 540, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToVehicleForCommandeer, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NumEarlyVehicleEntryDriversAllowed, PsoDataType.UByte, 548, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SafeTimeBeforeLeavingCover, PsoDataType.UInt, 552, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WaitTimeForJackingSlowedVehicle, PsoDataType.UInt, 556, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxInjuredTargetTimerVariation, PsoDataType.Float, 560, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxNumPedsChasingOnFoot, PsoDataType.UByte, 564, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FireTimeAfterStaticMovementAbort, PsoDataType.Float, 568, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinMovingToCoverTimeToStop, PsoDataType.Float, 572, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToCoverToStop, PsoDataType.Float, 576, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FireTimeAfterStoppingMoveToCover, PsoDataType.Float, 580, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ApproachingTargetVehicleHoldFireDistance, PsoDataType.Float, 584, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDefensiveAreaRadiusForWillAdvance, PsoDataType.Float, 588, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToHoldFireForArrest, PsoDataType.Float, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToDelayChaseOnFoot, PsoDataType.Float, 596, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FireTimeAfterChaseOnFoot, PsoDataType.Float, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToChangeChaseOnFootSpeed, PsoDataType.UInt, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableForcedFireForTargetProximity, PsoDataType.Bool, 608, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForceFiringStateTime, PsoDataType.Float, 612, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForceFiringStateTime, PsoDataType.Float, 616, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBeforeInitialForcedFire, PsoDataType.Float, 620, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenForcedFireStates, PsoDataType.Float, 624, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInStateForForcedFire, PsoDataType.Float, 628, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForceFiringDistance, PsoDataType.Float, 632, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForceFiringDistance, PsoDataType.Float, 636, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForAimIntro, PsoDataType.Float, 640, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3129989164, PsoDataType.Float, 644, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4269476531, PsoDataType.Float, 648, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmbientAnimLengthBuffer, PsoDataType.Float, 652, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenPlayerArrestAttempts, PsoDataType.UInt, 656, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenArmedMeleeAttemptsInMs, PsoDataType.UInt, 660, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowMovingArmedMeleeAttack, PsoDataType.Bool, 664, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToHoldFireAfterJack, PsoDataType.Float, 668, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenMeleeJackAttempts, PsoDataType.UInt, 672, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenMeleeJackAttemptsOnNetworkClone, PsoDataType.UInt, 676, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToHoldFireAtTaskInitialization, PsoDataType.Float, 680, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToRejectRespawnedTarget, PsoDataType.UInt, 684, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForLawToFleeFromCombat, PsoDataType.Float, 688, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForLawToReturnToCombatFromFlee, PsoDataType.Float, 692, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2186966715, PsoDataType.Float, 696, 0, 0)
                    );
                case (MetaName)4144542252:
                    return new PsoStructureInfo((MetaName)4144542252, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBeforeReact, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeBeforeReact, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeSinceShot, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistance, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)455065356:
                    return new PsoStructureInfo((MetaName)455065356, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.WantedLevel1, PsoDataType.Structure, 8, 0, (MetaName)2841666617),
                     new PsoStructureEntryInfo(MetaName.WantedLevel2, PsoDataType.Structure, 24, 0, (MetaName)2841666617),
                     new PsoStructureEntryInfo(MetaName.WantedLevel3, PsoDataType.Structure, 40, 0, (MetaName)2841666617),
                     new PsoStructureEntryInfo(MetaName.WantedLevel4, PsoDataType.Structure, 56, 0, (MetaName)2841666617),
                     new PsoStructureEntryInfo(MetaName.WantedLevel5, PsoDataType.Structure, 72, 0, (MetaName)2841666617),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForVehicle, PsoDataType.Float, 88, 0, 0)
                    );
                case (MetaName)2841666617:
                    return new PsoStructureInfo((MetaName)2841666617, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeSinceLastHostileAction, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)1818844971:
                    return new PsoStructureInfo((MetaName)1818844971, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.iMinNumEnemiesForScaling, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAccuracyReductionPerEnemy, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAccuracyReductionFloor, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)3736726978:
                    return new PsoStructureInfo((MetaName)3736726978, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.bChargeTargetEnabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uMaxNumActiveChargers, PsoDataType.UByte, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uConsiderRecentChargeAsActiveTimeoutMS, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uMinTimeBetweenChargesAtSameTargetMS, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uMinTimeForSamePedToChargeAgainMS, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uCheckForChargeTargetPeriodMS, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeInCombatSeconds, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceToTarget, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxDistanceToTarget, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistToNonTargetEnemy, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistBetweenTargetAndOtherEnemies, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDistToHidingTarget_Outer, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDistToHidingTarget_Inner, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fChargeGoalCompletionRadius, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCancelTargetOutOfCoverMovedDist, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCancelTargetInCoverMovedDist, PsoDataType.Float, 64, 0, 0)
                    );
                case (MetaName)1530415684:
                    return new PsoStructureInfo((MetaName)1530415684, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.bThrowSmokeGrenadeEnabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uMaxNumActiveThrowers, PsoDataType.UByte, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uConsiderRecentThrowAsActiveTimeoutMS, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uMinTimeBetweenThrowsAtSameTargetMS, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uMinTimeForSamePedToThrowAgainMS, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uCheckForSmokeThrowPeriodMS, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceToTarget, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxDistanceToTarget, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDotMinThrowerToTarget, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinLoiteringTimeSeconds, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CTaskReactAimWeapon__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskReactAimWeapon__Tunables, 0, 0, 680,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Professional, PsoDataType.Structure, 16, 0, (MetaName)3326118529),
                     new PsoStructureEntryInfo(MetaName.NotProfessional, PsoDataType.Structure, 344, 0, (MetaName)3326118529),
                     new PsoStructureEntryInfo(MetaName.Rate, PsoDataType.Float, 672, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRateVariance, PsoDataType.Float, 676, 0, 0)
                    );
                case (MetaName)3326118529:
                    return new PsoStructureInfo((MetaName)3326118529, 0, 0, 328,
                     new PsoStructureEntryInfo(MetaName.Flinch, PsoDataType.Structure, 8, 0, (MetaName)2563725968),
                     new PsoStructureEntryInfo(MetaName.Surprised, PsoDataType.Structure, 88, 0, (MetaName)2563725968),
                     new PsoStructureEntryInfo(MetaName.Sniper, PsoDataType.Structure, 168, 0, (MetaName)2563725968),
                     new PsoStructureEntryInfo(MetaName.None, PsoDataType.Structure, 248, 0, (MetaName)2563725968)
                    );
                case (MetaName)2563725968:
                    return new PsoStructureInfo((MetaName)2563725968, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Pistol, PsoDataType.Structure, 8, 0, (MetaName)3972028972),
                     new PsoStructureEntryInfo(MetaName.Rifle, PsoDataType.Structure, 32, 0, (MetaName)3972028972),
                     new PsoStructureEntryInfo(MetaName.MicroSMG, PsoDataType.Structure, 56, 0, (MetaName)3972028972)
                    );
                case (MetaName)3972028972:
                    return new PsoStructureInfo((MetaName)3972028972, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ClipSetId, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Rate, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HasSixDirections, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HasCreateWeaponTag, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HasInterruptTag, PsoDataType.Bool, 18, 0, 0)
                    );
                case MetaName.CTaskSearch__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSearch__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToStare, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MoveBlendRatio, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetReached, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeed, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CTaskSearchForUnknownThreat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSearchForUnknownThreat__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.iMinTimeBeforeSearchingForNewHidingPlace, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iMaxTimeBeforeSearchingForNewHidingPlace, PsoDataType.SInt, 20, 0, 0)
                    );
                case MetaName.CTaskSharkCircle__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSharkCircle__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AdvanceDistanceSquared, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MoveRateOverride, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskSharkAttack__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSharkAttack__Tunables, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SurfaceProjectionDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SurfaceZOffset, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDepthBelowSurface, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CirclingAngularSpeed, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToCircle, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinCircleRadius, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCircleRadius, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CirclingMBR, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DiveProjectionDistance, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DiveDepth, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DiveMBR, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinNumberFakeApproaches, PsoDataType.SInt, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxNumberFakeApproaches, PsoDataType.SInt, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FakeLungeOffset, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LungeForwardOffset, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LungeZOffset, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LungeChangeDistance, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LungeTargetRadius, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowTimeout, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowYOffset, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowZOffset, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LandProbeLength, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingVehicleVelocityThreshold, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SharkFleeDist, PsoDataType.Float, 108, 0, 0)
                    );
                case MetaName.CTaskCover__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCover__Tunables, 0, 0, 192,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3698046827, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1862046997, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1686691958, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3171144783, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1788624470, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1235126752, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1335048013, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2787523036, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PriorityCoverWeighting, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleToCameraWeighting, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleToDynamicCoverWeighting, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceWeighting, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleToCoverWeighting, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleOfCoverWeighting, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeWeighting, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NetworkBlendOutDurationRun, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NetworkBlendOutDurationRunStart, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NetworkBlendOutDuration, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPlayerToCoverDist, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1534623417, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularDiffBetweenDynamicAndStaticCover, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RangeToUseDynamicCoverPointMin, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RangeToUseDynamicCoverPointMax, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistToCoverAnyDir, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistToPriorityCoverToForce, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistToCoverSpecificDir, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BehindPedToCoverCosTolerance, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SearchToCoverCosTolerance, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CapsuleZOffset, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenTestSpheresIntersectingRoute, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistToCoverWhenPlayerIsClose, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinCoverToPlayerCoverDist, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinMoveToCoverDistForCoverMeAudio, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSecondsAsTopLevelTask, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceStreamingFailure, PsoDataType.Bool, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StreamedUnarmedCoverMovementClipSetId, PsoDataType.String, 156, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1804897490, PsoDataType.String, 160, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamedOneHandedCoverMovementClipSetId, PsoDataType.String, 164, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AIOneHandedAimingClipSetId, PsoDataType.String, 168, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AITwoHandedAimingClipSetId, PsoDataType.String, 172, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CoreWeaponClipSetId, PsoDataType.String, 176, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3902347552, PsoDataType.String, 180, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2620277506, PsoDataType.String, 184, 7, 0)
                    );
                case MetaName.CTaskEnterCover__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskEnterCover__Tunables, 0, 0, 368,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryRatePlayer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryRateAI, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryShortDistanceAI, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryShortDistancePlayer, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryStandDistance, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryStandStrafeDistance, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryMinDistance, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryMaxDistance, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryMinDistanceAI, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryMaxDistanceAI, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryMaxDirectDistance, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryMinTimeNavigatingAI, PsoDataType.SInt, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryMinAngleToScale, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryHeadingReachedTol, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverEntryPositionReachedTol, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FromCoverExitDistance, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NetworkBlendInDuration, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistFromCoverToAllowReloadCache, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WaitForFootPlant, PsoDataType.Bool, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableFootTagSyncing, PsoDataType.Bool, 89, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceToTarget, PsoDataType.Bool, 90, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableInitialHeadingBlend, PsoDataType.Bool, 91, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableTranslationScaling, PsoDataType.Bool, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableRotationScaling, PsoDataType.Bool, 93, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreventTranslationOvershoot, PsoDataType.Bool, 94, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreventRotationOvershoot, PsoDataType.Bool, 95, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoInitialHeadingBlend, PsoDataType.Bool, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoFinalHeadingFixUp, PsoDataType.Bool, 97, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistToPlayEntryAnim, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistToScale, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeed, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRotSpeed, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTransScale, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTransScale, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRotScale, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRotScale, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DeltaTolerance, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRotDelta, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleToSetDirectly, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AiEntryHalfAngleTolerance, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableNewAICoverEntry, PsoDataType.Bool, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableUseSwatClipSet, PsoDataType.Bool, 149, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseShortDistAngleRotation, PsoDataType.Bool, 150, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableAiCoverEntryStreamCheck, PsoDataType.Bool, 151, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistToUseShortestRotation, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InCoverTolerance, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DotThresholdForCenterEnter, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AiEntryMinRate, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AiEntryMaxRate, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerSprintEntryRate, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultPlayerStandEntryStartMovementPhase, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultPlayerStandEntryEndMovementPhase, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleToBeginRotationScale, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDefaultAngularVelocity, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverInterruptMinTime, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverInterruptDistanceTolerance, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverInterruptHeadingTolerance, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverAimInterruptDistanceTolerance, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverAimInterruptHeadingTolerance, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3920645611, PsoDataType.Structure, 216, 0, (MetaName)1285383250),
                     new PsoStructureEntryInfo((MetaName)80713034, PsoDataType.Structure, 256, 0, (MetaName)1285383250),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)711010751),
                     new PsoStructureEntryInfo(MetaName.AIEnterCoverClips, PsoDataType.Array, 296, 0, (MetaName)62),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3590219800),
                     new PsoStructureEntryInfo(MetaName.AIStandEnterCoverClips, PsoDataType.Array, 312, 0, (MetaName)64),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)711010751),
                     new PsoStructureEntryInfo(MetaName.AIEnterTransitionClips, PsoDataType.Array, 328, 0, (MetaName)66),
                     new PsoStructureEntryInfo(MetaName.EnterCoverAIAimingBase1H, PsoDataType.String, 344, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverAIAimingBase2H, PsoDataType.String, 348, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverAIAimingSwat1H, PsoDataType.String, 352, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverAIAimingSwat2H, PsoDataType.String, 356, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverAITransition1H, PsoDataType.String, 360, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnterCoverAITransition2H, PsoDataType.String, 364, 7, 0)
                    );
                case (MetaName)1285383250:
                    return new PsoStructureInfo((MetaName)1285383250, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo((MetaName)611524654, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo((MetaName)1120845094, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case (MetaName)3038580099:
                    return new PsoStructureInfo((MetaName)3038580099, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Clips, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1559228487),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 24, 0, (MetaName)2097154)
                    );
                case (MetaName)711010751:
                    return new PsoStructureInfo((MetaName)711010751, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.EnterClipId, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1559228487),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 12, 0, (MetaName)2097153)
                    );
                case (MetaName)3590219800:
                    return new PsoStructureInfo((MetaName)3590219800, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.StandClip0Id, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StandClip1Id, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StandClip2Id, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1559228487),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 20, 0, (MetaName)2097155)
                    );
                case MetaName.CTaskExitCover__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskExitCover__Tunables, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)4010326834),
                     new PsoStructureEntryInfo(MetaName.CornerExitClips, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo(MetaName.ExitCoverBaseClipSetId, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2123164199, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ExitCoverExtraClipSetId, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1757551615, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinInputToInterruptIdle, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CornerExitHeadingModifier, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitCornerZOffset, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitCornerYOffset, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitCornerDirOffset, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3920645611, PsoDataType.Structure, 72, 0, (MetaName)246682624),
                     new PsoStructureEntryInfo((MetaName)80713034, PsoDataType.Structure, 96, 0, (MetaName)246682624)
                    );
                case (MetaName)246682624:
                    return new PsoStructureInfo((MetaName)246682624, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo((MetaName)1123873278, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)4010326834:
                    return new PsoStructureInfo((MetaName)4010326834, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.ExitClipId, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1559228487),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 12, 0, (MetaName)2097153)
                    );
                case MetaName.CTaskInCover__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskInCover__Tunables, 0, 0, 536,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MovementClipRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurnClipRate, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxInputForIdleExit, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InputYAxisCornerExitValue, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlDebugXPos, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlDebugYPos, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlDebugRadius, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlDebugBeginAngle, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ControlDebugEndAngle, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStickInputToMoveInCover, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStickInputXAxisToTurnInCover, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)927329696, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InputYAxisQuitValue, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartExtendedProbeTime, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToSpendInTask, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredDistanceToCover, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredDistanceToCoverToRequestStep, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OptimumDistToRightCoverEdgeCrouched, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OptimumDistToLeftCoverEdgeCrouched, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OptimumDistToRightCoverEdge, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OptimumDistToLeftCoverEdge, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinMovingProbeOffset, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMovingProbeOffset, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTurnProbeOffset, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTurnProbeOffset, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultProbeOffset, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStoppingEdgeCheckProbeOffset, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStoppingEdgeCheckProbeOffset, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStoppingProbeOffset, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStoppingProbeOffset, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingChangeRate, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBeforeAllowingCornerMove, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CrouchedLeftFireOffset, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CrouchedRightFireOffset, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverLeftFireModifier, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverRightFireModifier, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverLeftFireModifierLow, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverRightFireModifierLow, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2625425324, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CFriendlyIssues, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverLeftIncreaseModifier, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverRightIncreaseModifier, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimTurnCosAngleTolerance, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InCoverMovementSpeedEnterCover, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InCoverMovementSpeed, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteppingMovementSpeed, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseAutoPeekAimFromCoverControls, PsoDataType.Bool, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ComeBackInWhenAimDirectChangeInHighCover, PsoDataType.Bool, 201, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AlternateControlStickInputThreshold, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeCapsuleRadius, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeStartXOffset, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeEndXOffset, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeStartYOffset, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeEndYOffset, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InsideEdgeStartYOffset, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InsideEdgeEndYOffset, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InsideEdgeStartXOffset, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InsideEdgeEndXOffset, PsoDataType.Float, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WallTestYOffset, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialLowEdgeWallTestYOffset, PsoDataType.Float, 248, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighCloseEdgeWallTestYOffset, PsoDataType.Float, 252, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WallTestStartXOffset, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WallTestEndXOffset, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WallHighTestZOffset, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingEdgeTestStartYOffset, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingEdgeTestEndYOffset, PsoDataType.Float, 272, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverEdgeTestStartYOffset, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverEdgeTestEndYOffset, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteppingEdgeTestStartYOffset, PsoDataType.Float, 284, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteppingEdgeTestEndYOffset, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialLowEdgeTestStartYOffset, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialLowEdgeTestEndYOffset, PsoDataType.Float, 296, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeHighZOffset, PsoDataType.Float, 300, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeLowZOffset, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeMinimumOffsetDiff, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeMaximumOffsetDiff, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedDownPeekChance, PsoDataType.Float, 316, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedDownBlindFireChance, PsoDataType.Float, 320, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBeforeAllowingAutoPeek, PsoDataType.Float, 324, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableAimDirectlyIntros, PsoDataType.Bool, 328, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedDirToPedCoverCosAngleTol, PsoDataType.Float, 332, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CamToPedDirCosAngleTol, PsoDataType.Float, 336, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CamToCoverDirCosAngleTol, PsoDataType.Float, 340, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToTargetForPeek, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenPeeksWithoutLOS, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1466791495),
                     new PsoStructureEntryInfo(MetaName.ThrowProjectileClips, PsoDataType.Array, 352, 0, (MetaName)86),
                     new PsoStructureEntryInfo(MetaName.RecreateWeaponTime, PsoDataType.SInt, 368, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlindFireHighCoverMinPitchLimit, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlindFireHighCoverMaxPitchLimit, PsoDataType.Float, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableLeftHandIkInCover, PsoDataType.Bool, 380, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableReloadingWhilstMovingInCover, PsoDataType.Bool, 381, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimIntroRateForAi, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimOutroRateForAi, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinReactToFireRate, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxReactToFireRate, PsoDataType.Float, 396, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxReactToFireDelay, PsoDataType.Float, 400, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeUntilReturnToIdleFromAimAfterAimedAt, PsoDataType.Float, 404, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeUntilReturnToIdleFromAimAfterAimedAt, PsoDataType.Float, 408, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeUntilReturnToIdleFromAimDefault, PsoDataType.Float, 412, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeUntilReturnToIdleFromAimDefault, PsoDataType.Float, 416, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GlobalLateralTorsoOffsetInLeftCover, PsoDataType.Float, 420, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3668359056, PsoDataType.Float, 424, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponLongBlockingOffsetInLeftCover, PsoDataType.Float, 428, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2805643991, PsoDataType.Float, 432, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponBlockingOffsetInLeftCover, PsoDataType.Float, 436, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3824828934, PsoDataType.Float, 440, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponBlockingOffsetInRightCover, PsoDataType.Float, 444, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3593181124, PsoDataType.Float, 448, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1214766569, PsoDataType.Float, 452, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2018978213, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponBlockingLengthOffset, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2717483602, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverStepClipSetId, PsoDataType.String, 468, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedDownTakeCoverAmount, PsoDataType.Float, 472, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmountPinnedDownByDamage, PsoDataType.Float, 476, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmountPinnedDownByBullet, PsoDataType.Float, 480, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmountPinnedDownByWitnessKill, PsoDataType.Float, 484, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedDownByBulletRange, PsoDataType.Float, 488, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedDownDecreaseAmountPerSecond, PsoDataType.Float, 492, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimIntroTaskBlendOutDuration, PsoDataType.Float, 496, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToBePinnedDown, PsoDataType.UInt, 500, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenBurstsMaxRandomPercent, PsoDataType.Float, 504, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimOutroDelayTime, PsoDataType.UInt, 508, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3531438516, PsoDataType.UInt, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)796325569, PsoDataType.UInt, 516, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2152868269, PsoDataType.Bool, 520, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1536785516, PsoDataType.Bool, 521, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThrowProjectileClipSetId, PsoDataType.String, 524, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3576155857, PsoDataType.String, 528, 7, 0)
                    );
                case (MetaName)1466791495:
                    return new PsoStructureInfo((MetaName)1466791495, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.IntroClipId, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PullPinClipId, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BaseClipId, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ThrowLongClipId, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ThrowShortClipId, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1925480184, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3487669030, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1559228487),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 36, 0, (MetaName)2097159)
                    );
                case MetaName.CTaskMotionInCover__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionInCover__Tunables, 0, 0, 1112,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverClipRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableCoverToCoverTranslationScaling, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableCoverToCoverRotationScaling, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseButtonToMoveAroundCorner, PsoDataType.Bool, 22, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableCoverToCover, PsoDataType.Bool, 23, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableWalkStops, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableCoverPeekingVariations, PsoDataType.Bool, 25, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableCoverPinnedVariations, PsoDataType.Bool, 26, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableCoverIdleVariations, PsoDataType.Bool, 27, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseSprintButtonForCoverToCover, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VerifyCoverInterval, PsoDataType.UByte, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeForCornerMove, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultSettleBlendDuration, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeightChangeSettleBlendDuration, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeStayPinned, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeStayPinned, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedDownThreshold, PsoDataType.UInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForcePinnedDown, PsoDataType.Bool, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToTargetForIdleVariations, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenIdleVariations, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeBetweenIdleVariations, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinWaitTimeToPlayPlayerIdleVariations, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenPlayerIdleVariations, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeBetweenPlayerIdleVariations, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverDuration, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMinScalePhase, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMaxScalePhase, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMinRotScalePhase, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMaxRotScalePhase, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRotationalSpeedScale, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRotationalSpeed, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStickInputToEnableMoveAroundCorner, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStickInputToEnableCoverToCover, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStickInputToMoveAroundCorner, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStoppingDuration, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStoppingDist, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToScale, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCDepthDistanceCompletionOffset, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeLowCoverMoveTime, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToStandUp, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMinDistToScale, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMinAngToScale, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMinAng, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverDistTol, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMaxDistToStep, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverAngTol, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMaxAngToStep, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMaxAccel, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForwardDistToStartSideScale, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverMinDepthToScale, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverSmallAnimDist, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingReachedTolerance, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendToIdleTime, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InsideCornerStopDistance, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverHeadingCloseEnough, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverHeadingCloseEnoughTurn, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverPositionCloseEnough, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultStillToTurnBlendDuration, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultEdgeTurnBlendDuration, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PeekToEdgeTurnBlendDuration, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMoveSpeedInCover, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinEdgeDistanceForStoppingAnim, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseNewStepAndWalkStarts, PsoDataType.Bool, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseNewTurns, PsoDataType.Bool, 237, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseNewTurnWalkStarts, PsoDataType.Bool, 238, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoreMotionClipSetId, PsoDataType.String, 240, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1642317087, PsoDataType.String, 244, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CoreAIMotionClipSetId, PsoDataType.String, 248, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.PeekingVariationAnimStateInfos, PsoDataType.Array, 256, 0, (MetaName)69),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PeekingLow1HVariationClipsets, PsoDataType.Array, 272, 0, (MetaName)71),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PeekingLow2HVariationClipsets, PsoDataType.Array, 288, 0, (MetaName)73),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PeekingHigh1HVariationClipsets, PsoDataType.Array, 304, 0, (MetaName)75),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PeekingHigh2HVariationClipsets, PsoDataType.Array, 320, 0, (MetaName)77),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedLow1HVariationClipsets, PsoDataType.Array, 336, 0, (MetaName)79),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedLow2HVariationClipsets, PsoDataType.Array, 352, 0, (MetaName)81),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedHigh1HVariationClipsets, PsoDataType.Array, 368, 0, (MetaName)83),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PinnedHigh2HVariationClipsets, PsoDataType.Array, 384, 0, (MetaName)85),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OutroReact1HVariationClipsets, PsoDataType.Array, 400, 0, (MetaName)87),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OutroReact2HVariationClipsets, PsoDataType.Array, 416, 0, (MetaName)89),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IdleLow1HVariationClipsets, PsoDataType.Array, 432, 0, (MetaName)91),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IdleLow2HVariationClipsets, PsoDataType.Array, 448, 0, (MetaName)93),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IdleHigh1HVariationClipsets, PsoDataType.Array, 464, 0, (MetaName)95),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IdleHigh2HVariationClipsets, PsoDataType.Array, 480, 0, (MetaName)97),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerIdleLow0HVariationClipsets, PsoDataType.Array, 496, 0, (MetaName)99),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerIdleLow1HVariationClipsets, PsoDataType.Array, 512, 0, (MetaName)101),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerIdleLow2HVariationClipsets, PsoDataType.Array, 528, 0, (MetaName)103),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerIdleHigh0HVariationClipsets, PsoDataType.Array, 544, 0, (MetaName)105),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerIdleHigh1HVariationClipsets, PsoDataType.Array, 560, 0, (MetaName)107),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerIdleHigh2HVariationClipsets, PsoDataType.Array, 576, 0, (MetaName)109),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.PinnedIntroAnimStateInfos, PsoDataType.Array, 592, 0, (MetaName)111),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.PinnedIdleAnimStateInfos, PsoDataType.Array, 608, 0, (MetaName)113),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.PinnedOutroAnimStateInfos, PsoDataType.Array, 624, 0, (MetaName)115),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.IdleVariationAnimStateInfos, PsoDataType.Array, 640, 0, (MetaName)117),
                     new PsoStructureEntryInfo((MetaName)3920645611, PsoDataType.Structure, 656, 0, (MetaName)2625948332),
                     new PsoStructureEntryInfo((MetaName)80713034, PsoDataType.Structure, 880, 0, (MetaName)2625948332),
                     new PsoStructureEntryInfo((MetaName)1230607954, PsoDataType.Bool, 1104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1949874056, PsoDataType.Bool, 1105, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3098272438, PsoDataType.Bool, 1106, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2038512667, PsoDataType.Bool, 1107, 0, 0)
                    );
                case (MetaName)2625948332:
                    return new PsoStructureInfo((MetaName)2625948332, 0, 0, 224,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.IdleAnimStateInfos, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.AtEdgeAnimStateInfos, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.PeekingAnimStateInfos, PsoDataType.Array, 40, 0, (MetaName)4),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.StoppingAnimStateInfos, PsoDataType.Array, 56, 0, (MetaName)6),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.MovingAnimStateInfos, PsoDataType.Array, 72, 0, (MetaName)8),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.EdgeTurnAnimStateInfos, PsoDataType.Array, 88, 0, (MetaName)10),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.CoverToCoverAnimStateInfos, PsoDataType.Array, 104, 0, (MetaName)MetaTypeName.PsoPOINTER),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.SteppingAnimStateInfos, PsoDataType.Array, 120, 0, (MetaName)14),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.WalkStartAnimStateInfos, PsoDataType.Array, 136, 0, (MetaName)MetaTypeName.STRING),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.SettleAnimStateInfos, PsoDataType.Array, 152, 0, (MetaName)18),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.TurnEnterAnimStateInfos, PsoDataType.Array, 168, 0, (MetaName)20),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.TurnEndAnimStateInfos, PsoDataType.Array, 184, 0, (MetaName)22),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo(MetaName.TurnWalkStartAnimStateInfos, PsoDataType.Array, 200, 0, (MetaName)24),
                     new PsoStructureEntryInfo((MetaName)3535710087, PsoDataType.String, 216, 7, 0)
                    );
                case MetaName.CTaskAimGunFromCoverIntro__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAimGunFromCoverIntro__Tunables, 0, 0, 312,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.UpperBodyAimBlendInDuration, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IntroMovementDuration, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableIntroScaling, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableRotationScaling, PsoDataType.Bool, 25, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableIntroOverShootCheck, PsoDataType.Bool, 26, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseConstantIntroScaling, PsoDataType.Bool, 27, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RenderArcsAtCoverPosition, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RenderAimArcDebug, PsoDataType.Bool, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseMoverPositionWhilePeeking, PsoDataType.Bool, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableWeaponBlocking, PsoDataType.Bool, 31, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableTranslationScaling, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArcRadius, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IntroScalingDefaultStartPhase, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IntroScalingDefaultEndPhase, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IntroRotScalingDefaultStartPhase, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IntroRotScalingDefaultEndPhase, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AiAimIntroCloseEnoughTolerance, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStepBackDist, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStepOutDist, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStepOutDist, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IntroMaxScale, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IntroRate, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2875886470, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OutroRate, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2575356761, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteppingApproachRateSlow, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteppingApproachRate, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteppingApproachRateFast, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteppingHeadingApproachRate, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRotationalSpeedScale, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRotationalSpeedScale, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingReachedTolerance, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StepOutCapsuleRadiusScale, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimDirectlyMaxAngle, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StepOutLeftX, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StepOutRightX, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StepOutY, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowXClearOffsetCapsuleTest, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowXOffsetCapsuleTest, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowYOffsetCapsuleTest, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowZOffsetCapsuleTest, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowOffsetCapsuleLength, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowOffsetCapsuleRadius, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowLeftStep, PsoDataType.Float2, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowRightStep, PsoDataType.Float2, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowBlockedBlend, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowStepOutLeftXBlocked, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowStepOutLeftYBlocked, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowStepBackLeftXBlocked, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowStepBackLeftYBlocked, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowStepOutRightXBlocked, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowStepOutRightYBlocked, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowStepBackRightXBlocked, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowStepBackRightYBlocked, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowSideZOffset, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistConsideredAtAimPosition, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPhaseToApplyExtraHeadingAi, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPhaseToApplyExtraHeadingPlayer, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularHeadingVelocityAi, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularHeadingVelocityPlayer, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularHeadingVelocityPlayerForcedStandAim, PsoDataType.Float, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighLeftAimStepInfoSet, PsoDataType.Structure, 248, 0, (MetaName)791301518),
                     new PsoStructureEntryInfo(MetaName.HighRightAimStepInfoSet, PsoDataType.Structure, 272, 0, (MetaName)791301518),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)4162893736),
                     new PsoStructureEntryInfo(MetaName.AimIntroClips, PsoDataType.Array, 296, 0, (MetaName)64)
                    );
                case (MetaName)791301518:
                    return new PsoStructureInfo((MetaName)791301518, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2996437115),
                     new PsoStructureEntryInfo(MetaName.StepInfos, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)2996437115:
                    return new PsoStructureInfo((MetaName)2996437115, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.StepOutX, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StepOutY, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StepTransitionMinAngle, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StepTransitionMaxAngle, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreviousTransitionExtraScalar, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NextTransitionExtraScalar, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreviousTransitionClipId, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.NextTransitionClipId, PsoDataType.String, 36, 7, 0)
                    );
                case (MetaName)4162893736:
                    return new PsoStructureInfo((MetaName)4162893736, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Clips, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1559228487),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 24, 0, (MetaName)2097154)
                    );
                case MetaName.CTaskAimGunFromCoverOutro__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAimGunFromCoverOutro__Tunables, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DisableOutroScaling, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableRotationScaling, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableOutroOverShootCheck, PsoDataType.Bool, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseConstantOutroScaling, PsoDataType.Bool, 19, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OutroRotationScalingDefaultStartPhase, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OutroRotationScalingDefaultEndPhase, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OutroScalingDefaultStartPhase, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OutroScalingDefaultEndPhase, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OutroMaxScale, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AdditionalModifier, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EndHeadingTolerance, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredDistanceToCover, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InCoverMovementSpeed, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OutroMovementDuration, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpperBodyAimBlendOutDuration, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularHeadingVelocityAi, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularHeadingVelocityPlayer, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2412106599, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularHeadingVelocityPlayerForcedStandAim, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3920645611, PsoDataType.Structure, 80, 0, (MetaName)1280972310),
                     new PsoStructureEntryInfo((MetaName)80713034, PsoDataType.Structure, 104, 0, (MetaName)1280972310)
                    );
                case (MetaName)1280972310:
                    return new PsoStructureInfo((MetaName)1280972310, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3038580099),
                     new PsoStructureEntryInfo((MetaName)2828309925, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.CTaskAimGunBlindFire__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAimGunBlindFire__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RemoveReticuleDuringBlindFire, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DontRemoveReticuleDuringBlindFireNew, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowBlindFireAimingDirectlyLimitAngle, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighBlindFireAimingDirectlyLimitAngle, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)994532516),
                     new PsoStructureEntryInfo(MetaName.BlindFireAnimStateNewInfos, PsoDataType.Array, 32, 0, (MetaName)5)
                    );
                case (MetaName)994532516:
                    return new PsoStructureInfo((MetaName)994532516, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.IntroClip0Id, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IntroClip1Id, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SweepClip0Id, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SweepClip1Id, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SweepClip2Id, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OutroClip0Id, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2291961208, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3610728325, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1018316975, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CockGunWeaponClipId, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1559228487),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 48, 0, (MetaName)2097162),
                     new PsoStructureEntryInfo(MetaName.MinHeadingAngle, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeadingAngle, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3127121455, PsoDataType.Structure, 64, 0, (MetaName)4209697792),
                     new PsoStructureEntryInfo((MetaName)465339689, PsoDataType.Structure, 88, 0, (MetaName)4209697792)
                    );
                case (MetaName)4209697792:
                    return new PsoStructureInfo((MetaName)4209697792, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.MinPitchAngle, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitchAngle, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPitchAngle2H, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitchAngle2H, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CAiCoverClipVariationHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CAiCoverClipVariationHelper__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinUsesForPeekingVariationChange, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxUsesForPeekingVariationChange, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinUsesForPinnedVariationChange, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxUsesForPinnedVariationChange, PsoDataType.UInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinUsesForOutroReactVariationChange, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxUsesForOutroReactVariationChange, PsoDataType.UInt, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinUsesForIdleVariationChange, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxUsesForIdleVariationChange, PsoDataType.UInt, 44, 0, 0)
                    );
                case MetaName.CPlayerCoverClipVariationHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CPlayerCoverClipVariationHelper__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinUsesForIdleVariationChange, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxUsesForIdleVariationChange, PsoDataType.UInt, 20, 0, 0)
                    );
                case MetaName.CDynamicCoverHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CDynamicCoverHelper__Tunables, 0, 0, 176,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnableConflictingNormalCollisionRemoval, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseStickHistoryForCoverSearch, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StickDownDuration, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StickDownMinRange, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseCameraOrientationWeighting, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseCameraOrientationWhenStill, PsoDataType.Bool, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseCameraOrientationForBackwardsDirection, PsoDataType.Bool, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BehindThreshold, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToWallStanding, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToWallCrouching, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToWallCoverToCover, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OCMCrouchedForwardClearanceOffset, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OCMStandingForwardClearanceOffset, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OCMSideClearanceDepth, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OCMClearanceCapsuleRadius, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OCMSideTestDepth, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OCMCrouchedHeightOffset, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OCMStandingHeightOffset, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCSideOffset, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCProbeDepth, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCForwardOffset, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCSpacingOffset, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCCapsuleRadius, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCHeightOffset, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowCoverProbeHeight, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighCoverProbeHeight, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCClearanceCapsuleRadius, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCClearanceCapsuleStartForwardOffset, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCClearanceCapsuleEndForwardOffset, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCClearanceCapsuleStartZOffset, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCClearanceCapsuleEndZOffset, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CTCMinDistance, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleEdgeProbeXOffset, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleEdgeProbeZOffset, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZDiffBetweenCoverPoints, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZDiffBetweenPedPos, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeadingDiffBetweenCTCPoints, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedToCoverCapsuleRadius, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedToCoverEndPullBackDistance, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedToCoverEndZOffset, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStickInputAngleInfluence, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdleYStartOffset, PsoDataType.Float, 168, 0, 0)
                    );
                case MetaName.CClipScalingHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CClipScalingHelper__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DisableRotationScaling, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableRotationOvershoot, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableTranslationScaling, PsoDataType.Bool, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableTranslationOvershoot, PsoDataType.Bool, 19, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelocityToScale, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTransVelocity, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRemainingAnimDurationToScale, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAnimRotationDeltaToScale, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAnimTranslationDeltaToScale, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinCurrentRotationDeltaToScale, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultMinRotationScalingValue, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultMaxRotationScalingValue, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultMinTranslationScalingValue, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultMaxTranslationScalingValue, PsoDataType.Float, 56, 0, 0)
                    );
                case MetaName.CTaskAdvance__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAdvance__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToWaitAtPosition, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenPointUpdates, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenSeekChecksAtTacticalPoint, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskAimFromGround__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAimFromGround__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAimFromGroundTime, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskBoatChase__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskBoatChase__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceForPursue, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskBoatCombat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskBoatCombat__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 16, 0, (MetaName)384450892),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForChase, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLookAheadForCollision, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DepthForLandProbe, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToWait, PsoDataType.Float, 44, 0, 0)
                    );
                case (MetaName)384450892:
                    return new PsoStructureInfo((MetaName)384450892, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionProbe, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LandProbe, PsoDataType.Bool, 10, 0, 0)
                    );
                case MetaName.CTaskBoatStrafe__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskBoatStrafe__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AdditionalDistanceForApproach, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AdditionalDistanceForStrafe, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedForStrafe, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RotationLookAhead, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAdjustmentLookAhead, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.CTaskDraggingToSafety__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskDraggingToSafety__Tunables, 0, 0, 208,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ObstructionProbe, PsoDataType.Structure, 16, 0, (MetaName)1469164696),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 40, 0, (MetaName)512689970),
                     new PsoStructureEntryInfo(MetaName.MaxTimeForStream, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverMinDistance, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverMaxDistance, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookAtUpdateTime, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookAtTime, PsoDataType.SInt, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverWeightDistance, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverWeightUsage, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverWeightValue, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SeparationPickup, PsoDataType.Float3, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SeparationDrag, PsoDataType.Float3, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SeparationPutdown, PsoDataType.Float3, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AbortAimedAtMinDistance, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverResponseTimeout, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotForPickupDirection, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForHolster, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForPedToBeVeryCloseToCover, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxNumPedsAllowedToBeVeryCloseToCover, PsoDataType.SInt, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenCoverPointSearches, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToSetApproachPosition, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToConsiderTooClose, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToAlwaysLookAtTarget, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeightDifferenceToApproachTarget, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxXYDistanceToApproachTarget, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToBeObstructed, PsoDataType.Float, 192, 0, 0)
                    );
                case (MetaName)1469164696:
                    return new PsoStructureInfo((MetaName)1469164696, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Height, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Radius, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraHeightForGround, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)512689970:
                    return new PsoStructureInfo((MetaName)512689970, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ObstructionProbe, PsoDataType.Bool, 9, 0, 0)
                    );
                case MetaName.CTaskHeliChase__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskHeliChase__Tunables, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DriftX, PsoDataType.Structure, 16, 0, (MetaName)1407012688),
                     new PsoStructureEntryInfo(MetaName.DriftY, PsoDataType.Structure, 40, 0, (MetaName)1407012688),
                     new PsoStructureEntryInfo(MetaName.DriftZ, PsoDataType.Structure, 64, 0, (MetaName)1407012688),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveTerrain, PsoDataType.SInt, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowDownDistanceMin, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowDownDistanceMax, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeed, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForOrientation, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NearDistanceForOrientation, PsoDataType.Float, 108, 0, 0)
                    );
                case (MetaName)1407012688:
                    return new PsoStructureInfo((MetaName)1407012688, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.MinValueForCorrection, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxValueForCorrection, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRate, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskHeliCombat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskHeliCombat__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Chase, PsoDataType.Structure, 16, 0, (MetaName)960897278)
                    );
                case (MetaName)960897278:
                    return new PsoStructureInfo((MetaName)960897278, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.MinSpeed, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetOffsetX, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTargetOffsetX, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetOffsetY, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTargetOffsetY, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetOffsetZ, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTargetOffsetZ, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetOffsetZ_TargetInAir, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTargetOffsetZ_TargetInAir, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CTaskPlaneChase__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskPlaneChase__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CTaskReactToBuddyShot__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskReactToBuddyShot__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CTaskVariedAimPose__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVariedAimPose__Tunables, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBeforeCanChooseNewPose, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBeforeNewPose, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeBeforeNewPose, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceForMinTimeBeforeNewPose, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceForMaxTimeBeforeNewPose, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidNearbyPedHorizontal, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidNearbyPedVertical, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidNearbyPedDotThreshold, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRadius, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenReactions, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAnimOffsetMagnitude, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Rate, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToCareAboutBlockingLineOfSight, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToUseUrgentTransitions, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToUseUrgentTransitionsWhenThreatened, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenReactionChecksForGunAimedAt, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToReactForGunAimedAt, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxClipsToCheckPerFrame, PsoDataType.SInt, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DebugDraw, PsoDataType.Bool, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultStandingPose, PsoDataType.String, 92, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultCrouchingPose, PsoDataType.String, 96, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)896772506),
                     new PsoStructureEntryInfo(MetaName.AimPoses, PsoDataType.Array, 104, 0, (MetaName)22)
                    );
                case (MetaName)896772506:
                    return new PsoStructureInfo((MetaName)896772506, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IsCrouching, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsStationary, PsoDataType.Bool, 13, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LoopClipSetId, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LoopClipId, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2263463390),
                     new PsoStructureEntryInfo(MetaName.Transitions, PsoDataType.Array, 24, 0, (MetaName)5)
                    );
                case (MetaName)2263463390:
                    return new PsoStructureInfo((MetaName)2263463390, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ToPose, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ClipSetId, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ClipId, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Rate, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)966624441),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 24, 2, (MetaName)524292)
                    );
                case MetaName.CTaskVehicleChase__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleChase__Tunables, 0, 0, 264,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CloseDistance, PsoDataType.Structure, 16, 0, (MetaName)653343499),
                     new PsoStructureEntryInfo(MetaName.Block, PsoDataType.Structure, 32, 0, (MetaName)45194904),
                     new PsoStructureEntryInfo(MetaName.Pursue, PsoDataType.Structure, 56, 0, (MetaName)2956630463),
                     new PsoStructureEntryInfo(MetaName.Ram, PsoDataType.Structure, 80, 0, (MetaName)533527745),
                     new PsoStructureEntryInfo(MetaName.SpinOut, PsoDataType.Structure, 104, 0, (MetaName)2379651736),
                     new PsoStructureEntryInfo(MetaName.PullAlongside, PsoDataType.Structure, 128, 0, (MetaName)3309127383),
                     new PsoStructureEntryInfo(MetaName.AggressiveMove, PsoDataType.Structure, 152, 0, (MetaName)2082149421),
                     new PsoStructureEntryInfo(MetaName.Cheat, PsoDataType.Structure, 216, 0, (MetaName)3284002922),
                     new PsoStructureEntryInfo(MetaName.MaxDotForHandBrake, PsoDataType.Float, 248, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenCarChaseShockingEvents, PsoDataType.Float, 252, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceForCarChaseShockingEvents, PsoDataType.Float, 256, 0, 0)
                    );
                case (MetaName)653343499:
                    return new PsoStructureInfo((MetaName)653343499, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MinDistanceToStart, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToContinue, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)45194904:
                    return new PsoStructureInfo((MetaName)45194904, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.MaxDotToStartFromAnalyze, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDotToContinueFromAnalyze, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetSpeedToStartFromPursue, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetSpeedToContinueFromPursue, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)2956630463:
                    return new PsoStructureInfo((MetaName)2956630463, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.MinDotToStartFromAnalyze, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToContinueFromAnalyze, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistance, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)533527745:
                    return new PsoStructureInfo((MetaName)533527745, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.StraightLineDistance, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetSpeedToStartFromPursue, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetSpeedToContinueFromPursue, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)2379651736:
                    return new PsoStructureInfo((MetaName)2379651736, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.StraightLineDistance, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetSpeedToStartFromPursue, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetSpeedToContinueFromPursue, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)3309127383:
                    return new PsoStructureInfo((MetaName)3309127383, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.StraightLineDistance, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetSpeedToStartFromPursue, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetSpeedToContinueFromPursue, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)2082149421:
                    return new PsoStructureInfo((MetaName)2082149421, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToStartFromPursue, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToStartFromPursue, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedLeewayToStartFromPursue, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTargetSteerAngleToStartFromPursue, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToContinueFromPursue, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToContinueFromPursue, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeInStateToContinueFromPursue, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTargetSteerAngleToContinueFromPursue, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelay, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelay, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeightToRamFromPursue, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeightToBlockFromPursue, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeightToSpinOutFromPursue, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeightToPullAlongsideFromPursue, PsoDataType.Float, 60, 0, 0)
                    );
                case (MetaName)3284002922:
                    return new PsoStructureInfo((MetaName)3284002922, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MinSpeedDifferenceForPowerAdjustment, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedDifferenceForPowerAdjustment, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PowerForMinAdjustment, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PowerForMaxAdjustment, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.destroySuppressor, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskVehiclePersuit__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehiclePersuit__Tunables, 0, 0, 184,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ApproachTarget, PsoDataType.Structure, 16, 0, (MetaName)2385857407),
                     new PsoStructureEntryInfo(MetaName.ObstructionProbeAngleA, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ObstructionProbeAngleB, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ObstructionProbeAngleC, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceOnBikeAndTargetUnarmed, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceOnBikeAndTargetArmed, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceInVehicleAndTargetUnarmed, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceInVehicleAndTargetArmed, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceShotAt, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceCouldLeaveCar, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToStopMultiplier, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToStopMassIdeal, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToStopMassWeight, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDriverTimeToLeaveVehicle, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDriverTimeToLeaveVehicle, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPassengerTimeToLeaveVehicle, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPassengerTimeToLeaveVehicle, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForEarlyCombatExit, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToJumpOutOfVehicle, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBoatOutOfWaterForExit, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidanceMarginForOtherLawEnforcementVehicles, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToWaitForOtherPedToExit, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayExitTime, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayExitTime, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreventShufflingExtraRange, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeWaitForExitBeforeWarp, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTargetStandingOnTrainSpeed, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToFollowInCar, PsoDataType.SInt, 176, 0, 0)
                    );
                case (MetaName)2385857407:
                    return new PsoStructureInfo((MetaName)2385857407, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.TargetArriveDist, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeed, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToConsiderClose, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedWhenClose, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedWhenObstructedByLawEnforcementPed, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedWhenObstructedByLawEnforcementVehicle, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1208093369, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4019091857, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2510151940, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1955661253, PsoDataType.SInt, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2422052515, PsoDataType.SInt, 48, 0, 0)
                    );
                case MetaName.CTaskVehicleCombat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleCombat__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenShootOutTiresGlobal, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeBetweenShootOutTiresGlobal, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInCombatToShootOutTires, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeInCombatToShootOutTires, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToApplyReactionWhenShootingOutTire, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToPrepareWeapon, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToPrepareWeapon, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1726607124, PsoDataType.UInt, 44, 0, 0)
                    );
                case MetaName.CTaskSearchBase__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSearchBase__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToGiveUp, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPositionVariance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDirectionVariance, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskSearchInAutomobile__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSearchInAutomobile__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FleeOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeed, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskSearchInBoat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSearchInBoat__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CTaskSearchInHeli__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSearchInHeli__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FleeOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeightAboveTerrain, PsoDataType.SInt, 24, 0, 0)
                    );
                case MetaName.CTaskSearchOnFoot__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSearchOnFoot__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FleeOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRadius, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CompletionRadius, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowDownDistance, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeSafeDistance, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MoveBlendRatio, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.CTaskShootOutTire__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShootOutTire__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeoutToAcquireLineOfSight, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeoutToAcquireLineOfSight, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenLineOfSightChecks, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToWaitForShot, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToWaitForShot, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWaitForShotFailures, PsoDataType.SInt, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToApplyReaction, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CTaskTargetUnreachable__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskTargetUnreachable__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeBetweenRouteSearches, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskTargetUnreachableInInterior__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskTargetUnreachableInInterior__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fDirectionTestProbeLength, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskTargetUnreachableInExterior__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskTargetUnreachableInExterior__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RangePercentage, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceFromNavMesh, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRadius, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MoveBlendRatio, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CompletionRadius, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToWait, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToWait, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CPedTargetting__Tunables:
                    return new PsoStructureInfo(MetaName.CPedTargetting__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fExistingTargetScoreWeight, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTargetingInactiveDisableTime, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fBlockedLosWeighting, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeToIgnoreBlockedLosWeighting, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPlayerHighThreatWeighting, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1742395925, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iTargetNotSeenIgnoreTimeMs, PsoDataType.SInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPlayerThreatDistance, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPlayerDirectThreatDistance, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPlayerBeingTargetedExtraDistance, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iPlayerDirectThreatTimeMs, PsoDataType.SInt, 56, 0, 0)
                    );
                case MetaName.CTaskPursueCriminal__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskPursueCriminal__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToFindVehicle, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToFindVehicle, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeightDifference, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DotProductFacing, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DotProductBehind, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToFollowVehicleBeforeFlee, PsoDataType.SInt, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToSignalVehiclePursuitToCriminal, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToSignalVehiclePursuitToCriminalMin, PsoDataType.SInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToSignalVehiclePursuitToCriminalMax, PsoDataType.SInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrawDebug, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowPursuePlayer, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CriminalVehicleMinStartSpeed, PsoDataType.Float, 56, 0, 0)
                    );
                case MetaName.CTaskReactToPursuit__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskReactToPursuit__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToFleeInVehicle, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToFleeInVehicle, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeSpeedInVehicle, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskStealVehicle__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskStealVehicle__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToFindVehicle, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToPursueVehicle, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToRunToVehicle, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CanStealPlayersVehicle, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CanStealCarsAtLights, PsoDataType.Bool, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CanStealParkedCars, PsoDataType.Bool, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CanStealStationaryCars, PsoDataType.Bool, 31, 0, 0)
                    );
                case MetaName.CLookAtHistory__Tunables:
                    return new PsoStructureInfo(MetaName.CLookAtHistory__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HistoryCosineThreshold, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MemoryDuration, PsoDataType.UInt, 20, 0, 0)
                    );
                case MetaName.CAmbientLookAt__Tunables:
                    return new PsoStructureInfo(MetaName.CAmbientLookAt__Tunables, 0, 0, 416,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultLookAtThreshold, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultDistanceFromWorldCenter, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtendedDistanceFromWorldCenter, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToScanLookAts, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BaseTimeToLook, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AITimeBetweenLookAtsFailureMin, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AITimeBetweenLookAtsFailureMax, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerTimeBetweenLookAtsMin, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerTimeBetweenLookAtsMax, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerTimeMyVehicleLookAtsMin, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerTimeMyVehicleLookAtsMax, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenScenarioScans, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioScanOffsetDistance, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioScanRadius, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPlayerScore, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasicPedScore, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasicVehicleScore, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BasicObjectScore, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BehindPedModifier, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerPedModifier, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WalkingRoundPedModifier, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RunningPedModifier, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClimbingOrJumpingPedModifier, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FightingModifier, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.JackingModifier, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HangingAroundVehicleModifier, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioToScenarioPedModifier, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GangScenarioPedToPlayerModifier, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ApproachingPlayerModifier, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClosePlayerModifier, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InRangePlayerModifier, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InRangeDrivingPlayerModifier, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HoldingWeaponPlayerModifier, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoveredInBloodPlayerModifier, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollingModifier, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PickingUpBikeModifier, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecklessCarModifier, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecentlyLookedAtPlayerModifier, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecentlyLookedAtEntityModifier, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighImportanceModifier, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MediumImportanceModifier, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowImportanceModifier, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ModelNamesToConsiderPlayersForScoringPurposes, PsoDataType.Array, 184, 0, (MetaName)43),
                     new PsoStructureEntryInfo(MetaName.RecklessCarSpeedMin, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RecklessCarSpeedMax, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CarSirenModifier, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerCopModifier, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerSexyPedModifier, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerSwankyCarModifier, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerCopCarModifier, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerHasslingModifier, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HotPedMinDistance, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HotPedMaxDistance, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HotPedMinDotAngle, PsoDataType.Float, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HotPedMaxDotAngle, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HotPedMaxHeightDifference, PsoDataType.Float, 248, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InRangePlayerDistanceThreshold, PsoDataType.Float, 252, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InRangePlayerInRaceDistanceThreshold, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClosePlayerDistanceThreshold, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ApproachingPlayerDistanceThreshold, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ApproachingPlayerCosineThreshold, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollPlayerDistanceThreshold, PsoDataType.Float, 272, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookingInRangePlayerMaxDotAngle, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVelocityForVehicleLookAtSqr, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerSwankyCarMin, PsoDataType.UByte, 284, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerSwankyCarMax, PsoDataType.UByte, 285, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HotPedRenderDebug, PsoDataType.Bool, 286, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HotPedDisableSexinessFlagChecks, PsoDataType.Bool, 287, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBeforeSwitchLookAt, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLookBackAngle, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTurnSpeedMotionOverPOI, PsoDataType.Float, 296, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpeedForNarrowestAnglePickPOI, PsoDataType.Float, 300, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAnglePickPOI, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAnglePickPOI, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitchingAnglePickPOI, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerLookAtDebugDraw, PsoDataType.Bool, 316, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CameraLookAtTurnRate, PsoDataType.Enum, 320, 0, MetaName.LookIkTurnRate),
                     new PsoStructureEntryInfo(MetaName.POILookAtTurnRate, PsoDataType.Enum, 324, 0, MetaName.LookIkTurnRate),
                     new PsoStructureEntryInfo(MetaName.MotionLookAtTurnRate, PsoDataType.Enum, 328, 0, MetaName.LookIkTurnRate),
                     new PsoStructureEntryInfo(MetaName.VehicleJumpLookAtTurnRate, PsoDataType.Enum, 332, 0, MetaName.LookIkTurnRate),
                     new PsoStructureEntryInfo(MetaName.CameraLookAtBlendRate, PsoDataType.Enum, 336, 0, MetaName.LookIkBlendRate),
                     new PsoStructureEntryInfo(MetaName.POILookAtBlendRate, PsoDataType.Enum, 340, 0, MetaName.LookIkBlendRate),
                     new PsoStructureEntryInfo(MetaName.MotionLookAtBlendRate, PsoDataType.Enum, 344, 0, MetaName.LookIkBlendRate),
                     new PsoStructureEntryInfo(MetaName.VehicleJumpLookAtBlendRate, PsoDataType.Enum, 348, 0, MetaName.LookIkBlendRate),
                     new PsoStructureEntryInfo(MetaName.CameraLookAtRotationLimit, PsoDataType.Enum, 352, 0, MetaName.LookIkRotationLimit),
                     new PsoStructureEntryInfo(MetaName.POILookAtRotationLimit, PsoDataType.Enum, 356, 0, MetaName.LookIkRotationLimit),
                     new PsoStructureEntryInfo(MetaName.MotionLookAtRotationLimit, PsoDataType.Enum, 360, 0, MetaName.LookIkRotationLimit),
                     new PsoStructureEntryInfo(MetaName.AITimeWaitingToCrossRoadMin, PsoDataType.Float, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AITimeWaitingToCrossRoadMax, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAIGreetingDistanceMin, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAIGreetingDistanceMax, PsoDataType.Float, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uAITimeBetweenGreeting, PsoDataType.UInt, 380, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAIGreetingPedModifier, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uTimeBetweenLookBacks, PsoDataType.UInt, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uTimeToLookBack, PsoDataType.UInt, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uAimToIdleLookAtTime, PsoDataType.UInt, 396, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAimToIdleBreakOutAngle, PsoDataType.Float, 400, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAimToIdleAngleLimitLeft, PsoDataType.Float, 404, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAimToIdleAngleLimitRight, PsoDataType.Float, 408, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3044687592, PsoDataType.UInt, 412, 0, 0)
                    );
                case MetaName.CTaskAmbientClips__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAmbientClips__Tunables, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LowLodBaseClipSetId, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultChanceOfStandingWhileMoving, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTimeBetweenIdles, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAfterGunshotToPlayIdles, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAfterGunshotForPlayerToPlayIdles, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.playerNearToHangoutDistanceInMetersSquared, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.minSecondsNearPlayerUntilHangoutQuit, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxSecondsNearPlayerUntilHangoutQuit, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxHangoutChatDistSq, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VFXCullRangeScaleNotVisible, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SecondsSinceInWaterThatCountsAsWet, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVehicleVelocityForAmbientIdles, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSteeringAngleForAmbientIdles, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeSinceGetUpForAmbientIdles, PsoDataType.UInt, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)110858295, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)990057437, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fArgumentProbability, PsoDataType.Float, 80, 0, 0)
                    );
                case MetaName.CTaskChat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskChat__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingToleranceDegrees, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWaitTime, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskCuffed__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCuffed__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HandcuffedClipSetId, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ActionsHandcuffedNetworkId, PsoDataType.String, 20, 7, 0)
                    );
                case MetaName.CTaskInCustody__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskInCustody__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AbandonDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowRadius, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FollowOffset, PsoDataType.Float3, 32, 0, 0)
                    );
                case MetaName.CTaskFlyingWander__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskFlyingWander__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RangeOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingWanderChange, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRadius, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskPlayerOnFoot__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskPlayerOnFoot__Tunables, 0, 0, 368,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ParachutePack, PsoDataType.Structure, 16, 0, (MetaName)810561649),
                     new PsoStructureEntryInfo(MetaName.ScubaGear, PsoDataType.Structure, 88, 0, (MetaName)4252665724),
                     new PsoStructureEntryInfo((MetaName)973020790, PsoDataType.Structure, 152, 0, (MetaName)2770060121),
                     new PsoStructureEntryInfo(MetaName.EvaluateThreatFromCoverPoints, PsoDataType.Bool, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseThreatWeighting, PsoDataType.Bool, 209, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1978058747, PsoDataType.Bool, 210, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3319528567, PsoDataType.Bool, 211, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArrestDistance, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArrestDot, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxEncumberedClimbHeight, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTrainClimbHeight, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TakeCustodyDistance, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UncuffDistance, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToTalk, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToTalk, PsoDataType.Float, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenPlayerEvents, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceBetweenAiPedsCoverAndPlayersCover, PsoDataType.Float, 248, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceAiPedFromTheirCoverToAbortPlayerEnterCover, PsoDataType.Float, 252, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SmallCapsuleCoverPenalty, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SmallCapsuleCoverRadius, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PriorityCoverWeight, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EdgeCoverWeight, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistToCoverWeightThreat, PsoDataType.Float, 272, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistToCoverWeight, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistToCoverWeightNoStickBonus, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VeryCloseToCoverDist, PsoDataType.Float, 284, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VeryCloseToCoverWeight, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredDirToCoverWeight, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredDirToCoverAimingWeight, PsoDataType.Float, 296, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThreatDirWeight, PsoDataType.Float, 300, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThreatEngageDirWeight, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverDirToCameraWeightMin, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverDirToCameraWeightMax, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverDirToCameraWeightMaxAimGun, PsoDataType.Float, 316, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverDirToCameraWeightMaxScaleDist, PsoDataType.Float, 320, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredDirToCoverMinDot, PsoDataType.Float, 324, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CameraDirToCoverMinDot, PsoDataType.Float, 328, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StaticLosTest1Offset, PsoDataType.Float, 332, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StaticLosTest2Offset, PsoDataType.Float, 336, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CollisionLosHeightOffset, PsoDataType.Float, 340, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VeryCloseIgnoreDesAndCamToleranceDist, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VeryCloseIgnoreDesAndCamToleranceDistAimGun, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DeadZoneStickNorm, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SearchThreatMaxDot, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3328482583, PsoDataType.String, 360, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1221942053, PsoDataType.Bool, 364, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1138836382, PsoDataType.Bool, 365, 0, 0)
                    );
                case (MetaName)810561649:
                    return new PsoStructureInfo((MetaName)810561649, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.VelocityInheritance, PsoDataType.Structure, 8, 0, (MetaName)1276417684),
                     new PsoStructureEntryInfo(MetaName.AttachOffsetX, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOffsetY, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOffsetZ, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOrientationX, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOrientationY, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOrientationZ, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendInDeltaForPed, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendInDeltaForProp, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PhaseToBlendOut, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutDelta, PsoDataType.Float, 68, 0, 0)
                    );
                case (MetaName)1276417684:
                    return new PsoStructureInfo((MetaName)1276417684, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.X, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Y, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Z, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)4252665724:
                    return new PsoStructureInfo((MetaName)4252665724, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.VelocityInheritance, PsoDataType.Structure, 8, 0, (MetaName)2201360486),
                     new PsoStructureEntryInfo(MetaName.AttachOffsetX, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOffsetY, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOffsetZ, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOrientationX, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOrientationY, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachOrientationZ, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PhaseToBlendOut, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutDelta, PsoDataType.Float, 60, 0, 0)
                    );
                case (MetaName)2201360486:
                    return new PsoStructureInfo((MetaName)2201360486, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.X, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Y, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Z, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)2770060121:
                    return new PsoStructureInfo((MetaName)2770060121, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.VelocityInheritance, PsoDataType.Structure, 8, 0, (MetaName)549752042),
                     new PsoStructureEntryInfo(MetaName.PhaseToBlendOut, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutDelta, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3586541171, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4160387967, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2801649672, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2028074979, PsoDataType.String, 52, 7, 0)
                    );
                case (MetaName)549752042:
                    return new PsoStructureInfo((MetaName)549752042, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.X, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Y, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Z, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskSwimmingWander__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSwimmingWander__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SurfaceSkimmerDepth, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NormalPreferredDepth, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidanceProbeLength, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidanceProbePullback, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidanceProbeInterval, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AvoidanceSteerAngleDegrees, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InstantProbeDurationMin, PsoDataType.UInt, 40, 0, 0)
                    );
                case MetaName.CTaskUnalerted__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskUnalerted__Tunables, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioDelayAfterFailureMin, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioDelayAfterFailureMax, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioDelayAfterFailureWhenStationary, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioDelayAfterNotAbleToSearch, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioDelayAfterSuccessMin, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioDelayAfterSuccessMax, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioDelayInitialMin, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioDelayInitialMax, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBeforeDriverAnimCheck, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenSearchesForNextScenarioInChain, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeMinBeforeLastPoint, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeMinBeforeLastPointType, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PavementFloodFillSearchRadius, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WaitTimeAfterFailedVehExit, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToReturnToLastUsedVehicle, PsoDataType.Float, 72, 0, 0)
                    );
                case MetaName.CTaskWander__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskWander__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.uNumPedsToTransitionToRainPerPeriod, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSecondsInRainTransitionPeriod, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskWanderInArea__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskWanderInArea__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinWaitTime, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWaitTime, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CDispatchSpawnHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CDispatchSpawnHelper__Tunables, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Restrictions, PsoDataType.Structure, 16, 0, (MetaName)2369454477),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 32, 0, (MetaName)3541869149),
                     new PsoStructureEntryInfo(MetaName.IdealSpawnDistance, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotForInFront, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceTraveledMultiplier, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToBeConsideredEscapingInVehicle, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForDispatchPosition, PsoDataType.Float, 64, 0, 0)
                    );
                case (MetaName)2369454477:
                    return new PsoStructureInfo((MetaName)2369454477, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MaxDistanceFromCameraForViewportChecks, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusForViewportCheck, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)3541869149:
                    return new PsoStructureInfo((MetaName)3541869149, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DispatchNode, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FindSpawnPointInDirection, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IncidentLocation, PsoDataType.Bool, 11, 0, 0)
                    );
                case MetaName.CDispatchAdvancedSpawnHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CDispatchAdvancedSpawnHelper__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 16, 0, (MetaName)3744572044),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenInvalidateInvalidDispatchVehicles, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenMarkDispatchVehiclesForDespawn, PsoDataType.Float, 36, 0, 0)
                    );
                case (MetaName)3744572044:
                    return new PsoStructureInfo((MetaName)3744572044, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0)
                    );
                case MetaName.CDispatchHelperSearchOnFoot__Tunables:
                    return new PsoStructureInfo(MetaName.CDispatchHelperSearchOnFoot__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2350715434),
                     new PsoStructureEntryInfo(MetaName.Constraints, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceFromNavMesh, PsoDataType.Float, 32, 0, 0)
                    );
                case (MetaName)2350715434:
                    return new PsoStructureInfo((MetaName)2350715434, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.MinTimeSinceLastSpotted, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeSinceLastSpotted, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRadiusForMinTimeSinceLastSpotted, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRadiusForMaxTimeSinceLastSpotted, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeight, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngle, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseLastSeenPosition, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseByDefault, PsoDataType.Bool, 33, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseEnclosedSearchRegions, PsoDataType.Bool, 34, 0, 0)
                    );
                case MetaName.CDispatchHelperSearchInAutomobile__Tunables:
                    return new PsoStructureInfo(MetaName.CDispatchHelperSearchInAutomobile__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2350715434),
                     new PsoStructureEntryInfo(MetaName.Constraints, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceFromRoadNode, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeed, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.CDispatchHelperSearchInBoat__Tunables:
                    return new PsoStructureInfo(MetaName.CDispatchHelperSearchInBoat__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2350715434),
                     new PsoStructureEntryInfo(MetaName.Constraints, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeed, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.CDispatchHelperSearchInHeli__Tunables:
                    return new PsoStructureInfo(MetaName.CDispatchHelperSearchInHeli__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2350715434),
                     new PsoStructureEntryInfo(MetaName.Constraints, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case MetaName.CDispatchHelperVolumes__Tunables:
                    return new PsoStructureInfo(MetaName.CDispatchHelperVolumes__Tunables, 0, 0, 152,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 16, 0, (MetaName)3452608618),
                     new PsoStructureEntryInfo(MetaName.LocationForNearestCarNodeOverrides, PsoDataType.Structure, 32, 0, (MetaName)3249772165),
                     new PsoStructureEntryInfo(MetaName.EnclosedSearchRegions, PsoDataType.Structure, 72, 0, (MetaName)898276026),
                     new PsoStructureEntryInfo(MetaName.BlockingAreas, PsoDataType.Structure, 112, 0, (MetaName)2312163633)
                    );
                case (MetaName)3452608618:
                    return new PsoStructureInfo((MetaName)3452608618, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LocationForNearestCarNodeOverrides, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnclosedSearchRegions, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlockingAreas, PsoDataType.Bool, 11, 0, 0)
                    );
                case (MetaName)3249772165:
                    return new PsoStructureInfo((MetaName)3249772165, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1419890352),
                     new PsoStructureEntryInfo(MetaName.AngledAreas, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3809247475),
                     new PsoStructureEntryInfo(MetaName.Spheres, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case (MetaName)898276026:
                    return new PsoStructureInfo((MetaName)898276026, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2129766809),
                     new PsoStructureEntryInfo(MetaName.AngledAreas, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)253441159),
                     new PsoStructureEntryInfo(MetaName.Spheres, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case (MetaName)2312163633:
                    return new PsoStructureInfo((MetaName)2312163633, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3558216355),
                     new PsoStructureEntryInfo(MetaName.AngledAreas, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1335944774),
                     new PsoStructureEntryInfo(MetaName.Spheres, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case (MetaName)1419890352:
                    return new PsoStructureInfo((MetaName)1419890352, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.AngledArea, PsoDataType.Structure, 8, 0, (MetaName)3558216355),
                     new PsoStructureEntryInfo(MetaName.Position, PsoDataType.Structure, 72, 0, (MetaName)2431379611)
                    );
                case (MetaName)3558216355:
                    return new PsoStructureInfo((MetaName)3558216355, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 8, 0, (MetaName)2431379611),
                     new PsoStructureEntryInfo(MetaName.End, PsoDataType.Structure, 32, 0, (MetaName)2431379611),
                     new PsoStructureEntryInfo(MetaName.Width, PsoDataType.Float, 56, 0, 0)
                    );
                case (MetaName)2431379611:
                    return new PsoStructureInfo((MetaName)2431379611, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.X, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Y, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Z, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)2129766809:
                    return new PsoStructureInfo((MetaName)2129766809, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.AngledArea, PsoDataType.Structure, 8, 0, (MetaName)3558216355),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)4002001992),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 72, 0, (MetaName)2097153)
                    );
                case MetaName.CWantedHelicopterDispatch__Tunables:
                    return new PsoStructureInfo(MetaName.CWantedHelicopterDispatch__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenSpawnAttempts, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpawnTimeForPoliceHeliAfterDestroyed, PsoDataType.UInt, 20, 0, 0)
                    );
                case MetaName.CPoliceBoatDispatch__Tunables:
                    return new PsoStructureInfo(MetaName.CPoliceBoatDispatch__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenSpawnAttempts, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CEventPedJackingMyVehicle__Tunables:
                    return new PsoStructureInfo(MetaName.CEventPedJackingMyVehicle__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimer, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CEventEncroachingPed__Tunables:
                    return new PsoStructureInfo(MetaName.CEventEncroachingPed__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CEventRequestHelp__Tunables:
                    return new PsoStructureInfo(MetaName.CEventRequestHelp__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRangeWithoutRadioForFistFights, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimer, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimer, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CEventCrimeCryForHelp__Tunables:
                    return new PsoStructureInfo(MetaName.CEventCrimeCryForHelp__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1405855513, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4108300187, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1818772152, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2579479361, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)977600457, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.CEventSuspiciousActivity__Tunables:
                    return new PsoStructureInfo(MetaName.CEventSuspiciousActivity__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceToBeConsideredSameEvent, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CEventAgitated__Tunables:
                    return new PsoStructureInfo(MetaName.CEventAgitated__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLive, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmbientEventLifetime, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TriggerAmbientReactionChances, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForAmbientReaction, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForAmbientReaction, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeForAmbientReaction, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeForAmbientReaction, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmbientEventType, PsoDataType.Enum, 44, 0, MetaName.AmbientEventType)
                    );
                case MetaName.CEventRespondedToThreat__Tunables:
                    return new PsoStructureInfo(MetaName.CEventRespondedToThreat__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimer, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskCallPolice__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCallPolice__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeMovingAwayToGiveToWitness, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeSinceTalkingEndedToMakeCall, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeSinceTargetTalkingEndedToMakeCall, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeTargetHasBeenTalkingToMakeCall, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeSinceTalkingEndedToSayContextForCall, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeSpentInEarLoopToSayContextForCall, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToSpendInEarLoopToPutDownPhone, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSpendInEarLoopToPutDownPhone, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.CTaskConversationHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskConversationHelper__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fMinSecondsDelayBetweenPhoneLines, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxSecondsDelayBetweenPhoneLines, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinSecondsDelayBetweenChatLines, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxSecondsDelayBetweenChatLines, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceSquaredToPlayerForAudio, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fChanceOfConversationRant, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fChanceOfArgumentChallenge, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fChanceOfArgumentChallengeBeingAccepted, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uTimeInMSUntilNewWeirdPedComment, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uMaxTimeInMSToPlayRingTone, PsoDataType.UInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uTimeToWaitAfterNewSayFailureInSeconds, PsoDataType.UByte, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uTicksUntilHangoutConversationCheck, PsoDataType.UByte, 57, 0, 0)
                    );
                case MetaName.CLegIkSolver__Tunables:
                    return new PsoStructureInfo(MetaName.CLegIkSolver__Tunables, 0, 0, 392,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PelvisInterp, PsoDataType.Structure, 16, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.PelvisInterpMoving, PsoDataType.Structure, 40, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.PelvisInterpOnDynamic, PsoDataType.Structure, 64, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.FootInterp, PsoDataType.Structure, 88, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.FootInterpIntersecting, PsoDataType.Structure, 112, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.FootInterpMoving, PsoDataType.Structure, 136, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.FootInterpIntersectingMoving, PsoDataType.Structure, 160, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.FootInterpOnDynamic, PsoDataType.Structure, 184, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.StairsPelvisInterp, PsoDataType.Structure, 208, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.StairsPelvisInterpMoving, PsoDataType.Structure, 232, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.StairsPelvisInterpCoverAim, PsoDataType.Structure, 256, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.StairsFootInterp, PsoDataType.Structure, 280, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.StairsFootInterpIntersecting, PsoDataType.Structure, 304, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.StairsFootInterpCoverAim, PsoDataType.Structure, 328, 0, (MetaName)1560218317),
                     new PsoStructureEntryInfo(MetaName.UpStairsPelvisMaxDeltaZMoving, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpStairsPelvisMaxNegativeDeltaZMoving, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DownStairsPelvisMaxDeltaZMoving, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DownStairsPelvisMaxNegativeDeltaZMoving, PsoDataType.Float, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StairsPelvisMaxNegativeDeltaZCoverAim, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelMagStairsSpringMin, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelMagStairsSpringMax, PsoDataType.Float, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StairsSpringMultiplierMin, PsoDataType.Float, 380, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StairsSpringMultiplierMax, PsoDataType.Float, 384, 0, 0)
                    );
                case (MetaName)1560218317:
                    return new PsoStructureInfo((MetaName)1560218317, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Rate, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AccelerationBased, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleAccelWithDelta, PsoDataType.Bool, 13, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZeroRateOnDirectionChange, PsoDataType.Bool, 14, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AccelRate, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CMiniMap__Tunables:
                    return new PsoStructureInfo(MetaName.CMiniMap__Tunables, 0, 0, 352,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Sonar, PsoDataType.Structure, 16, 0, (MetaName)53366421),
                     new PsoStructureEntryInfo(MetaName.HealthBar, PsoDataType.Structure, 132, 0, (MetaName)1902464991),
                     new PsoStructureEntryInfo(MetaName.Bitmap, PsoDataType.Structure, 140, 0, (MetaName)2026887270),
                     new PsoStructureEntryInfo((MetaName)1791333322, PsoDataType.Structure, 172, 0, (MetaName)1690619373),
                     new PsoStructureEntryInfo(MetaName.Camera, PsoDataType.Structure, 200, 0, (MetaName)283649933),
                     new PsoStructureEntryInfo(MetaName.Tiles, PsoDataType.Structure, 300, 0, (MetaName)3031405081),
                     new PsoStructureEntryInfo(MetaName.Overlay, PsoDataType.Structure, 320, 0, (MetaName)2449081679),
                     new PsoStructureEntryInfo(MetaName.Display, PsoDataType.Structure, 340, 0, (MetaName)1640186987)
                    );
                case (MetaName)53366421:
                    return new PsoStructureInfo((MetaName)53366421, 0, 0, 116,
                     new PsoStructureEntryInfo(MetaName.fSoundRange_BarelyAudible, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_MostlyAudible, PsoDataType.Float, 4, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_ClearlyAudible, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_Whisper, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_Talking, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_Shouting, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_Megaphone, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_FootstepBase, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_HeavyFootstep, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_FootstepFoliage, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_LandFromFall, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_WeaponSpinUp, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_Gunshot, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_SilencedGunshot, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)60903513, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_Explosion, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_ObjectCollision, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_GlassBreak, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_CarHorn, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_CarLowSpeed, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_CarHighSpeed, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_WaterSplashSmall, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSoundRange_WaterSplashLarge, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinListenerRangeToDrawSonarBlips, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRainSnowSoundReductionAmount, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRadioSoundReductionAmount, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRadioSoundReductionDistance, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fConversationSoundReductionAmount, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fConversationSoundReductionDistance, PsoDataType.Float, 112, 0, 0)
                    );
                case (MetaName)1902464991:
                    return new PsoStructureInfo((MetaName)1902464991, 0, 0, 8,
                     new PsoStructureEntryInfo(MetaName.fStaminaDepletionBlinkPercentage, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iHealthDepletionBlinkPercentage, PsoDataType.SInt, 4, 0, 0)
                    );
                case (MetaName)2026887270:
                    return new PsoStructureInfo((MetaName)2026887270, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)2262853888, PsoDataType.UShort, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)735818480, PsoDataType.UShort, 2, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4138858726, PsoDataType.Float2, 4, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3205558428, PsoDataType.Float2, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1388742013, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3712817443, PsoDataType.Enum, 24, 0, (MetaName)1049598006),
                     new PsoStructureEntryInfo((MetaName)1737696498, PsoDataType.Enum, 28, 0, (MetaName)1049598006)
                    );
                case (MetaName)1690619373:
                    return new PsoStructureInfo((MetaName)1690619373, 0, 0, 28,
                     new PsoStructureEntryInfo(MetaName.fWorldX, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fWorldY, PsoDataType.Float, 4, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fWorldW, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fWorldH, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fBaseAlpha, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2170322240, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFowWaterHeight, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)283649933:
                    return new PsoStructureInfo((MetaName)283649933, 0, 0, 100,
                     new PsoStructureEntryInfo((MetaName)3708284943, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1172300147, PsoDataType.Float, 4, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fExteriorFootZoom, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fExteriorFootZoomRunning, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fExteriorFootZoomWanted, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fExteriorFootZoomWantedRunning, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1296848312, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)187030077, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInteriorFootZoom, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2652932444, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2075786872, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fParachutingZoom, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVehicleStaticZoom, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVehicleStaticWantedZoom, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVehicleMovingZoom, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVehicleMovingWantedZoom, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVehicleSpeedZoomScalar, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVehicleTilt, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVehicleOffset, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1912611271, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4158832075, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2229181525, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1723892124, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRangeZoomedScalarStandard, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRangeZoomedScalarPlane, PsoDataType.Float, 96, 0, 0)
                    );
                case (MetaName)3031405081:
                    return new PsoStructureInfo((MetaName)3031405081, 0, 0, 20,
                     new PsoStructureEntryInfo((MetaName)2152230135, PsoDataType.Float2, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4201231105, PsoDataType.Float2, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)142519492, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)866622976, PsoDataType.Bool, 17, 0, 0)
                    );
                case (MetaName)2449081679:
                    return new PsoStructureInfo((MetaName)2449081679, 0, 0, 20,
                     new PsoStructureEntryInfo(MetaName.vPos, PsoDataType.Float2, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vScale, PsoDataType.Float2, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDisplayZ, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)1640186987:
                    return new PsoStructureInfo((MetaName)1640186987, 0, 0, 12,
                     new PsoStructureEntryInfo((MetaName)717614933, PsoDataType.UShort, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)81490677, PsoDataType.UShort, 2, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2014573657, PsoDataType.Float, 4, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3269987179, PsoDataType.Float, 8, 0, 0)
                    );
                case MetaName.CTaskMotionAiming__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionAiming__Tunables, 0, 0, 144,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerMoveAccel, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerMoveDecel, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedMoveAccel, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedMoveDecel, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FromOnFootAccelerationMod, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WalkAngAccel, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RunAngAccel, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1722146938, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1350763236, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3375971933, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3611401160, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Turn180ActivationAngle, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Turn180ConsistentAngleTolerance, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Turn, PsoDataType.Structure, 72, 0, (MetaName)150934045),
                     new PsoStructureEntryInfo(MetaName.PitchChangeRate, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchChangeRateAcceleration, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OverwriteMaxPitch, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimIntroMaxAngleChangeRate, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimIntroMinPhaseChangeRate, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimIntroMaxPhaseChangeRate, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimIntroMaxTimedOutPhaseChangeRate, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerIdleIntroAnimRate, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingWalkAnimRateMin, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingWalkAnimRateMax, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingWalkAnimRateAcceleration, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoPostCameraClipUpdateForPlayer, PsoDataType.Bool, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableIkForAI, PsoDataType.Bool, 133, 0, 0),
                     new PsoStructureEntryInfo((MetaName)122800807, PsoDataType.Bool, 134, 0, 0),
                     new PsoStructureEntryInfo((MetaName)259313148, PsoDataType.Bool, 135, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2900191858, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)827074596, PsoDataType.Bool, 140, 0, 0)
                    );
                case (MetaName)150934045:
                    return new PsoStructureInfo((MetaName)150934045, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MaxVariationForCurrentPitch, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVariationForDesiredPitch, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.CTaskBirdLocomotion__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskBirdLocomotion__Tunables, 0, 0, 136,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinWaitTimeBetweenTakeOffsMS, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWaitTimeBetweenTakeOffsMS, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTakeOffRate, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTakeOffRate, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTakeOffHeadingChangeRate, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTakeOffHeadingChangeRate, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2759154614, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToFlapMin, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToFlapMax, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1388183900, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceNoTimeslicingHeadingDiff, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceFromPlayerToDeleteStuckBird, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeUntilDeletionWhenStuckOffscreen, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeWhenStuckToIgnoreBird, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighLodWalkHeadingLerpRate, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowLodWalkHeadingLerpRate, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)325409255, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2032071225, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)171073592, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2122727094, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)578605354, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)961953658, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)291845288, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1918070033, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)677105748, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2691956573, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)27278755, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2575647854, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3370383430, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)951644869, PsoDataType.String, 132, 7, 0)
                    );
                case MetaName.CTaskMotionSwimming__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionSwimming__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinStruggleTime, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStruggleTime, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)4258712195),
                     new PsoStructureEntryInfo(MetaName.ScubaGearVariations, PsoDataType.Array, 24, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.fCargoFuel),
                     new PsoStructureEntryInfo(MetaName.ScubaMaskProps, PsoDataType.Array, 40, 0, (MetaName)5)
                    );
                case (MetaName)4258712195:
                    return new PsoStructureInfo((MetaName)4258712195, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ModelName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3020467458),
                     new PsoStructureEntryInfo(MetaName.Variations, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case (MetaName)3020467458:
                    return new PsoStructureInfo((MetaName)3020467458, 0, 0, 72,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2469984956),
                     new PsoStructureEntryInfo(MetaName.Wearing, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScubaGearWithLightsOn, PsoDataType.Structure, 24, 0, (MetaName)2354064210),
                     new PsoStructureEntryInfo(MetaName.ScubaGearWithLightsOff, PsoDataType.Structure, 48, 0, (MetaName)2354064210)
                    );
                case (MetaName)2354064210:
                    return new PsoStructureInfo((MetaName)2354064210, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Component, PsoDataType.Enum, 8, 0, MetaName.ePedVarComp),
                     new PsoStructureEntryInfo(MetaName.DrawableId, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrawableAltId, PsoDataType.UInt, 16, 0, 0)
                    );
                case (MetaName)2469984956:
                    return new PsoStructureInfo((MetaName)2469984956, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Component, PsoDataType.Enum, 8, 0, MetaName.ePedVarComp),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrawableIds, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case MetaName.fCargoFuel:
                    return new PsoStructureInfo(MetaName.fCargoFuel, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.ModelName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Index, PsoDataType.SInt, 12, 0, 0)
                    );
                case MetaName.CTaskMotionBasicLocomotionLowLod__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionBasicLocomotionLowLod__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MovingExtraHeadingChangeRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingExtraHeadingChangeRateAcceleration, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceUpdatesWhenTurningStartThresholdRadians, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceUpdatesWhenTurningStopThresholdRadians, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CTaskFishLocomotion__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskFishLocomotion__Tunables, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StartTurnThresholdDegrees, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StopTurnThresholdDegrees, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTurnApproachRate, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealTurnApproachRate, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)190400066, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurnAcceleration, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2035076014, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchAcceleration, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2691956573, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1242076965, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2945548240, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3350282721, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FishOutOfWaterDelay, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1246801857, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AssistanceAngle, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraHeadingRate, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SurfaceProbeHead, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SurfaceProbeTail, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SurfacePitchLerpRate, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SurfaceHeightFallingLerpRate, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SurfaceHeightRisingLerpRate, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SurfaceHeightFollowingTriggerRange, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2334883967, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)487169156, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)951644869, PsoDataType.String, 112, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1984559896, PsoDataType.Float, 116, 0, 0)
                    );
                case MetaName.CTaskHumanLocomotion__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskHumanLocomotion__Tunables, 0, 0, 864,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Player_MBRAcceleration, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Player_MBRDeceleration, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AI_MBRAcceleration, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AI_MBRDeceleration, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FromStrafeAccelerationMod, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastWalkRateMin, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastWalkRateMax, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowRunRateMin, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowRunRateMax, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastRunRateMin, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastRunRateMax, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Turn180ActivationAngle, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Turn180ConsistentAngleTolerance, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdleHeadingLerpRate, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Player_IdleTurnRate, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AI_IdleTurnRate, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FromStrafe_WeightRate, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FromStrafe_MovingBlendOutTime, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdleTransitionBlendTimeFromActionMode, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdleTransitionBlendTimeFromStealth, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2557519583),
                     new PsoStructureEntryInfo(MetaName.MovingVarsSet, PsoDataType.Array, 96, 4, (MetaName)196629)
                    );
                case (MetaName)2557519583:
                    return new PsoStructureInfo((MetaName)2557519583, 0, 0, 256,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Standard, PsoDataType.Structure, 16, 0, (MetaName)1246324819),
                     new PsoStructureEntryInfo(MetaName.StandardAI, PsoDataType.Structure, 96, 0, (MetaName)1246324819),
                     new PsoStructureEntryInfo(MetaName.TighterTurn, PsoDataType.Structure, 176, 0, (MetaName)1246324819)
                    );
                case (MetaName)1246324819:
                    return new PsoStructureInfo((MetaName)1246324819, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingAngleMin, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingAngleMax, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingRateMin, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingRateMaxWalk, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingRateMaxRun, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingRateAccelerationMin, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingRateAccelerationMax, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingForwardAngleWalk, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingForwardAngleRun, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingForwardRateMod, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingDirectionSmoothingForwardRateAccelerationMod, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingExtraHeadingChangeAngleMin, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingExtraHeadingChangeAngleMax, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingExtraHeadingChangeRateWalk, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingExtraHeadingChangeRateRun, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingExtraHeadingChangeRateAccelerationMin, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingExtraHeadingChangeRateAccelerationMax, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseExtraHeading, PsoDataType.Bool, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseMovingDirectionDiff, PsoDataType.Bool, 77, 0, 0)
                    );
                case MetaName.CTaskMotionTennis__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionTennis__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StrafeDirectionLerpRateMinAI, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StrafeDirectionLerpRateMaxAI, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StrafeDirectionLerpRateMinPlayer, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StrafeDirectionLerpRateMaxPlayer, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CTaskQuadLocomotion__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskQuadLocomotion__Tunables, 0, 0, 184,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StartAnimatedTurnsD, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StopAnimatedTurnsD, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurnTransitionDelay, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurnToIdleTransitionDelay, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteepSlopeStartAnimatedTurnsD, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteepSlopeStopAnimatedTurnsD, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteepSlopeThresholdD, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InMotionAlignmentVelocityTolerance, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InMotionTighterTurnsVelocityTolerance, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InPlaceAlignmentVelocityTolerance, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurnSpeedMBRThreshold, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowMinTurnApproachRate, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastMinTurnApproachRate, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowTurnApproachRate, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastTurnApproachRate, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowTurnAcceleration, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastTurnAcceleration, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurnResetThresholdD, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ProcessPhysicsApproachRate, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableTimeslicingHeadingThresholdD, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowLodExtraHeadingAdjustmentRate, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartLocomotionBlendoutThreshold, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartLocomotionHeadingDeltaBlendoutThreshold, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartLocomotionDefaultBlendDuration, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartLocomotionDefaultBlendOutDuration, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartLocomotionEarlyOutBlendOutDuration, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartLocomotionWalkRunBoundary, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartToIdleDirectlyPhaseThreshold, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovementAcceleration, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinMBRToStop, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PursuitModeGallopRateFactor, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PursuitModeExtraHeadingRate, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StoppingDistanceWalkMBR, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StoppingDistanceRunMBR, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StoppingDistanceGallopMBR, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StoppingGotoPointRemainingDist, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StopPhaseThreshold, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStopPhaseToResumeMovement, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStopPhaseToResumeMovement, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo((MetaName)951644869, PsoDataType.String, 172, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3492404675, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3330274310, PsoDataType.Float, 180, 0, 0)
                    );
                case MetaName.CEventPotentialBeWalkedInto__Tunables:
                    return new PsoStructureInfo(MetaName.CEventPotentialBeWalkedInto__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3702961077, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3689187240, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1651660816, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CEventPotentialGetRunOver__Tunables:
                    return new PsoStructureInfo(MetaName.CEventPotentialGetRunOver__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToDive, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpeedToAlwaysDive, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedToDive, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToDive, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4246045187, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelay, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelay, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CEventPotentialBlast__Tunables:
                    return new PsoStructureInfo(MetaName.CEventPotentialBlast__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CTaskComplexEvasiveStep__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskComplexEvasiveStep__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutDelta, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskFall__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskFall__Tunables, 0, 0, 176,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ImmediateHighFallSpeedPlayer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ImmediateHighFallSpeedAi, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighFallProbeLength, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1694135928, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ContinuousGapHighFallTime, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3844027585, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3197570465, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)391736247, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2495576534, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InAirHeadingRate, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InAirMovementRate, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InAirMovementApproachRate, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LandHeadingModifier, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StandingLandHeadingModifier, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FallLandThreshold, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReenterFallLandThreshold, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PadShakeMinIntensity, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PadShakeMaxIntensity, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PadShakeMinHeight, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PadShakeMaxHeight, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PadShakeMinDuration, PsoDataType.UInt, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PadShakeMaxDuration, PsoDataType.UInt, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2719065213, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)358840176, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VaultFallTestAngle, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.JumpFallTestAngle, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FallTestAngleBlendOutTime, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DiveControlMaxFallDistance, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DiveControlExtraDistanceForDiveFromVehicle, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DiveControlExtraDistanceBlendOutSpeed, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DiveWaterOffsetToHitFullyInControlWeight, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LandRollHeightFromJump, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LandRollHeightFromVault, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LandRollHeight, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4229439207, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2581310221, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3807299066, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2776570354, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4036569890, PsoDataType.Float, 168, 0, 0)
                    );
                case MetaName.CTaskFlyToPoint__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskFlyToPoint__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HeightMapDelta, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeightMapLookAheadDist, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialTerrainAvoidanceAngleD, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ProgressiveTerrainAvoidanceAngleD, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenIncreasingAvoidanceAngle, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.CTaskGetUp__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskGetUp__Tunables, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fPreferInjuredGetupPlayerHealthThreshold, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInjuredGetupImpulseMag2, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeInGetUpToAllowCover, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowNonPlayerHighFallAbort, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowOffScreenHighFallAbort, PsoDataType.Bool, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FallTimeBeforeHighFallAbort, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFallSpeedForHighFallAbort, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)866980885, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerMoverFixupMaxExtraHeadingChange, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AiMoverFixupMaxExtraHeadingChange, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartClipWaitTimePlayer, PsoDataType.SInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartClipWaitTime, PsoDataType.SInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StuckWaitTime, PsoDataType.SInt, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StuckWaitTimeMp, PsoDataType.SInt, 64, 0, 0)
                    );
                case MetaName.CTaskGoToScenario__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskGoToScenario__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ClosePointDistanceSquared, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClosePointCounterMax, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingDiffStartBlendDegrees, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PositionDiffStartBlend, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExactStopTargetRadius, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreferNearWaterSurfaceArrivalRadius, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenBrokenPointChecks, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CTaskMoveToTacticalPoint__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMoveToTacticalPoint__Tunables, 0, 0, 224,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Scoring, PsoDataType.Structure, 16, 0, (MetaName)2164459263),
                     new PsoStructureEntryInfo(MetaName.TargetRadiusForMoveToPosition, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeUntilRelease, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToConsiderCloseToPositionToMoveTo, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenInfluenceSphereChecks, PsoDataType.Float, 220, 0, 0)
                    );
                case (MetaName)2164459263:
                    return new PsoStructureInfo((MetaName)2164459263, 0, 0, 192,
                     new PsoStructureEntryInfo(MetaName.CoverPoint, PsoDataType.Structure, 8, 0, (MetaName)3018405043),
                     new PsoStructureEntryInfo(MetaName.NavMeshPoint, PsoDataType.Structure, 88, 0, (MetaName)1211016943),
                     new PsoStructureEntryInfo(MetaName.Position, PsoDataType.Structure, 160, 0, (MetaName)3791189200)
                    );
                case (MetaName)3018405043:
                    return new PsoStructureInfo((MetaName)3018405043, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Base, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Bonus, PsoDataType.Structure, 16, 0, (MetaName)1890577968),
                     new PsoStructureEntryInfo(MetaName.Penalty, PsoDataType.Structure, 32, 0, (MetaName)3436208238)
                    );
                case (MetaName)1890577968:
                    return new PsoStructureInfo((MetaName)1890577968, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Current, PsoDataType.Float, 8, 0, 0)
                    );
                case (MetaName)3436208238:
                    return new PsoStructureInfo((MetaName)3436208238, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.BadRoute, PsoDataType.Structure, 8, 0, (MetaName)532562236),
                     new PsoStructureEntryInfo(MetaName.Arc, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LineOfSight, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Nearby, PsoDataType.Float, 40, 0, 0)
                    );
                case (MetaName)532562236:
                    return new PsoStructureInfo((MetaName)532562236, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ValueForMin, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMax, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Min, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Max, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)1211016943:
                    return new PsoStructureInfo((MetaName)1211016943, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Base, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Bonus, PsoDataType.Structure, 16, 0, (MetaName)1762528765),
                     new PsoStructureEntryInfo(MetaName.Penalty, PsoDataType.Structure, 32, 0, (MetaName)2194913513)
                    );
                case (MetaName)1762528765:
                    return new PsoStructureInfo((MetaName)1762528765, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Current, PsoDataType.Float, 8, 0, 0)
                    );
                case (MetaName)2194913513:
                    return new PsoStructureInfo((MetaName)2194913513, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.BadRoute, PsoDataType.Structure, 8, 0, (MetaName)333617339),
                     new PsoStructureEntryInfo(MetaName.LineOfSight, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Nearby, PsoDataType.Float, 36, 0, 0)
                    );
                case (MetaName)333617339:
                    return new PsoStructureInfo((MetaName)333617339, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ValueForMin, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMax, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Min, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Max, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)3791189200:
                    return new PsoStructureInfo((MetaName)3791189200, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MaxDistanceFromPed, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMaxDistanceFromPed, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMinDistanceFromPed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceFromOptimal, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMaxDistanceFromOptimal, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMinDistanceFromOptimal, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CTaskMoveCrossRoadAtTrafficLights__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMoveCrossRoadAtTrafficLights__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.bTrafficLightPositioning, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)645072887),
                     new PsoStructureEntryInfo(MetaName.WaitingOffsets, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo(MetaName.iMaxPedsAtTrafficLights, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceBetweenPeds, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDecideToRunChance, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPlayerObstructionCheckRadius, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPlayerObstructionRadius, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bDebugRender, PsoDataType.Bool, 60, 0, 0)
                    );
                case (MetaName)645072887:
                    return new PsoStructureInfo((MetaName)645072887, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Pos, PsoDataType.Float3, 16, 0, 0)
                    );
                case MetaName.CTaskMoveWithinAttackWindow__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMoveWithinAttackWindow__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxAngleOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinAlliesForMaxAngleOffset, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxAllyDistance, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxRandomAdditionalOffset, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxRouteDistanceModifier, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeToWait, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxTimeToWait, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CTaskMoveFollowNavMesh__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMoveFollowNavMesh__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.uRepeatedAttemptsBeforeTeleportToLeader, PsoDataType.UByte, 16, 0, 0)
                    );
                case MetaName.CTaskParachute__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskParachute__Tunables, 0, 0, 1312,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ChangeRatesForSkydiving, PsoDataType.Structure, 16, 0, (MetaName)2342773550),
                     new PsoStructureEntryInfo(MetaName.ChangeRatesForParachuting, PsoDataType.Structure, 40, 0, (MetaName)1918103038),
                     new PsoStructureEntryInfo(MetaName.FlightAngleLimitsForSkydiving, PsoDataType.Structure, 64, 0, (MetaName)3788337195),
                     new PsoStructureEntryInfo(MetaName.FlightAngleLimitsForParachutingNormal, PsoDataType.Structure, 96, 0, (MetaName)4209854321),
                     new PsoStructureEntryInfo(MetaName.FlightAngleLimitsForParachutingBraking, PsoDataType.Structure, 144, 0, (MetaName)4209854321),
                     new PsoStructureEntryInfo(MetaName.PedAngleLimitsForSkydiving, PsoDataType.Structure, 192, 0, (MetaName)3009185102),
                     new PsoStructureEntryInfo(MetaName.MoveParameters, PsoDataType.Structure, 208, 0, (MetaName)560894723),
                     new PsoStructureEntryInfo(MetaName.ForcesForSkydiving, PsoDataType.Structure, 264, 0, (MetaName)2934924657),
                     new PsoStructureEntryInfo(MetaName.ParachutingAi, PsoDataType.Structure, 280, 0, (MetaName)2046463254),
                     new PsoStructureEntryInfo(MetaName.Landing, PsoDataType.Structure, 464, 0, (MetaName)2206940602),
                     new PsoStructureEntryInfo(MetaName.CrashLanding, PsoDataType.Structure, 552, 0, (MetaName)954219518),
                     new PsoStructureEntryInfo(MetaName.Allow, PsoDataType.Structure, 592, 0, (MetaName)202257215),
                     new PsoStructureEntryInfo(MetaName.CameraSettings, PsoDataType.Structure, 616, 0, (MetaName)892940524),
                     new PsoStructureEntryInfo(MetaName.ParachutePhysics, PsoDataType.Structure, 640, 0, (MetaName)3266862316),
                     new PsoStructureEntryInfo(MetaName.ExtraForces, PsoDataType.Structure, 656, 0, (MetaName)889782022),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 1008, 0, (MetaName)629460530),
                     new PsoStructureEntryInfo(MetaName.LowLod, PsoDataType.Structure, 1024, 0, (MetaName)196547756),
                     new PsoStructureEntryInfo(MetaName.ParachuteBones, PsoDataType.Structure, 1040, 0, (MetaName)1078023650),
                     new PsoStructureEntryInfo(MetaName.Aiming, PsoDataType.Structure, 1144, 0, (MetaName)2787675897),
                     new PsoStructureEntryInfo(MetaName.PadShake, PsoDataType.Structure, 1160, 0, (MetaName)947557087),
                     new PsoStructureEntryInfo(MetaName.BrakingDifferenceForLinearVZMin, PsoDataType.Float, 1216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BrakingDifferenceForLinearVZMax, PsoDataType.Float, 1220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LinearVZForBrakingDifferenceMin, PsoDataType.Float, 1224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LinearVZForBrakingDifferenceMax, PsoDataType.Float, 1228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchRatioForLinearVZMin, PsoDataType.Float, 1232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchRatioForLinearVZMax, PsoDataType.Float, 1236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LinearVZForPitchRatioMin, PsoDataType.Float, 1240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LinearVZForPitchRatioMax, PsoDataType.Float, 1244, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBrakeForCloseUpCamera, PsoDataType.Float, 1248, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParachuteMass, PsoDataType.Float, 1252, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParachuteMassReduced, PsoDataType.Float, 1256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToLookAheadForFutureTargetPosition, PsoDataType.Float, 1260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDifferenceToAverageBrakes, PsoDataType.Float, 1264, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ModelForParachuteInSP, PsoDataType.String, 1268, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ModelForParachuteInMP, PsoDataType.String, 1272, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)974228846),
                     new PsoStructureEntryInfo(MetaName.ParachutePackVariations, PsoDataType.Array, 1280, 0, (MetaName)36),
                     new PsoStructureEntryInfo(MetaName.FirstPersonDriveByIKOffset, PsoDataType.Float3, 1296, 0, 0)
                    );
                case (MetaName)2342773550:
                    return new PsoStructureInfo((MetaName)2342773550, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Pitch, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Roll, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Yaw, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Heading, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)1918103038:
                    return new PsoStructureInfo((MetaName)1918103038, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Pitch, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Roll, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Yaw, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Brake, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)3788337195:
                    return new PsoStructureInfo((MetaName)3788337195, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MinPitch, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InflectionPitch, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxYaw, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)4209854321:
                    return new PsoStructureInfo((MetaName)4209854321, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.MinPitch, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRollFromStick, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRollFromBrake, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRoll, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxYawFromStick, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxYawFromRoll, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RollForMinYaw, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RollForMaxYaw, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxYaw, PsoDataType.Float, 44, 0, 0)
                    );
                case (MetaName)3009185102:
                    return new PsoStructureInfo((MetaName)3009185102, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MinPitch, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitch, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)560894723:
                    return new PsoStructureInfo((MetaName)560894723, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Parachuting, PsoDataType.Structure, 8, 0, (MetaName)801836577)
                    );
                case (MetaName)801836577:
                    return new PsoStructureInfo((MetaName)801836577, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.InterpRates, PsoDataType.Structure, 8, 0, (MetaName)2238420428),
                     new PsoStructureEntryInfo(MetaName.MinParachutePitch, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxParachutePitch, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxParachuteRoll, PsoDataType.Float, 40, 0, 0)
                    );
                case (MetaName)2238420428:
                    return new PsoStructureInfo((MetaName)2238420428, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.StickX, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StickY, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TotalStickInput, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CurrentHeading, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)2934924657:
                    return new PsoStructureInfo((MetaName)2934924657, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MaxThrust, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLift, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)2046463254:
                    return new PsoStructureInfo((MetaName)2046463254, 0, 0, 184,
                     new PsoStructureEntryInfo(MetaName.Target, PsoDataType.Structure, 8, 0, (MetaName)4228543633),
                     new PsoStructureEntryInfo(MetaName.Brake, PsoDataType.Structure, 24, 0, (MetaName)1021020250),
                     new PsoStructureEntryInfo(MetaName.RollForNormal, PsoDataType.Structure, 56, 0, (MetaName)1305267132),
                     new PsoStructureEntryInfo(MetaName.RollForBraking, PsoDataType.Structure, 80, 0, (MetaName)1305267132),
                     new PsoStructureEntryInfo(MetaName.PitchForNormal, PsoDataType.Structure, 104, 0, (MetaName)2131363216),
                     new PsoStructureEntryInfo(MetaName.PitchForBraking, PsoDataType.Structure, 128, 0, (MetaName)2131363216),
                     new PsoStructureEntryInfo(MetaName.Drop, PsoDataType.Structure, 152, 0, (MetaName)209420440)
                    );
                case (MetaName)4228543633:
                    return new PsoStructureInfo((MetaName)4228543633, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MinDistanceToAdjust, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Adjustment, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)1021020250:
                    return new PsoStructureInfo((MetaName)1021020250, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MaxDistance, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToStart, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceForFull, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleForMin, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleForMax, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)1305267132:
                    return new PsoStructureInfo((MetaName)1305267132, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.AngleDifferenceForMin, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleDifferenceForMax, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StickValueForMin, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StickValueForMax, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)2131363216:
                    return new PsoStructureInfo((MetaName)2131363216, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.DesiredTimeToResolveAngleDifference, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DeltaForMaxStickChange, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStickChangePerSecond, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)209420440:
                    return new PsoStructureInfo((MetaName)209420440, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MinDistance, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistance, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeight, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeight, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDot, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)2206940602:
                    return new PsoStructureInfo((MetaName)2206940602, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.NormalThresholds, PsoDataType.Structure, 8, 0, (MetaName)3954319942),
                     new PsoStructureEntryInfo(MetaName.MaxVelocityForSlow, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelocityForFast, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParachuteProbeRadius, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinStickMagnitudeForEarlyOutMovement, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FramesToLookAheadForProbe, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendDurationForEarlyOut, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngleForRunway, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LookAheadForRunway, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DropForRunway, PsoDataType.Float, 80, 0, 0)
                    );
                case (MetaName)3954319942:
                    return new PsoStructureInfo((MetaName)3954319942, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Normal, PsoDataType.Structure, 8, 0, (MetaName)1558938460),
                     new PsoStructureEntryInfo(MetaName.Braking, PsoDataType.Structure, 24, 0, (MetaName)1558938460)
                    );
                case (MetaName)1558938460:
                    return new PsoStructureInfo((MetaName)1558938460, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Forward, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Collision, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)954219518:
                    return new PsoStructureInfo((MetaName)954219518, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.NoParachuteTimeForMinCollisionNormalThreshold, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoParachuteMaxCollisionNormalThreshold, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoParachuteMinCollisionNormalThreshold, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoParachuteMaxPitch, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParachuteProbeRadius, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParachuteUpThreshold, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FramesToLookAheadForProbe, PsoDataType.Float, 32, 0, 0)
                    );
                case (MetaName)202257215:
                    return new PsoStructureInfo((MetaName)202257215, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.MinClearDistanceBelow, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFallingSpeedInRagdoll, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInRagdoll, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)892940524:
                    return new PsoStructureInfo((MetaName)892940524, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.SkyDivingCamera, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ParachuteCamera, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ParachuteCloseUpCamera, PsoDataType.String, 16, 7, 0)
                    );
                case (MetaName)3266862316:
                    return new PsoStructureInfo((MetaName)3266862316, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.ParachuteInitialVelocityY, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParachuteInitialVelocityZ, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)889782022:
                    return new PsoStructureInfo((MetaName)889782022, 0, 0, 352,
                     new PsoStructureEntryInfo(MetaName.Parachuting, PsoDataType.Structure, 8, 0, (MetaName)1386112053)
                    );
                case (MetaName)1386112053:
                    return new PsoStructureInfo((MetaName)1386112053, 0, 0, 344,
                     new PsoStructureEntryInfo(MetaName.Normal, PsoDataType.Structure, 8, 0, (MetaName)899207522),
                     new PsoStructureEntryInfo(MetaName.Braking, PsoDataType.Structure, 224, 0, (MetaName)3310740500)
                    );
                case (MetaName)899207522:
                    return new PsoStructureInfo((MetaName)899207522, 0, 0, 216,
                     new PsoStructureEntryInfo(MetaName.TurnFromStick, PsoDataType.Structure, 8, 0, (MetaName)3387028743),
                     new PsoStructureEntryInfo(MetaName.TurnFromRoll, PsoDataType.Structure, 120, 0, (MetaName)3597070589)
                    );
                case (MetaName)3387028743:
                    return new PsoStructureInfo((MetaName)3387028743, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.FromValue, PsoDataType.Structure, 8, 0, (MetaName)3597070589),
                     new PsoStructureEntryInfo(MetaName.UseVerticalAxis, PsoDataType.Bool, 104, 0, 0)
                    );
                case (MetaName)3597070589:
                    return new PsoStructureInfo((MetaName)3597070589, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.ValueForMin, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMax, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinValue, PsoDataType.Structure, 16, 0, (MetaName)162458425),
                     new PsoStructureEntryInfo(MetaName.ZeroValue, PsoDataType.Structure, 40, 0, (MetaName)162458425),
                     new PsoStructureEntryInfo(MetaName.MaxValue, PsoDataType.Structure, 64, 0, (MetaName)162458425),
                     new PsoStructureEntryInfo(MetaName.IsLocal, PsoDataType.Bool, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClearZ, PsoDataType.Bool, 89, 0, 0)
                    );
                case (MetaName)162458425:
                    return new PsoStructureInfo((MetaName)162458425, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.X, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Y, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Z, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)3310740500:
                    return new PsoStructureInfo((MetaName)3310740500, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.TurnFromStick, PsoDataType.Structure, 8, 0, (MetaName)3387028743)
                    );
                case (MetaName)629460530:
                    return new PsoStructureInfo((MetaName)629460530, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RunwayProbes, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValidityProbes, PsoDataType.Bool, 10, 0, 0)
                    );
                case (MetaName)196547756:
                    return new PsoStructureInfo((MetaName)196547756, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MinDistance, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AlwaysUse, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NeverUse, PsoDataType.Bool, 13, 0, 0)
                    );
                case (MetaName)1078023650:
                    return new PsoStructureInfo((MetaName)1078023650, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.LeftGrip, PsoDataType.Structure, 8, 0, (MetaName)2874150626),
                     new PsoStructureEntryInfo(MetaName.RightGrip, PsoDataType.Structure, 32, 0, (MetaName)2874150626),
                     new PsoStructureEntryInfo(MetaName.LeftWire, PsoDataType.Structure, 56, 0, (MetaName)2874150626),
                     new PsoStructureEntryInfo(MetaName.RightWire, PsoDataType.Structure, 80, 0, (MetaName)2874150626)
                    );
                case (MetaName)2874150626:
                    return new PsoStructureInfo((MetaName)2874150626, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.X, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Y, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Z, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseOrientationFromParachuteBone, PsoDataType.Bool, 20, 0, 0)
                    );
                case (MetaName)2787675897:
                    return new PsoStructureInfo((MetaName)2787675897, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Disabled, PsoDataType.Bool, 8, 0, 0)
                    );
                case (MetaName)947557087:
                    return new PsoStructureInfo((MetaName)947557087, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Falling, PsoDataType.Structure, 8, 0, (MetaName)3452040071),
                     new PsoStructureEntryInfo(MetaName.Deploy, PsoDataType.Structure, 40, 0, (MetaName)2930279050)
                    );
                case (MetaName)3452040071:
                    return new PsoStructureInfo((MetaName)3452040071, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchForMinIntensity, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchForMaxIntensity, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinIntensity, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxIntensity, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)2930279050:
                    return new PsoStructureInfo((MetaName)2930279050, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Intensity, PsoDataType.Float, 12, 0, 0)
                    );
                case (MetaName)974228846:
                    return new PsoStructureInfo((MetaName)974228846, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ModelName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1192596872),
                     new PsoStructureEntryInfo(MetaName.Variations, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case (MetaName)1192596872:
                    return new PsoStructureInfo((MetaName)1192596872, 0, 0, 48,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3788247444),
                     new PsoStructureEntryInfo(MetaName.Wearing, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParachutePack, PsoDataType.Structure, 24, 0, (MetaName)3901274641)
                    );
                case (MetaName)3901274641:
                    return new PsoStructureInfo((MetaName)3901274641, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Component, PsoDataType.Enum, 8, 0, MetaName.ePedVarComp),
                     new PsoStructureEntryInfo(MetaName.DrawableId, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrawableAltId, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TexId, PsoDataType.UInt, 20, 0, 0)
                    );
                case (MetaName)3788247444:
                    return new PsoStructureInfo((MetaName)3788247444, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Component, PsoDataType.Enum, 8, 0, MetaName.ePedVarComp),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DrawableIds, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case MetaName.CTaskParachuteObject__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskParachuteObject__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PhaseDuringDeployToConsiderOut, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskTakeOffPedVariation__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskTakeOffPedVariation__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CTaskHeliPassengerRappel__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskHeliPassengerRappel__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fDefaultRopeLength, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fExtraRopeLength, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fExitDescendRate, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDefaultDescendRate, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStartDescendingDistToTargetSq, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRopeUnwindRate, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinHeightToRappel, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxHeliSpeedForRappel, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.CTaskRappel__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskRappel__Tunables, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fJumpDescendRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fLongJumpDescendRate, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fJumpToSmashWindowPhaseChange, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinJumpPhaseAllowDescend, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxJumpPhaseAllowDescend, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinJumpPhaseAllowSmashWindow, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxJumpPhaseAllowSmashWindow, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinSmashWindowPhase, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fGlassBreakRadius, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fGlassDamage, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinDistanceToBreakWindow, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinStickValueAllowDescend, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bAllowSmashDuringJump, PsoDataType.Bool, 64, 0, 0)
                    );
                case MetaName.CTaskVault__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVault__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AngledClimbTheshold, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAngleForScaleVelocityExtension, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleForScaleVelocityExtension, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AngledClimbScaleVelocityExtensionMax, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableVaultForwardDot, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlideWalkAnimRate, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.CTaskJump__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskJump__Tunables, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinSuperJumpInitialVelocity, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSuperJumpInitialVelocity, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3568425813, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1477129396, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighJumpMinAngleForVelScale, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighJumpMaxAngleForVelScale, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighJumpMinVelScale, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighJumpMaxVelScale, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableJumpOnSteepStairs, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStairsJumpAngle, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bEnableJumpCollisions, PsoDataType.Bool, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bEnableJumpCollisionsMp, PsoDataType.Bool, 57, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bBlockJumpCollisionAgainstRagdollBlocked, PsoDataType.Bool, 58, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveProbeZOffset, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveBraceStartDelay, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveBraceProbeLength, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveBraceBlendInDuration, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveBraceBlendOutDuration, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveBraceMaxUpDotSlope, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveRagdollIntersectionDot, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveRagdollStartDelay, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveRagdollProbeLength, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveRagdollProbeRadius, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PredictiveRagdollRequiredVelocityMag, PsoDataType.Float, 100, 0, 0)
                    );
                case MetaName.CPedTargetEvaluator__Tunables:
                    return new PsoStructureInfo(MetaName.CPedTargetEvaluator__Tunables, 0, 0, 232,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAngularLimitMelee, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAngularLimitMeleeLockOnNoStick, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetDistanceWeightMelee, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2485794835, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetDistanceWeightMeleeRunning, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2082223604, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetHeadingWeightMelee, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3532059041, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetHeadingWeightMeleeRunning, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1974328161, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAngularLimitVehicleWeapon, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MeleeLockOnStickWeighting, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MeleeLockOnCameraWeighting, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MeleeLockOnCameraWeightingNoStick, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MeleeLockOnPedWeightingNoStick, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioHarmless, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioNeutral, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioNeutralInjured, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioIngangOrFriend, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioPotentialThreat, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioMissionThreat, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioMissionThreatCuffed, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DownedThreatModifier, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioPlayer2PlayerEveryone, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioPlayer2PlayerStrangers, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioPlayer2PlayerAttackers, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioPlayer2Player, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioPlayer2PlayerCuffed, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioScriptedHighPriority, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioMeleeDead, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioMeleeCombatThreat, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioMeleeDownedCombatThreat, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioMeleeInjured, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PrioMeleePotentialThreat, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InCoverScoreModifier, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClosestPointToLineDist, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClosestPointToLineBonusModifier, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MeleeHeadingOverride, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MeleeHeadingOverrideRunning, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MeleeHeadingFalloffPowerRunning, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultMeleeRange, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetDistanceWeightingMin, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetDistanceWeightingMax, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetHeadingWeighting, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetDistanceMaxWeightingAimTime, PsoDataType.UInt, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetDistanceFallOffMin, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetDistanceFallOffMax, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RejectLockonHeadingTheshold, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingScoreForCoverLockOnRejection, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo((MetaName)680689338, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeForTakedownTargetAcquiry, PsoDataType.UInt, 216, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3508597107, PsoDataType.UInt, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseMeleeHeadingOverride, PsoDataType.Bool, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LimitMeleeRangeToDefault, PsoDataType.Bool, 225, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DebugTargetting, PsoDataType.Bool, 226, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseNonNormalisedScoringForPlayer, PsoDataType.Bool, 227, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RejectLockIfBestTargetIsInCover, PsoDataType.Bool, 228, 0, 0)
                    );
                case MetaName.CGrabHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CGrabHelper__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Sets, PsoDataType.Structure, 16, 0, (MetaName)2070432509)
                    );
                case (MetaName)2070432509:
                    return new PsoStructureInfo((MetaName)2070432509, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.sets, PsoDataType.Map, 8, 1, (MetaName)1)
                    );
                case MetaName.CNmTuningSet:
                    return new PsoStructureInfo(MetaName.CNmTuningSet, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Id, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Priority, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.Messages, PsoDataType.Array, 24, 0, (MetaName)3)
                    );
                case MetaName.CNmMessage:
                    return new PsoStructureInfo(MetaName.CNmMessage, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.Params, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo(MetaName.ForceNewMessage, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TaskMessage, PsoDataType.Bool, 33, 0, 0)
                    );
                case MetaName.CNmParameterBool:
                    return new PsoStructureInfo(MetaName.CNmParameterBool, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.Bool, 16, 0, 0)
                    );
                case MetaName.CNmParameterFloat:
                    return new PsoStructureInfo(MetaName.CNmParameterFloat, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CNmParameterRandomFloat:
                    return new PsoStructureInfo(MetaName.CNmParameterRandomFloat, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Min, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Max, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CNmParameterInt:
                    return new PsoStructureInfo(MetaName.CNmParameterInt, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.SInt, 16, 0, 0)
                    );
                case MetaName.CTaskNMControl__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMControl__Tunables, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DriveToGetup, PsoDataType.Structure, 16, 0, (MetaName)3336326859),
                     new PsoStructureEntryInfo(MetaName.OnEnableDriveToGetup, PsoDataType.Structure, 48, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnDisableDriveToGetup, PsoDataType.Structure, 88, 0, MetaName.CNmTuningSet)
                    );
                case (MetaName)3336326859:
                    return new PsoStructureInfo((MetaName)3336326859, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.AllowDriveToGetup, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OnlyAllowForShot, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowWhenBalanced, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHealth, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxUprightRatio, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MatchTimer, PsoDataType.UInt, 24, 0, 0)
                    );
                case MetaName.CNmParameterString:
                    return new PsoStructureInfo(MetaName.CNmParameterString, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.String, 16, 8, 0)
                    );
                case MetaName.CTaskNMBehaviour__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMBehaviour__Tunables, 0, 0, 936,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnableRagdollPooling, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxGameplayNmAgents, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRageRagdolls, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReserveLocalPlayerNmAgent, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableRagdollPoolingMp, PsoDataType.Bool, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxGameplayNmAgentsMp, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRageRagdollsMp, PsoDataType.SInt, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReserveLocalPlayerNmAgentMp, PsoDataType.Bool, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlockOffscreenShotReactions, PsoDataType.Bool, 41, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UsePreEmptiveEdgeActivation, PsoDataType.Bool, 42, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UsePreEmptiveEdgeActivationMp, PsoDataType.Bool, 43, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseBalanceForEdgeActivation, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreEmptiveEdgeActivationMaxVel, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreEmptiveEdgeActivationMaxHeadingDiff, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreEmptiveEdgeActivationMinDotVel, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreEmptiveEdgeActivationMaxDistance, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreEmptiveEdgeActivationMinDesiredMBR2, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StandardBlendOutThresholds, PsoDataType.Structure, 72, 0, (MetaName)3979439077),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)4115910995),
                     new PsoStructureEntryInfo(MetaName.CamAttachPositionWeights, PsoDataType.Array, 176, 0, (MetaName)MetaTypeName.USHORT),
                     new PsoStructureEntryInfo(MetaName.SpActivationModifiers, PsoDataType.Structure, 192, 0, (MetaName)2720813484),
                     new PsoStructureEntryInfo(MetaName.MpActivationModifiers, PsoDataType.Structure, 272, 0, (MetaName)2720813484),
                     new PsoStructureEntryInfo(MetaName.PlayerBumpedByCloneCarActivationModifier, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClonePlayerBumpedByCarActivationModifier, PsoDataType.Float, 356, 0, 0),
                     //new PsoStructureEntryInfo(MetaName.ClonePlayerBumpedByCloneCarActivationModifier, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2048881690, PsoDataType.Float, 360, 0, 0), //seems to be newer than above..
                     new PsoStructureEntryInfo(MetaName.MaxVehicleCapsulePushTimeForRagdollActivation, PsoDataType.Float, 364, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVehicleCapsulePushTimeForPlayerRagdollActivation, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleMinSpeedForContinuousPushActivation, PsoDataType.Float, 372, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinContactDepthForContinuousPushActivation, PsoDataType.Float, 376, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DurationRampDownCapsulePushedByVehicle, PsoDataType.Float, 380, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleMinSpeedForAiActivation, PsoDataType.Float, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleMinSpeedForStationaryAiActivation, PsoDataType.Float, 388, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleMinSpeedForPlayerActivation, PsoDataType.Float, 392, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleMinSpeedForStationaryPlayerActivation, PsoDataType.Float, 396, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleMinSpeedForWarningActivation, PsoDataType.Float, 400, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleFallingSpeedWeight, PsoDataType.Float, 404, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleActivationForceMultiplierDefault, PsoDataType.Float, 408, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleActivationForceMultiplierBicycle, PsoDataType.Float, 412, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleActivationForceMultiplierBike, PsoDataType.Float, 416, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleActivationForceMultiplierBoat, PsoDataType.Float, 420, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleActivationForceMultiplierPlane, PsoDataType.Float, 424, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleActivationForceMultiplierQuadBike, PsoDataType.Float, 428, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleActivationForceMultiplierHeli, PsoDataType.Float, 432, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleActivationForceMultiplierTrain, PsoDataType.Float, 436, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExcludePedBumpAngleFromPushCalculation, PsoDataType.Bool, 440, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedActivationForceMultiplier, PsoDataType.Float, 444, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ObjectMinSpeedForActivation, PsoDataType.Float, 448, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ObjectActivationForceMultiplier, PsoDataType.Float, 452, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPlayerCapsulePushTimeForRagdollActivation, PsoDataType.SInt, 456, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerCapsuleMinSpeedForContinuousPushActivation, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StuckOnVehicleMaxTime, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StuckOnVehicleBlendOutThresholds, PsoDataType.Structure, 472, 0, (MetaName)3979439077),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 576, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.TeeterControl, PsoDataType.Structure, 616, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.animPoseDefault, PsoDataType.Structure, 656, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.animPoseAttachDefault, PsoDataType.Structure, 696, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.animPoseAttachToVehicle, PsoDataType.Structure, 736, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.animPoseHandsCuffed, PsoDataType.Structure, 776, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.forceFall, PsoDataType.Structure, 816, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.RagdollUnderWheelTuning, PsoDataType.Structure, 856, 0, (MetaName)267001783),
                     new PsoStructureEntryInfo(MetaName.KickOnGroundTuning, PsoDataType.Structure, 880, 0, (MetaName)2578171686),
                     new PsoStructureEntryInfo(MetaName.CapsuleVehicleHitTuning, PsoDataType.Structure, 896, 0, (MetaName)1040069698)
                    );
                case (MetaName)3979439077:
                    return new PsoStructureInfo((MetaName)3979439077, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.Ai, PsoDataType.Structure, 8, 0, (MetaName)3749122641),
                     new PsoStructureEntryInfo(MetaName.Player, PsoDataType.Structure, 40, 0, (MetaName)3749122641),
                     new PsoStructureEntryInfo(MetaName.PlayerMp, PsoDataType.Structure, 72, 0, (MetaName)3749122641)
                    );
                case (MetaName)3749122641:
                    return new PsoStructureInfo((MetaName)3749122641, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.MaxLinearVelocity, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularVelocity, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SettledTimeMS, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RandomiseSettledTime, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SettledTimeMinMS, PsoDataType.SInt, 24, 0, 0)
                    );
                case (MetaName)2720813484:
                    return new PsoStructureInfo((MetaName)2720813484, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.BumpedByCar, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByCarFriendly, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerBumpedByCar, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVehicleWarning, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByPedMinVel, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByPedMinDotVel, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByPed, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByPlayerRagdoll, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByPedRagdoll, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByPedFriendly, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByPedIsQuadruped, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumpedByObject, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Walking, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Running, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Sprinting, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPlayerActivationLimit, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAiActivationLimit, PsoDataType.Float, 72, 0, 0)
                    );
                case (MetaName)267001783:
                    return new PsoStructureInfo((MetaName)267001783, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.fImpulseMultLimbs, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fImpulseMultSpine, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinSpeedForPush, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFastCarPushImpulseMult, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)2578171686:
                    return new PsoStructureInfo((MetaName)2578171686, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.fPronePedKickImpulse, PsoDataType.Float, 8, 0, 0)
                    );
                case (MetaName)1040069698:
                    return new PsoStructureInfo((MetaName)1040069698, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.EnableActivationsFromCapsuleImpacts, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleVelToImpactNormalMinDot, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableSideSwipeActivations, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2302810051, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSideNormalForSideSwipe, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelThroughNormalForSideSwipe, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAccumulatedImpactForSideSwipe, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVehVelMagForSideSwipe, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVehVelMagForBicycleSideSwipe, PsoDataType.Float, 36, 0, 0)
                    );
                case (MetaName)4115910995:
                    return new PsoStructureInfo((MetaName)4115910995, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Bound, PsoDataType.Enum, 8, 0, (MetaName)3509057257),
                     new PsoStructureEntryInfo(MetaName.Weight, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.CTaskNMBalance__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMBalance__Tunables, 0, 0, 912,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.InitialBumpComponent, PsoDataType.Enum, 16, 0, (MetaName)3509057257),
                     new PsoStructureEntryInfo(MetaName.InitialBumpForce, PsoDataType.Structure, 24, 0, (MetaName)760338219),
                     new PsoStructureEntryInfo(MetaName.InitialBumpOffset, PsoDataType.Float3, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleStayUprightWithVel, PsoDataType.Bool, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StayUprightAtMaxVel, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVel, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StayUprightAtMinVel, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVel, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lookAtVelProbIfNoBumpTarget, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxTargetDistToUpdateFlinch, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxTargetDistToUpdateFlinchOnGround, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFlinchTargetZOffset, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFlinchTargetZOffsetOnGround, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinForwardVectorToFlinch, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinForwardVectorToFlinchOnGround, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fHeadLookZOffset, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fHeadLookZOffsetOnGround, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSteps, PsoDataType.SInt, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.timeToCatchfallMS, PsoDataType.SInt, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartWeak, PsoDataType.Structure, 176, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StartAggressive, PsoDataType.Structure, 216, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StartDefault, PsoDataType.Structure, 256, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.BumpedByPed, PsoDataType.Structure, 296, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnStairs, PsoDataType.Structure, 336, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnSteepSlope, PsoDataType.Structure, 376, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnMovingGround, PsoDataType.Structure, 416, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.LostBalanceAndGrabbing, PsoDataType.Structure, 456, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Teeter, PsoDataType.Structure, 496, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.FallOffAMovingCar, PsoDataType.Structure, 536, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.RollingFall, PsoDataType.Structure, 576, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.CatchFall, PsoDataType.Structure, 616, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailed, PsoDataType.Structure, 656, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.NotBeingPushedDelayMS, PsoDataType.SInt, 696, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NotBeingPushedOnGroundDelayMS, PsoDataType.SInt, 700, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BeingPushedOnGroundTooLongMS, PsoDataType.SInt, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OnBeingPushed, PsoDataType.Structure, 712, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBeingPushedOnGround, PsoDataType.Structure, 752, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnNotBeingPushed, PsoDataType.Structure, 792, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBeingPushedOnGroundTooLong, PsoDataType.Structure, 832, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.PushedThresholdOnGround, PsoDataType.Structure, 872, 0, (MetaName)3749122641)
                    );
                case (MetaName)760338219:
                    return new PsoStructureInfo((MetaName)760338219, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Enable, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Mag, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleWithVelocity, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityMin, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityMax, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceAtMinVelocity, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceAtMaxVelocity, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClampImpulse, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinImpulse, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxImpulse, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Delay, PsoDataType.SInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.SInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleWithUpright, PsoDataType.Bool, 56, 0, 0)
                    );
                case MetaName.CNmParameterVector:
                    return new PsoStructureInfo(MetaName.CNmParameterVector, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.Float3, 16, 0, 0)
                    );
                case MetaName.CTaskNMBrace__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMBrace__Tunables, 0, 0, 1568,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleOverrides, PsoDataType.Structure, 16, 0, (MetaName)2216090795),
                     new PsoStructureEntryInfo(MetaName.InverseMassScales, PsoDataType.Structure, 40, 0, (MetaName)1267291753),
                     new PsoStructureEntryInfo(MetaName.AngularVelocityLimits, PsoDataType.Structure, 64, 0, (MetaName)238499603),
                     new PsoStructureEntryInfo(MetaName.InitialForce, PsoDataType.Structure, 144, 0, (MetaName)2808943365),
                     new PsoStructureEntryInfo(MetaName.ChestForce, PsoDataType.Structure, 176, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.FeetLiftForce, PsoDataType.Structure, 232, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.RootLiftForce, PsoDataType.Structure, 288, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.FlipForce, PsoDataType.Structure, 344, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.CapsuleHitForce, PsoDataType.Structure, 400, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.SideSwipeForce, PsoDataType.Structure, 456, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.ForceUnderVehicle, PsoDataType.Bool, 512, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceOverVehicle, PsoDataType.Bool, 513, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChestForcePitch, PsoDataType.Float, 516, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowWarningActivations, PsoDataType.Bool, 520, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowVelocityReactionThreshold, PsoDataType.Float, 524, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FallingSpeedForHighFall, PsoDataType.Float, 528, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleCollisionElasticityMult, PsoDataType.Float, 532, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleCollisionFrictionMult, PsoDataType.Float, 536, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleCollisionNormalPitchOverVehicle, PsoDataType.Float, 540, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleCollisionNormalPitchUnderVehicle, PsoDataType.Float, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AiClearedVehicleDelay, PsoDataType.SInt, 548, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AiClearedVehicleSmartFallDelay, PsoDataType.SInt, 552, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerClearedVehicleDelay, PsoDataType.SInt, 556, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerClearedVehicleSmartFallDelay, PsoDataType.SInt, 560, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 568, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnStairs, PsoDataType.Structure, 608, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Weak, PsoDataType.Structure, 648, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailed, PsoDataType.Structure, 688, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailedStairs, PsoDataType.Structure, 728, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HighVelocity, PsoDataType.Structure, 768, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StuckUnderVehicleMaxUpright, PsoDataType.Float, 808, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StuckOnVehicle, PsoDataType.Structure, 816, 0, (MetaName)1649426998),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 1256, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Dead, PsoDataType.Structure, 1296, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OverVehicle, PsoDataType.Structure, 1336, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.UnderVehicle, PsoDataType.Structure, 1376, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.ClearedVehicle, PsoDataType.Structure, 1416, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HighVelocityBlendOut, PsoDataType.Structure, 1456, 0, (MetaName)3979439077)
                    );
                case (MetaName)2216090795:
                    return new PsoStructureInfo((MetaName)2216090795, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1768633733),
                     new PsoStructureEntryInfo(MetaName.sets, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)1267291753:
                    return new PsoStructureInfo((MetaName)1267291753, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ApplyVehicleScale, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleScale, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ApplyPedScale, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedScale, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)238499603:
                    return new PsoStructureInfo((MetaName)238499603, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Apply, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Constant, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Velocity, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Velocity2, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Max, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Delay, PsoDataType.SInt, 68, 0, 0)
                    );
                case (MetaName)2808943365:
                    return new PsoStructureInfo((MetaName)2808943365, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.VelocityMin, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityMax, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceAtMinVelocity, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceAtMaxVelocity, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleWithUpright, PsoDataType.Bool, 24, 0, 0)
                    );
                case (MetaName)3816043274:
                    return new PsoStructureInfo((MetaName)3816043274, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Apply, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleWithVelocity, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelThreshold, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVelThreshold, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelMag, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVelMag, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleWithUpright, PsoDataType.Bool, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleWithMass, PsoDataType.Bool, 29, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReduceWithPedVelocity, PsoDataType.Bool, 30, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReduceWithPedAngularVelocity, PsoDataType.Bool, 31, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OnlyInContact, PsoDataType.Bool, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OnlyNotInContact, PsoDataType.Bool, 33, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceMag, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinMag, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxMag, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.SInt, 48, 0, 0)
                    );
                case (MetaName)1649426998:
                    return new PsoStructureInfo((MetaName)1649426998, 0, 0, 440,
                     new PsoStructureEntryInfo(MetaName.InitialDelay, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UnderVehicleInitialDelay, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityThreshold, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ContinuousContactTime, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UnderVehicleVelocityThreshold, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UnderVehicleContinuousContactTime, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UnderCarMaxVelocity, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StuckOnVehicle, PsoDataType.Structure, 40, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.EndStuckOnVehicle, PsoDataType.Structure, 80, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.UpdateOnVehicle, PsoDataType.Structure, 120, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StuckUnderVehicle, PsoDataType.Structure, 160, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.EndStuckUnderVehicle, PsoDataType.Structure, 200, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StuckOnVehiclePlayer, PsoDataType.Structure, 240, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.EndStuckOnVehiclePlayer, PsoDataType.Structure, 280, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.UpdateOnVehiclePlayer, PsoDataType.Structure, 320, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StuckUnderVehiclePlayer, PsoDataType.Structure, 360, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.EndStuckUnderVehiclePlayer, PsoDataType.Structure, 400, 0, MetaName.CNmTuningSet)
                    );
                case (MetaName)1768633733:
                    return new PsoStructureInfo((MetaName)1768633733, 0, 0, 736,
                     new PsoStructureEntryInfo(MetaName.Id, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OverrideInverseMassScales, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InverseMassScales, PsoDataType.Structure, 16, 0, (MetaName)1267291753),
                     new PsoStructureEntryInfo(MetaName.OverrideReactionType, PsoDataType.Bool, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceUnderVehicle, PsoDataType.Bool, 41, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceOverVehicle, PsoDataType.Bool, 42, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleCentreZOffset, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OverrideRootLiftForce, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RootLiftForce, PsoDataType.Structure, 56, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.OverrideFlipForce, PsoDataType.Bool, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FlipForce, PsoDataType.Structure, 120, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.OverrideInitialForce, PsoDataType.Bool, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialForce, PsoDataType.Structure, 184, 0, (MetaName)2808943365),
                     new PsoStructureEntryInfo(MetaName.OverrideElasticity, PsoDataType.Bool, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleCollisionElasticityMult, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OverrideFriction, PsoDataType.Bool, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleCollisionFrictionMult, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LateralForce, PsoDataType.Structure, 232, 0, (MetaName)3816043274),
                     new PsoStructureEntryInfo(MetaName.OverrideStuckOnVehicleSets, PsoDataType.Bool, 288, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AddToBaseStuckOnVehicleSets, PsoDataType.Bool, 289, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StuckOnVehicle, PsoDataType.Structure, 296, 0, (MetaName)1649426998)
                    );
                case MetaName.CNmParameterResetMessage:
                    return new PsoStructureInfo(MetaName.CNmParameterResetMessage, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0)
                    );
                case MetaName.CTaskNMDraggingToSafety__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMDraggingToSafety__Tunables, 0, 0, 160,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Stiffness, PsoDataType.Structure, 16, 0, (MetaName)2241487041),
                     new PsoStructureEntryInfo(MetaName.DraggerArmIk, PsoDataType.Structure, 48, 0, (MetaName)1214269741),
                     new PsoStructureEntryInfo(MetaName.Constraints, PsoDataType.Structure, 96, 0, (MetaName)2410537644),
                     new PsoStructureEntryInfo(MetaName.Forces, PsoDataType.Structure, 112, 0, (MetaName)2059054874)
                    );
                case (MetaName)2241487041:
                    return new PsoStructureInfo((MetaName)2241487041, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Relaxation, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadAndNeck, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AnkleAndWrist, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)1214269741:
                    return new PsoStructureInfo((MetaName)1214269741, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftBoneOffset, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightBoneOffset, PsoDataType.Float3, 32, 0, 0)
                    );
                case (MetaName)2410537644:
                    return new PsoStructureInfo((MetaName)2410537644, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.MaxDistance, PsoDataType.Float, 8, 0, 0)
                    );
                case (MetaName)2059054874:
                    return new PsoStructureInfo((MetaName)2059054874, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftHandOffset, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightHandOffset, PsoDataType.Float3, 32, 0, 0)
                    );
                case MetaName.CTaskNMDrunk__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMDrunk__Tunables, 0, 0, 256,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fMinHeadingDeltaToFixTurn, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fHeadingRandomizationRange, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iHeadingRandomizationTimeMin, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iHeadingRandomizationTimeMax, PsoDataType.SInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fForceLeanInDirectionAmountMin, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fForceLeanInDirectionAmountMax, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fForceRampMinSpeed, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fForceRampMaxSpeed, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fHeadLookHeadingRandomizationRange, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fHeadLookPitchRandomizationRange, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iHeadLookRandomizationTimeMin, PsoDataType.SInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iHeadlookRandomizationTimeMax, PsoDataType.SInt, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinHeadingDeltaToIdleTurn, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iRunningTimeForVelocityBasedStayupright, PsoDataType.SInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStayUprightForceNonVelocityBased, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStayUprightForceMoving, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStayUprightForceIdle, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFallingSpeedForHighFall, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bUseStayUpright, PsoDataType.Bool, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bDrawIdleHeadLookTarget, PsoDataType.Bool, 89, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 96, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Base, PsoDataType.Structure, 136, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Moving, PsoDataType.Structure, 176, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Idle, PsoDataType.Structure, 216, 0, MetaName.CNmTuningSet)
                    );
                case MetaName.CTaskNMElectrocute__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMElectrocute__Tunables, 0, 0, 400,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.InitialForceComponent, PsoDataType.Enum, 16, 0, (MetaName)3509057257),
                     new PsoStructureEntryInfo(MetaName.InitialForce, PsoDataType.Structure, 24, 0, (MetaName)760338219),
                     new PsoStructureEntryInfo(MetaName.InitialForceOffset, PsoDataType.Float3, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FallingSpeedForHighFall, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 120, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Walking, PsoDataType.Structure, 160, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Running, PsoDataType.Structure, 200, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Sprinting, PsoDataType.Structure, 240, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailed, PsoDataType.Structure, 280, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnCatchFallSuccess, PsoDataType.Structure, 320, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnElectrocuteFinished, PsoDataType.Structure, 360, 0, MetaName.CNmTuningSet)
                    );
                case MetaName.CTaskNMExplosion__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMExplosion__Tunables, 0, 0, 344,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinStunnedTime, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStunnedTime, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowPlayerStunned, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseRelaxBehaviour, PsoDataType.Bool, 25, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RollUpHeightThreshold, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CatchFallHeightThresholdRollUp, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CatchFallHeightThresholdWindmill, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CatchFallHeightThresholdClipPose, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToStartCatchFall, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToStartCatchFallPlayer, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoCatchFallRelax, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CatchFallRelaxHeight, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeightToStartWrithe, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeForInitialState, PsoDataType.SInt, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeForInitialState, PsoDataType.SInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinWritheTime, PsoDataType.SInt, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWritheTime, PsoDataType.SInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceRollUp, PsoDataType.Bool, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceWindmill, PsoDataType.Bool, 81, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartWindmill, PsoDataType.Structure, 88, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StartCatchFall, PsoDataType.Structure, 128, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StartRollDownStairs, PsoDataType.Structure, 168, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 208, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Explosion, PsoDataType.Structure, 248, 0, (MetaName)3490848491)
                    );
                case (MetaName)3490848491:
                    return new PsoStructureInfo((MetaName)3490848491, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.NMBodyScale, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HumanBodyScale, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HumanPelvisScale, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HumanSpine0Scale, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HumanSpine1Scale, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AnimalBodyScale, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AnimalPelvisScale, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StrongBlastMagnitude, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastMovingPedSpeed, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceAbovePedPositionToClampPitch, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchClampMin, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchClampMax, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MagnitudeClamp, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SideScale, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchSideAngle, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchTorqueMin, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PitchTorqueMax, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlanketForceScale, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraTorqueTwistMax, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableInjuredBehaviorImpulseLimit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableInjuredBehaviorDistLimit, PsoDataType.Float, 88, 0, 0)
                    );
                case MetaName.CTaskNMFlinch__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMFlinch__Tunables, 0, 0, 624,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 16, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Passive, PsoDataType.Structure, 56, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.WaterCannon, PsoDataType.Structure, 96, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Armoured, PsoDataType.Structure, 136, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnStairs, PsoDataType.Structure, 176, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.BoundAnkles, PsoDataType.Structure, 216, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.FatallyInjured, PsoDataType.Structure, 256, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.PlayerDeath, PsoDataType.Structure, 296, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HoldingTwoHandedWeapon, PsoDataType.Structure, 336, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HoldingSingleHandedWeapon, PsoDataType.Structure, 376, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 416, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailed, PsoDataType.Structure, 456, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailedStairs, PsoDataType.Structure, 496, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.WeaponSets, PsoDataType.Structure, 536, 0, (MetaName)2070432509),
                     new PsoStructureEntryInfo(MetaName.ActionSets, PsoDataType.Structure, 568, 0, (MetaName)2070432509),
                     new PsoStructureEntryInfo(MetaName.RandomiseLeadingHand, PsoDataType.Bool, 600, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinLeanInDirectionTime, PsoDataType.SInt, 604, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLeanInDirectionTime, PsoDataType.SInt, 608, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fImpulseReductionScaleMax, PsoDataType.Float, 612, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSpecialAbilityRageKickImpulseModifier, PsoDataType.Float, 616, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCounterImpulseScale, PsoDataType.Float, 620, 0, 0)
                    );
                case MetaName.CNmParameterRandomInt:
                    return new PsoStructureInfo(MetaName.CNmParameterRandomInt, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Min, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Max, PsoDataType.SInt, 20, 0, 0)
                    );
                case MetaName.CTaskNMHighFall__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMHighFall__Tunables, 0, 0, 1112,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PitchInDirectionForce, PsoDataType.Structure, 16, 0, (MetaName)760338219),
                     new PsoStructureEntryInfo(MetaName.PitchInDirectionComponent, PsoDataType.SInt, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StuntJumpPitchInDirectionForce, PsoDataType.Structure, 88, 0, (MetaName)760338219),
                     new PsoStructureEntryInfo(MetaName.StuntJumpPitchInDirectionComponent, PsoDataType.SInt, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighFallTimeToBlockInjuredOnGround, PsoDataType.SInt, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 160, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.InAir, PsoDataType.Structure, 200, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Vault, PsoDataType.Structure, 240, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.FromCarHit, PsoDataType.Structure, 280, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SlopeSlide, PsoDataType.Structure, 320, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.TeeterEdge, PsoDataType.Structure, 360, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SprintExhausted, PsoDataType.Structure, 400, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.JumpCollision, PsoDataType.Structure, 440, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StuntJump, PsoDataType.Structure, 480, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailedSprintExhausted, PsoDataType.Structure, 520, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.DisableStartMessageForSprintExhausted, PsoDataType.Bool, 560, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 568, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.BlendOut, PsoDataType.Structure, 608, 0, (MetaName)3979439077),
                     new PsoStructureEntryInfo(MetaName.PlayerQuickBlendOut, PsoDataType.Structure, 712, 0, (MetaName)3749122641),
                     new PsoStructureEntryInfo(MetaName.MpPlayerQuickBlendOut, PsoDataType.Structure, 744, 0, (MetaName)3749122641),
                     new PsoStructureEntryInfo(MetaName.MaxHealthLossForQuickGetup, PsoDataType.Float, 776, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHealthForQuickGetup, PsoDataType.Float, 780, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MpMaxHealthLossForQuickGetup, PsoDataType.Float, 784, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MpMinHealthForQuickGetup, PsoDataType.Float, 788, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseRemainingMinTimeForGroundWrithe, PsoDataType.Bool, 792, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeRemainingForGroundWrithe, PsoDataType.SInt, 796, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeElapsedForGroundWrithe, PsoDataType.SInt, 800, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighHighFallStart, PsoDataType.Structure, 808, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SuperHighFallStart, PsoDataType.Structure, 848, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HighHighFallEnd, PsoDataType.Structure, 888, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.AirResistanceOption, PsoDataType.UByte, 928, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceZThresholdForHighHighFall, PsoDataType.Float, 932, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityZThresholdForHighHighFall, PsoDataType.Float, 936, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityZThresholdForSuperHighFall, PsoDataType.Float, 940, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollComponentAirResistanceForce, PsoDataType.Array, 944, 4, (MetaName)1376291),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollComponentAirResistanceMinStiffness, PsoDataType.Array, 1028, 4, (MetaName)1310757)
                    );
                case MetaName.CTaskNMInjuredOnGround__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMInjuredOnGround__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fDoInjuredOnGroundChance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFallingSpeedThreshold, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iRandomDurationMin, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iRandomDurationMax, PsoDataType.SInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iMaxNumInjuredOnGroundAgents, PsoDataType.SInt, 32, 0, 0)
                    );
                case MetaName.CTaskNMJumpRollFromRoadVehicle__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMJumpRollFromRoadVehicle__Tunables, 0, 0, 304,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GravityScale, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartForceDownHeight, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 24, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.EntryPointSets, PsoDataType.Structure, 64, 0, (MetaName)2070432509),
                     new PsoStructureEntryInfo(MetaName.BlendOut, PsoDataType.Structure, 96, 0, (MetaName)3979439077),
                     new PsoStructureEntryInfo(MetaName.QuickBlendOut, PsoDataType.Structure, 200, 0, (MetaName)3979439077)
                    );
                case MetaName.CTaskNMOnFire__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMOnFire__Tunables, 0, 0, 176,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 16, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Weak, PsoDataType.Structure, 56, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 96, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailed, PsoDataType.Structure, 136, 0, MetaName.CNmTuningSet)
                    );
                case MetaName.CTaskNMPrototype__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMPrototype__Tunables, 0, 0, 288,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RunForever, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SimulationTimeInMs, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CheckForMovingGround, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 32, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 72, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailed, PsoDataType.Structure, 112, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CTaskNMPrototype__Tunables__TimedTuning),
                     new PsoStructureEntryInfo(MetaName.TimedMessages, PsoDataType.Array, 152, 0, (MetaName)MetaTypeName.POINTER),
                     new PsoStructureEntryInfo(MetaName.DynamicSet1, PsoDataType.Structure, 168, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.DynamicSet2, PsoDataType.Structure, 208, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.DynamicSet3, PsoDataType.Structure, 248, 0, MetaName.CNmTuningSet)
                    );
                case MetaName.CTaskNMPrototype__Tunables__TimedTuning:
                    return new PsoStructureInfo(MetaName.CTaskNMPrototype__Tunables__TimedTuning, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.TimeInSeconds, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Periodic, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Messages, PsoDataType.Structure, 16, 0, MetaName.CNmTuningSet)
                    );
                case MetaName.CTaskNMRiverRapids__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMRiverRapids__Tunables, 0, 0, 272,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fMinRiverFlowForRapids, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinRiverGroundClearanceForRapids, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bHorizontalRighting, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fHorizontalRightingStrength, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fHorizontalRightingTime, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bVerticalRighting, PsoDataType.Bool, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVerticalRightingStrength, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVerticalRightingTime, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRagdollComponentBuoyancy, PsoDataType.Array, 48, 4, (MetaName)1376265),
                     new PsoStructureEntryInfo(MetaName.BodyWrithe, PsoDataType.Structure, 136, 0, (MetaName)2361842970),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 192, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 232, 0, MetaName.CNmTuningSet)
                    );
                case (MetaName)2361842970:
                    return new PsoStructureInfo((MetaName)2361842970, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.bControlledByPlayerSprintInput, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinArmAmplitude, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxArmAmplitude, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinArmStiffness, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxArmStiffness, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinArmPeriod, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxArmPeriod, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinStroke, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxStroke, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinBuoyancy, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxBuoyancy, PsoDataType.Float, 48, 0, 0)
                    );
                case MetaName.CTaskNMShot__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMShot__Tunables, 0, 0, 2304,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinimumShotReactionTimePlayerMS, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinimumShotReactionTimeAIMS, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bUseClipPoseHelper, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bEnableDebugDraw, PsoDataType.Bool, 25, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fImpactConeAngleFront, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fImpactConeAngleBack, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.eImpactConeRagdollComponent, PsoDataType.Enum, 36, 0, (MetaName)3509057257),
                     new PsoStructureEntryInfo(MetaName.iShotMinTimeBeforeGunThreaten, PsoDataType.SInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iShotMaxTimeBeforeGunThreaten, PsoDataType.SInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iShotMinTimeBetweenFireGun, PsoDataType.SInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iShotMaxTimeBetweenFireGun, PsoDataType.SInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iShotMaxBlindFireTimeL, PsoDataType.SInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iShotMaxBlindFireTimeH, PsoDataType.SInt, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutDelayStanding, PsoDataType.SInt, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutDelayBalanceFailed, PsoDataType.SInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fShotBlindFireProbability, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fShotWeaponAngleToFireGun, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fShotHeadAngleToFireGun, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFireWeaponStrengthForceMultiplier, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinFinisherShotgunTotalImpulseNormal, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinFinisherShotgunTotalImpulseBraced, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFinisherShotgunBonusArmedSpeedModifier, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleSnapWithSpineOrientation, PsoDataType.Bool, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSnap, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShotAgainstWall, PsoDataType.Structure, 112, 0, (MetaName)2342621138),
                     new PsoStructureEntryInfo(MetaName.BCRExclusionZone, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Impulses, PsoDataType.Structure, 152, 0, (MetaName)209088469),
                     new PsoStructureEntryInfo(MetaName.HitRandomisation, PsoDataType.Structure, 352, 0, (MetaName)996969243),
                     new PsoStructureEntryInfo(MetaName.HitRandomisationAutomatic, PsoDataType.Structure, 392, 0, (MetaName)996969243),
                     new PsoStructureEntryInfo(MetaName.StayUpright, PsoDataType.Structure, 432, 0, (MetaName)1001943080),
                     new PsoStructureEntryInfo(MetaName.ArmShot, PsoDataType.Structure, 464, 0, (MetaName)2149258416),
                     new PsoStructureEntryInfo(MetaName.FallingSpeedForHighFall, PsoDataType.Float, 504, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReduceDownedTimeByPerformanceTime, PsoDataType.Bool, 508, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinimumDownedTime, PsoDataType.SInt, 512, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChanceOfFallToKneesOnCollapse, PsoDataType.Float, 516, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChanceOfFallToKneesAfterLastStand, PsoDataType.Float, 520, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChanceForGutShotKnockdown, PsoDataType.Float, 524, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LastStandMaxTotalTime, PsoDataType.Float, 528, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LastStandMaxArmouredTotalTime, PsoDataType.Float, 532, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RapidHitCount, PsoDataType.SInt, 536, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArmouredRapidHitCount, PsoDataType.SInt, 540, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowArmouredLegShot, PsoDataType.Bool, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowArmouredKnockdown, PsoDataType.Bool, 545, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableReachForWoundOnHeadShot, PsoDataType.Bool, 546, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableReachForWoundOnHeadShotMinDelay, PsoDataType.SInt, 548, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableReachForWoundOnHeadShotMaxDelay, PsoDataType.SInt, 552, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableReachForWoundOnNeckShot, PsoDataType.Bool, 556, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableReachForWoundOnNeckShotMinDelay, PsoDataType.SInt, 560, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableReachForWoundOnNeckShotMaxDelay, PsoDataType.SInt, 564, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParamSets, PsoDataType.Structure, 568, 0, (MetaName)436100430),
                     new PsoStructureEntryInfo(MetaName.WeaponSets, PsoDataType.Structure, 2136, 0, (MetaName)2070432509),
                     new PsoStructureEntryInfo(MetaName.BlendOutThreshold, PsoDataType.Structure, 2168, 0, (MetaName)3979439077),
                     new PsoStructureEntryInfo(MetaName.SubmergedBlendOutThreshold, PsoDataType.Structure, 2272, 0, (MetaName)3749122641)
                    );
                case (MetaName)2342621138:
                    return new PsoStructureInfo((MetaName)2342621138, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.HealthRatioLimit, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WallProbeRadius, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WallProbeDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ProbeHeightAbovePelvis, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ImpulseMult, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWallAngle, PsoDataType.Float, 28, 0, 0)
                    );
                case (MetaName)209088469:
                    return new PsoStructureInfo((MetaName)209088469, 0, 0, 200,
                     new PsoStructureEntryInfo(MetaName.MaxArmourImpulseMult, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinArmourImpulseMult, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHealthImpulseMult, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHealthImpulseMult, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDamageTakenImpulseMult, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDamageTakenImpulseMult, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDamageTakenThreshold, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDamageTakenThreshold, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultKillShotImpulseMult, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultRapidFireKillShotImpulseMult, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultMPKillShotImpulseMult, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultMPRapidFireKillShotImpulseMult, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShotgunMaxSpeedForLiftImpulse, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShotgunMaxLiftImpulse, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShotgunLiftNearThreshold, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShotgunChanceToMoveSpine3ImpulseToSpine2, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShotgunChanceToMoveNeckImpulseToSpine2, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShotgunChanceToMoveHeadImpulseToSpine2, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RapidFireBoostShotImpulseMult, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RapidFireBoostShotMinRandom, PsoDataType.SInt, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RapidFireBoostShotMaxRandom, PsoDataType.SInt, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EqualizeAmount, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.COMImpulseScale, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)3509057257),
                     new PsoStructureEntryInfo(MetaName.COMImpulseComponent, PsoDataType.Flags, 100, 0, (MetaName)2097175),
                     new PsoStructureEntryInfo(MetaName.COMImpulseMaxRootVelocityMagnitude, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.COMImpulseOnlyWhileBalancing, PsoDataType.Bool, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadShotImpulseMultiplier, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadShotMPImpulseMultiplier, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ScaleHeadShotImpulseWithSpineOrientation, PsoDataType.Bool, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeadShotImpulseMultiplier, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AutomaticInitialSnapMult, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BurstFireInitialSnapMult, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FinalShotImpulseClampMax, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RunningAgainstBulletImpulseMult, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RunningAgainstBulletImpulseMultMax, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RunningWithBulletImpulseMult, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LegShotFallRootImpulseMinUpright, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LegShotFallRootImpulseMult, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SniperImpulses, PsoDataType.Structure, 160, 0, (MetaName)3192684380)
                    );
                case (MetaName)3192684380:
                    return new PsoStructureInfo((MetaName)3192684380, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.MaxHealthImpulseMult, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHealthImpulseMult, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDamageTakenImpulseMult, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDamageTakenImpulseMult, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDamageTakenThreshold, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDamageTakenThreshold, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultKillShotImpulseMult, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultMPKillShotImpulseMult, PsoDataType.Float, 36, 0, 0)
                    );
                case (MetaName)996969243:
                    return new PsoStructureInfo((MetaName)996969243, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Enable, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TopSpread, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BottomSpread, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Blend, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TopComponent, PsoDataType.Enum, 24, 0, (MetaName)3509057257),
                     new PsoStructureEntryInfo(MetaName.BottomComponent, PsoDataType.Enum, 28, 0, (MetaName)3509057257),
                     new PsoStructureEntryInfo(MetaName.BiasSide, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BiasSideTime, PsoDataType.UInt, 36, 0, 0)
                    );
                case (MetaName)1001943080:
                    return new PsoStructureInfo((MetaName)1001943080, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.HoldingWeaponBonus, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UnarmedBonus, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArmouredBonus, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingMultiplierBonus, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HealthMultiplierBonus, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)2149258416:
                    return new PsoStructureInfo((MetaName)2149258416, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.MinLookAtArmWoundTime, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLookAtArmWoundTime, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpperArmImpulseCap, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowerArmImpulseCap, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClavicleImpulseScale, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpperArmNoTorsoHitImpulseCap, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowerArmNoTorseHitImpulseCap, PsoDataType.Float, 32, 0, 0)
                    );
                case (MetaName)436100430:
                    return new PsoStructureInfo((MetaName)436100430, 0, 0, 1568,
                     new PsoStructureEntryInfo(MetaName.Base, PsoDataType.Structure, 8, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Melee, PsoDataType.Structure, 48, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Electrocute, PsoDataType.Structure, 88, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SprintingLegShot, PsoDataType.Structure, 128, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SprintingDeath, PsoDataType.Structure, 168, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Sprinting, PsoDataType.Structure, 208, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.AutomaticHeadShot, PsoDataType.Structure, 248, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HeadShot, PsoDataType.Structure, 288, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.AutomaticNeckShot, PsoDataType.Structure, 328, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.NeckShot, PsoDataType.Structure, 368, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SniperLegShot, PsoDataType.Structure, 408, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.LegShot, PsoDataType.Structure, 448, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.ArmShot, PsoDataType.Structure, 488, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.BackShot, PsoDataType.Structure, 528, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Underwater, PsoDataType.Structure, 568, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.UnderwaterRelax, PsoDataType.Structure, 608, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Armoured, PsoDataType.Structure, 648, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.BoundAnkles, PsoDataType.Structure, 688, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.FatallyInjured, PsoDataType.Structure, 728, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.PlayerDeathSP, PsoDataType.Structure, 768, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.PlayerDeathMP, PsoDataType.Structure, 808, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnStairs, PsoDataType.Structure, 848, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.ShotAgainstWall, PsoDataType.Structure, 888, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.LastStand, PsoDataType.Structure, 928, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.LastStandArmoured, PsoDataType.Structure, 968, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HeadLook, PsoDataType.Structure, 1008, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.FallToKnees, PsoDataType.Structure, 1048, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.StaggerFall, PsoDataType.Structure, 1088, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.CatchFall, PsoDataType.Structure, 1128, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SetFallingReactionHealthy, PsoDataType.Structure, 1168, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SetFallingReactionInjured, PsoDataType.Structure, 1208, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SetFallingReactionFallOverWall, PsoDataType.Structure, 1248, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.SetFallingReactionFallOverVehicle, PsoDataType.Structure, 1288, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.RubberBulletKnockdown, PsoDataType.Structure, 1328, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Teeter, PsoDataType.Structure, 1368, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HoldingTwoHandedWeapon, PsoDataType.Structure, 1408, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.HoldingSingleHandedWeapon, PsoDataType.Structure, 1448, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.CrouchedOrLowCover, PsoDataType.Structure, 1488, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Female, PsoDataType.Structure, 1528, 0, MetaName.CNmTuningSet)
                    );
                case MetaName.CTaskNMSimple__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMSimple__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2610655349),
                     new PsoStructureEntryInfo(MetaName.Tuning, PsoDataType.Map, 16, 1, (MetaName)65538)
                    );
                case (MetaName)2610655349:
                    return new PsoStructureInfo((MetaName)2610655349, 0, 0, 176,
                     new PsoStructureEntryInfo(MetaName.iMinTime, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iMaxTime, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRagdollScore, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 24, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 64, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.OnBalanceFailure, PsoDataType.Structure, 104, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.BlendOutThreshold, PsoDataType.Structure, 144, 0, (MetaName)3749122641)
                    );
                case MetaName.CTaskNMThroughWindscreen__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMThroughWindscreen__Tunables, 0, 0, 232,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GravityScale, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartForceDownHeight, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeForwardMinComponent, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeForwardMaxComponent, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeUpMinComponent, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeUpMaxComponent, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikePitchMinComponent, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikePitchMaxComponent, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeMinSpeed, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeMaxSpeed, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeMinUpright, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeMaxUpright, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeEjectMaxImpactDepth, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.KnockOffBikeEjectImpactFriction, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClearVehicleTimeMS, PsoDataType.SInt, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultInverseMassScales, PsoDataType.Structure, 80, 0, (MetaName)1267291753),
                     new PsoStructureEntryInfo(MetaName.BicycleInverseMassScales, PsoDataType.Structure, 104, 0, (MetaName)1267291753),
                     new PsoStructureEntryInfo(MetaName.BikeInverseMassScales, PsoDataType.Structure, 128, 0, (MetaName)1267291753),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Structure, 152, 0, MetaName.CNmTuningSet),
                     new PsoStructureEntryInfo(MetaName.Update, PsoDataType.Structure, 192, 0, MetaName.CNmTuningSet)
                    );
                case MetaName.CTaskRageRagdoll__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskRageRagdoll__Tunables, 0, 0, 168,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SpineStrengthTuning, PsoDataType.Structure, 16, 0, (MetaName)2579680791),
                     new PsoStructureEntryInfo(MetaName.NeckStrengthTuning, PsoDataType.Structure, 48, 0, (MetaName)2579680791),
                     new PsoStructureEntryInfo(MetaName.LimbStrengthTuning, PsoDataType.Structure, 80, 0, (MetaName)2579680791),
                     new PsoStructureEntryInfo(MetaName.RageRagdollImpulseTuning, PsoDataType.Structure, 112, 0, (MetaName)2715073781),
                     new PsoStructureEntryInfo(MetaName.fMuscleAngleStrengthRampDownRate, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMuscleSpeedStrengthRampDownRate, PsoDataType.Float, 164, 0, 0)
                    );
                case (MetaName)2579680791:
                    return new PsoStructureInfo((MetaName)2579680791, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.fInitialDelay, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStartStrength, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMidStrength, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fEndStrength, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDurationStage1, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDurationStage2, PsoDataType.Float, 28, 0, 0)
                    );
                case (MetaName)2715073781:
                    return new PsoStructureInfo((MetaName)2715073781, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.fImpulseReductionPerShot, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fImpulseRecoveryPerSecond, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxImpulseModifier, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinImpulseModifier, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCounterImpulseRatio, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTempInitialStiffnessWhenShot, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAnimalMassMult, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAnimalImpulseMultMin, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fAnimalImpulseMultMax, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInitialHitImpulseMult, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.CPlayerInfo__Tunables:
                    return new PsoStructureInfo(MetaName.CPlayerInfo__Tunables, 0, 0, 992,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnemyCharging, PsoDataType.Structure, 16, 0, (MetaName)2968011987),
                     new PsoStructureEntryInfo(MetaName.CombatLoitering, PsoDataType.Structure, 48, 0, (MetaName)2859355160),
                     new PsoStructureEntryInfo(MetaName.MinVehicleCollisionDamageScale, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVehicleCollisionDamageScale, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngleConsidered, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToConsiderVehicleValid, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistToConsiderVehicleValid, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SprintReplenishFinishedPercentage, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3763989739, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SprintReplenishRateMultiplier, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SprintReplenishRateMultiplierBike, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWorldLimitsPlayerX, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWorldLimitsPlayerY, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinWorldLimitsPlayerX, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinWorldLimitsPlayerY, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToTrespassWhileSwimmingBeforeDeath, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovementAwayWeighting, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceWeighting, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingWeighting, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CameraWeighting, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceWeightingNoStick, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingWeightingNoStick, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OnFireWeightingMult, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeMaxRestoreDuration, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeMinRestoreDuration, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleDepletionMinMult, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleDepletionMidMult, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleDepletionMaxMult, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleMinDepletionLimit, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleMidDepletionLimit, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleMaxDepletionLimit, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenSwitchToClearTasks, PsoDataType.UInt, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenShoutTargetPosition, PsoDataType.UInt, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TrespassGuardModelName, PsoDataType.String, 188, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GuardWorldExtents, PsoDataType.Bool, 192, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPlayerInfo__sSprintControlData),
                     new PsoStructureEntryInfo(MetaName.SprintControlData, PsoDataType.Array, 200, 1, (MetaName)262180),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPlayerInfo__sPlayerStatInfo),
                     new PsoStructureEntryInfo(MetaName.PlayerStatInfos, PsoDataType.Array, 336, 1, (MetaName)524326),
                     new PsoStructureEntryInfo((MetaName)4178328957, PsoDataType.Float, 984, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1447073898, PsoDataType.Float, 988, 0, 0)
                    );
                case (MetaName)2968011987:
                    return new PsoStructureInfo((MetaName)2968011987, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.fChargeGoalBehindCoverCentralOffset, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fChargeGoalLateralOffset, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fChargeGoalRearOffset, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fChargeGoalMaxAdjustRadius, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPlayerMoveDistToResetChargeGoals, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)2859355160:
                    return new PsoStructureInfo((MetaName)2859355160, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.fPlayerMoveDistToResetLoiterPosition, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uDistanceCheckPeriodMS, PsoDataType.UInt, 12, 0, 0)
                    );
                case MetaName.CPlayerInfo__sSprintControlData:
                    return new PsoStructureInfo(MetaName.CPlayerInfo__sSprintControlData, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.TapAdd, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HoldSub, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReleaseSub, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Threshhold, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLimit, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ResultMult, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CPlayerInfo__sPlayerStatInfo:
                    return new PsoStructureInfo(MetaName.CPlayerInfo__sPlayerStatInfo, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinStaminaDuration, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxStaminaDuration, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHoldBreathDuration, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHoldBreathDuration, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinWheelieAbility, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWheelieAbility, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPlaneControlAbility, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPlaneControlAbility, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPlaneDamping, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPlaneDamping, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeliDamping, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeliDamping, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFallHeight, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxFallHeight, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDiveHeight, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDiveHeight, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DiveRampPow, PsoDataType.Float, 76, 0, 0)
                    );
                case MetaName.CAnimSpeedUps__Tunables:
                    return new PsoStructureInfo(MetaName.CAnimSpeedUps__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplayerClimbStandRateModifier, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplayerClimbRunningRateModifier, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplayerClimbClamberRateModifier, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplayerEnterExitJackVehicleRateModifier, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplayerLadderRateModifier, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplayerReloadRateModifier, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplayerCoverIntroRateModifier, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplayerIdleTurnRateModifier, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceMPAnimRatesInSP, PsoDataType.Bool, 48, 0, 0)
                    );
                case MetaName.CPlayerPedTargeting__Tunables:
                    return new PsoStructureInfo(MetaName.CPlayerPedTargeting__Tunables, 0, 0, 3776,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fTargetableDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTargetThreatOverride, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArrestHardLockDistance, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UnarmedInCoverTargetingDistance, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1109599945, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MeleeLostLOSBreakTime, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToAllowCachedStickInputForMelee, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoAynchronousProbesWhenFindingFreeAimAssistTarget, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowDriverLockOnToAmbientPeds, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowDriverLockOnToAmbientPedsInSP, PsoDataType.Bool, 46, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisplayAimAssistIntersections, PsoDataType.Bool, 47, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2235947157, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisplayAimAssistCurves, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisplayLockOnDistRanges, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisplayLockOnAngularRanges, PsoDataType.Bool, 51, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisplaySoftLockDebug, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisplayFreeAimTargetDebug, PsoDataType.Bool, 53, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DebugLockOnTargets, PsoDataType.Bool, 54, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseRagdollTargetIfNoAssistTarget, PsoDataType.Bool, 55, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseReticuleSlowDownStrafeClamp, PsoDataType.Bool, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EasyTargettingDifficultyInfo, PsoDataType.Structure, 64, 0, MetaName.CTargettingDifficultyInfo),
                     new PsoStructureEntryInfo(MetaName.NormalTargettingDifficultyInfo, PsoDataType.Structure, 528, 0, MetaName.CTargettingDifficultyInfo),
                     new PsoStructureEntryInfo((MetaName)2843442022, PsoDataType.Structure, 992, 0, MetaName.CTargettingDifficultyInfo),
                     new PsoStructureEntryInfo((MetaName)933644170, PsoDataType.Structure, 1456, 0, MetaName.CTargettingDifficultyInfo),
                     new PsoStructureEntryInfo((MetaName)1051298047, PsoDataType.Structure, 1920, 0, MetaName.CTargettingDifficultyInfo),
                     new PsoStructureEntryInfo((MetaName)1712651720, PsoDataType.Structure, 2384, 0, MetaName.CTargettingDifficultyInfo),
                     new PsoStructureEntryInfo((MetaName)3769742750, PsoDataType.Structure, 2848, 0, MetaName.CTargettingDifficultyInfo),
                     new PsoStructureEntryInfo((MetaName)1604857901, PsoDataType.Structure, 3312, 0, MetaName.CTargettingDifficultyInfo)
                    );
                case MetaName.CTargettingDifficultyInfo:
                    return new PsoStructureInfo(MetaName.CTargettingDifficultyInfo, 0, 0, 464,
                     new PsoStructureEntryInfo(MetaName.LockType, PsoDataType.Enum, 8, 0, MetaName.LockType),
                     new PsoStructureEntryInfo((MetaName)3742855719, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseLockOnTargetSwitching, PsoDataType.Bool, 13, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseReticuleSlowDownForRunAndGun, PsoDataType.Bool, 14, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseReticuleSlowDown, PsoDataType.Bool, 15, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableBulletBending, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowSoftLockFineAim, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseFineAimSpring, PsoDataType.Bool, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseNewSlowDownCode, PsoDataType.Bool, 19, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseCapsuleTests, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseDriveByAssistedAim, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3708089905, PsoDataType.Bool, 22, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnRangeModifier, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReticuleSlowDownRadius, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReticuleSlowDownCapsuleRadius, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReticuleSlowDownCapsuleLength, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAngularLimit, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAngularLimitClose, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAngularLimitCloseDistMin, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAngularLimitCloseDistMax, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WideTargetAngularLimit, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CycleTargetAngularLimit, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CycleTargetAngularLimitMelee, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAimPitchMin, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTargetAimPitchMax, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoReticuleLockOnRangeModifier, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoReticuleMaxLockOnRange, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoReticuleTargetAngularLimit, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoReticuleTargetAngularLimitClose, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoReticuleTargetAngularLimitCloseDistMin, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoReticuleTargetAngularLimitCloseDistMax, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NoReticuleTargetAimPitchLimit, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelocityForDriveByAssistedAim, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnDistanceRejectionModifier, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimVerticalMovement, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimDownwardsVerticalMovement, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimSidewaysScale, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSoftLockBreakTime, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSoftLockBreakTimeCloseRange, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSoftLockBreakAtMaxXStickTime, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockBreakDistanceMin, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockBreakDistanceMax, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFineAimTime, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinFineAimTimeHoldingStick, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinNoReticuleAimTime, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimAssistCapsuleRadius, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2152729841, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimAssistCapsuleMaxLength, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1099784311, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimAssistBlendInTime, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimAssistBlendOutTime, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockFineAimBreakXYValue, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockFineAimBreakZValue, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockFineAimXYAbsoluteValue, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockFineAimXYAbsoluteValueClose, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockBreakValue, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockTime, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SoftLockTimeToAcquireTarget, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2167138628, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimHorSpeedMin, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimHorSpeedMax, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimVerSpeed, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimSpeedMultiplier, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimHorWeightSpeedMultiplier, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimHorSpeedPower, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimSpeedMultiplierClose, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimSpeedMultiplierCloseDistMin, PsoDataType.Float, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FineAimSpeedMultiplierCloseDistMax, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1715236167, PsoDataType.Float, 248, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1204531501, PsoDataType.Float, 252, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2894104098, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1976544803, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo((MetaName)922516025, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3926385305, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo((MetaName)493048149, PsoDataType.Float, 272, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2549641570, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnSwitchTimeExtensionBreakLock, PsoDataType.UInt, 280, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LockOnSwitchTimeExtensionKillTarget, PsoDataType.UInt, 284, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3957255740, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3799655237, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CCurveSet),
                     new PsoStructureEntryInfo(MetaName.CurveSets, PsoDataType.Array, 296, 1, (MetaName)262224),
                     new PsoStructureEntryInfo(MetaName.AimAssistDistanceCurve, PsoDataType.Structure, 432, 0, MetaName.CCurveSet)
                    );
                case MetaName.CCurveSet:
                    return new PsoStructureInfo(MetaName.CCurveSet, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CCurve),
                     new PsoStructureEntryInfo(MetaName.curves, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case MetaName.CCurve:
                    return new PsoStructureInfo(MetaName.CCurve, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.fInputMax, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fResultMax, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPow, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CRandomEventManager__Tunables:
                    return new PsoStructureInfo(MetaName.CRandomEventManager__Tunables, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RenderDebug, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceCrime, PsoDataType.Bool, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EventInterval, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EventInitInterval, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1980648971),
                     new PsoStructureEntryInfo(MetaName.RandomEventType, PsoDataType.Array, 32, 0, (MetaName)6),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2790883352),
                     new PsoStructureEntryInfo(MetaName.RandomEventData, PsoDataType.Array, 48, 0, (MetaName)8),
                     new PsoStructureEntryInfo(MetaName.SpawningChasesEnabled, PsoDataType.Bool, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxNumberCopVehiclesInChase, PsoDataType.SInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ProbSpawnHeli, PsoDataType.SInt, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAmbientVehiclesToSpawnChase, PsoDataType.SInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPlayerMoveDistanceToSpawnChase, PsoDataType.SInt, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeliVehicleModelId, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HeliPedModelId, PsoDataType.String, 88, 7, 0)
                    );
                case (MetaName)1980648971:
                    return new PsoStructureInfo((MetaName)1980648971, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.RandomEventTypeName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RandomEventTimeIntervalMin, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RandomEventTimeIntervalMax, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DeltaScaleWhenPlayerStationary, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)2790883352:
                    return new PsoStructureInfo((MetaName)2790883352, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.RandomEventName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RandomEventType, PsoDataType.Enum, 12, 0, (MetaName)3799765345)
                    );
                case MetaName.CEventExplosionHeard__Tunables:
                    return new PsoStructureInfo(MetaName.CEventExplosionHeard__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCombineDistThresholdSquared, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CRelationshipManager__Tunables:
                    return new PsoStructureInfo(MetaName.CRelationshipManager__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DisplayRemovedGroups, PsoDataType.Bool, 16, 0, 0)
                    );
                case MetaName.CTaskAgitated__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAgitated__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 16, 0, (MetaName)588501678),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenLookAts, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovingAwayVelocityMSThreshold, PsoDataType.Float, 36, 0, 0)
                    );
                case (MetaName)588501678:
                    return new PsoStructureInfo((MetaName)588501678, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Info, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Hashes, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.History, PsoDataType.Bool, 11, 0, 0)
                    );
                case MetaName.CTaskConfront__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskConfront__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceIfUnarmed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistanceIfArmed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToMove, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRadius, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToIntimidateArmedTarget, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToIntimidateUnarmedTarget, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.CTaskSmartFlee__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSmartFlee__Tunables, 0, 0, 136,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2976849785, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2923989237, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo((MetaName)964747565, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2588997912, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3847831913, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)182226933, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3494668944, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)55450154, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1156561437, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2850307001, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4172340609, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)916666022, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1653710895, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitVehicleMaxDistance, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitVehicleRouteMinDistance, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenHandsUpChecks, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenExitVehicleDueToRouteChecks, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToCower, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeForHandsUp, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeForHandsUp, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimeForExitVehicle, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimeForExitVehicle, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChanceToDeleteOnExitVehicle, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistFromPlayerToDeleteOnExitVehicle, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRouteLengthForCower, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistFromTargetWhenCoweringToCheckForExit, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeTargetTooCloseDistance, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)945475130, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)909291779, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRouteSizeForCower, PsoDataType.SInt, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceCower, PsoDataType.Bool, 132, 0, 0)
                    );
                case MetaName.CTaskScenarioFlee__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskScenarioFlee__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fFleeProjectRange, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInitialSearchRadius, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSearchScaler, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSearchRangeMax, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFleeRange, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1817496588, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTargetScenarioRadius, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fProbeLength, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3328108206, PsoDataType.UInt, 48, 0, 0)
                    );
                case MetaName.CTaskExhaustedFlee__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskExhaustedFlee__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StartingEnergy, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnergyLostPerSecond, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OuterDistanceThreshold, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InnerDistanceThreshold, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CTaskWalkAway__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskWalkAway__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SafeDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenRouteAdjustments, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskGrowlAndFlee__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskGrowlAndFlee__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FleeMBR, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskIntimidate__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskIntimidate__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)920671881, PsoDataType.Bool, 16, 0, 0)
                    );
                case MetaName.CTaskReactAndFlee__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskReactAndFlee__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinFleeMoveBlendRatio, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxFleeMoveBlendRatio, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OverrideDirections, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OverrideReactDirection, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OverrideFleeDirection, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxReactionTime, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRate, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRate, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingChangeRate, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToRepeatLastAnimation, PsoDataType.Float, 52, 0, 0)
                    );
                case MetaName.CTaskReactInDirection__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskReactInDirection__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CTaskReactToExplosion__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskReactToExplosion__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxShellShockedDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxFlinchDistance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLookAtDistance, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskReactToImminentExplosion__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskReactToImminentExplosion__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxEscapeDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxFlinchDistance, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskShockingEvent__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingEvent__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinRemainingRotationForScaling, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAngularVelocityScaleFactor, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAngularVelocityScaleFactor, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskShockingEventGoto__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingEventGoto__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DistSquaredThresholdAtCrowdRoundPos, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistSquaredThresholdMovingToCrowdRoundPos, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistVicinityOfCrowd, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraDistForGoto, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistFromOtherPeds, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MoveBlendRatioForFarGoto, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRadiusForCloseNavMeshTask, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraToleranceForStopWatchDistance, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.CTaskShockingEventHurryAway__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingEventHurryAway__Tunables, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LookAheadDistanceForBackAway, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToCallPolice, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToCallPolice, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToCallPolice, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancePlayingInitalTurnAnimSmallReact, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancePlayingCustomBackAwayAnimSmallReact, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancePlayingInitalTurnAnimBigReact, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancePlayingCustomBackAwayAnimBigReact, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldFleeDistance, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldFleeVehicleDistance, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldFleeFilmingDistance, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EvasionThreshold, PsoDataType.UInt, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClosePlayerSpeedupDistanceSquaredThreshold, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClosePlayerSpeedupTimeThreshold, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceFromPlayerToDeleteHurriedPed, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeUntilDeletionWhenHurrying, PsoDataType.Float, 76, 0, 0)
                    );
                case MetaName.CTaskShockingEventWatch__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingEventWatch__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTargetAngularMovementForWatch, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThresholdWatchAfterFace, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThresholdWatchStop, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceBetweenFilmingPeds, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1440458015, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceAwayForFilming, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.CTaskShockingEventReact__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingEventReact__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TurningTolerance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurningRate, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurningEnergyUpperThreshold, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurningEnergyLowerThreshold, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenReactionIdlesMin, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenReactionIdlesMax, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendoutPhase, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CTaskShockingEventBackAway__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingEventBackAway__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeadingAdjustmentRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeadingAlignmentCosThreshold, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeadingAlignmentCosThreshold, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MoveNetworkBlendoutDuration, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBackwardsProjectionRange, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AxesFacingTolerance, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForBackAway, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForBackAway, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWptAngle, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutPhase, PsoDataType.Float, 52, 0, 0)
                    );
                case MetaName.CTaskShockingEventReactToAircraft__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingEventReactToAircraft__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ThresholdWatch, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThresholdRun, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskShockingPoliceInvestigate__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingPoliceInvestigate__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraDistForGoto, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MoveBlendRatioForFarGoto, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistFromPlayerToDeleteOffscreen, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DeleteOffscreenTimeMS_MIN, PsoDataType.UInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DeleteOffscreenTimeMS_MAX, PsoDataType.UInt, 32, 0, 0)
                    );
                case MetaName.CTaskShockingEventStopAndStare__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShockingEventStopAndStare__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BringVehicleToHaltDistance, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskShove__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShove__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 16, 0, (MetaName)838054604),
                     new PsoStructureEntryInfo(MetaName.MaxDistance, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDot, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusForContact, PsoDataType.Float, 40, 0, 0)
                    );
                case (MetaName)838054604:
                    return new PsoStructureInfo((MetaName)838054604, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Contact, PsoDataType.Bool, 9, 0, 0)
                    );
                case MetaName.CTaskShoved__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskShoved__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CScenarioClipHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CScenarioClipHelper__Tunables, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CVehicleScenarioManager__AttractorTuning__Tunables:
                    return new PsoStructureInfo(MetaName.CVehicleScenarioManager__AttractorTuning__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ForwardDirectionThresholdCosSquared, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistToPathDefault, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistToVehicle, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistToVehicle, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NumToUpdatePerFrame, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAfterAttractionMs, PsoDataType.UInt, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAfterChainTestFailedMs, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAfterFailedConditionsMs, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAfterNoBoundsMs, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPassengersForAttraction, PsoDataType.UShort, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPassengersForAttraction, PsoDataType.UShort, 54, 0, 0)
                    );
                case MetaName.CTaskCoupleScenario__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCoupleScenario__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ResumeDistSq, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StopDistSq, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetDistance, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskCowerScenario__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCowerScenario__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EventDecayTimeMS, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReturnToNormalDistanceSq, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BackHeadingInterpRate, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EventlessSwitchStateTimeRequirement, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EventlessSwitchInactivityTimeRequirement, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EventlessSwitchDistanceRequirement, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistFromPlayerToDeleteCoweringForever, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoweringForeverDeleteOffscreenTimeMS_MIN, PsoDataType.UInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoweringForeverDeleteOffscreenTimeMS_MAX, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FlinchDecayTime, PsoDataType.UInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenFlinches, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.CTaskUseScenario__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskUseScenario__Tunables, 0, 0, 152,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AdvanceUseTimeRandomMaxProportion, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BreakAttachmentMoveSpeedThreshold, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BreakAttachmentOrientationThreshold, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitAttachmentMoveSpeedThreshold, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RouteLengthThresholdForFinalApproach, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZThresholdForApproachOffset, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DetachExitDefaultPhaseThreshold, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastExitDefaultPhaseThreshold, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RouteLengthThresholdForApproachOffset, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraFleeDistance, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FindPropInEnvironmentDist, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRateToPlayCowerReaction, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRateToPlayCowerReaction, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDifferenceBetweenCowerReactionRates, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReactAndFleeBlendOutPhase, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RegularExitDefaultPhaseThreshold, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeOfDayRandomnessHours, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLeaveMinBetweenAnybody, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLeaveRandomAmount, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLeaveRandomFraction, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PavementFloodFillSearchRadius, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DelayBetweenPavementFloodFillSearches, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeMBRMin, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeMBRMax, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPathLengthForValidExit, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceNavmeshMayAdjustPath, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenChecksToLeaveCowering, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SkipGotoXYDist, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SkipGotoZDist, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SkipGotoHeadingDeltaDegrees, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinExtraMoney, PsoDataType.SInt, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxExtraMoney, PsoDataType.SInt, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpdatesBeforeShiftingBounds, PsoDataType.SByte, 144, 0, 0)
                    );
                case MetaName.CTaskUseVehicleScenario__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskUseVehicleScenario__Tunables, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BringVehicleToHaltDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdleTimeRandomFactor, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowDownDist, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowDownSpeed, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SwitchToStraightLineDist, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetArriveDist, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlaneTargetArriveDist, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeliTargetArriveDist, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BoatTargetArriveDist, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlaneTargetArriveDistTaxiOnGround, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlaneDrivingSubtaskArrivalDist, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BoatMaxAvoidanceAngle, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSearchDistance, PsoDataType.UShort, 64, 0, 0)
                    );
                case MetaName.CTaskWanderingScenario__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskWanderingScenario__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeWaitingForBlockingArea, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SwitchToNextPointDistWalking, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SwitchToNextPointDistJogging, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreferNearWaterSurfaceArrivalRadius, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenBlockingAreaChecksMS, PsoDataType.UInt, 32, 0, 0)
                    );
                case MetaName.CTaskPoliceOrderResponse__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskPoliceOrderResponse__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToWait, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForVehicleMovingSlowly, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForVehicleMovingQuickly, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBeforeOvertakeToMatchSpeedWhenPulledOver, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBeforeOvertakeToMatchSpeedWhenCruising, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CheatPowerIncreaseForMatchSpeed, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HashOrdering, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)683308669, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)527303977, PsoDataType.Float, 48, 0, 0)
                    );
                case MetaName.CTaskArrestPed__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskArrestPed__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AimDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArrestDistance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArrestInVehicleDistance, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MoveToDistanceInVehicle, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetDistanceFromVehicleEntry, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1425198943, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4218266421, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)305536610, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RegainComet, PsoDataType.UInt, 48, 0, 0)
                    );
                case MetaName.CTaskSwatOrderResponse__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSwatOrderResponse__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HashOrdering, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)683308669, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskHeliOrderResponse__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskHeliOrderResponse__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3905663903, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)143141277, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)490178156, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2892876798, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3618878970, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3555374966, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.CTaskWitness__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskWitness__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeMoveNearCrimeMs, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeMoveToLawMs, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeSearchMs, PsoDataType.UInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeMoveToLawFailedPathfindingMs, PsoDataType.UInt, 28, 0, 0)
                    );
                case MetaName.CEventShocking__Tunables:
                    return new PsoStructureInfo(MetaName.CEventShocking__Tunables, 0, 0, 224,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LifeTime, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VisualReactionRange, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CopInVehicleVisualReactionRange, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AudioReactionRange, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AIOnlyReactionRangeScaleFactor, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DuckAndCoverCanTriggerForPlayerTime, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GotoWatchRange, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StopWatchDistance, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HurryAwayMBRChangeDelay, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HurryAwayMBRChangeRange, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HurryAwayInitialMBR, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HurryAwayMoveBlendRatioWhenFar, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HurryAwayMoveBlendRatioWhenNear, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinWatchTime, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWatchTime, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinWatchTimeHurryAway, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxWatchTimeHurryAway, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChanceOfWatchRatherThanHurryAway, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPhoneFilmTime, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPhoneFilmTime, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPhoneFilmTimeHurryAway, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPhoneFilmTimeHurryAway, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChanceOfFilmingEventOnPhoneIfWatching, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedGenBlockedAreaMinRadius, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WanderInfluenceSphereRadius, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TriggerAmbientReactionChances, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForAmbientReaction, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForAmbientReaction, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmbientEventLifetime, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeForAmbientReaction, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeForAmbientReaction, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedFearImpact, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShockingSpeechChance, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimer, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimer, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DuplicateDistanceCheck, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeForAudioReaction, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToUseGunfireReactAndFleeAnimations, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedGenBlockingAreaLifeTimeMS, PsoDataType.UInt, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DuplicateTimeCheck, PsoDataType.UInt, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShockingSpeechHash, PsoDataType.String, 176, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShockingFilmSpeechHash, PsoDataType.String, 180, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Priority, PsoDataType.SInt, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AmbientEventType, PsoDataType.Enum, 188, 0, MetaName.AmbientEventType),
                     new PsoStructureEntryInfo(MetaName.AddPedGenBlockedArea, PsoDataType.Bool, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CausesVehicleAvoidance, PsoDataType.Bool, 193, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowIgnoreAsLowPriority, PsoDataType.Bool, 194, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DebugDisplayAlwaysUseEventPosition, PsoDataType.Bool, 195, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DebugDisplayListPlayerInfo, PsoDataType.Bool, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HurryAwayWatchFirst, PsoDataType.Bool, 197, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MobileChatScenarioMayAbort, PsoDataType.Bool, 198, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WatchSayFightCheers, PsoDataType.Bool, 199, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WatchSayShocked, PsoDataType.Bool, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleSlowDown, PsoDataType.Bool, 201, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IgnoreIfSensingPedIsOtherEntity, PsoDataType.Bool, 202, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IgnorePavementChecks, PsoDataType.Bool, 203, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AllowScanningEvenIfPreviouslyReacted, PsoDataType.Bool, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReactionMode, PsoDataType.Enum, 208, 0, (MetaName)1809724872),
                     new PsoStructureEntryInfo(MetaName.StopResponseWhenExpired, PsoDataType.Bool, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeIfApproachedByOtherEntity, PsoDataType.Bool, 213, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeIfApproachedBySourceEntity, PsoDataType.Bool, 214, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CanCallPolice, PsoDataType.Bool, 215, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IgnoreFovForHeadIk, PsoDataType.Bool, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReactToOtherEntity, PsoDataType.Bool, 217, 0, 0)
                    );
                case MetaName.CSituationalClipSetStreamer__Tunables:
                    return new PsoStructureInfo(MetaName.CSituationalClipSetStreamer__Tunables, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Avoids, PsoDataType.Structure, 16, 0, (MetaName)2341712397),
                     new PsoStructureEntryInfo(MetaName.FleeReactions, PsoDataType.Structure, 48, 0, MetaName.iVectorDefault)
                    );
                case (MetaName)2341712397:
                    return new PsoStructureInfo((MetaName)2341712397, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.ClipSet, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ClipSetForCasual, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1095110195),
                     new PsoStructureEntryInfo(MetaName.Variations, PsoDataType.Array, 16, 0, (MetaName)2)
                    );
                case MetaName.iVectorDefault:
                    return new PsoStructureInfo(MetaName.iVectorDefault, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.ClipSetForIntro, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3696584734, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4027846555, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ClipSetForRuns, PsoDataType.String, 20, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3482724306, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo((MetaName)551602790, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInCombatToNotStreamIn, PsoDataType.Float, 32, 0, 0)
                    );
                case (MetaName)1095110195:
                    return new PsoStructureInfo((MetaName)1095110195, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ClipSet, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.IsCasual, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Chances, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CEventFootStepHeard__Tunables:
                    return new PsoStructureInfo(MetaName.CEventFootStepHeard__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimer, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTacticalAnalysisNavMeshPoints__Tunables:
                    return new PsoStructureInfo(MetaName.CTacticalAnalysisNavMeshPoints__Tunables, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BufferDistance, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenAttemptsToFindNewPosition, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenLineOfSightTests, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenAttemptsToFindNearby, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSearchRadiusForNavMesh, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusForFindNearby, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceBetweenPositionsWithClearLineOfSight, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceBetweenPositionsWithoutClearLineOfSightInExteriors, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceBetweenPositionsWithoutClearLineOfSightInInteriors, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxXYDistanceForNewPosition, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxNearbyToFindPerFrame, PsoDataType.SInt, 64, 0, 0)
                    );
                case MetaName.CTacticalAnalysisCoverPointSearch__Tunables:
                    return new PsoStructureInfo(MetaName.CTacticalAnalysisCoverPointSearch__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Scoring, PsoDataType.Structure, 16, 0, (MetaName)850484111),
                     new PsoStructureEntryInfo(MetaName.ScoreCalculationsPerFrame, PsoDataType.SInt, 48, 0, 0)
                    );
                case (MetaName)850484111:
                    return new PsoStructureInfo((MetaName)850484111, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Occupied, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Scripted, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PointOnMap, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToBeConsideredOptimal, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToBeConsideredOptimal, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Optimal, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CTacticalAnalysisCoverPoints__Tunables:
                    return new PsoStructureInfo(MetaName.CTacticalAnalysisCoverPoints__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceMovedToStartSearch, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeBetweenSearches, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForSearch, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForSearch, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenLineOfSightTests, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenAttemptsToFindNearby, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenStatusUpdates, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RadiusForFindNearby, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxNearbyToFindPerFrame, PsoDataType.SInt, 48, 0, 0)
                    );
                case MetaName.CTacticalAnalysis__Tunables:
                    return new PsoStructureInfo(MetaName.CTacticalAnalysis__Tunables, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BadRoute, PsoDataType.Structure, 16, 0, (MetaName)1310298123),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 40, 0, (MetaName)1230349399),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedToActivate, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToDeactivate, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeWithNoReferences, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 76, 0, 0)
                    );
                case (MetaName)1310298123:
                    return new PsoStructureInfo((MetaName)1310298123, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ValueForUnableToFind, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForTooCloseToTarget, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForTaint, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DecayRate, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)1230349399:
                    return new PsoStructureInfo((MetaName)1230349399, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverPoints, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NavMeshPoints, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Position, PsoDataType.Bool, 11, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LineOfSightStatus, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArcStatus, PsoDataType.Bool, 13, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Reserved, PsoDataType.Bool, 14, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Nearby, PsoDataType.Bool, 15, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BadRouteValue, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Reservations, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LineOfSightTests, PsoDataType.Bool, 18, 0, 0)
                    );
                case MetaName.CTaskVehicleApproach__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleApproach__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceAroundClosestRoadNode, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskVehicleBlock__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleBlock__Tunables, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToCapSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToStartCappingSpeed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AdditionalSpeedCap, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceFromTargetToForceStraightLineMode, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLookAhead, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToLookAhead, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotTargetMovingTowardsUsToStartBackAndForth, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotTargetMovingTowardsOurSideToStartBackAndForth, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotTargetMovingTowardsUsToContinueBackAndForth, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotTargetMovingTowardsUsToStartBrakeInFront, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotMovingTowardsTargetToStartBrakeInFront, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotTargetMovingTowardsUsToContinueBrakeInFront, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotMovingTowardsTargetToContinueBrakeInFront, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotTargetMovingTowardsUsToStartCruiseInFront, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotMovingAwayFromTargetToStartCruiseInFront, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotTargetMovingTowardsUsToContinueCruiseInFront, PsoDataType.Float, 76, 0, 0)
                    );
                case MetaName.CTaskVehicleBlockCruiseInFront__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleBlockCruiseInFront__Tunables, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Probes, PsoDataType.Structure, 16, 0, (MetaName)3985328803),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 56, 0, (MetaName)999261493),
                     new PsoStructureEntryInfo(MetaName.StraightLineDistance, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLookAhead, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToLookAhead, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotForSlowdown, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForSlowdown, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForSlowdown, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedMultiplierForMinSlowdown, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedMultiplierForMaxSlowdown, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealDistance, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToAdjustSpeed, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToAdjustSpeed, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinCruiseSpeedMultiplier, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCruiseSpeedMultiplier, PsoDataType.Float, 120, 0, 0)
                    );
                case (MetaName)3985328803:
                    return new PsoStructureInfo((MetaName)3985328803, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Collision, PsoDataType.Structure, 8, 0, (MetaName)1883924946)
                    );
                case (MetaName)1883924946:
                    return new PsoStructureInfo((MetaName)1883924946, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)2241187664, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3879165256, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4154217054, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinLength, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxLength, PsoDataType.Float, 24, 0, 0)
                    );
                case (MetaName)999261493:
                    return new PsoStructureInfo((MetaName)999261493, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Probe, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ProbeResults, PsoDataType.Bool, 10, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2932215080, PsoDataType.Bool, 11, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1537468920, PsoDataType.Bool, 12, 0, 0)
                    );
                case MetaName.CTaskVehicleBlockBrakeInFront__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleBlockBrakeInFront__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAheadForGetInPosition, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinOffsetForGetInPosition, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAheadForBrake, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeAheadForBrakeOnWideRoads, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeForBrake, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FutureDistanceForMinSteerAngle, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FutureDistanceForMaxSteerAngle, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedToUseHandBrake, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToClampCruiseSpeed, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToClampCruiseSpeed, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCruiseSpeedWhenClamped, PsoDataType.Float, 56, 0, 0)
                    );
                case MetaName.CTaskVehicleBlockBackAndForth__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleBlockBackAndForth__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ThrottleMultiplier, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskVehicleFlee__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleFlee__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesForSwerve, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForSwerve, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToSwerve, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToSwerve, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesForHesitate, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForHesitate, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToHesitate, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToHesitate, PsoDataType.Float, 44, 0, 0)
                    );
                case MetaName.CTaskVehicleCruiseBoat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleCruiseBoat__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fTimeToPickNewPoint, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDistToPickNewPoint, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDistSearch, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskVehicleDeadDriver__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleDeadDriver__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SwerveTime, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteerAngleControl, PsoDataType.Enum, 20, 0, (MetaName)888350649),
                     new PsoStructureEntryInfo(MetaName.MinSteerAngle, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSteerAngle, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThrottleControl, PsoDataType.Enum, 32, 0, (MetaName)2055586340),
                     new PsoStructureEntryInfo(MetaName.MinThrottle, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxThrottle, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BrakeControl, PsoDataType.Enum, 44, 0, (MetaName)3374036042),
                     new PsoStructureEntryInfo(MetaName.MinBrake, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBrake, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HandBrakeControl, PsoDataType.Enum, 56, 0, (MetaName)4001255574)
                    );
                case MetaName.CTaskVehicleFleeBoat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleFleeBoat__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FleeDistance, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskVehicleCrash__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleCrash__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForWreck, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskVehicleGoToPointWithAvoidanceAutomobile__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleGoToPointWithAvoidanceAutomobile__Tunables, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TailgateDistanceMax, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TailgateIdealDistanceMin, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TailgateIdealDistanceMax, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TailgateSpeedMultiplierMin, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TailgateSpeedMultiplierMax, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TailgateVelocityMin, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChanceOfPedSeeingCarFromBehind, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForAvoid, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForAvoid, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForAvoid, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForAvoid, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceForAvoidDirected, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForAvoidDirected, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceForAvoidDirected, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForAvoidDirected, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAbsDotForAvoidDirected, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForBrace, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForDive, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToConsiderDangerousDriving, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MultiplierForDangerousDriving, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToSideOnPavement, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToSideOnPavement, PsoDataType.Float, 100, 0, 0)
                    );
                case MetaName.CTaskVehicleGoToBoat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleGoToBoat__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SlowdownDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RouteArrivalDistance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RouteLookAheadDistance, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskVehicleGoToHelicopter__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleGoToHelicopter__Tunables, 0, 0, 152,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.slowDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxCruiseSpeed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxPitchRoll, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxThrottle, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leanKp, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leanKi, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.leanKd, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.yawKp, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.yawKi, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.yawKd, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.throttleKp, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.throttleKi, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.throttleKd, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.whiskerForwardTestDistance, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.whiskerForwardSpeedScale, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.whiskerLateralTestDistance, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.whiskerVerticalTestDistance, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.whiskerTestAngle, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.avoidHeadingChangeSpeed, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.avoidHeadingJump, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.avoidPitchChangeSpeed, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.avoidPitchJump, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.avoidLockDuration, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2436197334, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1438517312, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)444775360, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3623227512, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3009403254, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.numHeightmapFutureSamples, PsoDataType.SInt, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.futureHeightmapSampleTime, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceXYToUseHeightMapAvoidance, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimesliceMinDistToTarget, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimesliceTimeAfterAvoidanceMs, PsoDataType.UInt, 144, 0, 0)
                    );
                case MetaName.CTaskVehicleGoToPlane__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleGoToPlane__Tunables, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.numFutureSamples, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.futureSampleTime, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxDesiredAngleYawDegrees, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxDesiredAnglePitchDegrees, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxDesiredAngleRollDegrees, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.angleToTargetDegreesToNotUseMinRadius, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.minMinDistanceForRollComputation, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxMinDistanceForRollComputation, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxYaw, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxPitch, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxRoll, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxThrottle, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.yawKp, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.yawKi, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.yawKd, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pitchKp, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pitchKi, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pitchKd, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rollKp, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rollKi, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rollKd, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.throttleKp, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.throttleKi, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.throttleKd, PsoDataType.Float, 108, 0, 0)
                    );
                case MetaName.CTaskVehicleLandPlane__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleLandPlane__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SlowDownDistance, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeOnGroundToDrive, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeightToStartLanding, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LandSlopeNoseUpMin, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LandSlopeNoseUpMax, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.CTaskVehicleMissionBase__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleMissionBase__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToKeepEngineAndLightsOnWhileParked, PsoDataType.UInt, 16, 0, 0)
                    );
                case MetaName.CTaskVehicleParkNew__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleParkNew__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ParkingSpaceBlockedWaitTimePerAttempt, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ParkingSpaceBlockedMaxAttempts, PsoDataType.UByte, 20, 0, 0)
                    );
                case MetaName.CTaskVehiclePlaneChase__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehiclePlaneChase__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeed, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskVehiclePullAlongside__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehiclePullAlongside__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLookAhead, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToLookAhead, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OverlapSpeedMultiplier, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedDifference, PsoDataType.Float, 28, 0, 0)
                    );
                case MetaName.CTaskVehiclePursue__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehiclePursue__Tunables, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DriftX, PsoDataType.Structure, 16, 0, (MetaName)42267357),
                     new PsoStructureEntryInfo(MetaName.DriftY, PsoDataType.Structure, 40, 0, (MetaName)42267357),
                     new PsoStructureEntryInfo(MetaName.TimeToLookBehind, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToLookBehind, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpeedDifferenceForMinDistanceToStartMatchingSpeed, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpeedDifferenceForMaxDistanceToStartMatchingSpeed, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToStartMatchingSpeed, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToStartMatchingSpeed, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedMultiplierForBackOff, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DotToClampSpeedToMinimum, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DotToClampSpeedToMaximum, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpeedForMinimumDot, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenLineOfSightChecks, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceForStraightLineModeAlways, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceForStraightLineModeIfLos, PsoDataType.Float, 112, 0, 0)
                    );
                case (MetaName)42267357:
                    return new PsoStructureInfo((MetaName)42267357, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.MinValueForCorrection, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxValueForCorrection, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRate, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskVehicleRam__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleRam__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BackOffTimer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBackOffDistance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxBackOffDistance, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedMultiplierForMinBackOffDistance, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseSpeedMultiplierForMaxBackOffDistance, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.CTaskVehicleShotTire__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleShotTire__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeInSwerve, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedInSwerve, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToApplyTorque, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDotToApplyTorque, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TorqueMultiplier, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.CTaskVehicleSpinOut__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleSpinOut__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToLookAhead, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToLookAhead, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumperOverlapForMaxSpeed, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumperOverlapForMinSpeed, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CatchUpSpeed, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BumperOverlapToBeInPosition, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSidePaddingForTurn, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TurnTime, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InvMassScale, PsoDataType.Float, 48, 0, 0)
                    );
                case MetaName.CPrioritizedClipSetRequestManager__Tunables:
                    return new PsoStructureInfo(MetaName.CPrioritizedClipSetRequestManager__Tunables, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RenderDebugDraw, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vScroll, PsoDataType.Float2, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fIndent, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxNumRequestsPerContext, PsoDataType.SInt, 32, 0, 0)
                    );
                case MetaName.CVehicleClipRequestHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CVehicleClipRequestHelper__Tunables, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToScanForNearbyVehicle, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToScanForNearbyVehicle, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistUpdateFrequency, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistUpdateFrequency, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistPercentageToScaleScanArc, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistScanArc, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistScanArc, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableVehicleDependencies, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableStreamedVehicleAnimRequestHelper, PsoDataType.Bool, 45, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableStreamedEntryAnims, PsoDataType.Bool, 46, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableStreamedInVehicleAnims, PsoDataType.Bool, 47, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableStreamedEntryVariationAnims, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StreamConnectedSeatAnims, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StreamInVehicleAndEntryAnimsTogether, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StreamEntryAndInVehicleAnimsTogether, PsoDataType.Bool, 51, 0, 0)
                    );
                case MetaName.CTaskInVehicleBasic__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskInVehicleBasic__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.fSecondsInAirBeforePassengerComment, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.CTaskPlayerDrive__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskPlayerDrive__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StealthNoisePeriodMS, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthSpeedThresholdLow, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthSpeedThresholdHigh, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StealthVehicleTypeFactorBicycles, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPlayerJumpOutSpeedBike, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPlayerJumpOutSpeedCar, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenAddingDangerousVehicleEvents, PsoDataType.Float, 40, 0, 0)
                    );
                case MetaName.CTaskCarReactToVehicleCollision__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCarReactToVehicleCollision__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SlowDown, PsoDataType.Structure, 16, 0, (MetaName)3933935552),
                     new PsoStructureEntryInfo(MetaName.MaxDamageToIgnore, PsoDataType.Float, 56, 0, 0)
                    );
                case (MetaName)3933935552:
                    return new PsoStructureInfo((MetaName)3933935552, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.MinTimeToReact, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToReact, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxCruiseSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToHonk, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToHonkHeldDown, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ChancesToFlipOff, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTime, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTime, PsoDataType.Float, 36, 0, 0)
                    );
                case MetaName.CTaskEnterVehicle__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskEnterVehicle__Tunables, 0, 0, 408,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.UseCombatEntryForAiJack, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableJackRateOverride, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableDoorHandleArmIk, PsoDataType.Bool, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableBikeHandleArmIk, PsoDataType.Bool, 19, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableSeatBoneArmIk, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableTagSyncIntoAlign, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableMoverFixups, PsoDataType.Bool, 22, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableBikePickPullUpOffsetScale, PsoDataType.Bool, 23, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableNewBikeEntry, PsoDataType.Bool, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForcedDoorHandleArmIk, PsoDataType.Bool, 25, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IgnoreRotationBlend, PsoDataType.Bool, 26, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EnableBikePickUpAlign, PsoDataType.Bool, 27, 0, 0),
                     new PsoStructureEntryInfo((MetaName)192421777, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)509809270, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikePickUpAlignBlendDuration, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GetInRate, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinMagForBikeToBeOnSide, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistanceToEvaluateDoors, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NetworkBlendDuration, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NetworkBlendDurationOpenDoorCombat, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoorRatioToConsiderDoorOpenSteps, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoorRatioToConsiderDoorOpen, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoorRatioToConsiderDoorOpenCombat, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DoorRatioToConsiderDoorClosed, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DistToEntryToAllowForcedActionMode, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VaultDepth, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VaultHorizClearance, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VaultVertClearance, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftPickUpTargetLerpPhaseStart, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftPickUpTargetLerpPhaseEnd, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftPullUpTargetLerpPhaseStart, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftPullUpTargetLerpPhaseEnd, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightPickUpTargetLerpPhaseStart, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightPickUpTargetLerpPhaseEnd, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightPullUpTargetLerpPhaseStart, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightPullUpTargetLerpPhaseEnd, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftPickUpTargetLerpPhaseStartBicycle, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftPickUpTargetLerpPhaseEndBicycle, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftPullUpTargetLerpPhaseStartBicycle, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftPullUpTargetLerpPhaseEndBicycle, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightPickUpTargetLerpPhaseStartBicycle, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightPickUpTargetLerpPhaseEndBicycle, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightPullUpTargetLerpPhaseStartBicycle, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RightPullUpTargetLerpPhaseEndBicycle, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3408785399, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToAbortOpenDoor, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToAbortOpenDoorCombat, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToAbortOpenDoorPlayer, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToRagdollOpenDoor, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToRagdollOpenDoorCombat, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToRagdollOpenDoorPlayer, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackRate, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeEnterForce, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleEnterForce, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastEnterExitRate, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRearDoorOpenRatio, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxOpenRatioForOpenDoorInitialOutside, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxOpenRatioForOpenDoorOutside, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxOscillationDisplacementOutside, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxOpenRatioForOpenDoorInitialInside, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxOpenRatioForOpenDoorInside, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxOscillationDisplacementInside, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeEnterLeanAngleOvershootAmt, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeEnterLeanAngleOvershootRate, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToCheckEntryCollisionWhenIgnoring, PsoDataType.Float, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CombatEntryBlendDuration, PsoDataType.Float, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToReactToJackForGoToDoor, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeStreamClipSetInBeforeWarpSP, PsoDataType.Float, 248, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeStreamClipSetInBeforeWarpMP, PsoDataType.Float, 252, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeStreamClipSetInBeforeSkippingCloseDoor, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeStreamShuffleClipSetInBeforeWarp, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ClimbAlignTolerance, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenDoorChecks, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseSlowInOut, PsoDataType.Bool, 272, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OpenDoorBlendDurationFromNormalAlign, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OpenDoorBlendDurationFromOnVehicleAlign, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OpenDoorToJackBlendDuration, PsoDataType.Float, 284, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroupMemberWaitMinTime, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroupMemberSlowDownDistance, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroupMemberWalkCloseDistance, PsoDataType.Float, 296, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroupMemberWaitDistance, PsoDataType.Float, 300, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SecondsBeforeWarpToLeader, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo((MetaName)119275101, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2217815195, PsoDataType.UInt, 312, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DurationHeldDownEnterButtonToJackFriendly, PsoDataType.UInt, 316, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackAlivePedFromOutsideClipId, PsoDataType.String, 320, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3431653792, PsoDataType.String, 324, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackDeadPedFromOutsideClipId, PsoDataType.String, 328, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1983954745, PsoDataType.String, 332, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackAlivePedFromWaterClipId, PsoDataType.String, 336, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackDeadPedFromWaterClipId, PsoDataType.String, 340, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackPedFromOnVehicleClipId, PsoDataType.String, 344, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackDeadPedFromOnVehicleClipId, PsoDataType.String, 348, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackPedOnVehicleIntoWaterClipId, PsoDataType.String, 352, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJackDeadPedOnVehicleIntoWaterClipId, PsoDataType.String, 356, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultClimbUpClipId, PsoDataType.String, 360, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultClimbUpNoDoorClipId, PsoDataType.String, 364, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1857752806, PsoDataType.String, 368, 7, 0),
                     new PsoStructureEntryInfo((MetaName)536807372, PsoDataType.String, 372, 7, 0),
                     new PsoStructureEntryInfo((MetaName)298521545, PsoDataType.String, 376, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3045129247, PsoDataType.String, 380, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2509236171, PsoDataType.String, 384, 7, 0),
                     new PsoStructureEntryInfo((MetaName)939217889, PsoDataType.String, 388, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3905015036, PsoDataType.String, 392, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3309036906, PsoDataType.String, 396, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3854657159, PsoDataType.Bool, 400, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1405627810, PsoDataType.Bool, 401, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2067860370, PsoDataType.Float, 404, 0, 0)
                    );
                case MetaName.CTaskEnterVehicleAlign__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskEnterVehicleAlign__Tunables, 0, 0, 136,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.UseAttachDuringAlign, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RenderDebugToTTY, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ApplyRotationScaling, PsoDataType.Bool, 18, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ApplyTranslationScaling, PsoDataType.Bool, 19, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableRotationOvershootCheck, PsoDataType.Bool, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableTranslationOvershootCheck, PsoDataType.Bool, 21, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReverseLeftFootAlignAnims, PsoDataType.Bool, 22, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceStandEnterOnly, PsoDataType.Bool, 23, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TranslationChangeRate, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RotationChangeRate, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultAlignRate, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastAlignRate, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CombatAlignRate, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ActionCombatAlignRate, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StandAlignMaxDist, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AlignSuccessMaxDist, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultAlignStartFixupPhase, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultAlignEndFixupPhase, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRadiusForOrientatedAlignWalk, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRadiusForOrientatedAlignRun, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRotationalSpeedScale, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRotationalSpeedScale, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRotationalSpeed, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTranslationalScale, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTranslationalScale, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTranslationalStandSpeed, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTranslationalMoveSpeed, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeadingReachedTolerance, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StdVehicleMinPhaseToStartRotFixup, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeVehicleMinPhaseToStartRotFixup, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VaultExtraZGroundTest, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSqdDistToSetPos, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistAwayFromEntryPointToConsiderFinished, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPedFwdToEntryDotToClampInitialOrientation, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistToAlwaysClampInitialOrientation, PsoDataType.Float, 128, 0, 0)
                    );
                case MetaName.CTaskOpenVehicleDoorFromOutside__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskOpenVehicleDoorFromOutside__Tunables, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnableOpenDoorHandIk, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultOpenDoorStartPhase, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultOpenDoorEndPhase, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultOpenDoorStartIkPhase, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultOpenDoorEndIkPhase, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendWeightToUseHighClipEvents, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultOpenDoorRate, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHandleHeightDiffVan, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHandleHeightDiffVan, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHandleHeightDiff, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultOpenDoorClipId, PsoDataType.String, 56, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HighOpenDoorClipId, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CombatOpenDoorClipId, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultTryLockedDoorClipId, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultForcedEntryClipId, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo((MetaName)26377711, PsoDataType.String, 76, 7, 0)
                    );
                case MetaName.CTaskCloseVehicleDoorFromInside__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskCloseVehicleDoorFromInside__Tunables, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnableCloseDoorHandIk, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultCloseDoorStartPhase, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultCloseDoorEndPhase, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultCloseDoorStartIkPhase, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultCloseDoorEndIkPhase, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinBlendWeightToUseFarClipEvents, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CloseDoorForceMultiplier, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VehicleSpeedToAbortCloseDoor, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedTestXOffset, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedTestYOffset, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedTestZStartOffset, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedTestZOffset, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedTestRadius, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinOpenDoorRatioToUseArmIk, PsoDataType.Float, 68, 0, 0)
                    );
                case MetaName.CTaskEnterVehicleSeat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskEnterVehicleSeat__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelocityToRagdollPed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVelocityToEnterBike, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultGetInClipId, PsoDataType.String, 24, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GetOnQuickClipId, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GetInFromWaterClipId, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GetInStandOnClipId, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GetInCombatClipId, PsoDataType.String, 40, 7, 0)
                    );
                case MetaName.CTaskExitVehicle__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskExitVehicle__Tunables, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TimeSinceLastSpottedToLeaveEngineOn, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BeJackedBlendInDuration, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitVehicleBlendInDuration, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThroughWindScreenBlendInDuration, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitVehicleBlendOutDuration, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitVehicleUnderWaterBlendOutDuration, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitVehicleAttempToFireBlendOutDuration, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeExitVehicleBlendOutDuration, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeaderExitVehicleDistance, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitProbeDistance, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExitDistance, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RearExitSideOffset, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelocityToRagdollPed, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToReserveComponentBeforeWarp, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraOffsetForGroundCheck, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.JumpOutofSubNeutralBuoyancyTime, PsoDataType.UInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultClimbDownClipId, PsoDataType.String, 80, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultClimbDownNoDoorClipId, PsoDataType.String, 84, 7, 0)
                    );
                case MetaName.CTaskExitVehicleSeat__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskExitVehicleSeat__Tunables, 0, 0, 232,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3909498748, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AdditionalWindscreenRagdollForceFwd, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AdditionalWindscreenRagdollForceUp, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SkyDiveProbeDistance, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InAirProbeDistance, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ArrestProbeDistance, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InWaterExitDepth, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InWaterExitProbeLength, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeVelocityToUseAnimatedJumpOff, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleVelocityToUseAnimatedJumpOff, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultGetOutBlendDuration, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultGetOutNoWindBlendDuration, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeForArrestBreakout, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThroughWindscreenDamagePlayer, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ThroughWindscreenDamageAi, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultCrashExitOnSideClipId, PsoDataType.String, 76, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBeJackedAlivePedFromOutsideClipId, PsoDataType.String, 80, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2529144140, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBeJackedDeadPedFromOutsideClipId, PsoDataType.String, 88, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3459414413, PsoDataType.String, 92, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBeJackedAlivePedFromWaterClipId, PsoDataType.String, 96, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBeJackedDeadPedFromWaterClipId, PsoDataType.String, 100, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBeJackedAlivePedOnVehicleClipId, PsoDataType.String, 104, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBeJackedDeadPedOnVehicleClipId, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBeJackedAlivePedOnVehicleIntoWaterClipId, PsoDataType.String, 112, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultBeJackedDeadPedOnVehicleIntoWaterClipId, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultFleeExitClipId, PsoDataType.String, 120, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultGetOutClipId, PsoDataType.String, 124, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultGetOutToWaterClipId, PsoDataType.String, 128, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultGetOutOnToVehicleClipId, PsoDataType.String, 132, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultGetOutNoWingId, PsoDataType.String, 136, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultJumpOutClipId, PsoDataType.String, 140, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DeadFallOutClipId, PsoDataType.String, 144, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)609992288),
                     new PsoStructureEntryInfo(MetaName.ExitToAimClipSets, PsoDataType.Array, 152, 0, (MetaName)34),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3736014569),
                     new PsoStructureEntryInfo(MetaName.ExitToAimVehicleInfos, PsoDataType.Array, 168, 0, (MetaName)36),
                     new PsoStructureEntryInfo(MetaName.BikeExitForce, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RagdollIntoWaterVelocity, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroundFixupHeight, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroundFixupHeightLarge, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroundFixupHeightLargeOffset, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroundFixupHeightBoatInWaterInitial, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroundFixupHeightBoatInWater, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExtraWaterZGroundFixup, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeExitExtraRotationSpeed, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FleeExitExtraTranslationSpeed, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1345368814, PsoDataType.String, 224, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1495472466, PsoDataType.String, 228, 7, 0)
                    );
                case (MetaName)609992288:
                    return new PsoStructureInfo((MetaName)609992288, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Clips, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case (MetaName)3736014569:
                    return new PsoStructureInfo((MetaName)3736014569, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1118518303),
                     new PsoStructureEntryInfo(MetaName.Seats, PsoDataType.Array, 16, 0, (MetaName)1)
                    );
                case (MetaName)1118518303:
                    return new PsoStructureInfo((MetaName)1118518303, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ExitToAimClipsName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OneHandedClipSetName, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TwoHandedClipSetName, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SeatPosition, PsoDataType.Enum, 20, 0, (MetaName)894416699)
                    );
                case MetaName.CTaskReactToBeingAskedToLeaveVehicle__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskReactToBeingAskedToLeaveVehicle__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToWatchVehicle, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToWatchVehicle, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskRideTrain__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskRideTrain__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayForGetOff, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayForGetOff, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxWaitSeconds, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskTrainBase__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskTrainBase__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TargetRadius, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CompletionRadius, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowDownDistance, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskTryToGrabVehicleDoor__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskTryToGrabVehicleDoor__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinGrabTime, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxGrabTime, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHandToHandleDistance, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CTaskVehicleFSM__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskVehicleFSM__Tunables, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AllowEntryToMPWarpInSeats, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForceStreamingFailure, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PushAngleDotTolerance, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TowardsDoorPushAngleDotTolerance, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DeadZoneAnyInputDirection, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisallowGroundProbeVelocity, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPedSpeedToActivateRagdoll, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPhysSpeedToActivateRagdoll, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHoverHeightDistToWarpIntoHeli, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToConsiderPedGoingToDoorPriority, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeToConsiderPedGoingToDoorPriority, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistToConsiderPedGoingToDoorPriority, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1835160785, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1964009554, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2552082535, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3034610967, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeToConsiderEnterInputValid, PsoDataType.UInt, 76, 0, 0)
                    );
                case MetaName.CTaskMotionInVehicle__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionInVehicle__Tunables, 0, 0, 208,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DisableCloseDoor, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3306188437, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedForVehicleToBeConsideredStillSqr, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityDeltaThrownOut, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityDeltaThrownOutPlayerSP, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.VelocityDeltaThrownOutPlayerMP, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRateForInVehicleAnims, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRateForInVehicleAnims, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HeavyBrakeYAcceleration, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinRatioForClosingDoor, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InAirZAccelTrigger, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InAirProbeDistance, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InAirProbeForwardOffset, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPitchDefault, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitchDefault, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPitchInAir, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitchInAir, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultPitchSmoothingRate, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikePitchSmoothingRate, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikePitchSmoothingPassengerRate, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WheelieAccelerateControlThreshold, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WheelieMaxSpeedThreshold, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WheelieUpDownControlThreshold, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WheelieDesiredLeanAngleTol, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StillAccTol, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StillPitchAngleTol, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AccelerationSmoothing, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AccelerationSmoothingBike, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AccelerationScaleBike, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInCurrentStateForStill, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AccelerationToStartLeaning, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZAccelerationToStartLeaning, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAccelerationForLean, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxXYAccelerationForLeanBike, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZAccelerationForLeanBike, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StillDelayTime, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShuntAccelerateMag, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShuntAccelerateMagBike, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInShuntStateBeforeRestart, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAbsThrottleForCloseDoor, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVehSpeedToConsiderClosingDoor, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDoorSpeedToConsiderClosingDoor, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVehVelocityToGoThroughWindscreen, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVehVelocityToGoThroughWindscreenMP, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxZComponentForCollisionNormal, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeStreamInVehicleClipSetBeforeStartingEngine, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo((MetaName)599372255, PsoDataType.UInt, 200, 0, 0)
                    );
                case MetaName.CTaskMotionInAutomobile__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionInAutomobile__Tunables, 0, 0, 536,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TestLowLodIdle, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3645563068, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2286839301, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInHornState, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxVelocityForSitIdles, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSteeringAngleForSitIdles, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinCentredSteeringAngleTimeForSitIdles, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanSidewaysAngleSmoothingRateMin, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanSidewaysAngleSmoothingRateMax, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanSidewaysAngleSmoothingAcc, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanSidewaysAngleMinAccAngle, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanSidewaysAngleMaxAccAngle, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftRightStickInputSmoothingRate, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeftRightStickInputMin, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanForwardsAngleSmoothingRate, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpDownStickInputSmoothingRate, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpDownStickInputMin, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZAccForLowImpact, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZAccForMedImpact, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ZAccForHighImpact, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseLegIkOnBikes, PsoDataType.Bool, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LargeVerticalAccelerationDelta, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NumFramesToPersistLargeVerticalAcceleration, PsoDataType.SInt, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowLodIdleClipSetId, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SeatDisplacementSmoothingRateDriver, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SeatDisplacementSmoothingRatePassenger, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3038905403, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartEngineForce, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardsPitchSlope, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardsPitchSlope, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardsPitchSlopeBalance, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardsPitchSlopeBalance, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeInWheelieToEnforceMinPitch, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardsPitchWheelieBalance, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardsPitchWheelieBalance, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardsPitchWheelieBegin, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowFastSpeedThreshold, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardsPitchSlowSpeed, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardsPitchSlowSpeed, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardsPitchFastSpeed, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardsPitchFastSpeed, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowApproachRate, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FastApproachRate, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WheelieApproachRate, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NewLeanSteerApproachRate, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenCloseDoorAttempts, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShuntDamageMultiplierAI, PsoDataType.Float, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShuntDamageMultiplierPlayer, PsoDataType.Float, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDamageTakenToApplyDamageAI, PsoDataType.Float, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDamageTakenToApplyDamagePlayer, PsoDataType.Float, 212, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInTaskToCheckForDamage, PsoDataType.Float, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDamageToCheckForRandomDeath, PsoDataType.Float, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDamageToCheckForRandomDeath, PsoDataType.Float, 224, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinHeavyCrashDeathChance, PsoDataType.Float, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxHeavyCrashDeathChance, PsoDataType.Float, 232, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteeringDeadZoneCentreTimeMS, PsoDataType.UInt, 236, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteeringDeadZoneTimeMS, PsoDataType.UInt, 240, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteeringDeadZone, PsoDataType.Float, 244, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteeringChangeToStartProcessMoveSignals, PsoDataType.Float, 248, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SteeringChangeToStopProcessMoveSignals, PsoDataType.Float, 252, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SeatBlendLinSpeed, PsoDataType.Float, 256, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SeatBlendAngSpeed, PsoDataType.Float, 260, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HoldLegOutVelocity, PsoDataType.Float, 264, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelStillStart, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinVelStillStop, PsoDataType.Float, 272, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ForcedLegUpVelocity, PsoDataType.Float, 276, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BurnOutBlendInTol, PsoDataType.Float, 280, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BurnOutBlendInSpeed, PsoDataType.Float, 284, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BurnOutBlendOutSpeed, PsoDataType.Float, 288, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeInAirDriveToStandUpTimeMin, PsoDataType.Float, 292, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BikeInAirDriveToStandUpTimeMax, PsoDataType.Float, 296, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSpeedToBlendInDriveFastFacial, PsoDataType.Float, 300, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDisplacementScale, PsoDataType.Float, 304, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisplacementScaleApproachRateIn, PsoDataType.Float, 308, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisplacementScaleApproachRateOut, PsoDataType.Float, 312, 0, 0),
                     new PsoStructureEntryInfo((MetaName)640640576, PsoDataType.Float, 316, 0, 0),
                     new PsoStructureEntryInfo((MetaName)445721091, PsoDataType.Float, 320, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2001571571, PsoDataType.Float, 324, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1295618163, PsoDataType.Float, 328, 0, 0),
                     new PsoStructureEntryInfo((MetaName)622968896, PsoDataType.Float, 332, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2029080904, PsoDataType.Float, 336, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2895820001, PsoDataType.Float, 340, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3604323696, PsoDataType.Float, 344, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2175428814, PsoDataType.Float, 348, 0, 0),
                     new PsoStructureEntryInfo((MetaName)346108670, PsoDataType.Float, 352, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3123171906, PsoDataType.Float, 356, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3193879405, PsoDataType.Float, 360, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1771118662, PsoDataType.Float, 364, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1104842114, PsoDataType.Float, 368, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2239621028, PsoDataType.UInt, 372, 0, 0),
                     new PsoStructureEntryInfo((MetaName)7898446, PsoDataType.UInt, 376, 0, 0),
                     new PsoStructureEntryInfo((MetaName)56684479, PsoDataType.UInt, 380, 0, 0),
                     new PsoStructureEntryInfo((MetaName)345487015, PsoDataType.UInt, 384, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartEngineClipId, PsoDataType.String, 388, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2195918560, PsoDataType.String, 392, 7, 0),
                     new PsoStructureEntryInfo(MetaName.HotwireClipId, PsoDataType.String, 396, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1348667163, PsoDataType.String, 400, 7, 0),
                     new PsoStructureEntryInfo(MetaName.PutOnHelmetClipId, PsoDataType.String, 404, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1430534928, PsoDataType.String, 408, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1851449258, PsoDataType.String, 412, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2091354753, PsoDataType.String, 416, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1023392893, PsoDataType.String, 420, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2273702662, PsoDataType.String, 424, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ChangeStationClipId, PsoDataType.String, 428, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StillToSitClipId, PsoDataType.String, 432, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SitToStillClipId, PsoDataType.String, 436, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BurnOutClipId, PsoDataType.String, 440, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1920935636, PsoDataType.String, 444, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BikeHornClipId, PsoDataType.String, 448, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4017603423, PsoDataType.String, 452, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1797630730, PsoDataType.Float, 456, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2658529211, PsoDataType.Float, 460, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3474960988, PsoDataType.Float, 464, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3061913457, PsoDataType.Float, 468, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2713284751, PsoDataType.Float, 472, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2044390676, PsoDataType.Float, 476, 0, 0),
                     new PsoStructureEntryInfo((MetaName)137806659, PsoDataType.Float, 480, 0, 0),
                     new PsoStructureEntryInfo((MetaName)402428981, PsoDataType.Float, 484, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1039484179, PsoDataType.Float, 488, 0, 0),
                     new PsoStructureEntryInfo((MetaName)342194320, PsoDataType.Float, 492, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2083391331, PsoDataType.Float, 496, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2714021858, PsoDataType.Float, 500, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2266547610, PsoDataType.Float, 504, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1896779219, PsoDataType.Float, 508, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1233910025, PsoDataType.Float, 512, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1743777174, PsoDataType.Float, 516, 0, 0),
                     new PsoStructureEntryInfo((MetaName)868848233, PsoDataType.Float, 520, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1182962200, PsoDataType.Float, 524, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1798674235, PsoDataType.Float, 528, 0, 0),
                     new PsoStructureEntryInfo((MetaName)647907606, PsoDataType.Float, 532, 0, 0)
                    );
                case MetaName.CTaskMotionOnBicycle__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionOnBicycle__Tunables, 0, 0, 224,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LeanAngleSmoothingRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StillToSitPedalGearApproachRate, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PedalGearApproachRate, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinXYVelForWantsToMove, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForStill, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedForStillReverse, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxThrottleForStill, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultPedalToFreewheelBlendDuration, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SlowPedalToFreewheelBlendDuration, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxRateForSlowBlendDuration, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StillToSitLeanRate, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StillToSitApproachRate, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UpHillMinPitchToStandUp, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DownHillMinPitchToStandUp, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInStandState, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1621924028, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinSprintResultToStand, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeSinceShiftedWeightForwardToAllowWheelie, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WheelieShiftThreshold, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinPitchDefault, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitchDefault, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForwardsPitchSlope, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxForwardsPitchSlope, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OnSlopeThreshold, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxJumpHeightForSmallImpact, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LongitudinalBodyLeanApproachRateSlope, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LongitudinalBodyLeanApproachRate, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LongitudinalBodyLeanApproachRateSlow, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SideZoneThreshold, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReturnZoneThreshold, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxYIntentionToUseSlowApproach, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeToStayUprightAfterImpact, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultSitToStandBlendDuration, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WheelieSitToStandBlendDuration, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WheelieStickPullBackMinIntention, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeSinceNotWantingToTrackStandToAllowStillTransition, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInSitToStillStateToReverse, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PreventDirectTransitionToReverseFromSit, PsoDataType.Bool, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultSmallImpactCharClipId, PsoDataType.String, 168, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultImpactCharClipId, PsoDataType.String, 172, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultSmallImpactBikeClipId, PsoDataType.String, 176, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultImpactBikeClipId, PsoDataType.String, 180, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DownHillSmallImpactCharClipId, PsoDataType.String, 184, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DownHillImpactCharClipId, PsoDataType.String, 188, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DownHillSmallImpactBikeClipId, PsoDataType.String, 192, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DownHillImpactBikeClipId, PsoDataType.String, 196, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SitToStillCharClipId, PsoDataType.String, 200, 7, 0),
                     new PsoStructureEntryInfo(MetaName.SitToStillBikeClipId, PsoDataType.String, 204, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TrackStandToStillLeftCharClipId, PsoDataType.String, 208, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TrackStandToStillLeftBikeClipId, PsoDataType.String, 212, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TrackStandToStillRightCharClipId, PsoDataType.String, 216, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TrackStandToStillRightBikeClipId, PsoDataType.String, 220, 7, 0)
                    );
                case MetaName.CTaskMotionOnBicycleController__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionOnBicycleController__Tunables, 0, 0, 184,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInStateToAllowTransitionFromFixieSkid, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeStillToTransitionToTrackStand, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInPedalState, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeInFreewheelState, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAiSpeedForStandingUp, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedToTriggerTrackStandTransition, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxSpeedToTriggerFixieSkidTransition, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseDuckPrepLeftCharClipId, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseDuckPrepRightCharClipId, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseDuckPrepLeftBikeClipId, PsoDataType.String, 52, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CruiseDuckPrepRightBikeClipId, PsoDataType.String, 56, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FastDuckPrepLeftCharClipId, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FastDuckPrepRightCharClipId, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FastDuckPrepLeftBikeClipId, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FastDuckPrepRightBikeClipId, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LaunchLeftCharClipId, PsoDataType.String, 76, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LaunchRightCharClipId, PsoDataType.String, 80, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LaunchLeftBikeClipId, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo(MetaName.LaunchRightBikeClipId, PsoDataType.String, 88, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TrackStandLeftCharClipId, PsoDataType.String, 92, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TrackStandRightCharClipId, PsoDataType.String, 96, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TrackStandLeftBikeClipId, PsoDataType.String, 100, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TrackStandRightBikeClipId, PsoDataType.String, 104, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidLeftCharClip0Id, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidLeftCharClip1Id, PsoDataType.String, 112, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidRightCharClip0Id, PsoDataType.String, 116, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidRightCharClip1Id, PsoDataType.String, 120, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidLeftBikeClip0Id, PsoDataType.String, 124, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidLeftBikeClip1Id, PsoDataType.String, 128, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidRightBikeClip0Id, PsoDataType.String, 132, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidRightBikeClip1Id, PsoDataType.String, 136, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidToBalanceLeftCharClip1Id, PsoDataType.String, 140, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidToBalanceRightCharClip1Id, PsoDataType.String, 144, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidToBalanceLeftBikeClip1Id, PsoDataType.String, 148, 7, 0),
                     new PsoStructureEntryInfo(MetaName.FixieSkidToBalanceRightBikeClip1Id, PsoDataType.String, 152, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CruisePedalCharClipId, PsoDataType.String, 156, 7, 0),
                     new PsoStructureEntryInfo(MetaName.InAirFreeWheelCharClipId, PsoDataType.String, 160, 7, 0),
                     new PsoStructureEntryInfo(MetaName.InAirFreeWheelBikeClipId, PsoDataType.String, 164, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DownHillInAirFreeWheelCharClipId, PsoDataType.String, 168, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DownHillInAirFreeWheelBikeClipId, PsoDataType.String, 172, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TuckFreeWheelToTrackStandRightCharClipId, PsoDataType.String, 176, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TuckFreeWheelToTrackStandRightBikeClipId, PsoDataType.String, 180, 7, 0)
                    );
                case MetaName.CBikeLeanAngleHelper__Tunables:
                    return new PsoStructureInfo(MetaName.CBikeLeanAngleHelper__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.UseReturnOvershoot, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseInitialLeanForcing, PsoDataType.Bool, 17, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredLeanAngleTolToBringLegIn, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredSpeedToBringLegIn, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredLeanAngleRate, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredLeanAngleRateQuad, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanAngleReturnRate, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanAngleDefaultRate, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanAngleDefaultRatePassenger, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DesiredOvershootLeanAngle, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeanAngleReturnedTol, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HasStickInputThreshold, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LeaningExtremeThreshold, PsoDataType.Float, 60, 0, 0)
                    );
                case MetaName.CTaskMotionInTurret__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskMotionInTurret__Tunables, 0, 0, 248,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4093217060, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4102381397, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2540703340, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3512769803, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo((MetaName)171275597, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2073783636, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1235675432, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2737439830, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1027124968, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2183745831, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3114603561, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)943989780, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2556074717, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1811862646, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)216639291, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1460193378, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3098186043, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1012095486, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1934656831, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2982819675, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2624331049, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1324859893, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1057562451, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1546988642, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3067513344, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1707713717, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2395321832, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1059545904, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3062906436, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2949666078, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo((MetaName)705780320, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo((MetaName)459162028, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1797630730, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1465756410, PsoDataType.String, 148, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1144665981, PsoDataType.String, 152, 7, 0),
                     new PsoStructureEntryInfo((MetaName)790728667, PsoDataType.String, 156, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2997429282, PsoDataType.String, 160, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2148334725, PsoDataType.String, 164, 7, 0),
                     new PsoStructureEntryInfo((MetaName)356569692, PsoDataType.String, 168, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2567235955, PsoDataType.String, 172, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3042448153, PsoDataType.String, 176, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TurnLeftSlowClipId, PsoDataType.String, 180, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TurnLeftFastClipId, PsoDataType.String, 184, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1206230540, PsoDataType.String, 188, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3713053450, PsoDataType.String, 192, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2463362878, PsoDataType.String, 196, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3874294070, PsoDataType.String, 200, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TurnRightSlowClipId, PsoDataType.String, 204, 7, 0),
                     new PsoStructureEntryInfo(MetaName.TurnRightFastClipId, PsoDataType.String, 208, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1611862735, PsoDataType.String, 212, 7, 0),
                     new PsoStructureEntryInfo((MetaName)413072568, PsoDataType.String, 216, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1114189303, PsoDataType.String, 220, 7, 0),
                     new PsoStructureEntryInfo((MetaName)531096295, PsoDataType.String, 224, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2077576253, PsoDataType.String, 228, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3278999439, PsoDataType.String, 232, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4038706223, PsoDataType.String, 236, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1558752243, PsoDataType.String, 240, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4031618294, PsoDataType.String, 244, 7, 0)
                    );
                case MetaName.CWanted__Tunables:
                    return new PsoStructureInfo(MetaName.CWanted__Tunables, 0, 0, 2696,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WantedClean, PsoDataType.Structure, 16, 0, MetaName.CWanted__Tunables__WantedLevel),
                     new PsoStructureEntryInfo(MetaName.WantedLevel1, PsoDataType.Structure, 128, 0, MetaName.CWanted__Tunables__WantedLevel),
                     new PsoStructureEntryInfo(MetaName.WantedLevel2, PsoDataType.Structure, 240, 0, MetaName.CWanted__Tunables__WantedLevel),
                     new PsoStructureEntryInfo(MetaName.WantedLevel3, PsoDataType.Structure, 352, 0, MetaName.CWanted__Tunables__WantedLevel),
                     new PsoStructureEntryInfo(MetaName.WantedLevel4, PsoDataType.Structure, 464, 0, MetaName.CWanted__Tunables__WantedLevel),
                     new PsoStructureEntryInfo(MetaName.WantedLevel5, PsoDataType.Structure, 576, 0, MetaName.CWanted__Tunables__WantedLevel),
                     new PsoStructureEntryInfo(MetaName.Difficulty, PsoDataType.Structure, 688, 0, (MetaName)2685546187),
                     new PsoStructureEntryInfo(MetaName.Rendering, PsoDataType.Structure, 2640, 0, (MetaName)1581603972),
                     new PsoStructureEntryInfo(MetaName.Timers, PsoDataType.Structure, 2656, 0, (MetaName)4245193745),
                     new PsoStructureEntryInfo(MetaName.MaxTimeTargetVehicleMoving, PsoDataType.UInt, 2672, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultAmnestyTime, PsoDataType.UInt, 2676, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DefaultHiddenEvasionTimeReduction, PsoDataType.UInt, 2680, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialAreaTimeoutWhenSeen, PsoDataType.UInt, 2684, 0, 0),
                     new PsoStructureEntryInfo(MetaName.InitialAreaTimeoutWhenCrimeReported, PsoDataType.UInt, 2688, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2940774268, PsoDataType.Float, 2692, 0, 0)
                    );
                case MetaName.CWanted__Tunables__WantedLevel:
                    return new PsoStructureInfo(MetaName.CWanted__Tunables__WantedLevel, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.Difficulty, PsoDataType.Structure, 8, 0, (MetaName)3660423994)
                    );
                case (MetaName)3660423994:
                    return new PsoStructureInfo((MetaName)3660423994, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.Calculation, PsoDataType.Structure, 8, 0, (MetaName)1037717481),
                     new PsoStructureEntryInfo(MetaName.Helis, PsoDataType.Structure, 72, 0, (MetaName)3986648608)
                    );
                case (MetaName)1037717481:
                    return new PsoStructureInfo((MetaName)1037717481, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.FromWantedLevel, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Weights, PsoDataType.Structure, 16, 0, (MetaName)1310408174),
                     new PsoStructureEntryInfo(MetaName.Decay, PsoDataType.Structure, 40, 0, (MetaName)3442100148)
                    );
                case (MetaName)1310408174:
                    return new PsoStructureInfo((MetaName)1310408174, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.WantedLevel, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LastSpottedDistance, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Randomness, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)3442100148:
                    return new PsoStructureInfo((MetaName)3442100148, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.TimeEvadingForMaxValue, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxValue, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableWhenOffMission, PsoDataType.Bool, 16, 0, 0)
                    );
                case (MetaName)3986648608:
                    return new PsoStructureInfo((MetaName)3986648608, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Refuel, PsoDataType.Structure, 8, 0, (MetaName)2123007699)
                    );
                case (MetaName)2123007699:
                    return new PsoStructureInfo((MetaName)2123007699, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBefore, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Delay, PsoDataType.Float, 16, 0, 0)
                    );
                case (MetaName)2685546187:
                    return new PsoStructureInfo((MetaName)2685546187, 0, 0, 1952,
                     new PsoStructureEntryInfo(MetaName.Spawning, PsoDataType.Structure, 8, 0, (MetaName)2145739988),
                     new PsoStructureEntryInfo(MetaName.Despawning, PsoDataType.Structure, 152, 0, (MetaName)1602748523),
                     new PsoStructureEntryInfo(MetaName.Peds, PsoDataType.Structure, 280, 0, (MetaName)3209293959),
                     new PsoStructureEntryInfo(MetaName.Dispatch, PsoDataType.Structure, 1920, 0, (MetaName)3840146119)
                    );
                case (MetaName)2145739988:
                    return new PsoStructureInfo((MetaName)2145739988, 0, 0, 144,
                     new PsoStructureEntryInfo(MetaName.Scoring, PsoDataType.Structure, 8, 0, (MetaName)3801009121),
                     new PsoStructureEntryInfo(MetaName.IdealDistance, PsoDataType.Structure, 96, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.ChancesToForceWaitInFront, PsoDataType.Structure, 120, 0, (MetaName)2528332511)
                    );
                case (MetaName)3801009121:
                    return new PsoStructureInfo((MetaName)3801009121, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.Weights, PsoDataType.Structure, 8, 0, (MetaName)2379615087)
                    );
                case (MetaName)2379615087:
                    return new PsoStructureInfo((MetaName)2379615087, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Distance, PsoDataType.Structure, 8, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.Direction, PsoDataType.Structure, 32, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.Randomness, PsoDataType.Structure, 56, 0, (MetaName)2528332511)
                    );
                case (MetaName)2528332511:
                    return new PsoStructureInfo((MetaName)2528332511, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Min, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Max, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMin, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ValueForMax, PsoDataType.Float, 20, 0, 0)
                    );
                case (MetaName)1602748523:
                    return new PsoStructureInfo((MetaName)1602748523, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.MaxFacingThreshold, PsoDataType.Structure, 8, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.MaxMovingThreshold, PsoDataType.Structure, 32, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToBeConsideredLaggingBehind, PsoDataType.Structure, 56, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToCheckClumped, PsoDataType.Structure, 80, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToBeConsideredClumped, PsoDataType.Structure, 104, 0, (MetaName)2528332511)
                    );
                case (MetaName)3209293959:
                    return new PsoStructureInfo((MetaName)3209293959, 0, 0, 1640,
                     new PsoStructureEntryInfo(MetaName.Cops, PsoDataType.Structure, 8, 0, (MetaName)2022711333),
                     new PsoStructureEntryInfo(MetaName.Swat, PsoDataType.Structure, 552, 0, (MetaName)2022711333),
                     new PsoStructureEntryInfo(MetaName.Army, PsoDataType.Structure, 1096, 0, (MetaName)2022711333)
                    );
                case (MetaName)2022711333:
                    return new PsoStructureInfo((MetaName)2022711333, 0, 0, 544,
                     new PsoStructureEntryInfo(MetaName.Situations, PsoDataType.Structure, 8, 0, (MetaName)1181208064),
                     new PsoStructureEntryInfo(MetaName.AutomobileSpeedModifier, PsoDataType.Structure, 496, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.HeliSpeedModifier, PsoDataType.Structure, 520, 0, (MetaName)2528332511)
                    );
                case (MetaName)1181208064:
                    return new PsoStructureInfo((MetaName)1181208064, 0, 0, 488,
                     new PsoStructureEntryInfo(MetaName.Default, PsoDataType.Structure, 8, 0, (MetaName)3795144262),
                     new PsoStructureEntryInfo(MetaName.InVehicle, PsoDataType.Structure, 128, 0, (MetaName)3795144262),
                     new PsoStructureEntryInfo(MetaName.InHeli, PsoDataType.Structure, 248, 0, (MetaName)3795144262),
                     new PsoStructureEntryInfo(MetaName.InBoat, PsoDataType.Structure, 368, 0, (MetaName)3795144262)
                    );
                case (MetaName)3795144262:
                    return new PsoStructureInfo((MetaName)3795144262, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.SensesRange, PsoDataType.Structure, 8, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.IdentificationRange, PsoDataType.Structure, 32, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.ShootRateModifier, PsoDataType.Structure, 56, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.WeaponAccuracy, PsoDataType.Structure, 80, 0, (MetaName)2528332511),
                     new PsoStructureEntryInfo(MetaName.WeaponAccuracyModifierForEvasiveMovement, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponAccuracyModifierForOffScreen, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponAccuracyModifierForAimedAt, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinForDrivebys, PsoDataType.Float, 116, 0, 0)
                    );
                case (MetaName)3840146119:
                    return new PsoStructureInfo((MetaName)3840146119, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.TimeBetweenSpawnAttemptsModifier, PsoDataType.Structure, 8, 0, (MetaName)2528332511)
                    );
                case (MetaName)1581603972:
                    return new PsoStructureInfo((MetaName)1581603972, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.Enabled, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Witnesses, PsoDataType.Bool, 9, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Crimes, PsoDataType.Bool, 10, 0, 0)
                    );
                case (MetaName)4245193745:
                    return new PsoStructureInfo((MetaName)4245193745, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.TimeBetweenDifficultyUpdates, PsoDataType.Float, 8, 0, 0)
                    );
                case MetaName.CEventGunAimedAt__Tunables:
                    return new PsoStructureInfo(MetaName.CEventGunAimedAt__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimer, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CEventGunShot__Tunables:
                    return new PsoStructureInfo(MetaName.CEventGunShot__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimer, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GunShotThresholdDistance, PsoDataType.Float, 24, 0, 0)
                    );
                case MetaName.CEventMeleeAction__Tunables:
                    return new PsoStructureInfo(MetaName.CEventMeleeAction__Tunables, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinDelayTimer, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDelayTimer, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.CTaskAimAndThrowProjectile__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAimAndThrowProjectile__Tunables, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.bEnableGaitAdditive, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinHoldThrowPitch, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iMaxRandomExplosionTime, PsoDataType.UInt, 24, 0, 0)
                    );
                case MetaName.CTaskSwapWeapon__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskSwapWeapon__Tunables, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.OnFootClipRate, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.OnFootBlendInDuration, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowCoverClipRate, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LowCoverBlendInDuration, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighCoverClipRate, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.HighCoverBlendInDuration, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ActionClipRate, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ActionBlendInDuration, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutDuration, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2181619027, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DebugSwapInstantly, PsoDataType.Bool, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SkipHolsterWeapon, PsoDataType.Bool, 57, 0, 0)
                    );
                case MetaName.CTaskAimGunOnFoot__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAimGunOnFoot__Tunables, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenFiringVariations, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IdealPitchForFiringVariation, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxPitchDifferenceForFiringVariation, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AssistedAimOutroTime, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RunAndGunOutroTime, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimOutroTime, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimOutroTimeIfAimingOnStick, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimOutroMinTaskTimeWhenRunPressed, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimingOnStickExitCooldown, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeForRunAndGunOutroDelays, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DampenRootTargetWeight, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DampenRootTargetHeight, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AlternativeAnimBlockedHeight, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CoverAimOffsetFromBlocked, PsoDataType.Float3, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DelayTimeWhenOutOfAmmoInScope, PsoDataType.UInt, 96, 0, 0)
                    );
                case MetaName.CTaskGun__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskGun__Tunables, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.iMinLookAtTime, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iMaxLookAtTime, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeBetweenBulletReactions, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxTimeBetweenBulletReactions, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxDistForOverheadReactions, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxAboveHeadForOverheadReactions, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fBulletReactionPosAdjustmentZ, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMinTimeBetweenLookAt, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMaxTimeBetweenLookAt, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bDisable2HandedGetups, PsoDataType.Bool, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeForEyeIk, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenEyeIkProcesses, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToPointGunAtPositionWhenUnableToTurn, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AssistedAimCamera, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RunAndGunAimCamera, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AssistedAimInterpolateInDuration, PsoDataType.UInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RunAndGunInterpolateInDuration, PsoDataType.UInt, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenOverheadBulletReactions, PsoDataType.UInt, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxTimeInBulletReactionState, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3228655867, PsoDataType.UInt, 92, 0, 0),
                     new PsoStructureEntryInfo((MetaName)270252777, PsoDataType.UInt, 96, 0, 0)
                    );
                case MetaName.CTaskAimGunVehicleDriveBy__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskAimGunVehicleDriveBy__Tunables, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenInsults, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToInsult, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDotToInsult, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinAimTimeMs, PsoDataType.UInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxAimTimeOnStickMs, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AimingOnStickCooldownMs, PsoDataType.UInt, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BicycleDrivebyFilterId, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BikeDrivebyFilterId, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.JetskiDrivebyFilterId, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ParachutingFilterId, PsoDataType.String, 52, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2221302483, PsoDataType.String, 56, 7, 0),
                     new PsoStructureEntryInfo((MetaName)3323392529, PsoDataType.String, 60, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2684719351, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4246561663, PsoDataType.String, 68, 7, 0)
                    );
                case MetaName.CWildlifeManager__Tunables:
                    return new PsoStructureInfo(MetaName.CWildlifeManager__Tunables, 0, 0, 136,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BirdHeightMapDeltaMin, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BirdHeightMapDeltaMax, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BirdSpawnXYRangeMin, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BirdSpawnXYRangeMax, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IncreasedAerialSpawningFactor, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToSearchForGroundWildlifePoints, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToSearchForGroundWildlifePoints, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenGroundProbes, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroundMaterialProbeDepth, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroundMaterialProbeOffset, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.GroundMaterialSpawnCoordNormalZTolerance, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinDistanceToSearchForAquaticPoints, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxDistanceToSearchForAquaticPoints, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenWaterHeightMapChecks, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TimeBetweenWaterProbes, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WaterProbeDepth, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WaterProbeOffset, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AquaticSpawnDepth, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AquaticSpawnMaxHeightAbovePlayer, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IncreasedAquaticSpawningFactor, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CloseSpawningViewMultiplier, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IncreasedGroundWildlifeSpawningFactor, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SharkModelName, PsoDataType.String, 104, 7, 0),
                     new PsoStructureEntryInfo(MetaName.DeepWaterThreshold, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PlayerSwimTimeThreshold, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MinTimeBetweenSharkDispatches, PsoDataType.UInt, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SharkAddRangeInViewMin, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3653722296, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1235067475, PsoDataType.Float, 128, 0, 0)
                    );
                case MetaName.CPopGroupList:
                    return new PsoStructureInfo(MetaName.CPopGroupList, 0, 0, 56,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPopulationGroup),
                     new PsoStructureEntryInfo(MetaName.pedGroups, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPopulationGroup),
                     new PsoStructureEntryInfo(MetaName.vehGroups, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.wildlifeHabitats, PsoDataType.Array, 40, 0, (MetaName)4)
                    );
                case MetaName.CPopulationGroup:
                    return new PsoStructureInfo(MetaName.CPopulationGroup, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3852980612),
                     new PsoStructureEntryInfo(MetaName.models, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)4186696686),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.Flags, 32, 0, (MetaName)2097155)
                    );
                case MetaName.CDoorTuningFile:
                    return new PsoStructureInfo(MetaName.CDoorTuningFile, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)607870603),
                     new PsoStructureEntryInfo(MetaName.NamedTuningArray, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)934147895),
                     new PsoStructureEntryInfo(MetaName.ModelToTuneMapping, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case (MetaName)607870603:
                    return new PsoStructureInfo((MetaName)607870603, 0, 0, 144,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 2, 0),
                     new PsoStructureEntryInfo(MetaName.Tuning, PsoDataType.Structure, 16, 0, (MetaName)2175119364)
                    );
                case (MetaName)2175119364:
                    return new PsoStructureInfo((MetaName)2175119364, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.AutoOpenVolumeOffset, PsoDataType.Float3a, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)2809332931),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 32, 0, (MetaName)2097153),
                     new PsoStructureEntryInfo(MetaName.AutoOpenRadiusModifier, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AutoOpenRate, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AutoOpenCosineAngleBetweenThreshold, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AutoOpenCloseRateTaper, PsoDataType.Bool, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UseAutoOpenTriggerBox, PsoDataType.Bool, 49, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CustomTriggerBox, PsoDataType.Bool, 50, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TriggerBoxMinMax, PsoDataType.Structure, 64, 0, MetaName.rage__spdAABB),
                     new PsoStructureEntryInfo(MetaName.BreakableByVehicle, PsoDataType.Bool, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BreakingImpulse, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ShouldLatchShut, PsoDataType.Bool, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MassMultiplier, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WeaponImpulseMultiplier, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RotationLimitAngle, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TorqueAngularVelocityLimit, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StdDoorRotDir, PsoDataType.Enum, 124, 0, MetaName.StdDoorRotDir)
                    );
                case (MetaName)934147895:
                    return new PsoStructureInfo((MetaName)934147895, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ModelName, PsoDataType.String, 8, 2, 0),
                     new PsoStructureEntryInfo(MetaName.TuningName, PsoDataType.String, 16, 2, 0)
                    );
                case (MetaName)3852980612:
                    return new PsoStructureInfo((MetaName)3852980612, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Variations, PsoDataType.Structure, 16, 3, 0)
                    );
                case MetaName.CAmbientPedModelVariations:
                    return new PsoStructureInfo(MetaName.CAmbientPedModelVariations, 0, 0, 48,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3553377863),
                     new PsoStructureEntryInfo(MetaName.CompRestrictions, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3159204365),
                     new PsoStructureEntryInfo(MetaName.PropRestrictions, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo(MetaName.LoadOut, PsoDataType.String, 40, 7, 0)
                    );
                case (MetaName)3553377863:
                    return new PsoStructureInfo((MetaName)3553377863, 0, 0, 12,
                     new PsoStructureEntryInfo(MetaName.Component, PsoDataType.Enum, 0, 0, MetaName.ePedVarComp),
                     new PsoStructureEntryInfo(MetaName.DrawableIndex, PsoDataType.SInt, 4, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Restriction, PsoDataType.Enum, 8, 0, (MetaName)2065671281)
                    );
                case MetaName.CSlownessZoneManager:
                    return new PsoStructureInfo(MetaName.CSlownessZoneManager, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1778476350),
                     new PsoStructureEntryInfo(MetaName.aSlownessZone, PsoDataType.Array, 8, 0, 0)
                    );
                case (MetaName)1778476350:
                    return new PsoStructureInfo((MetaName)1778476350, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.bBox, PsoDataType.Structure, 16, 0, MetaName.rage__spdAABB)
                    );
                case MetaName.CScenarioPointManifest:
                    return new PsoStructureInfo(MetaName.CScenarioPointManifest, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.VersionNumber, PsoDataType.SInt, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.RegionDefs, PsoDataType.Array, 8, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.Groups, PsoDataType.Array, 24, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.InteriorNames, PsoDataType.Array, 40, 0, (MetaName)5)
                    );
                case MetaName.CScenarioPointRegionDef:
                    return new PsoStructureInfo(MetaName.CScenarioPointRegionDef, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 0, 8, 0),
                     new PsoStructureEntryInfo(MetaName.AABB, PsoDataType.Structure, 16, 0, MetaName.rage__spdAABB)
                    );
                case MetaName.CScenarioPointGroup:
                    return new PsoStructureInfo(MetaName.CScenarioPointGroup, 0, 0, 8,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.EnabledByDefault, PsoDataType.Bool, 4, 0, 0)
                    );
                case MetaName.strRequestRecording:
                    return new PsoStructureInfo(MetaName.strRequestRecording, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.strRecordedRequest),
                     new PsoStructureEntryInfo(MetaName.Requests, PsoDataType.Array, 0, 0, 0)
                    );
                case MetaName.strRecordedRequest:
                    return new PsoStructureInfo(MetaName.strRecordedRequest, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Resource, PsoDataType.Structure, 0, 0, MetaName.strResourceReference),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.SInt, 24, 0, 0)
                    );
                case MetaName.strResourceReference:
                    return new PsoStructureInfo(MetaName.strResourceReference, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.AssetName, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.Extension, PsoDataType.String, 16, 0, (MetaName)524288)
                    );
                case (MetaName)62557530:
                    return new PsoStructureInfo((MetaName)62557530, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3112305802),
                     new PsoStructureEntryInfo(MetaName.zones, PsoDataType.Array, 0, 0, 0)
                    );
                case (MetaName)3112305802:
                    return new PsoStructureInfo((MetaName)3112305802, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.zoneName, PsoDataType.String, 0, 2, 0),
                     new PsoStructureEntryInfo(MetaName.spName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.mpName, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.vfxRegion, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)1896983690, PsoDataType.UByte, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1856972365, PsoDataType.Enum, 24, 0, (MetaName)1756502932),
                     new PsoStructureEntryInfo(MetaName.lawResponseTime, PsoDataType.Enum, 28, 0, (MetaName)3204395397),
                     new PsoStructureEntryInfo(MetaName.lawResponseType, PsoDataType.Enum, 32, 0, (MetaName)4286852891),
                     new PsoStructureEntryInfo(MetaName.specialZoneAttribute, PsoDataType.Enum, 36, 0, (MetaName)4095090001),
                     new PsoStructureEntryInfo(MetaName.vehDirtMin, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vehDirtMax, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vehDirtGrowScale, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pedDirtMin, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pedDirtMax, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.dirtRed, PsoDataType.UByte, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.dirtGreen, PsoDataType.UByte, 61, 0, 0),
                     new PsoStructureEntryInfo(MetaName.dirtBlue, PsoDataType.UByte, 62, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1430479029, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2056852733, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1457405649, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2304154060, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1018913504, PsoDataType.SInt, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)9361713, PsoDataType.Bool, 84, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2931752685, PsoDataType.Bool, 85, 0, 0)
                    );
                case MetaName.CCoverTuningFile:
                    return new PsoStructureInfo(MetaName.CCoverTuningFile, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)912106952),
                     new PsoStructureEntryInfo(MetaName.NamedTuningArray, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)3821651387),
                     new PsoStructureEntryInfo(MetaName.ModelToTuneMapping, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case (MetaName)912106952:
                    return new PsoStructureInfo((MetaName)912106952, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 2, 0),
                     new PsoStructureEntryInfo(MetaName.Tuning, PsoDataType.Structure, 16, 0, (MetaName)61864478)
                    );
                case (MetaName)61864478:
                    return new PsoStructureInfo((MetaName)61864478, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)98189892),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 8, 0, (MetaName)2097152)
                    );
                case (MetaName)3821651387:
                    return new PsoStructureInfo((MetaName)3821651387, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ModelName, PsoDataType.String, 8, 2, 0),
                     new PsoStructureEntryInfo(MetaName.TuningName, PsoDataType.String, 16, 2, 0)
                    );
                case MetaName.CScenarioPointRegion:
                    return new PsoStructureInfo(MetaName.CScenarioPointRegion, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.VersionNumber, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CExtensionDefSpawnPoint),
                     new PsoStructureEntryInfo(MetaName.ScenarioPoints, PsoDataType.Array, 16, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)3696045377, PsoDataType.Structure, 32, 0, (MetaName)4023740759)
                    );
                case MetaName.CExtensionDefSpawnPoint:
                    return new PsoStructureInfo(MetaName.CExtensionDefSpawnPoint, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.offsetPosition, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.offsetRotation, PsoDataType.Float4, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.spawnType, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.pedType, PsoDataType.String, 52, 7, 0),
                     new PsoStructureEntryInfo(MetaName.group, PsoDataType.String, 56, 7, 0),
                     new PsoStructureEntryInfo(MetaName.availableInMpSp, PsoDataType.Enum, 60, 0, (MetaName)3573596290),
                     new PsoStructureEntryInfo(MetaName.timeTillPedLeaves, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.start, PsoDataType.UByte, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.end, PsoDataType.UByte, 69, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.CScenarioPointFlags__Flags),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.Flags, 72, 0, (MetaName)196618),
                     new PsoStructureEntryInfo(MetaName.highPri, PsoDataType.Bool, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.extendedRange, PsoDataType.Bool, 77, 0, 0)
                    );
                case (MetaName)4023740759:
                    return new PsoStructureInfo((MetaName)4023740759, 0, 0, 40,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CScenarioChainingNode),
                     new PsoStructureEntryInfo(MetaName.Nodes, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CScenarioChainingEdge),
                     new PsoStructureEntryInfo(MetaName.Edges, PsoDataType.Array, 24, 0, (MetaName)2)
                    );
                case MetaName.CScenarioChainingNode:
                    return new PsoStructureInfo(MetaName.CScenarioChainingNode, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Position, PsoDataType.Float3a, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2602393771, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ScenarioType, PsoDataType.String, 36, 7, 0),
                     new PsoStructureEntryInfo((MetaName)407126079, PsoDataType.Bool, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1308720135, PsoDataType.Bool, 41, 0, 0)
                    );
                case MetaName.CScenarioChainingEdge:
                    return new PsoStructureInfo(MetaName.CScenarioChainingEdge, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.NodeIndexFrom, PsoDataType.UShort, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NodeIndexTo, PsoDataType.UShort, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.NavMode, PsoDataType.Enum, 12, 2, MetaName.CScenarioChainingEdge__eNavMode),
                     new PsoStructureEntryInfo(MetaName.NavSpeed, PsoDataType.Enum, 13, 2, MetaName.CScenarioChainingEdge__eNavSpeed)
                    );
                case MetaName.CPedVariationInfo:
                    return new PsoStructureInfo(MetaName.CPedVariationInfo, 0, 0, 112,
                     new PsoStructureEntryInfo(MetaName.bHasTexVariations, PsoDataType.Bool, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bHasDrawblVariations, PsoDataType.Bool, 1, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bHasLowLODs, PsoDataType.Bool, 2, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bIsSuperLOD, PsoDataType.Bool, 3, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.availComp, PsoDataType.Array, 4, 4, (MetaName)786436),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPVComponentData),
                     new PsoStructureEntryInfo(MetaName.aComponentData3, PsoDataType.Array, 16, 0, (MetaName)6),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPedSelectionSet),
                     new PsoStructureEntryInfo(MetaName.aSelectionSets, PsoDataType.Array, 32, 0, (MetaName)8),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CComponentInfo),
                     new PsoStructureEntryInfo(MetaName.compInfos, PsoDataType.Array, 48, 0, (MetaName)10),
                     new PsoStructureEntryInfo(MetaName.propInfo, PsoDataType.Structure, 64, 0, MetaName.CPedPropInfo),
                     new PsoStructureEntryInfo(MetaName.dlcName, PsoDataType.String, 104, 8, 0)
                    );
                case MetaName.CPedPropInfo:
                    return new PsoStructureInfo(MetaName.CPedPropInfo, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.numAvailProps, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPedPropMetaData),
                     new PsoStructureEntryInfo(MetaName.aPropMetaData, PsoDataType.Array, 8, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CAnchorProps),
                     new PsoStructureEntryInfo(MetaName.aAnchors, PsoDataType.Array, 24, 0, (MetaName)3)
                    );
                case MetaName.CPVComponentData:
                    return new PsoStructureInfo(MetaName.CPVComponentData, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.numAvailTex, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPVDrawblData),
                     new PsoStructureEntryInfo(MetaName.aDrawblData3, PsoDataType.Array, 8, 0, (MetaName)1)
                    );
                case MetaName.CPVDrawblData:
                    return new PsoStructureInfo(MetaName.CPVDrawblData, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.propMask, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.numAlternatives, PsoDataType.UByte, 1, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPVTextureData),
                     new PsoStructureEntryInfo(MetaName.aTexData, PsoDataType.Array, 8, 0, (MetaName)2),
                     new PsoStructureEntryInfo(MetaName.clothData, PsoDataType.Structure, 24, 0, (MetaName)2236980467)
                    );
                case (MetaName)2236980467:
                    return new PsoStructureInfo((MetaName)2236980467, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ownsCloth, PsoDataType.Bool, 0, 0, 0)
                    );
                case MetaName.CPVTextureData:
                    return new PsoStructureInfo(MetaName.CPVTextureData, 0, 0, 3,
                     new PsoStructureEntryInfo(MetaName.texId, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.distribution, PsoDataType.UByte, 1, 0, 0)
                    );
                case MetaName.CComponentInfo:
                    return new PsoStructureInfo(MetaName.CComponentInfo, 0, 0, 48,
                     new PsoStructureEntryInfo((MetaName)802196719, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4233133352, PsoDataType.String, 4, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)128864925, PsoDataType.Array, 8, 4, (MetaName)327682),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.UInt, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.inclusions, PsoDataType.Flags, 32, 0, (MetaName)2101247),
                     new PsoStructureEntryInfo(MetaName.exclusions, PsoDataType.Flags, 36, 0, (MetaName)2101247),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.ePedVarComp),
                     new PsoStructureEntryInfo((MetaName)1613922652, PsoDataType.Flags, 40, 1, (MetaName)1048583),
                     new PsoStructureEntryInfo((MetaName)2114993291, PsoDataType.UShort, 42, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3509540765, PsoDataType.UByte, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4196345791, PsoDataType.UByte, 45, 0, 0)
                    );
                case MetaName.CCreatureMetaData:
                    return new PsoStructureInfo(MetaName.CCreatureMetaData, 0, 0, 56,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CShaderVariableComponent),
                     new PsoStructureEntryInfo(MetaName.shaderVariableComponents, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPedPropExpressionData),
                     new PsoStructureEntryInfo(MetaName.pedPropExpressions, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CPedCompExpressionData),
                     new PsoStructureEntryInfo(MetaName.pedCompExpressions, PsoDataType.Array, 40, 0, (MetaName)4)
                    );
                case MetaName.CShaderVariableComponent:
                    return new PsoStructureInfo(MetaName.CShaderVariableComponent, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.pedcompID, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maskID, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.shaderVariableHashString, PsoDataType.String, 16, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.tracks, PsoDataType.Array, 24, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UShort, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ids, PsoDataType.Array, 40, 0, (MetaName)5),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.components, PsoDataType.Array, 56, 0, (MetaName)MetaTypeName.POINTER)
                    );
                case MetaName.CPedPropExpressionData:
                    return new PsoStructureInfo(MetaName.CPedPropExpressionData, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.pedPropID, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pedPropVarIndex, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pedPropExpressionIndex, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.tracks, PsoDataType.Array, 24, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UShort, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ids, PsoDataType.Array, 40, 0, (MetaName)5),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.types, PsoDataType.Array, 56, 0, (MetaName)MetaTypeName.POINTER),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.components, PsoDataType.Array, 72, 0, (MetaName)9)
                    );
                case MetaName.CPedCompExpressionData:
                    return new PsoStructureInfo(MetaName.CPedCompExpressionData, 0, 0, 88,
                     new PsoStructureEntryInfo(MetaName.pedCompID, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pedCompVarIndex, PsoDataType.SInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pedCompExpressionIndex, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.tracks, PsoDataType.Array, 24, 0, (MetaName)3),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UShort, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ids, PsoDataType.Array, 40, 0, (MetaName)5),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.types, PsoDataType.Array, 56, 0, (MetaName)MetaTypeName.POINTER),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.components, PsoDataType.Array, 72, 0, (MetaName)9)
                    );
                case MetaName.rage__cutfCutsceneFile2:
                    return new PsoStructureInfo(MetaName.rage__cutfCutsceneFile2, 0, 0, 5296,
                     new PsoStructureEntryInfo(MetaName.fTotalDuration, PsoDataType.Float, 268, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cFaceDir, PsoDataType.String, 272, 0, (MetaName)16777216),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iCutsceneFlags, PsoDataType.Array, 528, 4, (MetaName)262146),
                     new PsoStructureEntryInfo(MetaName.vOffset, PsoDataType.Float3, 544, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRotation, PsoDataType.Float, 560, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cExtraRoom, PsoDataType.String, 564, 0, (MetaName)1572864),
                     new PsoStructureEntryInfo(MetaName.vExtraRoomPos, PsoDataType.Float3, 592, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vTriggerOffset, PsoDataType.Float3, 608, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.pCutsceneObjects, PsoDataType.Array, 624, 0, (MetaName)9),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.pCutsceneLoadEventList, PsoDataType.Array, 640, 0, (MetaName)11),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.pCutsceneEventList, PsoDataType.Array, 656, 0, (MetaName)13),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.pCutsceneEventArgsList, PsoDataType.Array, 672, 0, (MetaName)15),
                     new PsoStructureEntryInfo(MetaName.attributes, PsoDataType.Structure, 688, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 696, 4, 0),
                     new PsoStructureEntryInfo(MetaName.iRangeStart, PsoDataType.SInt, 704, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iRangeEnd, PsoDataType.SInt, 708, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iAltRangeEnd, PsoDataType.SInt, 712, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSectionByTimeSliceDuration, PsoDataType.Float, 716, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFadeOutCutsceneDuration, PsoDataType.Float, 720, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFadeInGameDuration, PsoDataType.Float, 724, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fadeInColor, PsoDataType.UInt, 728, 1, 0),
                     new PsoStructureEntryInfo(MetaName.iBlendOutCutsceneDuration, PsoDataType.SInt, 732, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iBlendOutCutsceneOffset, PsoDataType.SInt, 736, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFadeOutGameDuration, PsoDataType.Float, 740, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFadeInCutsceneDuration, PsoDataType.Float, 744, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fadeOutColor, PsoDataType.UInt, 748, 1, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cameraCutList, PsoDataType.Array, 752, 0, (MetaName)31),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.sectionSplitList, PsoDataType.Array, 768, 0, (MetaName)MetaTypeName.FLOAT),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.rage__cutfCutsceneFile2__SConcatData),
                     new PsoStructureEntryInfo(MetaName.concatDataList, PsoDataType.Array, 784, 1, (MetaName)2621475),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.vHaltFrequency),
                     new PsoStructureEntryInfo(MetaName.discardFrameList, PsoDataType.Array, 5280, 0, (MetaName)37)
                    );
                case MetaName.rage__parAttributeList:
                    return new PsoStructureInfo(MetaName.rage__parAttributeList, 0, 0, 12,
                     new PsoStructureEntryInfo(MetaName.UserData1, PsoDataType.UByte, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.UserData2, PsoDataType.UByte, 9, 0, 0)
                    );
                case MetaName.rage__cutfCutsceneFile2__SConcatData:
                    return new PsoStructureInfo(MetaName.rage__cutfCutsceneFile2__SConcatData, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.cSceneName, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.vOffset, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fStartTime, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRotation, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPitch, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRoll, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iRangeStart, PsoDataType.SInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iRangeEnd, PsoDataType.SInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bValidForPlayBack, PsoDataType.Bool, 56, 0, 0)
                    );
                case MetaName.rage__cutfAssetManagerObject:
                    return new PsoStructureInfo(MetaName.rage__cutfAssetManagerObject, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0)
                    );
                case MetaName.rage__cutfAnimationManagerObject:
                    return new PsoStructureInfo(MetaName.rage__cutfAnimationManagerObject, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0)
                    );
                case MetaName.rage__cutfCameraObject:
                    return new PsoStructureInfo(MetaName.rage__cutfCameraObject, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AnimStreamingBase, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fNearDrawDistance, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFarDrawDistance, PsoDataType.Float, 60, 0, 0)
                    );
                case MetaName.rage__cutfPedModelObject:
                    return new PsoStructureInfo(MetaName.rage__cutfPedModelObject, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamingName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AnimStreamingBase, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cAnimExportCtrlSpecFile, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cFaceExportCtrlSpecFile, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cAnimCompressionFile, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cHandle, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo(MetaName.typeFile, PsoDataType.String, 88, 7, 0),
                     new PsoStructureEntryInfo(MetaName.overrideFaceAnimationFilename, PsoDataType.String, 96, 7, 0),
                     new PsoStructureEntryInfo(MetaName.bFoundFaceAnimation, PsoDataType.Bool, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bFaceAndBodyAreMerged, PsoDataType.Bool, 105, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bOverrideFaceAnimation, PsoDataType.Bool, 106, 0, 0),
                     new PsoStructureEntryInfo(MetaName.faceAnimationNodeName, PsoDataType.String, 108, 7, 0),
                     new PsoStructureEntryInfo(MetaName.faceAttributesFilename, PsoDataType.String, 112, 7, 0)
                    );
                case MetaName.rage__cutfPropModelObject:
                    return new PsoStructureInfo(MetaName.rage__cutfPropModelObject, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamingName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AnimStreamingBase, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cAnimExportCtrlSpecFile, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cFaceExportCtrlSpecFile, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cAnimCompressionFile, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cHandle, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo(MetaName.typeFile, PsoDataType.String, 88, 7, 0)
                    );
                case MetaName.rage__cutfBlockingBoundsObject:
                    return new PsoStructureInfo(MetaName.rage__cutfBlockingBoundsObject, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float3, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vCorners, PsoDataType.Array, 48, 4, (MetaName)262148),
                     new PsoStructureEntryInfo(MetaName.fHeight, PsoDataType.Float, 112, 0, 0)
                    );
                case MetaName.rage__cutfAudioObject:
                    return new PsoStructureInfo(MetaName.rage__cutfAudioObject, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 3, 0),
                     new PsoStructureEntryInfo(MetaName.fOffset, PsoDataType.Float, 56, 0, 0)
                    );
                case MetaName.rage__cutfHiddenModelObject:
                    return new PsoStructureInfo(MetaName.rage__cutfHiddenModelObject, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.vPosition, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRadius, PsoDataType.Float, 64, 0, 0)
                    );
                case MetaName.rage__cutfOverlayObject:
                    return new PsoStructureInfo(MetaName.rage__cutfOverlayObject, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 3, 0),
                     new PsoStructureEntryInfo(MetaName.cRenderTargetName, PsoDataType.String, 56, 3, 0),
                     new PsoStructureEntryInfo(MetaName.iOverlayType, PsoDataType.UInt, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.modelHashName, PsoDataType.String, 76, 7, 0)
                    );
                case MetaName.rage__cutfSubtitleObject:
                    return new PsoStructureInfo(MetaName.rage__cutfSubtitleObject, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0)
                    );
                case MetaName.rage__cutfLightObject:
                    return new PsoStructureInfo(MetaName.rage__cutfLightObject, 0, 0, 240,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 16, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 32, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.vDirection, PsoDataType.Float3, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vColour, PsoDataType.Float3, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vPosition, PsoDataType.Float3, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fIntensity, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFallOff, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fConeAngle, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVolumeIntensity, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVolumeSizeScale, PsoDataType.Float, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCoronaSize, PsoDataType.Float, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCoronaIntensity, PsoDataType.Float, 184, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCoronaZBias, PsoDataType.Float, 188, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInnerConeAngle, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fExponentialFallOff, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iLightType, PsoDataType.SInt, 200, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iLightProperty, PsoDataType.SInt, 204, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TextureDictID, PsoDataType.SInt, 208, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TextureKey, PsoDataType.SInt, 212, 0, 0),
                     new PsoStructureEntryInfo((MetaName)34975788, PsoDataType.SInt, 216, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uLightFlags, PsoDataType.UInt, 220, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uHourFlags, PsoDataType.UInt, 224, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1437992521, PsoDataType.UShort, 228, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bStatic, PsoDataType.Bool, 230, 0, 0)
                    );
                case MetaName.rage__cutfAnimatedLightObject:
                    return new PsoStructureInfo(MetaName.rage__cutfAnimatedLightObject, 0, 0, 208,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.vDirection, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vColour, PsoDataType.Float3, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vPosition, PsoDataType.Float3, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fIntensity, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFallOff, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fConeAngle, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVolumeIntensity, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fVolumeSizeScale, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCoronaSize, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCoronaIntensity, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fCoronaZBias, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fInnerConeAngle, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fExponentialFallOff, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fShadowBlur, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iLightType, PsoDataType.SInt, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iLightProperty, PsoDataType.SInt, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TextureDictID, PsoDataType.SInt, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.TextureKey, PsoDataType.SInt, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uLightFlags, PsoDataType.UInt, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.uHourFlags, PsoDataType.UInt, 180, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bStatic, PsoDataType.Bool, 186, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AnimStreamingBase, PsoDataType.UInt, 192, 0, 0)
                    );
                case MetaName.rage__cutfObjectIdEvent:
                    return new PsoStructureInfo(MetaName.rage__cutfObjectIdEvent, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.fTime, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iEventId, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iEventArgsIndex, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pChildEvents, PsoDataType.Structure, 32, 3, 0),
                     new PsoStructureEntryInfo(MetaName.StickyId, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsChild, PsoDataType.Bool, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 48, 0, 0)
                    );
                case MetaName.rage__cutfObjectVariationEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfObjectVariationEventArgs, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iComponent, PsoDataType.SInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iDrawable, PsoDataType.SInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iTexture, PsoDataType.SInt, 48, 0, 0)
                    );
                case MetaName.rage__cutfEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfEventArgs, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0)
                    );
                case MetaName.rage__cutfAttributeList:
                    return new PsoStructureInfo(MetaName.rage__cutfAttributeList, 0, 0, 16,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.Items, PsoDataType.Array, 0, 0, 0)
                    );
                case (MetaName)1626675902:
                    return new PsoStructureInfo((MetaName)1626675902, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.SInt, 16, 0, 0)
                    );
                case (MetaName)1674696498:
                    return new PsoStructureInfo((MetaName)1674696498, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.Float, 16, 0, 0)
                    );
                case MetaName.rage__cutfLoadSceneEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfLoadSceneEventArgs, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.vOffset, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRotation, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fPitch, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRoll, PsoDataType.Float, 72, 0, 0)
                    );
                case MetaName.rage__cutfObjectIdEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfObjectIdEventArgs, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 32, 0, 0)
                    );
                case MetaName.rage__cutfObjectIdListEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfObjectIdListEventArgs, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.SInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iObjectIdList, PsoDataType.Array, 32, 0, (MetaName)2)
                    );
                case MetaName.rage__cutfNameEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfNameEventArgs, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 32, 7, 0)
                    );
                case MetaName.rage__cutfCameraCutEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfCameraCutEventArgs, 0, 0, 272,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.vPosition, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vRotationQuaternion, PsoDataType.Float4, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fNearDrawDistance, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fFarDrawDistance, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fMapLodScale, PsoDataType.Float, 88, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReflectionLodRangeStart, PsoDataType.Float, 92, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReflectionLodRangeEnd, PsoDataType.Float, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReflectionSLodRangeStart, PsoDataType.Float, 100, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ReflectionSLodRangeEnd, PsoDataType.Float, 104, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LodMultHD, PsoDataType.Float, 108, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LodMultOrphanedHD, PsoDataType.Float, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LodMultLod, PsoDataType.Float, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LodMultSLod1, PsoDataType.Float, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LodMultSLod2, PsoDataType.Float, 124, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LodMultSLod3, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LodMultSLod4, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.WaterReflectionFarClip, PsoDataType.Float, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SSAOLightInten, PsoDataType.Float, 140, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ExposurePush, PsoDataType.Float, 144, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LightFadeDistanceMult, PsoDataType.Float, 148, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LightShadowFadeDistanceMult, PsoDataType.Float, 152, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LightSpecularFadeDistMult, PsoDataType.Float, 156, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LightVolumetricFadeDistanceMult, PsoDataType.Float, 160, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DirectionalLightMultiplier, PsoDataType.Float, 164, 0, 0),
                     new PsoStructureEntryInfo(MetaName.LensArtefactMultiplier, PsoDataType.Float, 168, 0, 0),
                     new PsoStructureEntryInfo(MetaName.BloomMax, PsoDataType.Float, 172, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableHighQualityDof, PsoDataType.Bool, 176, 0, 0),
                     new PsoStructureEntryInfo(MetaName.FreezeReflectionMap, PsoDataType.Bool, 177, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DisableDirectionalLighting, PsoDataType.Bool, 178, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AbsoluteIntensityEnabled, PsoDataType.Bool, 179, 0, 0),
                     new PsoStructureEntryInfo(MetaName.CharacterLight, PsoDataType.Structure, 192, 0, MetaName.rage__cutfCameraCutCharacterLightParams),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.rage__cutfCameraCutTimeOfDayDofModifier),
                     new PsoStructureEntryInfo(MetaName.TimeOfDayDofModifers, PsoDataType.Array, 256, 0, (MetaName)34)
                    );
                case MetaName.rage__cutfCameraCutCharacterLightParams:
                    return new PsoStructureInfo(MetaName.rage__cutfCameraCutCharacterLightParams, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.bUseTimeCycleValues, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vDirection, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vColour, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fIntensity, PsoDataType.Float, 48, 0, 0)
                    );
                case MetaName.rage__cutfSubtitleEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfSubtitleEventArgs, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.iLanguageID, PsoDataType.SInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iTransitionIn, PsoDataType.SInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTransitionInDuration, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iTransitionOut, PsoDataType.SInt, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fTransitionOutDuration, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fSubtitleDuration, PsoDataType.Float, 60, 0, 0)
                    );
                case MetaName.rage__cutfFinalNameEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfFinalNameEventArgs, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 32, 3, 0)
                    );
                case MetaName.rage__cutfObjectIdNameEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfObjectIdNameEventArgs, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0)
                    );
                case MetaName.vHaltFrequency:
                    return new PsoStructureInfo(MetaName.vHaltFrequency, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.cSceneName, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.SInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.frames, PsoDataType.Array, 8, 0, (MetaName)1)
                    );
                case MetaName.rage__cutfVehicleModelObject:
                    return new PsoStructureInfo(MetaName.rage__cutfVehicleModelObject, 0, 0, 120,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamingName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AnimStreamingBase, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cAnimExportCtrlSpecFile, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cFaceExportCtrlSpecFile, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cAnimCompressionFile, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cHandle, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo(MetaName.typeFile, PsoDataType.String, 88, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.cRemoveBoneNameList, PsoDataType.Array, 96, 0, (MetaName)11),
                     new PsoStructureEntryInfo(MetaName.bCanApplyRealDamage, PsoDataType.Bool, 112, 0, 0)
                    );
                case MetaName.rage__cutfEvent:
                    return new PsoStructureInfo(MetaName.rage__cutfEvent, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.fTime, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iEventId, PsoDataType.SInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iEventArgsIndex, PsoDataType.SInt, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pChildEvents, PsoDataType.Structure, 32, 3, 0),
                     new PsoStructureEntryInfo(MetaName.StickyId, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsChild, PsoDataType.Bool, 44, 0, 0)
                    );
                case MetaName.rage__cutfCascadeShadowEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfCascadeShadowEventArgs, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cameraCutHashName, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo(MetaName.position, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.radius, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.interpTime, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cascadeIndex, PsoDataType.SInt, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.enabled, PsoDataType.Bool, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.interpolateToDisabled, PsoDataType.Bool, 77, 0, 0)
                    );
                case MetaName.rage__cutfFloatValueEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfFloatValueEventArgs, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.fValue, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.rage__cutfAnimatedParticleEffectObject:
                    return new PsoStructureInfo(MetaName.rage__cutfAnimatedParticleEffectObject, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamingName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AnimStreamingBase, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.athFxListHash, PsoDataType.String, 64, 7, 0)
                    );
                case MetaName.rage__cutfWeaponModelObject:
                    return new PsoStructureInfo(MetaName.rage__cutfWeaponModelObject, 0, 0, 104,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamingName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.AnimStreamingBase, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.cAnimExportCtrlSpecFile, PsoDataType.String, 64, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cFaceExportCtrlSpecFile, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cAnimCompressionFile, PsoDataType.String, 72, 7, 0),
                     new PsoStructureEntryInfo(MetaName.cHandle, PsoDataType.String, 84, 7, 0),
                     new PsoStructureEntryInfo(MetaName.typeFile, PsoDataType.String, 88, 7, 0),
                     new PsoStructureEntryInfo(MetaName.GenericWeaponType, PsoDataType.UInt, 96, 0, 0)
                    );
                case MetaName.rage__cutfPlayParticleEffectEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfPlayParticleEffectEventArgs, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.vInitialBoneRotation, PsoDataType.Float4, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vInitialBoneOffset, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iAttachParentId, PsoDataType.SInt, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iAttachBoneHash, PsoDataType.UShort, 68, 0, 0)
                    );
                case MetaName.rage__cutfBoolValueEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfBoolValueEventArgs, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.bValue, PsoDataType.Bool, 32, 0, 0)
                    );
                case MetaName.rage__cutfRayfireObject:
                    return new PsoStructureInfo(MetaName.rage__cutfRayfireObject, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamingName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.vStartPosition, PsoDataType.Float3, 64, 0, 0)
                    );
                case MetaName.rage__cutfParticleEffectObject:
                    return new PsoStructureInfo(MetaName.rage__cutfParticleEffectObject, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamingName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.athFxListHash, PsoDataType.String, 56, 7, 0)
                    );
                case MetaName.rage__cutfDecalObject:
                    return new PsoStructureInfo(MetaName.rage__cutfDecalObject, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.StreamingName, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo(MetaName.RenderId, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.rage__cutfDecalEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfDecalEventArgs, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.vPosition, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.vRotation, PsoDataType.Float4, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fWidth, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fHeight, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Colour, PsoDataType.UInt, 72, 1, 0),
                     new PsoStructureEntryInfo(MetaName.fLifeTime, PsoDataType.Float, 76, 0, 0)
                    );
                case MetaName.rage__cutfScreenFadeObject:
                    return new PsoStructureInfo(MetaName.rage__cutfScreenFadeObject, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0)
                    );
                case MetaName.rage__cutfVehicleVariationEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfVehicleVariationEventArgs, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iMainBodyColour, PsoDataType.SInt, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iSecondBodyColour, PsoDataType.SInt, 44, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iSpecularColour, PsoDataType.SInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iWheelTrimColour, PsoDataType.SInt, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2747538743, PsoDataType.SInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iLivery, PsoDataType.SInt, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iLivery2, PsoDataType.SInt, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fDirtLevel, PsoDataType.Float, 68, 0, 0)
                    );
                case (MetaName)557437386:
                    return new PsoStructureInfo((MetaName)557437386, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 8, 0),
                     new PsoStructureEntryInfo(MetaName.Value, PsoDataType.String, 16, 3, 0)
                    );
                case MetaName.rage__cutfScreenFadeEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfScreenFadeEventArgs, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.fValue, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.color, PsoDataType.UInt, 40, 1, 0)
                    );
                case MetaName.rage__cutfTriggerLightEffectEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfTriggerLightEffectEventArgs, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.iAttachParentId, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iAttachBoneHash, PsoDataType.UShort, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.AttachedParentName, PsoDataType.String, 40, 7, 0)
                    );
                case MetaName.rage__cutfVehicleExtraEventArgs:
                    return new PsoStructureInfo(MetaName.rage__cutfVehicleExtraEventArgs, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 12, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 24, 4, 0),
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.SInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.pExtraBoneIds, PsoDataType.Array, 40, 0, (MetaName)3)
                    );
                case MetaName.rage__cutfFixupModelObject:
                    return new PsoStructureInfo(MetaName.rage__cutfFixupModelObject, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.iObjectId, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.attributeList, PsoDataType.Structure, 20, 0, MetaName.rage__parAttributeList),
                     new PsoStructureEntryInfo(MetaName.cutfAttributes, PsoDataType.Structure, 32, 4, 0),
                     new PsoStructureEntryInfo(MetaName.cName, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo(MetaName.vPosition, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fRadius, PsoDataType.Float, 64, 0, 0)
                    );
                case MetaName.CPackFileMetaData:
                    return new PsoStructureInfo(MetaName.CPackFileMetaData, 0, 0, 96,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CMapDataGroup),
                     new PsoStructureEntryInfo(MetaName.MapDataGroups, PsoDataType.Array, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CHDTxdAssetBinding),
                     new PsoStructureEntryInfo(MetaName.HDTxdBindingArray, PsoDataType.Array, 16, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CImapDependency),
                     new PsoStructureEntryInfo(MetaName.imapDependencies, PsoDataType.Array, 32, 0, (MetaName)4),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CImapDependencies),
                     new PsoStructureEntryInfo(MetaName.imapDependencies_2, PsoDataType.Array, 48, 0, (MetaName)6),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CItypDependencies),
                     new PsoStructureEntryInfo(MetaName.itypDependencies_2, PsoDataType.Array, 64, 0, (MetaName)8),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CInteriorBoundsFiles),
                     new PsoStructureEntryInfo(MetaName.Interiors, PsoDataType.Array, 80, 0, (MetaName)10)
                    );
                case MetaName.CMapDataGroup:
                    return new PsoStructureInfo(MetaName.CMapDataGroup, 0, 0, 56,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Bounds, PsoDataType.Array, 8, 0, (MetaName)1),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, (MetaName)1193003611),
                     new PsoStructureEntryInfo(MetaName.Flags, PsoDataType.Flags, 24, 0, (MetaName)2097155),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.WeatherTypes, PsoDataType.Array, 32, 0, (MetaName)5),
                     new PsoStructureEntryInfo(MetaName.HoursOnOff, PsoDataType.UInt, 48, 0, 0)
                    );
                case MetaName.CHDTxdAssetBinding:
                    return new PsoStructureInfo(MetaName.CHDTxdAssetBinding, 0, 0, 132,
                     new PsoStructureEntryInfo(MetaName.assetType, PsoDataType.Enum, 0, 0, (MetaName)3387532954),
                     new PsoStructureEntryInfo(MetaName.targetAsset, PsoDataType.String, 4, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.HDTxd, PsoDataType.String, 68, 0, (MetaName)4194304)
                    );
                case MetaName.CImapDependencies:
                    return new PsoStructureInfo(MetaName.CImapDependencies, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.imapName, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.manifestFlags),
                     new PsoStructureEntryInfo(MetaName.manifestFlags, PsoDataType.Flags, 4, 0, (MetaName)2097153),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.itypDepArray, PsoDataType.Array, 8, 0, (MetaName)3)
                    );
                case MetaName.CItypDependencies:
                    return new PsoStructureInfo(MetaName.CItypDependencies, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.itypName, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Enum, 0, 0, MetaName.manifestFlags),
                     new PsoStructureEntryInfo(MetaName.manifestFlags, PsoDataType.Flags, 4, 0, (MetaName)2097153),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.itypDepArray, PsoDataType.Array, 8, 0, (MetaName)3)
                    );
                case MetaName.CInteriorBoundsFiles:
                    return new PsoStructureInfo(MetaName.CInteriorBoundsFiles, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Bounds, PsoDataType.Array, 8, 0, (MetaName)1)
                    );
                case MetaName.CMapTypes:
                    return new PsoStructureInfo(MetaName.CMapTypes, 0, 0, 96,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.extensions, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.archetypes, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 40, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.dependencies, PsoDataType.Array, 48, 0, (MetaName)5),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CTxdRelationship),
                     new PsoStructureEntryInfo(MetaName.txdRelationships, PsoDataType.Array, 64, 0, (MetaName)MetaTypeName.POINTER),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CCompositeEntityType),
                     new PsoStructureEntryInfo(MetaName.compositeEntityTypes, PsoDataType.Array, 80, 0, (MetaName)9)
                    );
                case MetaName.CTxdRelationship:
                    return new PsoStructureInfo(MetaName.CTxdRelationship, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.parent, PsoDataType.String, 8, 3, 0),
                     new PsoStructureEntryInfo(MetaName.child, PsoDataType.String, 24, 3, 0)
                    );
                case MetaName.CCompositeEntityType:
                    return new PsoStructureInfo(MetaName.CCompositeEntityType, 0, 0, 336,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 0, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.lodDist, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.UInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.specialAttribute, PsoDataType.UInt, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bbMin, PsoDataType.Float3, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bbMax, PsoDataType.Float3, 96, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bsCentre, PsoDataType.Float3, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bsRadius, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.StartModel, PsoDataType.String, 136, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.EndModel, PsoDataType.String, 200, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.StartImapFile, PsoDataType.String, 264, 3, 0),
                     new PsoStructureEntryInfo(MetaName.EndImapFile, PsoDataType.String, 280, 3, 0),
                     new PsoStructureEntryInfo(MetaName.PtFxAssetName, PsoDataType.String, 296, 3, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CCompEntityAnims),
                     new PsoStructureEntryInfo(MetaName.Animations, PsoDataType.Array, 312, 0, (MetaName)13)
                    );
                case MetaName.CCompEntityAnims:
                    return new PsoStructureInfo(MetaName.CCompEntityAnims, 0, 0, 216,
                     new PsoStructureEntryInfo(MetaName.AnimDict, PsoDataType.String, 0, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.AnimName, PsoDataType.String, 64, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.AnimatedModel, PsoDataType.String, 128, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.punchInPhase, PsoDataType.Float, 192, 0, 0),
                     new PsoStructureEntryInfo(MetaName.punchOutPhase, PsoDataType.Float, 196, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CCompEntityEffectsData),
                     new PsoStructureEntryInfo(MetaName.effectsData, PsoDataType.Array, 200, 0, (MetaName)5)
                    );
                case MetaName.CCompEntityEffectsData:
                    return new PsoStructureInfo(MetaName.CCompEntityEffectsData, 0, 0, 160,
                     new PsoStructureEntryInfo(MetaName.fxType, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fxOffsetPos, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fxOffsetRot, PsoDataType.Float4, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boneTag, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.startPhase, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.endPhase, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxIsTriggered, PsoDataType.Bool, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxTag, PsoDataType.String, 61, 0, (MetaName)4194304),
                     new PsoStructureEntryInfo(MetaName.ptFxScale, PsoDataType.Float, 128, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxProbability, PsoDataType.Float, 132, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxHasTint, PsoDataType.Bool, 136, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxTintR, PsoDataType.UByte, 137, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxTintG, PsoDataType.UByte, 138, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxTintB, PsoDataType.UByte, 139, 0, 0),
                     new PsoStructureEntryInfo(MetaName.ptFxSize, PsoDataType.Float3, 144, 0, 0)
                    );
                case MetaName.CExtensionDefParticleEffect:
                    return new PsoStructureInfo(MetaName.CExtensionDefParticleEffect, 0, 0, 96,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.offsetPosition, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.offsetRotation, PsoDataType.Float4, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.fxName, PsoDataType.String, 48, 3, 0),
                     new PsoStructureEntryInfo(MetaName.fxType, PsoDataType.SInt, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.boneTag, PsoDataType.SInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.scale, PsoDataType.Float, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.probability, PsoDataType.SInt, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.SInt, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.color, PsoDataType.UInt, 84, 1, 0)
                    );
                case MetaName.CBaseArchetypeDef:
                    return new PsoStructureInfo(MetaName.CBaseArchetypeDef, 0, 0, 176,
                     new PsoStructureEntryInfo(MetaName.lodDist, PsoDataType.Float, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.specialAttribute, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bbMin, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bbMax, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bsCentre, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bsRadius, PsoDataType.Float, 80, 0, 0),
                     new PsoStructureEntryInfo(MetaName.hdTextureDist, PsoDataType.Float, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 88, 3, 0),
                     new PsoStructureEntryInfo(MetaName.textureDictionary, PsoDataType.String, 104, 3, 0),
                     new PsoStructureEntryInfo(MetaName.clipDictionary, PsoDataType.String, 120, 3, 0),
                     new PsoStructureEntryInfo(MetaName.drawableDictionary, PsoDataType.String, 136, 3, 0),
                     new PsoStructureEntryInfo(MetaName.physicsDictionary, PsoDataType.String, 152, 7, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.extensions, PsoDataType.Array, 160, 0, (MetaName)13)
                    );
                case MetaName.CImapDependency:
                    return new PsoStructureInfo(MetaName.CImapDependency, 0, 0, 12,
                     new PsoStructureEntryInfo(MetaName.imapName, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.itypName, PsoDataType.String, 4, 7, 0),
                     new PsoStructureEntryInfo(MetaName.packFileName, PsoDataType.String, 8, 7, 0)
                    );
                case MetaName.CMapData:
                    return new PsoStructureInfo(MetaName.CMapData, 0, 0, 304,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.parent, PsoDataType.String, 12, 7, 0),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.contentFlags, PsoDataType.UInt, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.streamingExtentsMin, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.streamingExtentsMax, PsoDataType.Float3, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.entitiesExtentsMin, PsoDataType.Float3, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.entitiesExtentsMax, PsoDataType.Float3, 80, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.entities, PsoDataType.Array, 96, 0, (MetaName)8),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.rage__fwContainerLodDef),
                     new PsoStructureEntryInfo(MetaName.containerLods, PsoDataType.Array, 112, 0, (MetaName)10),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.BoxOccluder),
                     new PsoStructureEntryInfo(MetaName.boxOccluders, PsoDataType.Array, 128, 0, (MetaName)MetaTypeName.PsoPOINTER),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.OccludeModel),
                     new PsoStructureEntryInfo(MetaName.occludeModels, PsoDataType.Array, 144, 0, (MetaName)14),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.physicsDictionaries, PsoDataType.Array, 160, 0, (MetaName)MetaTypeName.STRING),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CTimeCycleModifier),
                     new PsoStructureEntryInfo(MetaName.timeCycleModifiers, PsoDataType.Array, 176, 0, (MetaName)18),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CCarGen),
                     new PsoStructureEntryInfo(MetaName.carGenerators, PsoDataType.Array, 192, 0, (MetaName)20),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.CLODLight),
                     new PsoStructureEntryInfo(MetaName.LODLights, PsoDataType.Array, 208, 0, (MetaName)22),
                     new PsoStructureEntryInfo(MetaName.block, PsoDataType.Structure, 224, 0, MetaName.CBlockDesc)
                    );
                case MetaName.rage__fwContainerLodDef:
                    return new PsoStructureInfo(MetaName.rage__fwContainerLodDef, 0, 0, 8,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 0, 7, 0),
                     new PsoStructureEntryInfo(MetaName.parentIndex, PsoDataType.UInt, 4, 0, 0)
                    );
                case MetaName.BoxOccluder:
                    return new PsoStructureInfo(MetaName.BoxOccluder, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.iCenterX, PsoDataType.SShort, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iCenterY, PsoDataType.SShort, 2, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iCenterZ, PsoDataType.SShort, 4, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iCosZ, PsoDataType.SShort, 6, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iLength, PsoDataType.SShort, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iWidth, PsoDataType.SShort, 10, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iHeight, PsoDataType.SShort, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.iSinZ, PsoDataType.SShort, 14, 0, 0)
                    );
                case MetaName.OccludeModel:
                    return new PsoStructureInfo(MetaName.OccludeModel, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.bmin, PsoDataType.Float3a, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bmax, PsoDataType.Float3a, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.dataSize, PsoDataType.UInt, 32, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.verts, PsoDataType.Array, 40, 6, (MetaName)2097155),
                     new PsoStructureEntryInfo(MetaName.numVertsInBytes, PsoDataType.UShort, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.numTris, PsoDataType.UShort, 50, 0, 0)
                    );
                case MetaName.CTimeCycleModifier:
                    return new PsoStructureInfo(MetaName.CTimeCycleModifier, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.minExtents, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.maxExtents, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.percentage, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.range, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo(MetaName.startHour, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.endHour, PsoDataType.UInt, 60, 0, 0)
                    );
                case MetaName.CCarGen:
                    return new PsoStructureInfo(MetaName.CCarGen, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.position, PsoDataType.Float3, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.orientX, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.orientY, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.perpendicularLength, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.carModel, PsoDataType.String, 44, 7, 0),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.UInt, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2431193454, PsoDataType.UInt, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)379378343, PsoDataType.UInt, 56, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bodyColorRemap1, PsoDataType.SInt, 60, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bodyColorRemap2, PsoDataType.SInt, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bodyColorRemap3, PsoDataType.SInt, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.bodyColorRemap4, PsoDataType.SInt, 72, 0, 0)
                    );
                case MetaName.CLODLight:
                    return new PsoStructureInfo(MetaName.CLODLight, 0, 0, 136,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.FloatXYZ),
                     new PsoStructureEntryInfo(MetaName.direction, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.falloff, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Float, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.falloffExponent, PsoDataType.Array, 40, 0, (MetaName)4),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.timeAndStateFlags, PsoDataType.Array, 56, 0, (MetaName)6),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.HASH, PsoDataType.Array, 72, 0, (MetaName)8),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.coneInnerAngle, PsoDataType.Array, 88, 0, (MetaName)10),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.coneOuterAngleOrCapExt, PsoDataType.Array, 104, 0, (MetaName)MetaTypeName.PsoPOINTER),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UByte, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.coronaIntensity, PsoDataType.Array, 120, 0, (MetaName)14)
                    );
                case MetaName.CBlockDesc:
                    return new PsoStructureInfo(MetaName.CBlockDesc, 0, 0, 72,
                     new PsoStructureEntryInfo(MetaName.version, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.UInt, 4, 0, 0),
                     new PsoStructureEntryInfo(MetaName.name, PsoDataType.String, 8, 3, 0),
                     new PsoStructureEntryInfo(MetaName.exportedBy, PsoDataType.String, 24, 3, 0),
                     new PsoStructureEntryInfo(MetaName.owner, PsoDataType.String, 40, 3, 0),
                     new PsoStructureEntryInfo(MetaName.time, PsoDataType.String, 56, 3, 0)
                    );
                case MetaName.CEntityDef:
                    return new PsoStructureInfo(MetaName.CEntityDef, 0, 0, 128,
                     new PsoStructureEntryInfo(MetaName.archetypeName, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.flags, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.guid, PsoDataType.UInt, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.position, PsoDataType.Float3, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.rotation, PsoDataType.Float4, 48, 0, 0),
                     new PsoStructureEntryInfo(MetaName.scaleXY, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo(MetaName.scaleZ, PsoDataType.Float, 68, 0, 0),
                     new PsoStructureEntryInfo(MetaName.parentIndex, PsoDataType.SInt, 72, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lodDist, PsoDataType.Float, 76, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lodLevel, PsoDataType.Enum, 80, 0, MetaName.rage__eLodType),
                     new PsoStructureEntryInfo(MetaName.numChildren, PsoDataType.UInt, 84, 0, 0),
                     new PsoStructureEntryInfo(MetaName.priorityLevel, PsoDataType.Enum, 88, 0, MetaName.rage__ePriorityLevel),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 3, 0),
                     new PsoStructureEntryInfo(MetaName.extensions, PsoDataType.Array, 96, 0, (MetaName)MetaTypeName.PsoPOINTER),
                     new PsoStructureEntryInfo(MetaName.ambientOcclusionMultiplier, PsoDataType.SInt, 112, 0, 0),
                     new PsoStructureEntryInfo(MetaName.artificialAmbientOcclusion, PsoDataType.SInt, 116, 0, 0),
                     new PsoStructureEntryInfo(MetaName.tintValue, PsoDataType.UInt, 120, 0, 0),
                     new PsoStructureEntryInfo(MetaName.lightGroup, PsoDataType.UInt, 124, 0, 0)
                    );
                case MetaName.CDistantLODLight:
                    return new PsoStructureInfo(MetaName.CDistantLODLight, 0, 0, 48,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.FloatXYZ),
                     new PsoStructureEntryInfo(MetaName.position, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RGBI, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo(MetaName.numStreetLights, PsoDataType.UShort, 40, 0, 0),
                     new PsoStructureEntryInfo(MetaName.category, PsoDataType.UShort, 42, 0, 0)
                    );
                case (MetaName)3730683074:
                    return new PsoStructureInfo((MetaName)3730683074, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2166096847),
                     new PsoStructureEntryInfo(MetaName.PortalInfoList, PsoDataType.Array, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)1998468471),
                     new PsoStructureEntryInfo(MetaName.PathNodeList, PsoDataType.Array, 16, 0, (MetaName)2)
                    );
                case (MetaName)2166096847:
                    return new PsoStructureInfo((MetaName)2166096847, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.InteriorProxyHash, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PortalIdx, PsoDataType.SInt, 4, 0, 0),
                     new PsoStructureEntryInfo(MetaName.RoomIdx, PsoDataType.SInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DestInteriorHash, PsoDataType.UInt, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DestRoomIdx, PsoDataType.SInt, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)4133637881),
                     new PsoStructureEntryInfo(MetaName.PortalEntityList, PsoDataType.Array, 24, 0, (MetaName)5)
                    );
                case (MetaName)4133637881:
                    return new PsoStructureInfo((MetaName)4133637881, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.LinkType, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MaxOcclusion, PsoDataType.Float, 4, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3815194629, PsoDataType.UInt, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsDoor, PsoDataType.Bool, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.IsGlass, PsoDataType.Bool, 13, 0, 0)
                    );
                case (MetaName)1998468471:
                    return new PsoStructureInfo((MetaName)1998468471, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.Key, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, (MetaName)2301425487),
                     new PsoStructureEntryInfo(MetaName.PathNodeChildList, PsoDataType.Array, 8, 0, (MetaName)1)
                    );
                case (MetaName)2301425487:
                    return new PsoStructureInfo((MetaName)2301425487, 0, 0, 8,
                     new PsoStructureEntryInfo(MetaName.PathNodeKey, PsoDataType.UInt, 0, 0, 0),
                     new PsoStructureEntryInfo(MetaName.PortalInfoIdx, PsoDataType.SInt, 4, 0, 0)
                    );
                case MetaName.camMotionBlurSettingsMetadata:
                    return new PsoStructureInfo(MetaName.camMotionBlurSettingsMetadata, 0, 0, 64,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.MovementMotionBlurMinSpeed, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovementMotionBlurMaxSpeed, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.MovementMotionBlurMaxStrength, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DamageMotionBlurMinDamage, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DamageMotionBlurMaxDamage, PsoDataType.Float, 32, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DamageMotionBlurMaxStrength, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo(MetaName.DamageMotionBlurDuration, PsoDataType.UInt, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3109021193, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2711268798, PsoDataType.Float, 48, 0, 0),
                     new PsoStructureEntryInfo((MetaName)960288180, PsoDataType.Float, 52, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3693477470, PsoDataType.UInt, 56, 0, 0)
                    );
                case MetaName.camVehicleRocketSettings:
                    return new PsoStructureInfo(MetaName.camVehicleRocketSettings, 0, 0, 16,
                     new PsoStructureEntryInfo(MetaName.ShakeRef, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.ShakeAmplitude, PsoDataType.Float, 12, 0, 0)
                    );
                case MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring:
                    return new PsoStructureInfo(MetaName.camThirdPersonCameraMetadataQuadrupedalHeightSpring, 0, 0, 24,
                     new PsoStructureEntryInfo(MetaName.ShouldApply, PsoDataType.Bool, 8, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringConstant, PsoDataType.Float, 12, 0, 0),
                     new PsoStructureEntryInfo(MetaName.SpringDampingRatio, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1065492607, PsoDataType.Float, 20, 0, 0)
                    );
                case MetaName.camInterpolatorMetadata:
                    return new PsoStructureInfo(MetaName.camInterpolatorMetadata, 0, 0, 40,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.Duration, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EaseIn, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo(MetaName.EaseOut, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo(MetaName.Start, PsoDataType.Float, 28, 0, 0),
                     new PsoStructureEntryInfo(MetaName.End, PsoDataType.Float, 32, 0, 0)
                    );
                case MetaName.camAnimSceneDirectorMetadata:
                    return new PsoStructureInfo(MetaName.camAnimSceneDirectorMetadata, 0, 0, 32,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.CanBePaused, PsoDataType.Bool, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2406434970, PsoDataType.String, 24, 7, 0)
                    );
                case MetaName.NonFlyableAreaArray:
                    return new PsoStructureInfo(MetaName.NonFlyableAreaArray, 0, 0, 24,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.Structure, 0, 0, MetaName.NonFlyableArea),
                     new PsoStructureEntryInfo(MetaName.areas, PsoDataType.Array, 8, 0, 0)
                    );
                case MetaName.NonFlyableArea:
                    return new PsoStructureInfo(MetaName.NonFlyableArea, 0, 0, 32,
                     new PsoStructureEntryInfo((MetaName)2473579981, PsoDataType.Float4, 16, 0, 0)
                    );
                case MetaName.CAssassinsDispatch__Tunables:
                    return new PsoStructureInfo(MetaName.CAssassinsDispatch__Tunables, 0, 0, 80,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2394623568, PsoDataType.Float, 16, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3368196277, PsoDataType.Float, 20, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2477967991, PsoDataType.Float, 24, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3376433583, PsoDataType.String, 28, 7, 0),
                     new PsoStructureEntryInfo((MetaName)689054964, PsoDataType.String, 32, 7, 0),
                     new PsoStructureEntryInfo((MetaName)147587300, PsoDataType.Float, 36, 0, 0),
                     new PsoStructureEntryInfo((MetaName)201820320, PsoDataType.Float, 40, 0, 0),
                     new PsoStructureEntryInfo((MetaName)4035657267, PsoDataType.Float, 44, 0, 0),
                     new PsoStructureEntryInfo((MetaName)3058832678, PsoDataType.String, 48, 7, 0),
                     new PsoStructureEntryInfo((MetaName)2649613364, PsoDataType.String, 52, 7, 0),
                     new PsoStructureEntryInfo((MetaName)4282874513, PsoDataType.Float, 56, 0, 0),
                     new PsoStructureEntryInfo((MetaName)969882415, PsoDataType.Float, 60, 0, 0),
                     new PsoStructureEntryInfo((MetaName)2143223731, PsoDataType.Float, 64, 0, 0),
                     new PsoStructureEntryInfo((MetaName)1149455399, PsoDataType.String, 68, 7, 0),
                     new PsoStructureEntryInfo((MetaName)835576115, PsoDataType.String, 72, 7, 0)
                    );
                case MetaName.CTaskNMBuoyancy__Tunables:
                    return new PsoStructureInfo(MetaName.CTaskNMBuoyancy__Tunables, 0, 0, 48,
                     new PsoStructureEntryInfo(MetaName.Name, PsoDataType.String, 8, 7, 0),
                     new PsoStructureEntryInfo(MetaName.BlendOutThreshold, PsoDataType.Structure, 16, 0, (MetaName)3749122641)
                    );
                case MetaName.fwProfanityFilter:
                    return new PsoStructureInfo(MetaName.fwProfanityFilter, 0, 0, 56,
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 2, 0),
                     new PsoStructureEntryInfo(MetaName.profaneTerms, PsoDataType.Array, 8, 0, 0),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 2, 0),
                     new PsoStructureEntryInfo(MetaName.reservedTerms, PsoDataType.Array, 24, 0, (MetaName)2),
                     new PsoStructureEntryInfo((MetaName)MetaTypeName.ARRAYINFO, PsoDataType.String, 0, 2, 0),
                     new PsoStructureEntryInfo(MetaName.reservedFSCharacters, PsoDataType.Array, 40, 0, (MetaName)4)
                    );
                default:
                    return null;
            }
        }
        public static PsoEnumInfo GetEnumInfo(MetaName name)
        {
            //to generate enuminfos
            switch (name)
            {
                case (MetaName)360458334: //vehicle mod color metallic id
                    return new PsoEnumInfo((MetaName)360458334, 1,
                     new PsoEnumEntryInfo(MetaName.none, -1),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_normal, 0),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_1, 1),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_2, 2),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_3, 3),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_4, 4),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_5, 5),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_6, 6),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_7, 7),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_8, 8),
                     new PsoEnumEntryInfo(MetaName.EVehicleModelColorMetallic_9, 9)
                    );
                case (MetaName)544262540: //vehicle mod color audio color
                    return new PsoEnumInfo((MetaName)544262540, 1,
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_black, 0),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_blue, 1),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_brown, 2),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_beige, 3),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_graphite, 4),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_green, 5),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_grey, 6),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_orange, 7),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_pink, 8),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_red, 9),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_silver, 10),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_white, 11),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_COLOUR_yellow, 12)
                    );
                case (MetaName)2065815796: //vehicle mod color audio prefix
                    return new PsoEnumInfo((MetaName)2065815796, 1,
                     new PsoEnumEntryInfo(MetaName.none, 0),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_PREFIX_bright, 1),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_PREFIX_light, 2),
                     new PsoEnumEntryInfo(MetaName.POLICE_SCANNER_PREFIX_dark, 3)
                    );
                case MetaName.eModKitType: //vehicle mod kit type
                    return new PsoEnumInfo(MetaName.eModKitType, 1,
                     new PsoEnumEntryInfo(MetaName.MKT_STANDARD, 0),
                     new PsoEnumEntryInfo(MetaName.MKT_SPORT, 1),
                     new PsoEnumEntryInfo(MetaName.MKT_SUV, 2),
                     new PsoEnumEntryInfo(MetaName.MKT_SPECIAL, 3)
                    );
                case MetaName.eVehicleModType: //vehicle mod type
                    return new PsoEnumInfo(MetaName.eVehicleModType, 1,
                     new PsoEnumEntryInfo(MetaName.VMT_SPOILER, 0),
                     new PsoEnumEntryInfo(MetaName.VMT_BUMPER_F, 1),
                     new PsoEnumEntryInfo(MetaName.VMT_BUMPER_R, 2),
                     new PsoEnumEntryInfo(MetaName.VMT_SKIRT, 3),
                     new PsoEnumEntryInfo(MetaName.VMT_EXHAUST, 4),
                     new PsoEnumEntryInfo(MetaName.VMT_CHASSIS, 5),
                     new PsoEnumEntryInfo(MetaName.VMT_GRILL, 6),
                     new PsoEnumEntryInfo(MetaName.VMT_BONNET, 7),
                     new PsoEnumEntryInfo(MetaName.VMT_WING_L, 8),
                     new PsoEnumEntryInfo(MetaName.VMT_WING_R, 9),
                     new PsoEnumEntryInfo(MetaName.VMT_ROOF, 10),
                     new PsoEnumEntryInfo(MetaName.VMT_PLTHOLDER, 11),
                     new PsoEnumEntryInfo(MetaName.VMT_PLTVANITY, 12),
                     new PsoEnumEntryInfo(MetaName.VMT_INTERIOR1, 13),
                     new PsoEnumEntryInfo(MetaName.VMT_INTERIOR2, 14),
                     new PsoEnumEntryInfo(MetaName.VMT_INTERIOR3, 15),
                     new PsoEnumEntryInfo(MetaName.VMT_INTERIOR4, 16),
                     new PsoEnumEntryInfo(MetaName.VMT_INTERIOR5, 17),
                     new PsoEnumEntryInfo(MetaName.VMT_SEATS, 18),
                     new PsoEnumEntryInfo(MetaName.VMT_STEERING, 19),
                     new PsoEnumEntryInfo(MetaName.VMT_KNOB, 20),
                     new PsoEnumEntryInfo(MetaName.VMT_PLAQUE, 21),
                     new PsoEnumEntryInfo(MetaName.VMT_ICE, 22),
                     new PsoEnumEntryInfo(MetaName.VMT_TRUNK, 23),
                     new PsoEnumEntryInfo(MetaName.VMT_HYDRO, 24),
                     new PsoEnumEntryInfo(MetaName.VMT_ENGINEBAY1, 25),
                     new PsoEnumEntryInfo(MetaName.VMT_ENGINEBAY2, 26),
                     new PsoEnumEntryInfo(MetaName.VMT_ENGINEBAY3, 27),
                     new PsoEnumEntryInfo(MetaName.VMT_CHASSIS2, 28),
                     new PsoEnumEntryInfo(MetaName.VMT_CHASSIS3, 29),
                     new PsoEnumEntryInfo(MetaName.VMT_CHASSIS4, 30),
                     new PsoEnumEntryInfo(MetaName.VMT_CHASSIS5, 31),
                     new PsoEnumEntryInfo(MetaName.VMT_DOOR_L, 32),
                     new PsoEnumEntryInfo(MetaName.VMT_DOOR_R, 33),
                     new PsoEnumEntryInfo(MetaName.VMT_LIVERY_MOD, 34),
                     new PsoEnumEntryInfo((MetaName)3409280882, 35),
                     new PsoEnumEntryInfo(MetaName.VMT_ENGINE, 36),
                     new PsoEnumEntryInfo(MetaName.VMT_BRAKES, 37),
                     new PsoEnumEntryInfo(MetaName.VMT_GEARBOX, 38),
                     new PsoEnumEntryInfo(MetaName.VMT_HORN, 39),
                     new PsoEnumEntryInfo(MetaName.VMT_SUSPENSION, 40),
                     new PsoEnumEntryInfo(MetaName.VMT_ARMOUR, 41),
                     new PsoEnumEntryInfo((MetaName)3278520444, 42),
                     new PsoEnumEntryInfo(MetaName.VMT_TURBO, 43),
                     new PsoEnumEntryInfo((MetaName)1675686396, 44),
                     new PsoEnumEntryInfo(MetaName.VMT_TYRE_SMOKE, 45),
                     new PsoEnumEntryInfo(MetaName.VMT_HYDRAULICS, 46),
                     new PsoEnumEntryInfo(MetaName.VMT_XENON_LIGHTS, 47),
                     new PsoEnumEntryInfo(MetaName.VMT_WHEELS, 48),
                     new PsoEnumEntryInfo(MetaName.VMT_WHEELS_REAR_OR_HYDRAULICS, 49)
                    );
                case (MetaName)3635907608: //vehicle mod bone
                    return new PsoEnumInfo((MetaName)3635907608, 1,
                     new PsoEnumEntryInfo(MetaName.none, -1),
                     new PsoEnumEntryInfo(MetaName.chassis, 0),
                     new PsoEnumEntryInfo(MetaName.bodyshell, 48),
                     new PsoEnumEntryInfo(MetaName.bumper_f, 49),
                     new PsoEnumEntryInfo(MetaName.bumper_r, 50),
                     new PsoEnumEntryInfo(MetaName.wing_rf, 51),
                     new PsoEnumEntryInfo(MetaName.wing_lf, 52),
                     new PsoEnumEntryInfo(MetaName.bonnet, 53),
                     new PsoEnumEntryInfo(MetaName.boot, 54),
                     new PsoEnumEntryInfo(MetaName.exhaust, 56),
                     new PsoEnumEntryInfo(MetaName.exhaust_2, 57),
                     new PsoEnumEntryInfo(MetaName.exhaust_3, 58),
                     new PsoEnumEntryInfo(MetaName.exhaust_4, 59),
                     new PsoEnumEntryInfo(MetaName.exhaust_5, 60),
                     new PsoEnumEntryInfo(MetaName.exhaust_6, 61),
                     new PsoEnumEntryInfo(MetaName.exhaust_7, 62),
                     new PsoEnumEntryInfo(MetaName.exhaust_8, 63),
                     new PsoEnumEntryInfo(MetaName.exhaust_9, 64),
                     new PsoEnumEntryInfo(MetaName.exhaust_10, 65),
                     new PsoEnumEntryInfo(MetaName.exhaust_11, 66),
                     new PsoEnumEntryInfo(MetaName.exhaust_12, 67),
                     new PsoEnumEntryInfo(MetaName.exhaust_13, 68),
                     new PsoEnumEntryInfo(MetaName.exhaust_14, 69),
                     new PsoEnumEntryInfo(MetaName.exhaust_15, 70),
                     new PsoEnumEntryInfo(MetaName.exhaust_16, 71),
                     new PsoEnumEntryInfo(MetaName.extra_1, 401),
                     new PsoEnumEntryInfo(MetaName.extra_2, 402),
                     new PsoEnumEntryInfo(MetaName.extra_3, 403),
                     new PsoEnumEntryInfo(MetaName.extra_4, 404),
                     new PsoEnumEntryInfo(MetaName.extra_5, 405),
                     new PsoEnumEntryInfo(MetaName.extra_6, 406),
                     new PsoEnumEntryInfo(MetaName.extra_7, 407),
                     new PsoEnumEntryInfo(MetaName.extra_8, 408),
                     new PsoEnumEntryInfo(MetaName.extra_9, 409),
                     new PsoEnumEntryInfo(MetaName.extra_10, 410),
                     new PsoEnumEntryInfo(MetaName.extra_11, 411),
                     new PsoEnumEntryInfo(MetaName.extra_12, 412),
                     new PsoEnumEntryInfo(MetaName.extra_13, 413),
                     new PsoEnumEntryInfo(MetaName.extra_14, 414),
                     new PsoEnumEntryInfo(MetaName.break_extra_1, 417),
                     new PsoEnumEntryInfo(MetaName.break_extra_2, 418),
                     new PsoEnumEntryInfo(MetaName.break_extra_3, 419),
                     new PsoEnumEntryInfo(MetaName.break_extra_4, 420),
                     new PsoEnumEntryInfo(MetaName.break_extra_5, 421),
                     new PsoEnumEntryInfo(MetaName.break_extra_6, 422),
                     new PsoEnumEntryInfo(MetaName.break_extra_7, 423),
                     new PsoEnumEntryInfo(MetaName.break_extra_8, 424),
                     new PsoEnumEntryInfo(MetaName.break_extra_9, 425),
                     new PsoEnumEntryInfo(MetaName.break_extra_10, 426),
                     new PsoEnumEntryInfo(MetaName.mod_col_1, 427),
                     new PsoEnumEntryInfo(MetaName.mod_col_2, 428),
                     new PsoEnumEntryInfo(MetaName.mod_col_3, 429),
                     new PsoEnumEntryInfo(MetaName.mod_col_4, 430),
                     new PsoEnumEntryInfo(MetaName.mod_col_5, 431),
                     new PsoEnumEntryInfo(MetaName.mod_col_6, 432),
                     new PsoEnumEntryInfo(MetaName.mod_col_7, 433),
                     new PsoEnumEntryInfo(MetaName.mod_col_8, 434),
                     new PsoEnumEntryInfo(MetaName.mod_col_9, 435),
                     new PsoEnumEntryInfo(MetaName.mod_col_10, 436),
                     new PsoEnumEntryInfo(MetaName.mod_col_11, 437),
                     new PsoEnumEntryInfo(MetaName.mod_col_12, 438),
                     new PsoEnumEntryInfo(MetaName.mod_col_13, 439),
                     new PsoEnumEntryInfo(MetaName.mod_col_14, 440),
                     new PsoEnumEntryInfo(MetaName.mod_col_15, 441),
                     new PsoEnumEntryInfo(MetaName.mod_col_16, 442),
                     new PsoEnumEntryInfo(MetaName.misc_a, 369),
                     new PsoEnumEntryInfo(MetaName.misc_b, 370),
                     new PsoEnumEntryInfo(MetaName.misc_c, 371),
                     new PsoEnumEntryInfo(MetaName.misc_d, 372),
                     new PsoEnumEntryInfo(MetaName.misc_e, 373),
                     new PsoEnumEntryInfo(MetaName.misc_f, 374),
                     new PsoEnumEntryInfo(MetaName.misc_g, 375),
                     new PsoEnumEntryInfo(MetaName.misc_h, 376),
                     new PsoEnumEntryInfo(MetaName.misc_i, 377),
                     new PsoEnumEntryInfo(MetaName.misc_j, 378),
                     new PsoEnumEntryInfo(MetaName.misc_k, 379),
                     new PsoEnumEntryInfo(MetaName.misc_l, 380),
                     new PsoEnumEntryInfo(MetaName.misc_m, 381),
                     new PsoEnumEntryInfo(MetaName.misc_n, 382),
                     new PsoEnumEntryInfo(MetaName.misc_o, 383),
                     new PsoEnumEntryInfo(MetaName.misc_p, 384),
                     new PsoEnumEntryInfo(MetaName.misc_q, 385),
                     new PsoEnumEntryInfo(MetaName.misc_r, 386),
                     new PsoEnumEntryInfo(MetaName.misc_s, 387),
                     new PsoEnumEntryInfo(MetaName.misc_t, 388),
                     new PsoEnumEntryInfo(MetaName.misc_u, 389),
                     new PsoEnumEntryInfo(MetaName.misc_v, 390),
                     new PsoEnumEntryInfo(MetaName.misc_w, 391),
                     new PsoEnumEntryInfo(MetaName.misc_x, 392),
                     new PsoEnumEntryInfo(MetaName.misc_y, 393),
                     new PsoEnumEntryInfo(MetaName.misc_z, 394),
                     new PsoEnumEntryInfo(MetaName.misc_1, 395),
                     new PsoEnumEntryInfo(MetaName.misc_2, 396),
                     new PsoEnumEntryInfo(MetaName.handlebars, 79),
                     new PsoEnumEntryInfo(MetaName.steeringwheel, 80),
                     new PsoEnumEntryInfo(MetaName.swingarm, 29),
                     new PsoEnumEntryInfo(MetaName.forks_u, 21),
                     new PsoEnumEntryInfo(MetaName.forks_l, 22),
                     new PsoEnumEntryInfo(MetaName.headlight_l, 91),
                     new PsoEnumEntryInfo(MetaName.headlight_r, 92),
                     new PsoEnumEntryInfo(MetaName.indicator_lr, 97),
                     new PsoEnumEntryInfo(MetaName.indicator_lf, 95),
                     new PsoEnumEntryInfo(MetaName.indicator_rr, 98),
                     new PsoEnumEntryInfo(MetaName.indicator_rf, 96),
                     new PsoEnumEntryInfo(MetaName.taillight_l, 93),
                     new PsoEnumEntryInfo(MetaName.taillight_r, 94),
                     new PsoEnumEntryInfo(MetaName.window_lf, 42),
                     new PsoEnumEntryInfo(MetaName.window_rf, 43),
                     new PsoEnumEntryInfo(MetaName.window_rr, 45),
                     new PsoEnumEntryInfo(MetaName.window_lr, 44),
                     new PsoEnumEntryInfo(MetaName.window_lm, 46),
                     new PsoEnumEntryInfo(MetaName.window_rm, 47),
                     new PsoEnumEntryInfo(MetaName.hub_lf, 30),
                     new PsoEnumEntryInfo(MetaName.hub_rf, 31),
                     new PsoEnumEntryInfo(MetaName.windscreen_r, 41),
                     new PsoEnumEntryInfo(MetaName.neon_l, 104),
                     new PsoEnumEntryInfo(MetaName.neon_r, 105),
                     new PsoEnumEntryInfo(MetaName.neon_f, 106),
                     new PsoEnumEntryInfo(MetaName.neon_b, 107),
                     new PsoEnumEntryInfo(MetaName.door_dside_f, 3),
                     new PsoEnumEntryInfo(MetaName.door_dside_r, 4),
                     new PsoEnumEntryInfo(MetaName.door_pside_f, 5),
                     new PsoEnumEntryInfo(MetaName.door_pside_r, 6),
                     new PsoEnumEntryInfo(MetaName.bobble_head, 361),
                     new PsoEnumEntryInfo(MetaName.bobble_base, 362),
                     new PsoEnumEntryInfo(MetaName.bobble_hand, 363),
                     new PsoEnumEntryInfo(MetaName.engineblock, 364),
                     new PsoEnumEntryInfo(MetaName.mod_a, 474),
                     new PsoEnumEntryInfo(MetaName.mod_b, 475),
                     new PsoEnumEntryInfo(MetaName.mod_c, 476),
                     new PsoEnumEntryInfo(MetaName.mod_d, 477),
                     new PsoEnumEntryInfo(MetaName.mod_e, 478),
                     new PsoEnumEntryInfo(MetaName.mod_f, 479),
                     new PsoEnumEntryInfo(MetaName.mod_g, 480),
                     new PsoEnumEntryInfo(MetaName.mod_h, 481),
                     new PsoEnumEntryInfo(MetaName.mod_i, 482),
                     new PsoEnumEntryInfo(MetaName.mod_j, 483),
                     new PsoEnumEntryInfo(MetaName.mod_k, 484),
                     new PsoEnumEntryInfo(MetaName.mod_l, 485),
                     new PsoEnumEntryInfo(MetaName.mod_m, 486),
                     new PsoEnumEntryInfo(MetaName.mod_n, 487),
                     new PsoEnumEntryInfo(MetaName.mod_o, 488),
                     new PsoEnumEntryInfo(MetaName.mod_p, 489),
                     new PsoEnumEntryInfo(MetaName.mod_q, 490),
                     new PsoEnumEntryInfo(MetaName.mod_r, 491),
                     new PsoEnumEntryInfo(MetaName.mod_s, 492),
                     new PsoEnumEntryInfo(MetaName.mod_t, 493),
                     new PsoEnumEntryInfo(MetaName.mod_u, 494),
                     new PsoEnumEntryInfo(MetaName.mod_v, 495),
                     new PsoEnumEntryInfo(MetaName.mod_w, 496),
                     new PsoEnumEntryInfo(MetaName.mod_x, 497),
                     new PsoEnumEntryInfo(MetaName.mod_y, 498),
                     new PsoEnumEntryInfo(MetaName.mod_z, 499),
                     new PsoEnumEntryInfo(MetaName.mod_aa, 500),
                     new PsoEnumEntryInfo(MetaName.mod_ab, 501),
                     new PsoEnumEntryInfo(MetaName.mod_ac, 502),
                     new PsoEnumEntryInfo(MetaName.mod_ad, 503),
                     new PsoEnumEntryInfo(MetaName.mod_ae, 504),
                     new PsoEnumEntryInfo(MetaName.mod_af, 505),
                     new PsoEnumEntryInfo(MetaName.mod_ag, 506),
                     new PsoEnumEntryInfo(MetaName.mod_ah, 507),
                     new PsoEnumEntryInfo(MetaName.mod_ai, 508),
                     new PsoEnumEntryInfo(MetaName.mod_aj, 509),
                     new PsoEnumEntryInfo(MetaName.mod_ak, 510),
                     new PsoEnumEntryInfo(MetaName.turret_a1, 511),
                     new PsoEnumEntryInfo(MetaName.turret_a2, 512),
                     new PsoEnumEntryInfo(MetaName.turret_a3, 513),
                     new PsoEnumEntryInfo(MetaName.turret_a4, 514),
                     new PsoEnumEntryInfo(MetaName.turret_b1, 524),
                     new PsoEnumEntryInfo(MetaName.turret_b2, 525),
                     new PsoEnumEntryInfo(MetaName.turret_b3, 526),
                     new PsoEnumEntryInfo(MetaName.turret_b4, 527),
                     new PsoEnumEntryInfo(MetaName.rblade_1mod, 560),
                     new PsoEnumEntryInfo(MetaName.rblade_1fast, 561),
                     new PsoEnumEntryInfo(MetaName.rblade_2mod, 562),
                     new PsoEnumEntryInfo(MetaName.rblade_2fast, 563),
                     new PsoEnumEntryInfo(MetaName.rblade_3mod, 564),
                     new PsoEnumEntryInfo(MetaName.rblade_3fast, 565),
                     new PsoEnumEntryInfo(MetaName.fblade_1mod, 566),
                     new PsoEnumEntryInfo(MetaName.fblade_1fast, 567),
                     new PsoEnumEntryInfo(MetaName.fblade_2mod, 568),
                     new PsoEnumEntryInfo(MetaName.fblade_2fast, 569),
                     new PsoEnumEntryInfo(MetaName.fblade_3mod, 570),
                     new PsoEnumEntryInfo(MetaName.fblade_3fast, 571),
                     new PsoEnumEntryInfo((MetaName)1086719913, 572),
                     new PsoEnumEntryInfo((MetaName)3237490897, 573),
                     new PsoEnumEntryInfo((MetaName)3375838140, 574),
                     new PsoEnumEntryInfo((MetaName)2381840182, 575),
                     new PsoEnumEntryInfo((MetaName)3607058940, 576),
                     new PsoEnumEntryInfo((MetaName)3607058940, 577),
                     new PsoEnumEntryInfo((MetaName)1208798824, 578),
                     new PsoEnumEntryInfo((MetaName)303656220, 579),
                     new PsoEnumEntryInfo((MetaName)660207018, 580),
                     new PsoEnumEntryInfo(MetaName.spike_1mod, 581),
                     new PsoEnumEntryInfo((MetaName)3045655218, 582),
                     new PsoEnumEntryInfo((MetaName)2017296145, 583),
                     new PsoEnumEntryInfo(MetaName.spike_2mod, 584),
                     new PsoEnumEntryInfo((MetaName)1122332083, 585),
                     new PsoEnumEntryInfo((MetaName)1123212214, 586),
                     new PsoEnumEntryInfo(MetaName.spike_3mod, 587),
                     new PsoEnumEntryInfo((MetaName)4011591561, 588),
                     new PsoEnumEntryInfo((MetaName)2320654166, 589),
                     new PsoEnumEntryInfo(MetaName.scoop_1mod, 590),
                     new PsoEnumEntryInfo(MetaName.scoop_2mod, 591),
                     new PsoEnumEntryInfo(MetaName.scoop_3mod, 592)
                    );
                case MetaName.eVehicleModCameraPos: //vehicle mod camera position
                    return new PsoEnumInfo(MetaName.eVehicleModCameraPos, 1,
                     new PsoEnumEntryInfo(MetaName.VMCP_DEFAULT, 0),
                     new PsoEnumEntryInfo(MetaName.VMCP_FRONT, 1),
                     new PsoEnumEntryInfo(MetaName.VMCP_FRONT_LEFT, 2),
                     new PsoEnumEntryInfo(MetaName.VMCP_FRONT_RIGHT, 3),
                     new PsoEnumEntryInfo(MetaName.VMCP_REAR, 4),
                     new PsoEnumEntryInfo(MetaName.VMCP_REAR_LEFT, 5),
                     new PsoEnumEntryInfo(MetaName.VMCP_REAR_RIGHT, 6),
                     new PsoEnumEntryInfo(MetaName.VMCP_LEFT, 7),
                     new PsoEnumEntryInfo(MetaName.VMCP_RIGHT, 8),
                     new PsoEnumEntryInfo(MetaName.VMCP_TOP, 9),
                     new PsoEnumEntryInfo(MetaName.VMCP_BOTTOM, 10)
                    );
                case (MetaName)1795685103: //credit item job line type
                    return new PsoEnumInfo((MetaName)1795685103, 1,
                     new PsoEnumEntryInfo(MetaName.JOB_BIG, 0),
                     new PsoEnumEntryInfo(MetaName.JOB_MED, 1),
                     new PsoEnumEntryInfo(MetaName.JOB_SMALL, 2),
                     new PsoEnumEntryInfo(MetaName.NAME_BIG, 3),
                     new PsoEnumEntryInfo(MetaName.NAME_MED, 4),
                     new PsoEnumEntryInfo(MetaName.NAME_SMALL, 5),
                     new PsoEnumEntryInfo(MetaName.SPACE_BIG, 6),
                     new PsoEnumEntryInfo(MetaName.SPACE_MED, 7),
                     new PsoEnumEntryInfo(MetaName.SPACE_SMALL, 8),
                     new PsoEnumEntryInfo(MetaName.SPACE_END, 9),
                     new PsoEnumEntryInfo(MetaName.SPRITE_1, 10),
                     new PsoEnumEntryInfo(MetaName.LEGALS, 11),
                     new PsoEnumEntryInfo(MetaName.AUDIO_NAME, 12),
                     new PsoEnumEntryInfo(MetaName.AUDIO_LEGALS, 13),
                     new PsoEnumEntryInfo(MetaName.JOB_AND_NAME_MED, 14)
                    );
                case MetaName.DataFileType:
                    return new PsoEnumInfo(MetaName.DataFileType, 1,
                     new PsoEnumEntryInfo(MetaName.RPF_FILE, 0),
                     new PsoEnumEntryInfo(MetaName.IDE_FILE, 1),
                     new PsoEnumEntryInfo(MetaName.DELAYED_IDE_FILE, 2),
                     new PsoEnumEntryInfo(MetaName.IPL_FILE, 3),
                     new PsoEnumEntryInfo(MetaName.PERMANENT_ITYP_FILE, 4),
                     new PsoEnumEntryInfo((MetaName)4175147664, 5),
                     new PsoEnumEntryInfo(MetaName.HANDLING_FILE, 6),
                     new PsoEnumEntryInfo(MetaName.VEHICLEEXTRAS_FILE, 7),
                     new PsoEnumEntryInfo(MetaName.CHINESE_SHUFFLING_CHECKS_FILE, 8),
                     new PsoEnumEntryInfo(MetaName.PEDSTREAM_FILE, 9),
                     new PsoEnumEntryInfo(MetaName.CARCOLS_FILE, 10),
                     new PsoEnumEntryInfo(MetaName.POPGRP_FILE, 11),
                     new PsoEnumEntryInfo((MetaName)337231870, 12),
                     new PsoEnumEntryInfo((MetaName)3783701680, 13),
                     new PsoEnumEntryInfo(MetaName.POPSCHED_FILE, 14),
                     new PsoEnumEntryInfo(MetaName.ZONEBIND_FILE, 15),
                     new PsoEnumEntryInfo(MetaName.RADIO_FILE, 16),
                     new PsoEnumEntryInfo(MetaName.EXTRAMAP_CONVERSION_OCC_FILE, 17),
                     new PsoEnumEntryInfo(MetaName.THROWNWEAPONINFO_FILE, 18),
                     new PsoEnumEntryInfo(MetaName.RMPTFX_FILE, 19),
                     new PsoEnumEntryInfo(MetaName.PED_PERSONALITY_FILE, 20),
                     new PsoEnumEntryInfo(MetaName.PED_PERCEPTION_FILE, 21),
                     new PsoEnumEntryInfo(MetaName.VEHICLE_CAMERA_OFFSETS_FILE, 22),
                     new PsoEnumEntryInfo(MetaName.FRONTEND_MENU_FILE, 23),
                     new PsoEnumEntryInfo(MetaName.LEADERBOARD_DATA_FILE, 24),
                     new PsoEnumEntryInfo(MetaName.LEADERBOARD_ICONS_FILE, 25),
                     new PsoEnumEntryInfo(MetaName.NETWORKOPTIONS_FILE, 26),
                     new PsoEnumEntryInfo(MetaName.TIMECYCLE_FILE, 27),
                     new PsoEnumEntryInfo(MetaName.TIMECYCLEMOD_FILE, 28),
                     new PsoEnumEntryInfo(MetaName.WEATHER_FILE, 29),
                     new PsoEnumEntryInfo((MetaName)2403189033, 30),
                     new PsoEnumEntryInfo((MetaName)2431127142, 31),
                     new PsoEnumEntryInfo(MetaName.PROCOBJ_FILE, 32),
                     new PsoEnumEntryInfo(MetaName.PROC_META_FILE, 33),
                     new PsoEnumEntryInfo(MetaName.VFX_SETTINGS_FILE, 34),
                     new PsoEnumEntryInfo(MetaName.SP_STATS_DISPLAY_LIST_FILE, 35),
                     new PsoEnumEntryInfo(MetaName.MP_STATS_DISPLAY_LIST_FILE, 36),
                     new PsoEnumEntryInfo(MetaName.PED_VARS_FILE, 37),
                     new PsoEnumEntryInfo(MetaName.DISABLE_FILE, 38),
                     new PsoEnumEntryInfo(MetaName.BUILDING_META_DISPLACEMENT_FILE, 39),
                     new PsoEnumEntryInfo(MetaName.HUD_TXD_FILE, 40),
                     new PsoEnumEntryInfo(MetaName.FRONTEND_DAT_FILE, 41),
                     new PsoEnumEntryInfo(MetaName.SCROLLBARS_FILE, 42),
                     new PsoEnumEntryInfo(MetaName.TIME_FILE, 43),
                     new PsoEnumEntryInfo(MetaName.BLOODFX_FILE, 44),
                     new PsoEnumEntryInfo(MetaName.ENTITYFX_FILE, 45),
                     new PsoEnumEntryInfo(MetaName.EXPLOSIONFX_FILE, 46),
                     new PsoEnumEntryInfo(MetaName.MATERIALFX_FILE, 47),
                     new PsoEnumEntryInfo(MetaName.MOTION_TASK_DATA_FILE, 48),
                     new PsoEnumEntryInfo(MetaName.DEFAULT_TASK_DATA_FILE, 49),
                     new PsoEnumEntryInfo(MetaName.MOUNT_TUNE_FILE, 50),
                     new PsoEnumEntryInfo(MetaName.PED_BOUNDS_FILE, 51),
                     new PsoEnumEntryInfo(MetaName.PED_HEALTH_FILE, 52),
                     new PsoEnumEntryInfo(MetaName.PED_COMPONENT_SETS_FILE, 53),
                     new PsoEnumEntryInfo(MetaName.PED_IK_SETTINGS_FILE, 54),
                     new PsoEnumEntryInfo(MetaName.PED_TASK_DATA_FILE, 55),
                     new PsoEnumEntryInfo(MetaName.PED_SPECIAL_ABILITIES_FILE, 56),
                     new PsoEnumEntryInfo(MetaName.WHEELFX_FILE, 57),
                     new PsoEnumEntryInfo(MetaName.WEAPONFX_FILE, 58),
                     new PsoEnumEntryInfo(MetaName.DECALS_FILE, 59),
                     new PsoEnumEntryInfo(MetaName.NAVMESH_INDEXREMAPPING_FILE, 60),
                     new PsoEnumEntryInfo(MetaName.NAVNODE_INDEXREMAPPING_FILE, 61),
                     new PsoEnumEntryInfo(MetaName.AUDIOMESH_INDEXREMAPPING_FILE, 62),
                     new PsoEnumEntryInfo(MetaName.JUNCTION_TEMPLATES_FILE, 63),
                     new PsoEnumEntryInfo(MetaName.PATH_ZONES_FILE, 64),
                     new PsoEnumEntryInfo(MetaName.DISTANT_LIGHTS_FILE, 65),
                     new PsoEnumEntryInfo(MetaName.DISTANT_LIGHTS_HD_FILE, 66),
                     new PsoEnumEntryInfo(MetaName.FLIGHTZONES_FILE, 67),
                     new PsoEnumEntryInfo(MetaName.WATER_FILE, 68),
                     new PsoEnumEntryInfo(MetaName.TRAINCONFIGS_FILE, 69),
                     new PsoEnumEntryInfo(MetaName.TRAINTRACK_FILE, 70),
                     new PsoEnumEntryInfo(MetaName.PED_METADATA_FILE, 71),
                     new PsoEnumEntryInfo(MetaName.WEAPON_METADATA_FILE, 72),
                     new PsoEnumEntryInfo(MetaName.VEHICLE_METADATA_FILE, 73),
                     new PsoEnumEntryInfo(MetaName.DISPATCH_DATA_FILE, 74),
                     new PsoEnumEntryInfo(MetaName.DEFORMABLE_OBJECTS_FILE, 75),
                     new PsoEnumEntryInfo(MetaName.TUNABLE_OBJECTS_FILE, 76),
                     new PsoEnumEntryInfo(MetaName.PED_NAV_CAPABILITES_FILE, 77),
                     new PsoEnumEntryInfo(MetaName.WEAPONINFO_FILE, 78),
                     new PsoEnumEntryInfo(MetaName.WEAPONCOMPONENTSINFO_FILE, 79),
                     new PsoEnumEntryInfo(MetaName.LOADOUTS_FILE, 80),
                     new PsoEnumEntryInfo(MetaName.FIRINGPATTERNS_FILE, 81),
                     new PsoEnumEntryInfo(MetaName.MOTIVATIONS_FILE, 82),
                     new PsoEnumEntryInfo(MetaName.SCENARIO_POINTS_FILE, 83),
                     new PsoEnumEntryInfo(MetaName.SCENARIO_POINTS_PSO_FILE, 84),
                     new PsoEnumEntryInfo(MetaName.STREAMING_FILE, 85),
                     new PsoEnumEntryInfo(MetaName.STREAMING_FILE_PLATFORM_PS3, 86),
                     new PsoEnumEntryInfo(MetaName.STREAMING_FILE_PLATFORM_XENON, 87),
                     new PsoEnumEntryInfo(MetaName.STREAMING_FILE_PLATFORM_OTHER, 88),
                     new PsoEnumEntryInfo(MetaName.PED_BRAWLING_STYLE_FILE, 89),
                     new PsoEnumEntryInfo(MetaName.AMBIENT_PED_MODEL_SET_FILE, 90),
                     new PsoEnumEntryInfo(MetaName.AMBIENT_PROP_MODEL_SET_FILE, 91),
                     new PsoEnumEntryInfo(MetaName.AMBIENT_VEHICLE_MODEL_SET_FILE, 92),
                     new PsoEnumEntryInfo(MetaName.LADDER_METADATA_FILE, 93),
                     new PsoEnumEntryInfo((MetaName)4108615940, 94),
                     new PsoEnumEntryInfo(MetaName.SLOWNESS_ZONES_FILE, 95),
                     new PsoEnumEntryInfo(MetaName.LIQUIDFX_FILE, 96),
                     new PsoEnumEntryInfo(MetaName.VFXVEHICLEINFO_FILE, 97),
                     new PsoEnumEntryInfo(MetaName.VFXPEDINFO_FILE, 98),
                     new PsoEnumEntryInfo(MetaName.DOOR_TUNING_FILE, 99),
                     new PsoEnumEntryInfo(MetaName.PTFXASSETINFO_FILE, 100),
                     new PsoEnumEntryInfo(MetaName.SCRIPTFX_FILE, 101),
                     new PsoEnumEntryInfo(MetaName.VFXREGIONINFO_FILE, 102),
                     new PsoEnumEntryInfo(MetaName.VFXINTERIORINFO_FILE, 103),
                     new PsoEnumEntryInfo(MetaName.CAMERA_METADATA_FILE, 104),
                     new PsoEnumEntryInfo(MetaName.STREET_VEHICLE_ASSOCIATION_FILE, 105),
                     new PsoEnumEntryInfo(MetaName.VFXWEAPONINFO_FILE, 106),
                     new PsoEnumEntryInfo(MetaName.EXPLOSION_INFO_FILE, 107),
                     new PsoEnumEntryInfo(MetaName.JUNCTION_TEMPLATES_PSO_FILE, 108),
                     new PsoEnumEntryInfo(MetaName.MAPZONES_FILE, 109),
                     new PsoEnumEntryInfo(MetaName.SP_STATS_UI_LIST_FILE, 110),
                     new PsoEnumEntryInfo(MetaName.MP_STATS_UI_LIST_FILE, 111),
                     new PsoEnumEntryInfo(MetaName.OBJ_COVER_TUNING_FILE, 112),
                     new PsoEnumEntryInfo(MetaName.STREAMING_REQUEST_LISTS_FILE, 113),
                     new PsoEnumEntryInfo(MetaName.PLAYER_CARD_SETUP, 114),
                     new PsoEnumEntryInfo(MetaName.WORLD_HEIGHTMAP_FILE, 115),
                     new PsoEnumEntryInfo(MetaName.WORLD_WATERHEIGHT_FILE, 116),
                     new PsoEnumEntryInfo(MetaName.PED_OVERLAY_FILE, 117),
                     new PsoEnumEntryInfo(MetaName.WEAPON_ANIMATIONS_FILE, 118),
                     new PsoEnumEntryInfo(MetaName.VEHICLE_POPULATION_FILE, 119),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_DEFINITIONS, 120),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_RESULTS, 121),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_IMPULSES, 122),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_RUMBLES, 123),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_INTERRELATIONS, 124),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_HOMINGS, 125),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_DAMAGES, 126),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_STRIKE_BONES, 127),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_BRANCHES, 128),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_STEALTH_KILLS, 129),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_VFX, 130),
                     new PsoEnumEntryInfo(MetaName.ACTION_TABLE_FACIAL_ANIM_SETS, 131),
                     new PsoEnumEntryInfo(MetaName.VEHGEN_MARKUP_FILE, 132),
                     new PsoEnumEntryInfo(MetaName.PED_COMPONENT_CLOTH_FILE, 133),
                     new PsoEnumEntryInfo(MetaName.TATTOO_SHOP_DLC_FILE, 134),
                     new PsoEnumEntryInfo(MetaName.VEHICLE_VARIATION_FILE, 135),
                     new PsoEnumEntryInfo(MetaName.CONTENT_UNLOCKING_META_FILE, 136),
                     new PsoEnumEntryInfo(MetaName.SHOP_PED_APPAREL_META_FILE, 137),
                     new PsoEnumEntryInfo(MetaName.AUDIO_SOUNDDATA, 138),
                     new PsoEnumEntryInfo(MetaName.AUDIO_CURVEDATA, 139),
                     new PsoEnumEntryInfo(MetaName.AUDIO_GAMEDATA, 140),
                     new PsoEnumEntryInfo(MetaName.AUDIO_DYNAMIXDATA, 141),
                     new PsoEnumEntryInfo(MetaName.AUDIO_SPEECHDATA, 142),
                     new PsoEnumEntryInfo(MetaName.AUDIO_SYNTHDATA, 143),
                     new PsoEnumEntryInfo(MetaName.AUDIO_WAVEPACK, 144),
                     new PsoEnumEntryInfo(MetaName.CLIP_SETS_FILE, 145),
                     new PsoEnumEntryInfo(MetaName.EXPRESSION_SETS_FILE, 146),
                     new PsoEnumEntryInfo(MetaName.FACIAL_CLIPSET_GROUPS_FILE, 147),
                     new PsoEnumEntryInfo((MetaName)1603688191, 148),
                     new PsoEnumEntryInfo(MetaName.VEHICLE_SHOP_DLC_FILE, 149),
                     new PsoEnumEntryInfo(MetaName.WEAPON_SHOP_INFO_METADATA_FILE, 150),
                     new PsoEnumEntryInfo(MetaName.SCALEFORM_PREALLOC_FILE, 151),
                     new PsoEnumEntryInfo(MetaName.CONTROLLER_LABELS_FILE, 152),
                     new PsoEnumEntryInfo(MetaName.CONTROLLER_LABELS_FILE_360, 153),
                     new PsoEnumEntryInfo(MetaName.CONTROLLER_LABELS_FILE_PS3, 154),
                     new PsoEnumEntryInfo(MetaName.CONTROLLER_LABELS_FILE_PS3_JPN, 155),
                     new PsoEnumEntryInfo(MetaName.CONTROLLER_LABELS_FILE_ORBIS, 156),
                     new PsoEnumEntryInfo(MetaName.CONTROLLER_LABELS_FILE_ORBIS_JPN, 157),
                     new PsoEnumEntryInfo(MetaName.CONTROLLER_LABELS_FILE_DURANGO, 158),
                     new PsoEnumEntryInfo((MetaName)25671749, 159),
                     new PsoEnumEntryInfo(MetaName.TEXTFILE_METAFILE, 160),
                     new PsoEnumEntryInfo(MetaName.NM_TUNING_FILE, 161),
                     new PsoEnumEntryInfo(MetaName.MOVE_NETWORK_DEFS, 162),
                     new PsoEnumEntryInfo(MetaName.WEAPONINFO_FILE_PATCH, 163),
                     new PsoEnumEntryInfo(MetaName.DLC_SCRIPT_METAFILE, 164),
                     new PsoEnumEntryInfo(MetaName.VEHICLE_LAYOUTS_FILE, 165),
                     new PsoEnumEntryInfo(MetaName.DLC_WEAPON_PICKUPS, 166),
                     new PsoEnumEntryInfo(MetaName.EXTRA_TITLE_UPDATE_DATA, 167),
                     new PsoEnumEntryInfo(MetaName.SCALEFORM_DLC_FILE, 168),
                     new PsoEnumEntryInfo(MetaName.OVERLAY_INFO_FILE, 169),
                     new PsoEnumEntryInfo(MetaName.ALTERNATE_VARIATIONS_FILE, 170),
                     new PsoEnumEntryInfo(MetaName.HORSE_REINS_FILE, 171),
                     new PsoEnumEntryInfo(MetaName.FIREFX_FILE, 172),
                     new PsoEnumEntryInfo(MetaName.INTERIOR_PROXY_ORDER_FILE, 173),
                     new PsoEnumEntryInfo(MetaName.DLC_ITYP_REQUEST, 174),
                     new PsoEnumEntryInfo(MetaName.EXTRA_FOLDER_MOUNT_DATA, 175),
                     new PsoEnumEntryInfo(MetaName.AMB_PROCEDURAL_BLOOD_FILE, 176),
                     new PsoEnumEntryInfo(MetaName.SCRIPT_BRAIN_FILE, 177),
                     new PsoEnumEntryInfo(MetaName.SCALEFORM_VALID_METHODS_FILE, 178),
                     new PsoEnumEntryInfo(MetaName.DLC_POP_GROUPS, 179),
                     new PsoEnumEntryInfo((MetaName)2145817864, 180),
                     new PsoEnumEntryInfo(MetaName.SCENARIO_INFO_FILE, 181),
                     new PsoEnumEntryInfo(MetaName.CONDITIONAL_ANIMS_FILE, 182),
                     new PsoEnumEntryInfo(MetaName.STATS_METADATA_PSO_FILE, 183),
                     new PsoEnumEntryInfo(MetaName.VFXFOGVOLUMEINFO_FILE, 184),
                     new PsoEnumEntryInfo(MetaName.RPF_FILE_PRE_INSTALL, 185),
                     new PsoEnumEntryInfo((MetaName)341493400, 186),
                     new PsoEnumEntryInfo(MetaName.LEVEL_STREAMING_FILE, 187),
                     new PsoEnumEntryInfo(MetaName.SCENARIO_POINTS_OVERRIDE_FILE, 188),
                     new PsoEnumEntryInfo((MetaName)2800880346, 189),
                     new PsoEnumEntryInfo(MetaName.DRIVER_RULES_STD_FILE, 190),
                     new PsoEnumEntryInfo(MetaName.PED_FIRST_PERSON_ASSET_DATA, 191),
                     new PsoEnumEntryInfo(MetaName.GTXD_PARENTING_DATA, 192),
                     new PsoEnumEntryInfo(MetaName.COMBAT_BEHAVIOUR_OVERRIDE_FILE, 193),
                     new PsoEnumEntryInfo(MetaName.EVENTS_OVERRIDE_FILE, 194),
                     new PsoEnumEntryInfo(MetaName.PED_DAMAGE_OVERRIDE_FILE, 195),
                     new PsoEnumEntryInfo(MetaName.PED_DAMAGE_APPEND_FILE, 196),
                     new PsoEnumEntryInfo(MetaName.BACKGROUND_SCRIPT_FILE, 197),
                     new PsoEnumEntryInfo(MetaName.PS3_SCRIPT_RPF, 198),
                     new PsoEnumEntryInfo(MetaName.X360_SCRIPT_RPF, 199),
                     new PsoEnumEntryInfo(MetaName.PED_FIRST_PERSON_ALTERNATE_DATA, 200)
                    );
                case MetaName.DataFileContents:
                    return new PsoEnumInfo(MetaName.DataFileContents, 1,
                     new PsoEnumEntryInfo(MetaName.CONTENTS_DEFAULT, 0),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_PROPS, 1),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_MAP, 2),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_LODS, 3),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_PEDS, 4),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_VEHICLES, 5),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_ANIMATION, 6),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_CUTSCENE, 7),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_DLC_MAP_DATA, 8),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_DEBUG_ONLY, 9),
                     new PsoEnumEntryInfo(MetaName.CONTENTS_MAX, 10)
                    );
                case MetaName.InstallPartition:
                    return new PsoEnumInfo(MetaName.InstallPartition, 1,
                     new PsoEnumEntryInfo(MetaName.PARTITION_NONE, -1),
                     new PsoEnumEntryInfo(MetaName.PARTITION_0, 0),
                     new PsoEnumEntryInfo(MetaName.PARTITION_1, 1),
                     new PsoEnumEntryInfo(MetaName.PARTITION_2, 2),
                     new PsoEnumEntryInfo(MetaName.PARTITION_MAX, 3)
                    );
                case MetaName.ePedRadioGenre:
                    return new PsoEnumInfo(MetaName.ePedRadioGenre, 1,
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_OFF, 0),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_MODERN_ROCK, 1),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_CLASSIC_ROCK, 2),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_POP, 3),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_MODERN_HIPHOP, 4),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_CLASSIC_HIPHOP, 5),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_PUNK, 6),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_LEFT_WING_TALK, 7),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_RIGHT_WING_TALK, 8),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_COUNTRY, 9),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_DANCE, 10),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_MEXICAN, 11),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_REGGAE, 12),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_JAZZ, 13),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_MOTOWN, 14),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_SURF, 15),
                     new PsoEnumEntryInfo(MetaName.RADIO_GENRE_UNSPECIFIED, 16)
                    );
                case MetaName.eSexinessFlags:
                    return new PsoEnumInfo(MetaName.eSexinessFlags, 1,
                     new PsoEnumEntryInfo(MetaName.SF_JEER_AT_HOT_PED, 0),
                     new PsoEnumEntryInfo((MetaName)296569367, 1),
                     new PsoEnumEntryInfo(MetaName.SF_HOT_PERSON, 2)
                    );
                case MetaName.eExternallyDrivenDOFs:
                    return new PsoEnumInfo(MetaName.eExternallyDrivenDOFs, 1,
                     new PsoEnumEntryInfo(MetaName.EMPTY, 0),
                     new PsoEnumEntryInfo(MetaName.HIGH_HEELS, 1),
                     new PsoEnumEntryInfo(MetaName.COLLAR, 2)
                    );
                case MetaName.SpecialAbilityType:   //CPedModelInfo__InitData AbilityType
                    return new PsoEnumInfo(MetaName.SpecialAbilityType, 1,
                     new PsoEnumEntryInfo(MetaName.SAT_NONE, -1),
                     new PsoEnumEntryInfo(MetaName.SAT_CAR_SLOWDOWN, 0),
                     new PsoEnumEntryInfo(MetaName.SAT_RAGE, 1),
                     new PsoEnumEntryInfo(MetaName.SAT_BULLET_TIME, 2),
                     new PsoEnumEntryInfo(MetaName.SAT_SNAPSHOT, 3),
                     new PsoEnumEntryInfo(MetaName.SAT_INSULT, 4)
                    );
                case MetaName.ThermalBehaviour:
                    return new PsoEnumInfo(MetaName.ThermalBehaviour, 1,
                     new PsoEnumEntryInfo(MetaName.TB_DEAD, 0),
                     new PsoEnumEntryInfo(MetaName.TB_COLD, 1),
                     new PsoEnumEntryInfo(MetaName.TB_WARM, 2),
                     new PsoEnumEntryInfo(MetaName.TB_HOT, 3)
                    );
                case MetaName.eSuperlodType:
                    return new PsoEnumInfo(MetaName.eSuperlodType, 1,
                     new PsoEnumEntryInfo(MetaName.SLOD_HUMAN, 0),
                     new PsoEnumEntryInfo(MetaName.SLOD_SMALL_QUADPED, 1),
                     new PsoEnumEntryInfo(MetaName.SLOD_LARGE_QUADPED, 2),
                     new PsoEnumEntryInfo(MetaName.SLOD_NULL, 3),
                     new PsoEnumEntryInfo(MetaName.SLOD_KEEP_LOWEST, 4)
                    );
                case MetaName.eScenarioPopStreamingSlot: //CPedModelInfo__InitData ScenarioPopStreamingSlot
                    return new PsoEnumInfo(MetaName.eScenarioPopStreamingSlot, 1,
                     new PsoEnumEntryInfo(MetaName.SCENARIO_POP_STREAMING_NORMAL, 0),
                     new PsoEnumEntryInfo(MetaName.SCENARIO_POP_STREAMING_SMALL, 1)
                    );
                case MetaName.DefaultSpawnPreference:  //CPedModelInfo__InitData DefaultSpawningPreference
                    return new PsoEnumInfo(MetaName.DefaultSpawnPreference, 1,
                     new PsoEnumEntryInfo(MetaName.DSP_AERIAL, 0),
                     new PsoEnumEntryInfo(MetaName.DSP_AQUATIC, 1),
                     new PsoEnumEntryInfo(MetaName.DSP_GROUND_WILDLIFE, 2),
                     new PsoEnumEntryInfo(MetaName.DSP_NORMAL, 3)
                    );
                case MetaName.eFadeCurveType:
                    return new PsoEnumInfo(MetaName.eFadeCurveType, 1,
                     new PsoEnumEntryInfo(MetaName.FCT_NONE, 0),
                     new PsoEnumEntryInfo(MetaName.FCT_LINEAR, 1),
                     new PsoEnumEntryInfo(MetaName.FCT_HALF_SIGMOID, 2),
                     new PsoEnumEntryInfo(MetaName.FCT_SIGMOID, 3)
                    );
                case MetaName.eAnimPlayerFlag: //animation clip flags
                    return new PsoEnumInfo(MetaName.eAnimPlayerFlag, 1,
                     new PsoEnumEntryInfo(MetaName.APF_USE_SECONDARY_SLOT, 0),
                     new PsoEnumEntryInfo((MetaName)892548521, 1),
                     new PsoEnumEntryInfo(MetaName.APF_UNUSED_3, 2),
                     new PsoEnumEntryInfo(MetaName.APF_UNUSED_4, 3),
                     new PsoEnumEntryInfo(MetaName.APF_ISPLAYING, 4),
                     new PsoEnumEntryInfo(MetaName.APF_ISLOOPED, 5),
                     new PsoEnumEntryInfo((MetaName)2715402776, 6),
                     new PsoEnumEntryInfo((MetaName)2559750026, 7),
                     new PsoEnumEntryInfo((MetaName)334669388, 8),
                     new PsoEnumEntryInfo(MetaName.APF_SCRIPT, 9),
                     new PsoEnumEntryInfo(MetaName.APF_UNUSED_6, 10),
                     new PsoEnumEntryInfo(MetaName.APF_UNUSED_7, 11),
                     new PsoEnumEntryInfo(MetaName.APF_UNUSED_8, 12),
                     new PsoEnumEntryInfo((MetaName)3730225937, 13),
                     new PsoEnumEntryInfo(MetaName.APF_ISBLENDAUTOREMOVE, 14),
                     new PsoEnumEntryInfo(MetaName.APF_ISFINISHAUTOREMOVE, 15),
                     new PsoEnumEntryInfo((MetaName)1984129772, 16),
                     new PsoEnumEntryInfo(MetaName.APF_ADDITIVE, 17),
                     new PsoEnumEntryInfo(MetaName.APF_FACIAL, 18),
                     new PsoEnumEntryInfo((MetaName)1964664646, 19),
                     new PsoEnumEntryInfo((MetaName)46138343, 20),
                     new PsoEnumEntryInfo(MetaName.APF_UPPERBODYONLY, 21),
                     new PsoEnumEntryInfo((MetaName)28056044, 22),
                     new PsoEnumEntryInfo((MetaName)2237335835, 23),
                     new PsoEnumEntryInfo((MetaName)1604795828, 24),
                     new PsoEnumEntryInfo((MetaName)1772376494, 25),
                     new PsoEnumEntryInfo(MetaName.APF_BLOCK_IK, 26),
                     new PsoEnumEntryInfo((MetaName)2310486390, 27),
                     new PsoEnumEntryInfo(MetaName.APF_BLOCK_HEAD_IK, 28),
                     new PsoEnumEntryInfo((MetaName)98009864, 29),
                     new PsoEnumEntryInfo(MetaName.APF_UNUSED_9, 30),
                     new PsoEnumEntryInfo((MetaName)3765996064, 31)
                    );
                case MetaName.eAnimPriority: //animation clip priority
                    return new PsoEnumInfo(MetaName.eAnimPriority, 1,
                     new PsoEnumEntryInfo(MetaName.AP_LOW, 0),
                     new PsoEnumEntryInfo(MetaName.AP_MEDIUM, 1),
                     new PsoEnumEntryInfo(MetaName.AP_HIGH, 2),
                     new PsoEnumEntryInfo(MetaName.AP_FACE_LOW, 3),
                     new PsoEnumEntryInfo(MetaName.AP_FACE_MEDIUM, 4),
                     new PsoEnumEntryInfo(MetaName.AP_FACE_HIGH, 5)
                    );
                case MetaName.eStreamingPolicy: //anim clip dictionary streaming policy
                    return new PsoEnumInfo(MetaName.eStreamingPolicy, 1,
                     new PsoEnumEntryInfo(MetaName.SP_STREAMING, 0),
                     new PsoEnumEntryInfo(MetaName.SP_SINGLEPLAYER_RESIDENT, 1),
                     new PsoEnumEntryInfo(MetaName.SP_MULTIPLAYER_RESIDENT, 2)
                    );
                case MetaName.eStreamingPriority: //anim clip dictionary streaming priority
                    return new PsoEnumInfo(MetaName.eStreamingPriority, 1,
                     new PsoEnumEntryInfo(MetaName.SP_Variation, 0),
                     new PsoEnumEntryInfo(MetaName.SP_Low, 1),
                     new PsoEnumEntryInfo(MetaName.SP_Medium, 2),
                     new PsoEnumEntryInfo(MetaName.SP_High, 3)
                    );
                case (MetaName)2476021071:
                    return new PsoEnumInfo((MetaName)2476021071, 1,
                     new PsoEnumEntryInfo(MetaName.POSTFX_IN_HOLD_OUT, 0),
                     new PsoEnumEntryInfo(MetaName.POSTFX_EASE_IN_HOLD_EASE_OUT, 1),
                     new PsoEnumEntryInfo(MetaName.POSTFX_EASE_IN, 2),
                     new PsoEnumEntryInfo(MetaName.POSTFX_EASE_OUT, 3)
                    );
                case (MetaName)3696221852:
                    return new PsoEnumInfo((MetaName)3696221852, 1,
                     new PsoEnumEntryInfo(MetaName.POSTFX_LOOP_ALL, 0),
                     new PsoEnumEntryInfo(MetaName.POSTFX_LOOP_HOLD_ONLY, 1),
                     new PsoEnumEntryInfo(MetaName.POSTFX_LOOP_NONE, 2)
                    );
                case (MetaName)2168155209:
                    return new PsoEnumInfo((MetaName)2168155209, 1,
                     new PsoEnumEntryInfo((MetaName)133697287, -1),
                     new PsoEnumEntryInfo((MetaName)3954742285, 0),
                     new PsoEnumEntryInfo((MetaName)201802975, 1),
                     new PsoEnumEntryInfo((MetaName)1916954442, 2),
                     new PsoEnumEntryInfo((MetaName)507840474, 3),
                     new PsoEnumEntryInfo((MetaName)1919284445, 4),
                     new PsoEnumEntryInfo((MetaName)4132752489, 5),
                     new PsoEnumEntryInfo((MetaName)499113587, 6)
                    );
                case (MetaName)3261591327:
                    return new PsoEnumInfo((MetaName)3261591327, 1,
                     new PsoEnumEntryInfo((MetaName)2736859683, 0),
                     new PsoEnumEntryInfo((MetaName)856837195, 1),
                     new PsoEnumEntryInfo((MetaName)1505659915, 2),
                     new PsoEnumEntryInfo((MetaName)2834501937, 3),
                     new PsoEnumEntryInfo((MetaName)373586892, 4)
                    );
                case (MetaName)516897642:
                    return new PsoEnumInfo((MetaName)516897642, 1,
                     new PsoEnumEntryInfo((MetaName)165337364, 0),
                     new PsoEnumEntryInfo((MetaName)385965901, 1),
                     new PsoEnumEntryInfo((MetaName)3779370608, 2)
                    );
                case (MetaName)2665868448:
                    return new PsoEnumInfo((MetaName)2665868448, 1,
                     new PsoEnumEntryInfo((MetaName)2470279016, 0),
                     new PsoEnumEntryInfo((MetaName)2411291289, 1),
                     new PsoEnumEntryInfo((MetaName)2103152910, 2),
                     new PsoEnumEntryInfo((MetaName)3535879124, 3),
                     new PsoEnumEntryInfo((MetaName)3456440271, 4)
                    );
                case (MetaName)2628996780:
                    return new PsoEnumInfo((MetaName)2628996780, 1,
                     new PsoEnumEntryInfo((MetaName)3708927680, 0),
                     new PsoEnumEntryInfo((MetaName)2136499167, 1),
                     new PsoEnumEntryInfo((MetaName)3568072118, 2)
                    );
                case MetaName.eAnimBoneTag:
                    return new PsoEnumInfo(MetaName.eAnimBoneTag, 1,
                     new PsoEnumEntryInfo(MetaName.BONETAG_INVALID, -1),
                     new PsoEnumEntryInfo(MetaName.BONETAG_ROOT, 0),
                     new PsoEnumEntryInfo((MetaName)1724691876, 17916),
                     new PsoEnumEntryInfo(MetaName.BONETAG_PELVIS, 11816),
                     new PsoEnumEntryInfo(MetaName.BONETAG_PELVIS1, 53251),
                     new PsoEnumEntryInfo(MetaName.BONETAG_SPINE_ROOT, 57597),
                     new PsoEnumEntryInfo(MetaName.BONETAG_SPINE0, 23553),
                     new PsoEnumEntryInfo(MetaName.BONETAG_SPINE1, 24816),
                     new PsoEnumEntryInfo(MetaName.BONETAG_SPINE2, 24817),
                     new PsoEnumEntryInfo(MetaName.BONETAG_SPINE3, 24818),
                     new PsoEnumEntryInfo(MetaName.BONETAG_NECK, 39317),
                     new PsoEnumEntryInfo(MetaName.BONETAG_NECK2, 24532),
                     new PsoEnumEntryInfo(MetaName.BONETAG_HEAD, 31086),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_CLAVICLE, 10706),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_UPPERARM, 40269),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FOREARM, 28252),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_HAND, 57005),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER0, 58866),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER01, 64016),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER02, 64017),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER1, 58867),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER11, 64096),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER12, 64097),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER2, 58868),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER21, 64112),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER22, 64113),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER3, 58869),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER31, 64064),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER32, 64065),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER4, 58870),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER41, 64080),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FINGER42, 64081),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_CLAVICLE, 64729),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_UPPERARM, 45509),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FOREARM, 61163),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_HAND, 18905),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER0, 26610),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER01, 4089),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER02, 4090),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER1, 26611),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER11, 4169),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER12, 4170),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER2, 26612),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER21, 4185),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER22, 4186),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER3, 26613),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER31, 4137),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER32, 4138),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER4, 26614),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER41, 4153),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FINGER42, 4154),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_THIGH, 58271),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_CALF, 63931),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_FOOT, 14201),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_TOE, 2108),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_TOE1, 7531),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_THIGH, 51826),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_CALF, 36864),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_FOOT, 52301),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_TOE, 20781),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_TOE1, 45631),
                     new PsoEnumEntryInfo((MetaName)1914315501, 35731),
                     new PsoEnumEntryInfo((MetaName)1648548119, 5232),
                     new PsoEnumEntryInfo((MetaName)533252846, 37119),
                     new PsoEnumEntryInfo((MetaName)2013152691, 61007),
                     new PsoEnumEntryInfo((MetaName)1274264606, 43810),
                     new PsoEnumEntryInfo((MetaName)2720329662, 23639),
                     new PsoEnumEntryInfo((MetaName)2841566254, 6442),
                     new PsoEnumEntryInfo(MetaName.BONETAG_L_PH_HAND, 60309),
                     new PsoEnumEntryInfo(MetaName.BONETAG_R_PH_HAND, 28422),
                     new PsoEnumEntryInfo((MetaName)3721004313, 41922),
                     new PsoEnumEntryInfo((MetaName)1646663439, 18212),
                     new PsoEnumEntryInfo((MetaName)2650851428, 33399),
                     new PsoEnumEntryInfo((MetaName)4289658079, 38180),
                     new PsoEnumEntryInfo((MetaName)2379061275, 39126),
                     new PsoEnumEntryInfo((MetaName)176026430, 39638),
                     new PsoEnumEntryInfo((MetaName)3870648046, 36029),
                     new PsoEnumEntryInfo((MetaName)3398180204, 6286),
                     new PsoEnumEntryInfo(MetaName.BONETAG_TAIL0, 838),
                     new PsoEnumEntryInfo(MetaName.BONETAG_TAIL1, 839),
                     new PsoEnumEntryInfo(MetaName.BONETAG_TAIL2, 840),
                     new PsoEnumEntryInfo(MetaName.BONETAG_TAIL3, 841),
                     new PsoEnumEntryInfo(MetaName.BONETAG_TAIL4, 842),
                     new PsoEnumEntryInfo(MetaName.BONETAG_TAIL5, 843),
                     new PsoEnumEntryInfo((MetaName)1109140959, 30992),
                     new PsoEnumEntryInfo((MetaName)803570034, 30993),
                     new PsoEnumEntryInfo((MetaName)631663860, 30994),
                     new PsoEnumEntryInfo((MetaName)322488345, 30995),
                     new PsoEnumEntryInfo((MetaName)5907036, 30996),
                     new PsoEnumEntryInfo((MetaName)4136373952, 30997),
                     new PsoEnumEntryInfo((MetaName)2411152202, 6986),
                     new PsoEnumEntryInfo((MetaName)2717247431, 6987),
                     new PsoEnumEntryInfo((MetaName)1883898992, 6988),
                     new PsoEnumEntryInfo((MetaName)601675397, 16351),
                     new PsoEnumEntryInfo((MetaName)62423933, 1805),
                     new PsoEnumEntryInfo((MetaName)2641889475, 39905),
                     new PsoEnumEntryInfo((MetaName)1949780854, 5956),
                     new PsoEnumEntryInfo((MetaName)3296017019, 6468),
                     new PsoEnumEntryInfo((MetaName)1556908436, 25260),
                     new PsoEnumEntryInfo((MetaName)3030868822, 27474),
                     new PsoEnumEntryInfo((MetaName)3452992357, 56604),
                     new PsoEnumEntryInfo((MetaName)3473614855, 12844),
                     new PsoEnumEntryInfo((MetaName)633057706, 57717),
                     new PsoEnumEntryInfo((MetaName)3570968304, 24806),
                     new PsoEnumEntryInfo((MetaName)2034893792, 65245),
                     new PsoEnumEntryInfo((MetaName)974447640, 35502),
                     new PsoEnumEntryInfo((MetaName)2263842006, 15570),
                     new PsoEnumEntryInfo((MetaName)1882359409, 28462),
                     new PsoEnumEntryInfo((MetaName)2421563540, 13201),
                     new PsoEnumEntryInfo((MetaName)2644806883, 48472),
                     new PsoEnumEntryInfo((MetaName)4089752637, 5930),
                     new PsoEnumEntryInfo((MetaName)1473028002, 44297),
                     new PsoEnumEntryInfo((MetaName)1823061586, 47158),
                     new PsoEnumEntryInfo((MetaName)3875905388, 19729),
                     new PsoEnumEntryInfo((MetaName)2560977519, 43885),
                     new PsoEnumEntryInfo((MetaName)4050972607, 2359),
                     new PsoEnumEntryInfo((MetaName)1028261740, 2449),
                     new PsoEnumEntryInfo((MetaName)3159424443, 19336),
                     new PsoEnumEntryInfo((MetaName)1939403075, 64744),
                     new PsoEnumEntryInfo((MetaName)1097108699, 64745),
                     new PsoEnumEntryInfo((MetaName)428522884, 64746),
                     new PsoEnumEntryInfo((MetaName)3131540134, 56194),
                     new PsoEnumEntryInfo((MetaName)1217317825, 45733),
                     new PsoEnumEntryInfo((MetaName)4198936370, 45734),
                     new PsoEnumEntryInfo((MetaName)1977853546, 45735),
                     new PsoEnumEntryInfo((MetaName)2601253047, 48389),
                     new PsoEnumEntryInfo((MetaName)2181973692, 48390),
                     new PsoEnumEntryInfo((MetaName)3140860170, 48391),
                     new PsoEnumEntryInfo((MetaName)455142432, 30929),
                     new PsoEnumEntryInfo((MetaName)3803060301, 33869),
                     new PsoEnumEntryInfo((MetaName)746815866, 4126)
                    );
                case MetaName.eVfxGroup:
                    return new PsoEnumInfo(MetaName.eVfxGroup, 1,
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_UNDEFINED, -1),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_VOID, 0),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_GENERIC, 1),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CONCRETE, 2),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CONCRETE_DUSTY, 3),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_TARMAC, 4),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_TARMAC_BRITTLE, 5),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_STONE, 6),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_BRICK, 7),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_MARBLE, 8),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PAVING, 9),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SANDSTONE, 10),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SANDSTONE_BRITTLE, 11),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SAND_LOOSE, 12),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SAND_COMPACT, 13),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SAND_WET, 14),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SAND_UNDERWATER, 15),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SAND_DEEP, 16),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SAND_WET_DEEP, 17),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_ICE, 18),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SNOW_LOOSE, 19),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_SNOW_COMPACT, 20),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_GRAVEL, 21),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_GRAVEL_DEEP, 22),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_DIRT_DRY, 23),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_MUD_SOFT, 24),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_MUD_DEEP, 25),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_MUD_UNDERWATER, 26),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CLAY, 27),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_GRASS, 28),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_GRASS_SHORT, 29),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_HAY, 30),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_BUSHES, 31),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_TREE_BARK, 32),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_LEAVES, 33),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_METAL, 34),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_WOOD, 35),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_WOOD_DUSTY, 36),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_WOOD_SPLINTER, 37),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CERAMIC, 38),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CARPET_FABRIC, 39),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CARPET_FABRIC_DUSTY, 40),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PLASTIC, 41),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PLASTIC_HOLLOW, 42),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_RUBBER, 43),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_LINOLEUM, 44),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PLASTER_BRITTLE, 45),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CARDBOARD, 46),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PAPER, 47),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_FOAM, 48),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_FEATHERS, 49),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_TVSCREEN, 50),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_GLASS, 51),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_GLASS_BULLETPROOF, 52),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CAR_METAL, 53),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CAR_PLASTIC, 54),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_CAR_GLASS, 55),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PUDDLE, 56),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_LIQUID_WATER, 57),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_LIQUID_BLOOD, 58),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_LIQUID_OIL, 59),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_LIQUID_PETROL, 60),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_LIQUID_MUD, 61),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_FRESH_MEAT, 62),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_DRIED_MEAT, 63),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PED_HEAD, 64),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PED_TORSO, 65),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PED_LIMB, 66),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PED_FOOT, 67),
                     new PsoEnumEntryInfo(MetaName.VFXGROUP_PED_CAPSULE, 68)
                    );
                case (MetaName)726948528:
                    return new PsoEnumInfo((MetaName)726948528, 1,
                     new PsoEnumEntryInfo((MetaName)2375349691, 0),
                     new PsoEnumEntryInfo((MetaName)978806254, 1),
                     new PsoEnumEntryInfo((MetaName)345174417, 2),
                     new PsoEnumEntryInfo((MetaName)2719994633, 3),
                     new PsoEnumEntryInfo((MetaName)1055203807, 4)
                    );
                case (MetaName)1599521754:
                    return new PsoEnumInfo((MetaName)1599521754, 1,
                     new PsoEnumEntryInfo((MetaName)1546538473, 0)
                    );
                case (MetaName)956281369:
                    return new PsoEnumInfo((MetaName)956281369, 1,
                     new PsoEnumEntryInfo(MetaName.PROCOBJ_ALIGN_OBJ, 0),
                     new PsoEnumEntryInfo(MetaName.PROCOBJ_USE_GRID, 1),
                     new PsoEnumEntryInfo(MetaName.PROCOBJ_USE_SEED, 2),
                     new PsoEnumEntryInfo(MetaName.PROCOBJ_IS_FLOATING, 3),
                     new PsoEnumEntryInfo(MetaName.PROCOBJ_CAST_SHADOW, 4),
                     new PsoEnumEntryInfo(MetaName.PROCOBJ_NETWORK_GAME, 5)
                    );
                case (MetaName)1247137385:
                    return new PsoEnumInfo((MetaName)1247137385, 1,
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_LOD0, 0),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_LOD1, 1),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_LOD2, 2),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_FURGRASS, 3),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_CAMERADONOTCULL, 4),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_UNDERWATER, 5),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_GROUNDSCALE1VERT, 6),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_NOGROUNDSKEW_LOD0, 7),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_NOGROUNDSKEW_LOD1, 8),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_NOGROUNDSKEW_LOD2, 9),
                     new PsoEnumEntryInfo(MetaName.PROCPLANT_NOSHADOW, 10)
                    );
                case (MetaName)2782414211:
                    return new PsoEnumInfo((MetaName)2782414211, 1,
                     new PsoEnumEntryInfo((MetaName)3117886768, 0),
                     new PsoEnumEntryInfo((MetaName)361913441, 1),
                     new PsoEnumEntryInfo((MetaName)1673907105, 2)
                    );
                case (MetaName)3404480226:
                    return new PsoEnumInfo((MetaName)3404480226, 1,
                     new PsoEnumEntryInfo((MetaName)643717225, 0),
                     new PsoEnumEntryInfo((MetaName)1755526414, 1),
                     new PsoEnumEntryInfo(MetaName.CVA_WHEEL_REAR_LEFT, 2),
                     new PsoEnumEntryInfo((MetaName)1950121897, 3),
                     new PsoEnumEntryInfo((MetaName)2065407336, 4),
                     new PsoEnumEntryInfo((MetaName)1249738073, 5),
                     new PsoEnumEntryInfo((MetaName)3037431481, 6),
                     new PsoEnumEntryInfo((MetaName)2009750623, 7),
                     new PsoEnumEntryInfo((MetaName)3659952659, 8),
                     new PsoEnumEntryInfo((MetaName)4024787971, 9),
                     new PsoEnumEntryInfo((MetaName)4081041159, 10),
                     new PsoEnumEntryInfo((MetaName)3123586287, 11),
                     new PsoEnumEntryInfo((MetaName)1678048274, 12),
                     new PsoEnumEntryInfo((MetaName)1275607205, 13),
                     new PsoEnumEntryInfo(MetaName.DEVINS_SNOWLIGHT, 14),
                     new PsoEnumEntryInfo((MetaName)2276229539, 15),
                     new PsoEnumEntryInfo((MetaName)493733281, 16),
                     new PsoEnumEntryInfo((MetaName)1037540057, 17),
                     new PsoEnumEntryInfo((MetaName)1162901510, 18),
                     new PsoEnumEntryInfo((MetaName)739467989, 19),
                     new PsoEnumEntryInfo((MetaName)677568593, 20),
                     new PsoEnumEntryInfo((MetaName)1739703625, 21),
                     new PsoEnumEntryInfo((MetaName)2932239149, 22),
                     new PsoEnumEntryInfo((MetaName)1612186711, 23),
                     new PsoEnumEntryInfo((MetaName)629489585, 24),
                     new PsoEnumEntryInfo((MetaName)3531030925, 25),
                     new PsoEnumEntryInfo((MetaName)1769219469, 26),
                     new PsoEnumEntryInfo((MetaName)1816463715, 27),
                     new PsoEnumEntryInfo((MetaName)3449512558, 28)
                    );
                case (MetaName)2136742399:
                    return new PsoEnumInfo((MetaName)2136742399, 1,
                     new PsoEnumEntryInfo(MetaName.LOOK_FORWARD_RELATIVE_TO_ATTACH, 0),
                     new PsoEnumEntryInfo((MetaName)4067031256, 1),
                     new PsoEnumEntryInfo((MetaName)3668340720, 2)
                    );
                case (MetaName)513957198:
                    return new PsoEnumInfo((MetaName)513957198, 1,
                     new PsoEnumEntryInfo(MetaName.ON_FOOT, 0),
                     new PsoEnumEntryInfo(MetaName.IN_VEHICLE, 1),
                     new PsoEnumEntryInfo(MetaName.ON_BIKE, 2),
                     new PsoEnumEntryInfo(MetaName.IN_BOAT, 3),
                     new PsoEnumEntryInfo(MetaName.IN_AIRCRAFT, 4),
                     new PsoEnumEntryInfo(MetaName.IN_SUBMARINE, 5),
                     new PsoEnumEntryInfo(MetaName.IN_HELI, 6),
                     new PsoEnumEntryInfo(MetaName.IN_TURRET, 7)
                    );
                case (MetaName)2703064232:
                    return new PsoEnumInfo((MetaName)2703064232, 1,
                     new PsoEnumEntryInfo(MetaName.THIRD_PERSON_NEAR, 0),
                     new PsoEnumEntryInfo(MetaName.THIRD_PERSON_MEDIUM, 1),
                     new PsoEnumEntryInfo(MetaName.THIRD_PERSON_FAR, 2),
                     new PsoEnumEntryInfo(MetaName.CINEMATIC, 3),
                     new PsoEnumEntryInfo(MetaName.FIRST_PERSON, 4)
                    );
                case MetaName.eCurveType:
                    return new PsoEnumInfo(MetaName.eCurveType, 1,
                     new PsoEnumEntryInfo((MetaName)3099024980, 0),
                     new PsoEnumEntryInfo((MetaName)3810146036, 1),
                     new PsoEnumEntryInfo((MetaName)1812191173, 2),
                     new PsoEnumEntryInfo((MetaName)2081939217, 3),
                     new PsoEnumEntryInfo((MetaName)2879490434, 4),
                     new PsoEnumEntryInfo((MetaName)3899422166, 5),
                     new PsoEnumEntryInfo((MetaName)4202435238, 6),
                     new PsoEnumEntryInfo((MetaName)1924959165, 7),
                     new PsoEnumEntryInfo((MetaName)2394366735, 8),
                     new PsoEnumEntryInfo((MetaName)2691765069, 9),
                     new PsoEnumEntryInfo((MetaName)3782218410, 10),
                     new PsoEnumEntryInfo((MetaName)1054471073, 11),
                     new PsoEnumEntryInfo((MetaName)3232508014, 12),
                     new PsoEnumEntryInfo((MetaName)3268408636, 13),
                     new PsoEnumEntryInfo((MetaName)156483360, 14),
                     new PsoEnumEntryInfo((MetaName)610568483, 15),
                     new PsoEnumEntryInfo((MetaName)1031574247, 16),
                     new PsoEnumEntryInfo((MetaName)1817346630, 17),
                     new PsoEnumEntryInfo((MetaName)3038768594, 18),
                     new PsoEnumEntryInfo((MetaName)3433776656, 19),
                     new PsoEnumEntryInfo((MetaName)3224106151, 20),
                     new PsoEnumEntryInfo((MetaName)3781585750, 21),
                     new PsoEnumEntryInfo((MetaName)728471947, 22),
                     new PsoEnumEntryInfo((MetaName)2868857018, 23),
                     new PsoEnumEntryInfo((MetaName)3394937778, 24),
                     new PsoEnumEntryInfo((MetaName)1572642455, 25),
                     new PsoEnumEntryInfo((MetaName)4148234847, 26)
                    );
                case (MetaName)4206293824:
                    return new PsoEnumInfo((MetaName)4206293824, 1,
                     new PsoEnumEntryInfo((MetaName)3090781844, 0),
                     new PsoEnumEntryInfo((MetaName)714439506, 1),
                     new PsoEnumEntryInfo((MetaName)3541453901, 2),
                     new PsoEnumEntryInfo((MetaName)3420181340, 3),
                     new PsoEnumEntryInfo((MetaName)3785293538, 4),
                     new PsoEnumEntryInfo((MetaName)2987466695, 5),
                     new PsoEnumEntryInfo((MetaName)2089898957, 6),
                     new PsoEnumEntryInfo((MetaName)776877896, 7),
                     new PsoEnumEntryInfo((MetaName)1192126664, 8),
                     new PsoEnumEntryInfo((MetaName)1989206149, 9),
                     new PsoEnumEntryInfo((MetaName)2132778492, 10),
                     new PsoEnumEntryInfo((MetaName)1459749185, 11),
                     new PsoEnumEntryInfo((MetaName)769806680, 12),
                     new PsoEnumEntryInfo((MetaName)344800761, 13),
                     new PsoEnumEntryInfo((MetaName)2607259232, 14)
                    );
                case (MetaName)902027429:
                    return new PsoEnumInfo((MetaName)902027429, 1,
                     new PsoEnumEntryInfo((MetaName)1166188847, 0),
                     new PsoEnumEntryInfo((MetaName)899750670, 1)
                    );
                case (MetaName)3529626013:
                    return new PsoEnumInfo((MetaName)3529626013, 1,
                     new PsoEnumEntryInfo(MetaName.CAR, 0),
                     new PsoEnumEntryInfo(MetaName.PLANE, 1),
                     new PsoEnumEntryInfo(MetaName.TRAILER, 2),
                     new PsoEnumEntryInfo(MetaName.QUADBIKE, 3),
                     new PsoEnumEntryInfo(MetaName.HELI, 4),
                     new PsoEnumEntryInfo(MetaName.AUTOGYRO, 5),
                     new PsoEnumEntryInfo(MetaName.BIKE, 6),
                     new PsoEnumEntryInfo(MetaName.BICYCLE, 7),
                     new PsoEnumEntryInfo(MetaName.BOAT, 8),
                     new PsoEnumEntryInfo(MetaName.TRAIN, 9),
                     new PsoEnumEntryInfo(MetaName.SUBMARINE, 10),
                     new PsoEnumEntryInfo(MetaName.ALL, 11)
                    );
                case (MetaName)3889593720:
                    return new PsoEnumInfo((MetaName)3889593720, 1,
                     new PsoEnumEntryInfo((MetaName)2388118614, 0),
                     new PsoEnumEntryInfo((MetaName)1309321028, 1),
                     new PsoEnumEntryInfo((MetaName)2409045263, 2),
                     new PsoEnumEntryInfo((MetaName)1405307635, 3),
                     new PsoEnumEntryInfo((MetaName)3183067451, 4)
                    );
                case (MetaName)2675594832:
                    return new PsoEnumInfo((MetaName)2675594832, 1,
                     new PsoEnumEntryInfo(MetaName.AF_CAN_DIVE, 0),
                     new PsoEnumEntryInfo((MetaName)665056449, 1),
                     new PsoEnumEntryInfo((MetaName)1838090773, 2),
                     new PsoEnumEntryInfo((MetaName)1429252732, 3),
                     new PsoEnumEntryInfo((MetaName)335156894, 4),
                     new PsoEnumEntryInfo((MetaName)1995789406, 5),
                     new PsoEnumEntryInfo((MetaName)301163182, 6),
                     new PsoEnumEntryInfo((MetaName)2727636155, 7),
                     new PsoEnumEntryInfo((MetaName)2154274359, 8),
                     new PsoEnumEntryInfo((MetaName)3233981661, 9),
                     new PsoEnumEntryInfo((MetaName)3036600251, 10)
                    );
                case (MetaName)2920442291:
                    return new PsoEnumInfo((MetaName)2920442291, 1,
                     new PsoEnumEntryInfo(MetaName.PED_DRIVES_POOR_CAR, 0),
                     new PsoEnumEntryInfo(MetaName.PED_DRIVES_AVERAGE_CAR, 1),
                     new PsoEnumEntryInfo(MetaName.PED_DRIVES_RICH_CAR, 2),
                     new PsoEnumEntryInfo(MetaName.PED_DRIVES_BIG_CAR, 3),
                     new PsoEnumEntryInfo(MetaName.PED_DRIVES_MOTORCYCLE, 4),
                     new PsoEnumEntryInfo(MetaName.PED_DRIVES_BOAT, 5)
                    );
                case MetaName.Affluence:
                    return new PsoEnumInfo(MetaName.Affluence, 1,
                     new PsoEnumEntryInfo(MetaName.AFF_POOR, 0),
                     new PsoEnumEntryInfo(MetaName.AFF_AVERAGE, 1),
                     new PsoEnumEntryInfo(MetaName.AFF_RICH, 2)
                    );
                case MetaName.TechSavvy:
                    return new PsoEnumInfo(MetaName.TechSavvy, 1,
                     new PsoEnumEntryInfo(MetaName.TS_LOW, 0),
                     new PsoEnumEntryInfo(MetaName.TS_HIGH, 1)
                    );
                case (MetaName)3079935216:
                    return new PsoEnumInfo((MetaName)3079935216, 1,
                     new PsoEnumEntryInfo(MetaName.BF_INTERVENE_ON_MELEE_ACTION, 0),
                     new PsoEnumEntryInfo(MetaName.BF_DONT_RUN_ON_MELEE_ATTACK, 1),
                     new PsoEnumEntryInfo(MetaName.BF_WATCH_ON_CAR_STOLEN, 2),
                     new PsoEnumEntryInfo(MetaName.BF_INTIMIDATE_PLAYER, 3),
                     new PsoEnumEntryInfo(MetaName.BF_GET_PISSED_WHEN_HIT_BY_CAR, 4),
                     new PsoEnumEntryInfo((MetaName)2870992946, 5),
                     new PsoEnumEntryInfo(MetaName.BF_DONT_SAY_PANIC_ON_FLEE, 6),
                     new PsoEnumEntryInfo(MetaName.BF_REACT_ON_COMBAT, 7),
                     new PsoEnumEntryInfo(MetaName.BF_PLAY_CAR_HORN, 8),
                     new PsoEnumEntryInfo(MetaName.BF_ARGUMENTATIVE, 9),
                     new PsoEnumEntryInfo(MetaName.BF_CONFRONTATIONAL, 10),
                     new PsoEnumEntryInfo(MetaName.BF_LIMIT_COMBATANTS, 11),
                     new PsoEnumEntryInfo(MetaName.BF_PURSUE_WHEN_HIT_BY_CAR, 12),
                     new PsoEnumEntryInfo(MetaName.BF_COWARDLY_FOR_SHOCKING_EVENTS, 13),
                     new PsoEnumEntryInfo(MetaName.BF_BOOST_BRAVERY_IN_GROUP, 14),
                     new PsoEnumEntryInfo(MetaName.BF_CAN_ACCELERATE_IN_CAR, 15),
                     new PsoEnumEntryInfo(MetaName.BF_CAN_GET_OUT_WHEN_HIT_BY_CAR, 16),
                     new PsoEnumEntryInfo(MetaName.BF_AGGRESSIVE_AFTER_RUNNING_PED_OVER, 17),
                     new PsoEnumEntryInfo(MetaName.BF_CAN_FLEE_WHEN_HIT_BY_CAR, 18),
                     new PsoEnumEntryInfo(MetaName.BF_ALLOW_CONFRONT_FOR_TERRITORY_REACTIONS, 19),
                     new PsoEnumEntryInfo(MetaName.BF_DONT_FORCE_FLEE_COMBAT, 20)
                    );
                case (MetaName)1639628591:
                    return new PsoEnumInfo((MetaName)1639628591, 1,
                     new PsoEnumEntryInfo(MetaName.CF_JACKING, 0),
                     new PsoEnumEntryInfo(MetaName.CF_ALLOWED_COP_PURSUIT, 1)
                    );
                case (MetaName)1559228487:
                    return new PsoEnumInfo((MetaName)1559228487, 1,
                     new PsoEnumEntryInfo(MetaName.AF_Low, 0),
                     new PsoEnumEntryInfo(MetaName.AF_EnterLeft, 1),
                     new PsoEnumEntryInfo(MetaName.AF_FaceLeft, 2),
                     new PsoEnumEntryInfo(MetaName.AF_AtEdge, 3),
                     new PsoEnumEntryInfo(MetaName.AF_ToLow, 4),
                     new PsoEnumEntryInfo(MetaName.AF_AimDirect, 5),
                     new PsoEnumEntryInfo(MetaName.AF_Center, 6),
                     new PsoEnumEntryInfo(MetaName.AF_ToPeek, 7),
                     new PsoEnumEntryInfo(MetaName.AF_Scope, 8)
                    );
                case (MetaName)966624441:
                    return new PsoEnumInfo((MetaName)966624441, 1,
                     new PsoEnumEntryInfo(MetaName.OnlyUseForReactions, 0),
                     new PsoEnumEntryInfo(MetaName.CanUseForReactions, 1),
                     new PsoEnumEntryInfo(MetaName.Urgent, 2),
                     new PsoEnumEntryInfo(MetaName.OnlyUseForLawEnforcementPeds, 3),
                     new PsoEnumEntryInfo(MetaName.OnlyUseForGangPeds, 4)
                    );
                case MetaName.LookIkTurnRate:
                    return new PsoEnumInfo(MetaName.LookIkTurnRate, 1,
                     new PsoEnumEntryInfo(MetaName.LOOKIK_TURN_RATE_SLOW, 0),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_TURN_RATE_NORMAL, 1),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_TURN_RATE_FAST, 2)
                    );
                case MetaName.LookIkBlendRate:
                    return new PsoEnumInfo(MetaName.LookIkBlendRate, 1,
                     new PsoEnumEntryInfo(MetaName.LOOKIK_BLEND_RATE_SLOWEST, 0),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_BLEND_RATE_SLOW, 1),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_BLEND_RATE_NORMAL, 2),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_BLEND_RATE_FAST, 3),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_BLEND_RATE_FASTEST, 4),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_BLEND_RATE_INSTANT, 5)
                    );
                case MetaName.LookIkRotationLimit:
                    return new PsoEnumInfo(MetaName.LookIkRotationLimit, 1,
                     new PsoEnumEntryInfo(MetaName.LOOKIK_ROT_LIM_OFF, 0),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_ROT_LIM_NARROWEST, 1),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_ROT_LIM_NARROW, 2),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_ROT_LIM_WIDE, 3),
                     new PsoEnumEntryInfo(MetaName.LOOKIK_ROT_LIM_WIDEST, 4)
                    );
                case (MetaName)4002001992:
                    return new PsoEnumInfo((MetaName)4002001992, 1,
                     new PsoEnumEntryInfo((MetaName)4238595643, 0)
                    );
                case MetaName.AmbientEventType:
                    return new PsoEnumInfo(MetaName.AmbientEventType, 1,
                     new PsoEnumEntryInfo((MetaName)1637555008, 0),
                     new PsoEnumEntryInfo(MetaName.AET_Interesting, 1),
                     new PsoEnumEntryInfo(MetaName.AET_Threatening, 2),
                     new PsoEnumEntryInfo(MetaName.AET_Threatened, 3),
                     new PsoEnumEntryInfo(MetaName.AET_In_Place, 4),
                     new PsoEnumEntryInfo(MetaName.AET_Directed_In_Place, 5),
                     new PsoEnumEntryInfo(MetaName.AET_Flinch, 6)
                    );
                case (MetaName)1049598006:
                    return new PsoEnumInfo((MetaName)1049598006, 1,
                     new PsoEnumEntryInfo((MetaName)1600411487, 0),
                     new PsoEnumEntryInfo((MetaName)3999177563, 1),
                     new PsoEnumEntryInfo((MetaName)3250863043, 2)
                    );
                case MetaName.ePedVarComp:
                    return new PsoEnumInfo(MetaName.ePedVarComp, 1,
                     new PsoEnumEntryInfo(MetaName.PV_COMP_INVALID, -1),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_HEAD, 0),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_BERD, 1),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_HAIR, 2),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_UPPR, 3),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_LOWR, 4),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_HAND, 5),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_FEET, 6),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_TEEF, 7),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_ACCS, 8),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_TASK, 9),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_DECL, 10),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_JBIB, 11),
                     new PsoEnumEntryInfo(MetaName.PV_COMP_MAX, 12)
                    );
                case (MetaName)3509057257:
                    return new PsoEnumInfo((MetaName)3509057257, 1,
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_BUTTOCKS, 0),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_THIGH_LEFT, 1),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_SHIN_LEFT, 2),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_FOOT_LEFT, 3),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_THIGH_RIGHT, 4),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_SHIN_RIGHT, 5),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_FOOT_RIGHT, 6),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_SPINE0, 7),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_SPINE1, 8),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_SPINE2, 9),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_SPINE3, 10),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_CLAVICLE_LEFT, 11),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_UPPER_ARM_LEFT, 12),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_LOWER_ARM_LEFT, 13),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_HAND_LEFT, 14),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_CLAVICLE_RIGHT, 15),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_UPPER_ARM_RIGHT, 16),
                     new PsoEnumEntryInfo((MetaName)937951079, 17),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_HAND_RIGHT, 18),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_NECK, 19),
                     new PsoEnumEntryInfo(MetaName.RAGDOLL_HEAD, 20)
                    );
                case MetaName.LockType:
                    return new PsoEnumInfo(MetaName.LockType, 1,
                     new PsoEnumEntryInfo((MetaName)40312212, 0),
                     new PsoEnumEntryInfo((MetaName)2263813480, 1),
                     new PsoEnumEntryInfo((MetaName)1924092278, 2)
                    );
                case (MetaName)3799765345:
                    return new PsoEnumInfo((MetaName)3799765345, 1,
                     new PsoEnumEntryInfo(MetaName.ET_INVALID, -1),
                     new PsoEnumEntryInfo(MetaName.ET_CRIME, 0),
                     new PsoEnumEntryInfo(MetaName.ET_JAYWALKING, 1),
                     new PsoEnumEntryInfo(MetaName.ET_COP_PURSUIT, 2),
                     new PsoEnumEntryInfo(MetaName.ET_SPAWNED_COP_PURSUIT, 3),
                     new PsoEnumEntryInfo(MetaName.ET_AMBIENT_COP, 4),
                     new PsoEnumEntryInfo(MetaName.ET_INTERESTING_DRIVER, 5),
                     new PsoEnumEntryInfo(MetaName.ET_AGGRESSIVE_DRIVER, 6)
                    );
                case (MetaName)1809724872:
                    return new PsoEnumInfo((MetaName)1809724872, 1,
                     new PsoEnumEntryInfo(MetaName.NO_REACTION, 0),
                     new PsoEnumEntryInfo(MetaName.SMALL_REACTION, 1),
                     new PsoEnumEntryInfo(MetaName.BIG_REACTION, 2)
                    );
                case (MetaName)888350649:
                    return new PsoEnumInfo((MetaName)888350649, 1,
                     new PsoEnumEntryInfo(MetaName.SAC_Retain, 0),
                     new PsoEnumEntryInfo((MetaName)485311651, 1),
                     new PsoEnumEntryInfo((MetaName)1413720671, 2),
                     new PsoEnumEntryInfo(MetaName.SAC_Randomize, 3)
                    );
                case (MetaName)2055586340:
                    return new PsoEnumInfo((MetaName)2055586340, 1,
                     new PsoEnumEntryInfo(MetaName.TC_Retain, 0),
                     new PsoEnumEntryInfo((MetaName)1410481380, 1),
                     new PsoEnumEntryInfo((MetaName)1541573135, 2),
                     new PsoEnumEntryInfo(MetaName.TC_Randomize, 3)
                    );
                case (MetaName)3374036042:
                    return new PsoEnumInfo((MetaName)3374036042, 1,
                     new PsoEnumEntryInfo(MetaName.BC_Retain, 0),
                     new PsoEnumEntryInfo((MetaName)1834753809, 1),
                     new PsoEnumEntryInfo((MetaName)2650352944, 2),
                     new PsoEnumEntryInfo(MetaName.BC_Randomize, 3)
                    );
                case (MetaName)4001255574:
                    return new PsoEnumInfo((MetaName)4001255574, 1,
                     new PsoEnumEntryInfo(MetaName.HBC_Retain, 0),
                     new PsoEnumEntryInfo((MetaName)2508718569, 1),
                     new PsoEnumEntryInfo((MetaName)1024930189, 2),
                     new PsoEnumEntryInfo(MetaName.HBC_Randomize, 3)
                    );
                case (MetaName)894416699:
                    return new PsoEnumInfo((MetaName)894416699, 1,
                     new PsoEnumEntryInfo(MetaName.SF_FrontDriverSide, 0),
                     new PsoEnumEntryInfo(MetaName.SF_FrontPassengerSide, 1),
                     new PsoEnumEntryInfo(MetaName.SF_BackDriverSide, 2),
                     new PsoEnumEntryInfo(MetaName.SF_BackPassengerSide, 3),
                     new PsoEnumEntryInfo(MetaName.SF_AltFrontDriverSide, 4),
                     new PsoEnumEntryInfo(MetaName.SF_AltFrontPassengerSide, 5),
                     new PsoEnumEntryInfo(MetaName.SF_AltBackDriverSide, 6),
                     new PsoEnumEntryInfo(MetaName.SF_AltBackPassengerSide, 7)
                    );
                case (MetaName)4186696686:
                    return new PsoEnumInfo((MetaName)4186696686, 1,
                     new PsoEnumEntryInfo(MetaName.POPGROUP_IS_GANG, 0),
                     new PsoEnumEntryInfo(MetaName.POPGROUP_AMBIENT, 1),
                     new PsoEnumEntryInfo(MetaName.POPGROUP_SCENARIO, 2),
                     new PsoEnumEntryInfo(MetaName.POPGROUP_RARE, 3),
                     new PsoEnumEntryInfo(MetaName.POPGROUP_NETWORK_COMMON, 4),
                     new PsoEnumEntryInfo(MetaName.POPGROUP_AERIAL, 5),
                     new PsoEnumEntryInfo(MetaName.POPGROUP_AQUATIC, 6),
                     new PsoEnumEntryInfo(MetaName.POPGROUP_WILDLIFE, 7),
                     new PsoEnumEntryInfo(MetaName.POPGROUP_IN_VEHICLE, 8)
                    );
                case (MetaName)2809332931:
                    return new PsoEnumInfo((MetaName)2809332931, 1,
                     new PsoEnumEntryInfo(MetaName.DontCloseWhenTouched, 0),
                     new PsoEnumEntryInfo(MetaName.AutoOpensForSPVehicleWithPedsOnly, 1),
                     new PsoEnumEntryInfo(MetaName.AutoOpensForSPPlayerPedsOnly, 2),
                     new PsoEnumEntryInfo(MetaName.AutoOpensForMPVehicleWithPedsOnly, 3),
                     new PsoEnumEntryInfo(MetaName.AutoOpensForMPPlayerPedsOnly, 4),
                     new PsoEnumEntryInfo(MetaName.DelayDoorClosingForPlayer, 5),
                     new PsoEnumEntryInfo(MetaName.AutoOpensForAllVehicles, 6),
                     new PsoEnumEntryInfo(MetaName.IgnoreOpenDoorTaskEdgeLerp, 7),
                     new PsoEnumEntryInfo(MetaName.AutoOpensForLawEnforcement, 8)
                    );
                case MetaName.StdDoorRotDir:
                    return new PsoEnumInfo(MetaName.StdDoorRotDir, 1,
                     new PsoEnumEntryInfo(MetaName.StdDoorOpenBothDir, 0),
                     new PsoEnumEntryInfo(MetaName.StdDoorOpenNegDir, 1),
                     new PsoEnumEntryInfo(MetaName.StdDoorOpenPosDir, 2)
                    );
                case (MetaName)2065671281:
                    return new PsoEnumInfo((MetaName)2065671281, 1,
                     new PsoEnumEntryInfo(MetaName.CantUse, 0),
                     new PsoEnumEntryInfo(MetaName.MustUse, 1)
                    );
                case (MetaName)1756502932:
                    return new PsoEnumInfo((MetaName)1756502932, 1,
                     new PsoEnumEntryInfo((MetaName)1618156912, 0),
                     new PsoEnumEntryInfo((MetaName)325739119, 1),
                     new PsoEnumEntryInfo((MetaName)1129710621, 2),
                     new PsoEnumEntryInfo((MetaName)2784202857, 3),
                     new PsoEnumEntryInfo((MetaName)3198877297, 4),
                     new PsoEnumEntryInfo((MetaName)2664559559, 5)
                    );
                case (MetaName)3204395397:
                    return new PsoEnumInfo((MetaName)3204395397, 1,
                     new PsoEnumEntryInfo((MetaName)3124669982, 0),
                     new PsoEnumEntryInfo((MetaName)1679254647, 1),
                     new PsoEnumEntryInfo((MetaName)1875723229, 2)
                    );
                case (MetaName)4286852891:
                    return new PsoEnumInfo((MetaName)4286852891, 1,
                     new PsoEnumEntryInfo(MetaName.VEHICLE_RESPONSE_DEFAULT, 0),
                     new PsoEnumEntryInfo(MetaName.VEHICLE_RESPONSE_COUNTRYSIDE, 1),
                     new PsoEnumEntryInfo(MetaName.VEHICLE_RESPONSE_ARMY_BASE, 2),
                     new PsoEnumEntryInfo((MetaName)2910242310, 3)
                    );
                case (MetaName)4095090001:
                    return new PsoEnumInfo((MetaName)4095090001, 1,
                     new PsoEnumEntryInfo(MetaName.SPECIAL_NONE, 0),
                     new PsoEnumEntryInfo(MetaName.SPECIAL_AIRPORT, 1)
                    );
                case (MetaName)98189892:
                    return new PsoEnumInfo((MetaName)98189892, 1,
                     new PsoEnumEntryInfo(MetaName.NoCoverNorthFaceEast, 0),
                     new PsoEnumEntryInfo(MetaName.NoCoverNorthFaceWest, 1),
                     new PsoEnumEntryInfo(MetaName.NoCoverNorthFaceCenter, 2),
                     new PsoEnumEntryInfo(MetaName.NoCoverSouthFaceEast, 3),
                     new PsoEnumEntryInfo(MetaName.NoCoverSouthFaceWest, 4),
                     new PsoEnumEntryInfo(MetaName.NoCoverSouthFaceCenter, 5),
                     new PsoEnumEntryInfo(MetaName.NoCoverEastFaceNorth, 6),
                     new PsoEnumEntryInfo(MetaName.NoCoverEastFaceSouth, 7),
                     new PsoEnumEntryInfo(MetaName.NoCoverEastFaceCenter, 8),
                     new PsoEnumEntryInfo(MetaName.NoCoverWestFaceNorth, 9),
                     new PsoEnumEntryInfo(MetaName.NoCoverWestFaceSouth, 10),
                     new PsoEnumEntryInfo(MetaName.NoCoverWestFaceCenter, 11),
                     new PsoEnumEntryInfo(MetaName.ForceLowCornerNorthFaceEast, 12),
                     new PsoEnumEntryInfo(MetaName.ForceLowCornerNorthFaceWest, 13),
                     new PsoEnumEntryInfo(MetaName.ForceLowCornerSouthFaceEast, 14),
                     new PsoEnumEntryInfo(MetaName.ForceLowCornerSouthFaceWest, 15),
                     new PsoEnumEntryInfo(MetaName.ForceLowCornerEastFaceNorth, 16),
                     new PsoEnumEntryInfo(MetaName.ForceLowCornerEastFaceSouth, 17),
                     new PsoEnumEntryInfo(MetaName.ForceLowCornerWestFaceNorth, 18),
                     new PsoEnumEntryInfo(MetaName.ForceLowCornerWestFaceSouth, 19),
                     new PsoEnumEntryInfo(MetaName.NoCoverVehicleDoors, 20)
                    );
                case (MetaName)3573596290:
                    return new PsoEnumInfo((MetaName)3573596290, 1,
                     new PsoEnumEntryInfo(MetaName.kBoth, 0),
                     new PsoEnumEntryInfo(MetaName.kOnlySp, 1),
                     new PsoEnumEntryInfo(MetaName.kOnlyMp, 2)
                    );
                case MetaName.CScenarioPointFlags__Flags:
                    return new PsoEnumInfo(MetaName.CScenarioPointFlags__Flags, 1,
                     new PsoEnumEntryInfo(MetaName.IgnoreMaxInRange, 0),
                     new PsoEnumEntryInfo(MetaName.NoSpawn, 1),
                     new PsoEnumEntryInfo(MetaName.StationaryReactions, 2)
                    );
                case MetaName.CScenarioChainingEdge__eNavMode:
                    return new PsoEnumInfo(MetaName.CScenarioChainingEdge__eNavMode, 1,
                     new PsoEnumEntryInfo(MetaName.Direct, 0),
                     new PsoEnumEntryInfo(MetaName.NavMesh, 1),
                     new PsoEnumEntryInfo(MetaName.Roads, 2)
                    );
                case MetaName.CScenarioChainingEdge__eNavSpeed:
                    return new PsoEnumInfo(MetaName.CScenarioChainingEdge__eNavSpeed, 1,
                     new PsoEnumEntryInfo((MetaName)3279574318, 0),
                     new PsoEnumEntryInfo((MetaName)2212923970, 1),
                     new PsoEnumEntryInfo((MetaName)4022799658, 2),
                     new PsoEnumEntryInfo((MetaName)1425672334, 3),
                     new PsoEnumEntryInfo((MetaName)957720931, 4),
                     new PsoEnumEntryInfo((MetaName)3795195414, 5),
                     new PsoEnumEntryInfo((MetaName)2834622009, 6),
                     new PsoEnumEntryInfo((MetaName)1876554076, 7),
                     new PsoEnumEntryInfo((MetaName)698543797, 8),
                     new PsoEnumEntryInfo((MetaName)3911005380, 9),
                     new PsoEnumEntryInfo((MetaName)3717649022, 10),
                     new PsoEnumEntryInfo((MetaName)3356026130, 11)
                    );
                case (MetaName)1193003611:
                    return new PsoEnumInfo((MetaName)1193003611, 1,
                     new PsoEnumEntryInfo(MetaName.TIME_DEPENDENT, 0),
                     new PsoEnumEntryInfo(MetaName.WEATHER_DEPENDENT, 1)
                    );
                case (MetaName)3387532954:
                    return new PsoEnumInfo((MetaName)3387532954, 1,
                     new PsoEnumEntryInfo(MetaName.AT_TXD, 0),
                     new PsoEnumEntryInfo(MetaName.AT_DRB, 1),
                     new PsoEnumEntryInfo(MetaName.AT_DWD, 2),
                     new PsoEnumEntryInfo(MetaName.AT_FRG, 3)
                    );
                case MetaName.manifestFlags:
                    return new PsoEnumInfo(MetaName.manifestFlags, 1,
                     new PsoEnumEntryInfo(MetaName.INTERIOR_DATA, 0)
                    );
                case MetaName.rage__eLodType:
                    return new PsoEnumInfo(MetaName.rage__eLodType, 1,
                     new PsoEnumEntryInfo(MetaName.LODTYPES_DEPTH_HD, 0),
                     new PsoEnumEntryInfo(MetaName.LODTYPES_DEPTH_LOD, 1),
                     new PsoEnumEntryInfo(MetaName.LODTYPES_DEPTH_SLOD1, 2),
                     new PsoEnumEntryInfo(MetaName.LODTYPES_DEPTH_SLOD2, 3),
                     new PsoEnumEntryInfo(MetaName.LODTYPES_DEPTH_SLOD3, 4),
                     new PsoEnumEntryInfo(MetaName.LODTYPES_DEPTH_ORPHANHD, 5),
                     new PsoEnumEntryInfo(MetaName.LODTYPES_DEPTH_SLOD4, 6)
                    );
                case MetaName.rage__ePriorityLevel:
                    return new PsoEnumInfo(MetaName.rage__ePriorityLevel, 1,
                     new PsoEnumEntryInfo(MetaName.PRI_REQUIRED, 0),
                     new PsoEnumEntryInfo(MetaName.PRI_OPTIONAL_HIGH, 1),
                     new PsoEnumEntryInfo(MetaName.PRI_OPTIONAL_MEDIUM, 2),
                     new PsoEnumEntryInfo(MetaName.PRI_OPTIONAL_LOW, 3)
                    );
                default:
                    return null;
            }
        }




        private static string GetSafeName(MetaName namehash, uint key)
        {
            string name = namehash.ToString();
            if (string.IsNullOrEmpty(name))
            {
                name = "Unk_" + key;
            }
            if (!char.IsLetter(name[0]))
            {
                name = "Unk_" + name;
            }
            return name;
        }



        public static T ConvertDataRaw<T>(byte[] data) where T : struct
        {
            GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);
            var h = handle.AddrOfPinnedObject();
            var r = Marshal.PtrToStructure<T>(h);
            handle.Free();
            return r;
        }
        public static T ConvertDataRaw<T>(byte[] data, int offset) where T : struct
        {
            GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);
            var h = handle.AddrOfPinnedObject();
            var r = Marshal.PtrToStructure<T>(h + offset);
            handle.Free();
            return r;
        }
        public static T ConvertData<T>(byte[] data, int offset) where T : struct, IPsoSwapEnd
        {
            GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);
            var h = handle.AddrOfPinnedObject();
            var r = Marshal.PtrToStructure<T>(h + offset);
            handle.Free();
            r.SwapEnd();
            return r;
        }
        public static T[] ConvertDataArrayRaw<T>(byte[] data, int offset, int count) where T : struct
        {
            T[] items = new T[count];
            int itemsize = Marshal.SizeOf(typeof(T));
            for (int i = 0; i < count; i++)
            {
                int off = offset + i * itemsize;
                items[i] = ConvertDataRaw<T>(data, off);
            }
            return items;
        }


        public static T GetItem<T>(PsoFile pso, int offset) where T : struct, IPsoSwapEnd
        {
            return ConvertData<T>(pso.DataSection.Data, offset);
        }
        public static T GetRootItem<T>(PsoFile pso) where T : struct, IPsoSwapEnd
        {
            var i = pso.DataMapSection.RootId - 1;
            var e = pso.DataMapSection.Entries[i];
            return GetItem<T>(pso, e.Offset);
        }
        public static PsoDataMappingEntry GetRootEntry(PsoFile pso)
        {
            var i = pso.DataMapSection.RootId - 1;
            var e = pso.DataMapSection.Entries[i];
            return e;
        }

        public static T[] GetItemArrayRaw<T>(PsoFile pso, Array_Structure arr) where T : struct
        {
            if ((arr.Count1 > 0) && (arr.Pointer > 0))
            {
                var entry = pso.DataMapSection.Entries[arr.PointerDataIndex];
                return ConvertDataArrayRaw<T>(pso.DataSection.Data, entry.Offset, arr.Count1);
            }
            return null;
        }
        public static T[] GetItemArray<T>(PsoFile pso, Array_Structure arr) where T : struct, IPsoSwapEnd
        {
            if ((arr.Count1 > 0) && (arr.Pointer > 0))
            {
                var entry = pso.DataMapSection.Entries[arr.PointerDataIndex];
                var res = ConvertDataArrayRaw<T>(pso.DataSection.Data, entry.Offset, arr.Count1);
                if (res != null)
                {
                    for (int i = 0; i < res.Length; i++)
                    {
                        res[i].SwapEnd();
                    }
                }
                return res;
            }
            return null;
        }


        public static uint[] GetUintArrayRaw(PsoFile pso, Array_uint arr)
        {
            byte[] data = pso.DataSection.Data;
            var entryid = arr.PointerDataId;
            if ((entryid == 0) || (entryid > pso.DataMapSection.EntriesCount))
            {
                return null;
            }
            var entryoffset = arr.PointerDataOffset;
            var arrentry = pso.DataMapSection.Entries[(int)entryid - 1];
            int totoffset = arrentry.Offset + (int)entryoffset;
            uint[] readdata = ConvertDataArrayRaw<uint>(data, totoffset, arr.Count1);
            return readdata;
        }
        public static uint[] GetUintArray(PsoFile pso, Array_uint arr)
        {
            uint[] uints = GetUintArrayRaw(pso, arr);
            if (uints == null) return null;
            for (int i = 0; i < uints.Length; i++)
            {
                uints[i] = MetaTypes.SwapBytes(uints[i]);
            }
            return uints;
        }

        public static MetaHash[] GetHashArray(PsoFile pso, Array_uint arr)
        {
            uint[] uints = GetUintArrayRaw(pso, arr);
            if (uints == null) return null;
            MetaHash[] hashes = new MetaHash[uints.Length];
            for (int n = 0; n < uints.Length; n++)
            {
                hashes[n].Hash = MetaTypes.SwapBytes(uints[n]);
            }
            return hashes;
        }




        public static float[] GetFloatArrayRaw(PsoFile pso, Array_float arr)
        {
            byte[] data = pso.DataSection.Data;
            var entryid = arr.PointerDataId;
            if ((entryid == 0) || (entryid > pso.DataMapSection.EntriesCount))
            {
                return null;
            }
            var entryoffset = arr.PointerDataOffset;
            var arrentry = pso.DataMapSection.Entries[(int)entryid - 1];
            int totoffset = arrentry.Offset + (int)entryoffset;
            float[] readdata = ConvertDataArrayRaw<float>(data, totoffset, arr.Count1);
            return readdata;
        }
        public static float[] GetFloatArray(PsoFile pso, Array_float arr)
        {
            float[] floats = GetFloatArrayRaw(pso, arr);
            if (floats == null) return null;
            for (int i = 0; i < floats.Length; i++)
            {
                floats[i] = MetaTypes.SwapBytes(floats[i]);
            }
            return floats;
        }





        public static ushort[] GetUShortArrayRaw(PsoFile pso, Array_Structure arr)
        {
            byte[] data = pso.DataSection.Data;
            var entryid = arr.PointerDataId;
            if ((entryid == 0) || (entryid > pso.DataMapSection.EntriesCount))
            {
                return null;
            }
            var entryoffset = arr.PointerDataOffset;
            var arrentry = pso.DataMapSection.Entries[(int)entryid - 1];
            int totoffset = arrentry.Offset + (int)entryoffset;
            ushort[] readdata = ConvertDataArrayRaw<ushort>(data, totoffset, arr.Count1);
            return readdata;
        }
        public static ushort[] GetUShortArray(PsoFile pso, Array_Structure arr)
        {
            ushort[] ushorts = GetUShortArrayRaw(pso, arr);
            if (ushorts == null) return null;
            for (int i = 0; i < ushorts.Length; i++)
            {
                ushorts[i] = MetaTypes.SwapBytes(ushorts[i]);
            }
            return ushorts;
        }






        public static T[] GetObjectArray<T, U>(PsoFile pso, Array_Structure arr) where U : struct, IPsoSwapEnd where T : PsoClass<U>, new()
        {
            U[] items = GetItemArray<U>(pso, arr);
            if (items == null) return null;
            if (items.Length == 0) return null;
            T[] result = new T[items.Length];
            for (int i = 0; i < items.Length; i++)
            {
                T newitem = new T();
                newitem.Init(pso, ref items[i]);
                result[i] = newitem;
            }
            return result;
        }


        public static byte[] GetByteArray(PsoFile pso, PsoStructureEntryInfo entry, int offset)
        {
            var aCount = (entry.ReferenceKey >> 16) & 0x0000FFFF;
            var aBlockId = (int)entry.ReferenceKey & 0x0000FFFF;
            var block = pso.GetBlock(aBlockId);
            if (block == null) return null;

            //block.Offset

            return null;
        }






        public static PsoPOINTER[] GetPointerArray(PsoFile pso, Array_StructurePointer array)
        {
            uint count = array.Count1;
            if (count == 0) return null;

            int ptrsize = Marshal.SizeOf(typeof(MetaPOINTER));
            int itemsleft = (int)count; //large arrays get split into chunks...
            uint ptrindex = array.PointerDataIndex;
            uint ptroffset = array.PointerDataOffset;
            var ptrblock = (ptrindex < pso.DataMapSection.EntriesCount) ? pso.DataMapSection.Entries[ptrindex] : null;
            if ((ptrblock == null) || (ptrblock.NameHash != (MetaName)MetaTypeName.PsoPOINTER))
            { return null; }

            var offset = ptrblock.Offset;
            int boffset = (int)(offset + ptroffset);

            var ptrs = ConvertDataArrayRaw<PsoPOINTER>(pso.DataSection.Data, boffset, (int)count);
            if (ptrs != null)
            {
                for (int i = 0; i < ptrs.Length; i++)
                {
                    ptrs[i].SwapEnd();
                }
            }

            return ptrs;
        }


        public static T[] ConvertDataArray<T>(PsoFile pso, Array_StructurePointer array) where T : struct, IPsoSwapEnd
        {
            uint count = array.Count1;
            if (count == 0) return null;
            PsoPOINTER[] ptrs = GetPointerArray(pso, array);
            if (ptrs == null) return null;
            if (ptrs.Length < count)
            { return null; }

            T[] items = new T[count];
            int itemsize = Marshal.SizeOf(typeof(T));

            for (int i = 0; i < count; i++)
            {
                var sptr = ptrs[i];
                int blocki = sptr.BlockID - 1;
                int offset = (int)sptr.ItemOffset;// * 16;//block data size...
                if (blocki >= pso.DataMapSection.EntriesCount)
                { continue; }
                var block = pso.DataMapSection.Entries[blocki];

                if ((offset < 0) || (offset >= block.Length))
                { continue; }

                int boffset = block.Offset + offset;

                items[i] = ConvertData<T>(pso.DataSection.Data, boffset);
            }

            return items;
        }



        public static string GetString(PsoFile pso, CharPointer ptr)
        {
            if (ptr.Count1 == 0) return null;

            var blocki = (int)ptr.PointerDataId;// (ptr.Pointer & 0xFFF) - 1;
            var offset = (int)ptr.PointerDataOffset;// (ptr.Pointer >> 12) & 0xFFFFF;

            var block = pso.GetBlock(blocki); //nameHash = 1
            if (block == null)
            { return null; }

            //if (block.NameHash != (MetaName)1)
            //{ }

            var length = ptr.Count1;
            var lastbyte = offset + length;
            if (lastbyte >= block.Length)
            { return null; }

            var data = pso.DataSection?.Data;
            if (data == null)
            { return null; }

            var doffset = block.Offset + offset;

            string s = Encoding.ASCII.GetString(data, doffset, length);

            //if (meta.Strings == null) return null;
            //if (offset < 0) return null;
            //if (offset >= meta.Strings.Length) return null;
            //string s = meta.Strings[offset];

            return s;
        }
        public static string GetString(PsoFile pso, DataBlockPointer ptr)
        {
            var blocki = (int)ptr.PointerDataId;// (ptr.Pointer & 0xFFF) - 1;
            var offset = (int)ptr.PointerDataOffset;// (ptr.Pointer >> 12) & 0xFFFFF;

            var block = pso.GetBlock(blocki); //nameHash = 1
            if (block == null)
            { return null; }

            //if (block.NameHash != (MetaName)1)
            //{ }

            //var length = ptr.Count1;
            //var lastbyte = offset + length;
            //if (lastbyte >= block.Length)
            //{ return null; }

            var data = pso.DataSection?.Data;
            if (data == null)
            { return null; }

            //var doffset = block.Offset + offset;

            //string s = Encoding.ASCII.GetString(data, doffset, length);

            StringBuilder sb = new StringBuilder();
            var o = block.Offset + offset;
            char c = (char)data[o];
            while (c != 0)
            {
                sb.Append(c);
                o++;
                c = (char)data[o];
            }
            var s = sb.ToString();

            return s;
        }


    }


    public interface IPsoSwapEnd
    {
        void SwapEnd();
    }

    public abstract class PsoClass<T> where T : struct, IPsoSwapEnd
    {
        public abstract void Init(PsoFile pso, ref T v);
    }


    public struct PsoChar64
    {
        public byte b00, b01, b02, b03, b04, b05, b06, b07, b08, b09,
                    b10, b11, b12, b13, b14, b15, b16, b17, b18, b19,
                    b20, b21, b22, b23, b24, b25, b26, b27, b28, b29,
                    b30, b31, b32, b33, b34, b35, b36, b37, b38, b39,
                    b40, b41, b42, b43, b44, b45, b46, b47, b48, b49,
                    b50, b51, b52, b53, b54, b55, b56, b57, b58, b59,
                    b60, b61, b62, b63;

        public override string ToString()
        {
            byte[] bytes = new byte[]
            {
                b00, b01, b02, b03, b04, b05, b06, b07, b08, b09,
                b10, b11, b12, b13, b14, b15, b16, b17, b18, b19,
                b20, b21, b22, b23, b24, b25, b26, b27, b28, b29,
                b30, b31, b32, b33, b34, b35, b36, b37, b38, b39,
                b40, b41, b42, b43, b44, b45, b46, b47, b48, b49,
                b50, b51, b52, b53, b54, b55, b56, b57, b58, b59,
                b60, b61, b62, b63
            };
            return Encoding.ASCII.GetString(bytes).Replace("\0", string.Empty);
        }
    }
    public struct PsoChar32
    {
        public byte b00, b01, b02, b03, b04, b05, b06, b07, b08, b09,
                    b10, b11, b12, b13, b14, b15, b16, b17, b18, b19,
                    b20, b21, b22, b23, b24, b25, b26, b27, b28, b29,
                    b30, b31;

        public PsoChar32(string s)
        {
            s = s.PadRight(32, '\0');
            b00 = (byte)s[0];
            b01 = (byte)s[1];
            b02 = (byte)s[2];
            b03 = (byte)s[3];
            b04 = (byte)s[4];
            b05 = (byte)s[5];
            b06 = (byte)s[6];
            b07 = (byte)s[7];
            b08 = (byte)s[8];
            b09 = (byte)s[9];
            b10 = (byte)s[10];
            b11 = (byte)s[11];
            b12 = (byte)s[12];
            b13 = (byte)s[13];
            b14 = (byte)s[14];
            b15 = (byte)s[15];
            b16 = (byte)s[16];
            b17 = (byte)s[17];
            b18 = (byte)s[18];
            b19 = (byte)s[19];
            b20 = (byte)s[20];
            b21 = (byte)s[21];
            b22 = (byte)s[22];
            b23 = (byte)s[23];
            b24 = (byte)s[24];
            b25 = (byte)s[25];
            b26 = (byte)s[26];
            b27 = (byte)s[27];
            b28 = (byte)s[28];
            b29 = (byte)s[29];
            b30 = (byte)s[30];
            b31 = (byte)s[31];
        }
        public override string ToString()
        {
            byte[] bytes = new byte[]
            {
                b00, b01, b02, b03, b04, b05, b06, b07, b08, b09,
                b10, b11, b12, b13, b14, b15, b16, b17, b18, b19,
                b20, b21, b22, b23, b24, b25, b26, b27, b28, b29,
                b30, b31
            };
            return Encoding.ASCII.GetString(bytes).Replace("\0", string.Empty);
        }
    }




    [TC(typeof(EXP))] public struct PsoPOINTER : IPsoSwapEnd //8 bytes - pointer to data item
    {
        public ulong Pointer { get; set; }

        public ushort BlockID { get { return (ushort)(Pointer & 0xFFF); } } //1-based ID
        public uint ItemOffset { get { return (uint)((Pointer>>12) & 0xFFFFF); } } //byte offset
        public uint Unk0 { get { return (uint)((Pointer>>32) & 0xFFFFFFFF); } }


        public PsoPOINTER(int blockID, int itemOffset)
        {
            Pointer = (((uint)itemOffset << 12) & 0xFFFFF000) + ((uint)blockID & 0xFFF);
        }


        public override string ToString()
        {
            return BlockID.ToString() + ", " + ItemOffset.ToString();// + ", " + Unk2.ToString();
        }

        public void SwapEnd()
        {
            Pointer = MetaTypes.SwapBytes(Pointer);
        }
    }









    //Struct infos
    [TC(typeof(EXP))] public struct CPackFileMetaData : IPsoSwapEnd //96 bytes, Type:0
    {
        public Array_Structure MapDataGroups; //0   MapDataGroups: Array: 0  {256: Structure: 0: 3260758307}
        public Array_Structure HDTxdBindingArray; //16   HDTxdBindingArray: Array: 16: 2  {256: Structure: 0: CHDTxdAssetBinding}
        public Array_Structure imapDependencies; //32   imapDependencies: Array: 32: 4  {256: Structure: 0: 3501026914}
        public Array_Structure imapDependencies_2; //48   imapDependencies_2: Array: 48: 6  {256: Structure: 0: 3240050401}
        public Array_Structure itypDependencies_2; //64   itypDependencies_2: Array: 64: 8  {256: Structure: 0: 1515605584}
        public Array_Structure Interiors; //80   Interiors: Array: 80: 10  {256: Structure: 0: 741495440}

        public void SwapEnd()
        {
            MapDataGroups.SwapEnd();
            HDTxdBindingArray.SwapEnd();
            imapDependencies.SwapEnd();
            imapDependencies_2.SwapEnd();
            itypDependencies_2.SwapEnd();
            Interiors.SwapEnd();
        }
    }

    [TC(typeof(EXP))] public struct CMapDataGroup : IPsoSwapEnd //56 bytes, Type:0
    {
        public MetaHash Name { get; set; } //0   Name: INT_0Bh: 0
        public uint Unused0 { get; set; } //4
        public Array_uint Bounds { get; set; } //8   Bounds: Array: 8: 1  {256: INT_0Bh: 0}
        public ushort Flags { get; set; } //24   Flags: SHORT_0Fh: 24: 2097155
        public ushort Unused1 { get; set; }//26
        public uint Unused2 { get; set; }//28
        public Array_uint WeatherTypes { get; set; } //32   WeatherTypes: Array: 32: 5  {256: INT_0Bh: 0}
        public uint HoursOnOff { get; set; } //48   HoursOnOff//4190815249: INT_06h: 48
        public uint Unused3 { get; set; }//52

        public override string ToString()
        {
            return Name.ToString() + ": ybn:" + Bounds.Count1.ToString() + ", wt:" + WeatherTypes.Count1.ToString() + ", flags:" + Flags.ToString() + ", hours:" + HoursOnOff.ToString();
        }

        public void SwapEnd()
        {
            Name = new MetaHash(MetaTypes.SwapBytes(Name.Hash));
            var b = Bounds; b.SwapEnd(); Bounds = b;
            var w = WeatherTypes; w.SwapEnd(); WeatherTypes = w;
            HoursOnOff = MetaTypes.SwapBytes(HoursOnOff);
            Flags = MetaTypes.SwapBytes(Flags);
        }
    }

    [TC(typeof(EXP))] public struct CHDTxdAssetBinding : IPsoSwapEnd //132 bytes, Type:0
    {
        public byte assetType { get; set; } //0   assetType: BYTE_ENUM_VALUE: 0: 3387532954
        public byte Unused01 { get; set; }//1
        public ushort Unused02 { get; set; }//2
        public PsoChar64 targetAsset { get; set; } //4   targetAsset: INT_0Bh: 4: 4194304
        public PsoChar64 HDTxd { get; set; } //68   HDTxd: INT_0Bh: 68: 4194304

        public override string ToString()
        {
            return assetType.ToString() + ": " + targetAsset.ToString() + ": " + HDTxd.ToString();
        }
        public void SwapEnd()
        {
            //targetAsset.Hash = MetaTypes.SwapBytes(targetAsset.Hash);
            //HDTxd.Hash = MetaTypes.SwapBytes(HDTxd.Hash);
        }
    }

    [TC(typeof(EXP))] public struct CImapDependency : IPsoSwapEnd //12 bytes, Type:0  //  CImapDependency//3501026914
    {
        public MetaHash imapName { get; set; } //0   imapName: INT_0Bh: 0
        public MetaHash itypName { get; set; } //4   itypName//2890158180: INT_0Bh: 4
        public MetaHash packFileName { get; set; } //8   packFileName//4216494073: INT_0Bh: 8

        public override string ToString()
        {
            return imapName.ToString() + ", " + itypName.ToString() + ", " + packFileName.ToString();
        }
        public void SwapEnd()
        {
            imapName = new MetaHash(MetaTypes.SwapBytes(imapName.Hash));
            itypName = new MetaHash(MetaTypes.SwapBytes(itypName.Hash));
            packFileName = new MetaHash(MetaTypes.SwapBytes(packFileName.Hash));
        }
    }

    [TC(typeof(EXP))] public struct CImapDependencies : IPsoSwapEnd //24 bytes, Type:0  // CImapDependencies//3240050401 imapDependencies_2
    {
        public MetaHash imapName { get; set; } //0   imapName: INT_0Bh: 0  //name hash
        public ushort manifestFlags { get; set; } //4   manifestFlags//1683136603: SHORT_0Fh: 4: 2097153
        public ushort Unused0 { get; set; } //6
        public Array_uint itypDepArray { get; set; } //8   itypDepArray//2410949350: Array: 8: 3  {256: INT_0Bh: 0} //children...


        public override string ToString()
        {
            return imapName.ToString() + ": " + manifestFlags.ToString() + ": " + itypDepArray.ToString();
        }
        public void SwapEnd()
        {
            imapName = new MetaHash(MetaTypes.SwapBytes(imapName.Hash));
            manifestFlags = MetaTypes.SwapBytes(manifestFlags);
            var d = itypDepArray; d.SwapEnd(); itypDepArray = d;
        }
    }

    [TC(typeof(EXP))] public struct CItypDependencies : IPsoSwapEnd //24 bytes, Type:0 // CItypDependencies//1515605584  itypDependencies_2
    {
        public MetaHash itypName { get; set; } //0   itypName//2890158180: INT_0Bh: 0
        public ushort manifestFlags { get; set; } //4   manifestFlags//1683136603: SHORT_0Fh: 4: 2097153
        public ushort Unused0 { get; set; } //6
        public Array_uint itypDepArray { get; set; } //8   itypDepArray//2410949350: Array: 8: 3  {256: INT_0Bh: 0}

        public override string ToString()
        {
            return itypName.ToString() + ": " + manifestFlags.ToString() + ": " + itypDepArray.ToString();
        }
        public void SwapEnd()
        {
            itypName = new MetaHash(MetaTypes.SwapBytes(itypName.Hash));
            manifestFlags = MetaTypes.SwapBytes(manifestFlags);
            var d = itypDepArray; d.SwapEnd(); itypDepArray = d;
        }
    }

    [TC(typeof(EXP))] public struct CInteriorBoundsFiles : IPsoSwapEnd //24 bytes, Type:0   // Interiors
    {
        public MetaHash Name { get; set; } //0   Name: INT_0Bh: 0
        public uint Unused0 { get; set; } //4
        public Array_uint Bounds { get; set; } //8   Bounds: Array: 8: 1  {256: INT_0Bh: 0}

        public override string ToString()
        {
            return JenkIndex.GetString(Name);
        }
        public void SwapEnd()
        {
            Name = new MetaHash(MetaTypes.SwapBytes(Name.Hash));
            var b = Bounds; b.SwapEnd(); Bounds = b;
        }
    }






    [TC(typeof(EXP))] public struct CScenarioPointManifest : IPsoSwapEnd //56 bytes, Type:0
    {
        public int VersionNumber { get; set; } //0   VersionNumber: INT_05h: 0
        public uint Unused0 { get; set; } //4
        public Array_StructurePointer RegionDefs { get; set; } //8   RegionDefs: Array: 8: 1  {ARRAYINFO: Structure: 0}
        public Array_StructurePointer Groups { get; set; } //24   Groups: Array: 24: 3  {ARRAYINFO: Structure: 0}
        public Array_uint InteriorNames { get; set; } //40   InteriorNames: Array: 40: 5  {ARRAYINFO: INT_0Bh: 0}

        public override string ToString()
        {
            return VersionNumber.ToString();
        }
        public void SwapEnd()
        {
            VersionNumber = MetaTypes.SwapBytes(VersionNumber);
            var r = RegionDefs; r.SwapEnd(); RegionDefs = r;
            var g = Groups; g.SwapEnd(); Groups = g;
            var i = InteriorNames; i.SwapEnd(); InteriorNames = i;
        }
    }

    [TC(typeof(EXP))] public struct CScenarioPointRegionDef : IPsoSwapEnd //64 bytes, Type:0
    {
        public MetaHash Name { get; set; } //0   Name: INT_0Bh: 0
        public uint Unused0 { get; set; } //4
        public uint Unused1 { get; set; } //8
        public uint Unused2 { get; set; } //12
        public rage__spdAABB AABB { get; set; } //16   AABB: Structure: 16: rage__spdAABB
        public uint Unused3 { get; set; } //48
        public uint Unused4 { get; set; } //52
        public uint Unused5 { get; set; } //56
        public uint Unused6 { get; set; } //60

        public override string ToString()
        {
            return Name.ToString() + ", " + AABB.ToString();
        }
        public void SwapEnd()
        {
            Name = new MetaHash(MetaTypes.SwapBytes(Name.Hash));
            var aabb = AABB; aabb.SwapEnd(); AABB = aabb;
        }
    }

    [TC(typeof(EXP))] public struct CScenarioPointGroup : IPsoSwapEnd //8 bytes, Type:0
    {
        public MetaHash Name { get; set; } //0   Name: INT_0Bh: 0
        public byte EnabledByDefault { get; set; } //4   EnabledByDefault: BYTE_00h: 4
        public byte Unused0 { get; set; } //5
        public ushort Unused1 { get; set; } //6

        public override string ToString()
        {
            return Name.ToString();
        }
        public void SwapEnd()
        {
            Name = new MetaHash(MetaTypes.SwapBytes(Name.Hash));
        }
    }




}
